(*<*)

theory START_HERE
imports
  concurrent_buffer_example
begin
(*>*)

text{*
This is the technical report for the Isabelle/HOL~@{cite "isabelle"} formalization accompanying the dissertation of Nils Berg~@{cite "bergdiss"}.
For explanations with regard to content please refer to the dissertation.
The intention of this document is to give a mapping from the formalization in the dissertation to the formalization in Isabelle/HOL.
Formalized are the parts where user interaction is required, i.e., the first part of the dissertation,
where {\em Communicating Sequential Processes}~(CSP) processes and {\em Communicating Unstructured Code}~(CUC) programs
are related.
More specifically, the sections 5.2 to 5.5 of the dissertation are formalized.
Due to technical implementation reasons and because the Isabelle/HOL formalization is a bit older than the formalization in the dissertation, which has evolved since, there is not a one-to-one mapping from the formalization in the dissertation to the formalization in Isabelle/HOL.
Most notably, the operational semantics of CUC is not concurrent.
The reasoning (that it is sufficient to consider single components) is explained in Section 5.6 of the dissertation.
The concurrent cases of the correspondence proofs between the
operational and denotational semantics that are missing in this formalization are given in the Appendix A.1 of the dissertation.

This document is intended to be read in its source code form in an Isabelle IDE (e.g.~@{cite "isabelle2017"}).
The source code of this document can be found in the file \texttt{START\_HERE.thy}, which is published together with this document~@{cite "report"}.
It will load all necessary dependencies.
Please note that the 2017 version of Isabelle/HOL is required for the provided theory files.
In a suitable Isabelle IDE, all terms in antiquotations (like this @{theory concurrent_buffer_example}) are links to the
definitions, lemmas, theories, etc. they refer to.
To go to the link target, hold the Command Key on a Mac or the Control Key on Linux/Windows
and and click the term.
Hovering the link holding the modifier key (without clicking) usually gives a small preview of the linked formalization.

In the Isabelle/HOL formalization, we use the concept of a locale.
A common example for the application of a locale is group theory, where an arbitrary but fixed operator and a base set are assumed.
The group theory locale can then be instantiated for concrete base sets and operators.
In general, a locale allows to define abstract theories by fixing one or more abstract entities (in our case usually the unstructured code).
The locale groups assumptions, definitions and lemmas (similar to a namespace).
It allows for instantiation: the fixed abstract entities are replaced by a concrete entities.
The instantiation of a locale requires proofs for all the assumptions and then provides instantiated versions all the definitions and lemmas that are grouped by the locale. 


For the example in Section~\ref{example}, we use the CSP formalization provided by the CSP-Prover library~@{cite "csp-prover"} by Yoshinao Isobe and Markus Roggenbach (in version 5-1-2016), which is available at its website~@{cite "csp-prover-website"}.
We have included it with our source code.

In the remainder of this document, we give a mapping from the defintions, assumptions, lemmas, and theorems from dissertation to their counterparts in the Isabelle/HOL formalization.
The numbering is the same as in the dissertation.


 *}
section {* Syntax and Semantic States *} (* 5.2 *)

text{*

\begin{definition}[Basic Data Types] % 5.1
  The local states (LStates) are defined by the record @{typ "'data_store definitions.state"}.
  It is parametrized by the type @{typ 'data_store} of the data store (formerly register store) R.
  It contains the data store and a program counter of type @{typ definitions.label}.
  Concurrent states are defined in @{typ "'state denotational_parallel.parallelState"}

  Events can be of any type @{typ "'e definitions.event"}.
  The set of all events can be fixed within a locale, see for example
  @{locale stable_failures.SFsemSigma}.

\end{definition}


\begin{definition}[Instructions of CUC] % 5.2
  The instructions of CUC are formalized in @{typ "('state, 
  'event) instr"} and are parametrized over the type of the state and the type of events.

\end{definition}

\begin{assumption}[At Least One Successor State] % 5.1
  We assume this assumption implicitly by only using functions $f$ that map to non-empty sets for {\tt do} $f$.
\end{assumption}

\begin{definition}[Local Program {\it lp}]  % 5.3
  The local programs are of the type @{typ "('state, 'event) labeled_instruction_set"}.
\end{definition}

\begin{assumption}[Uniqueness of Labels]  % 5.2
  The uniqueness of labels is required in the respective lemmas.
  The notion of uniqueness of labels for unstructured programs is captured as an assumption of the locale @{const liSet}.
  The notion of uniqueness of labels for structured programs is captured by the definition @{const sc_wellformed}.
  
\end{assumption}

\begin{definition}[Labels of a Program labels]  % 5.4
  To extract the labels of an (unstructured) program, we use the definition @{const indom}.
\end{definition}

\begin{definition}[Concurrent Program {\it cp}]  % 5.5
  We do not define unstructured concurrent programs in this formalization.
  This is due to the fact that we relate structured concurrent programs to CSP processes.
  The type of structured concurrent programs is defined by @{typ "('state, 'event)structured_code_parallel "}.
\end{definition}

\begin{assumption}[Same Tree Structure]  % 5.3
  This is defined in the semantics, e.g., see @{const denotational_parallel.densem\<^sub>p}:
  We define successor states only for states whose tree structure matches the tree structure of the code.
\end{assumption}

\begin{definition}[Structured Program {\it sp}]  % 5.6
  The typ used for structured code is @{typ "('state, 'event)structured_code "} which is actually just a wrapper
  around the definition @{typ "'code structured_code_base"} which builds a tree with single lines of labeled instruction as leaves.
\end{definition}

\begin{definition}[Unstructuring Function $\mathcal{U}$]  % 5.7
  To obtain the unstructured code from structured code we use the function @{const sc_inj}.
\end{definition}

\begin{definition}[Labels of a Structured Program labels]  % 5.8
  The labels of a structured program are extracted directly in this formalization via @{const dom}.
\end{definition}

\begin{assumption}[Uniqueness of Labels for {\it sp}]  % 5.4
  As for the uniqueness of labels of unstructured code, we require this property in the locales.
  For the uniqueness of labels of structured code, we use the function @{const sc_wellformed}.
\end{assumption}

\begin{definition}[Structured Concurrent Program {\it scp}]  % 5.9
  The type of structured concurrent programs is defined by @{typ "('state, 'event)structured_code_parallel "}.
\end{definition}

*}

section {* Semantics *} (* 5.3 *)


subsection {* Operational Semantics *} (* 5.3.1 *)

text{*  

\begin{definition}[Operational Semantics of CUC]  % 5.10
  The operational semantics of CUC are distributed over several definitions.
  The smallstep semantics of CUC are defined by @{const smallstep}, which is an inductively defined set of
  @{typ "('state, 'event) trace_state "} pairs, describing trace-state pairs before and after the execution of an instruction,
  i.e., the transitions.
  @{const smallstep} is parametrized by a labeled instruction set.
  To require the wellformedness assumptions on unstructured programs (uniqueness of labels),
  we defined a wrapper named Smallstep (observe the capital 'S') @{const small_step.liSet.Smallstep} inside a locale. 

  The concurrent operational semantics {\em are not} formalized. See the dissertation for the explanation.
  The concurrent denotational semantics {\em are} formalized.

  The reflexive transitive hull of small step is defined in @{const small_step.liSet.multistep}.
  For historical reasons (see the next definition) multistep is defined by prepending to the execution traces.
  Which is in contrast to the dissertation, where it is defined by appending.

\end{definition}


\begin{definition}[Terminating Execution]  % 5.11
  Terminating executions are defined by @{const small_step.liSet.Exec}, which is similar to multistep.
  The difference is that the ``final empty step'' is only allowed in exec, if the program counter points outside of the code,
  i.e., the execution has terminated.
\end{definition}

*}


subsection {* Defining Denotational Semantics with Fixpoints *} (* 5.3.2 *)

text{*  
  For fixpoint theory, we use the existing HOL library @{const Complete_Partial_Order.ccpo_class} which provides most notably
  @{thm[display] "fixp_induct"}.
*}

subsection {* Traces Semantics *} (* 5.3.3 *)

text{*  

\begin{definition}[Traces Semantics of CUC]  % 5.12
  The sequential part of the traces semantics is defined in @{const denotational.densem}.
  The concurrent part of the traces semantics is defined in @{const denotational_parallel.densem\<^sub>p}.
\end{definition}

\begin{definition}[Operational Characterization of the Traces of CUC]  % 5.13
  This is not defined explicitly in the formalization. The operational and the traces semantics are related
  directly in a lemma.
  See the correspondence theorem below (Theorem~\ref{correspondence_traces}).
\end{definition}

\begin{assumption}[Empty Initial Traces]  % 5.5
  The assumption is only implicit in that instantiations only use an empty trace initially.
  See for example the preconditon in the example @{const concurrent_buffer_example.SingleBuffer.TPre}.
\end{assumption}

\begin{theorem}[Correspondence Between Operational Characterization and Traces Semantics]  % 5.1
  \label{correspondence_traces}
  The correspondence between the operational semantics and the traces semantics is captured by theorems,
  one for each direction.
  @{thm[display] denotational.liSet.densem_implies_multistep}
  @{thm[display] denotational.liSet.multistep_implies_densem}
\end{theorem}

\begin{corollary}[Invariance Under Structure]  % 5.1
  In the formalization, the independence of the denotational semantics from a particular structure on the unstructured code
  is proven as a step towards the proof of the previous theorem (in contrast to being a corollary of it).
  @{thm[display,names_short] denotational.oplus_commutative}
  @{thm[display,names_short] denotational.liSet.oplus_associative}
\end{corollary}

*}


subsection {* Stable Failures Semantics *} (* 5.3.4 *)

text{*  

\begin{definition}[CSP-like Stable States of CUC]  % 5.14
  The definition of stable states is implicit in the definition of stable failures @{const stable_failures.SFsemSigma.sfsem}.
  Only stable states are included in the semantics.
  That the stable failure semantics works as expected is shown in @{theory stable_failures_traces_conformant} by proving SF1 to SF4.
\end{definition}

\begin{definition}[Refusal Set of CUC]  % 5.15
  As for the stable states, the refusal sets are defined implicitly in the definition of stable failures.
\end{definition}

\begin{definition}[Operational Characterization of the Stable Failures of CUC]  % 5.16
  This is not defined explicitly in the formalization. The operational and the stable failures semantics are related
  directly in a lemma.
  See the correspondence theorem below (Theorem~\ref{correspondence_stable_failures}).
\end{definition}

\begin{definition}[Communication States]  % 5.17
  The communication and normal states are defined in @{typ "'state NCstate"}.
  In the formalization we use a sum datatype with explicit constructors.
  Thus, we do not need to append ``$\times \{c\}$''.
  The tree of concurrent NCStates (named CNCStates) is defined in @{typ "'state parallelState"}.
\end{definition}

\begin{definition}[Test for Normal State and Conversions]  % 5.18
  The test for normal state is defined by @{const N}.
  The conversion from normal state to communication is implemented by @{const NCC}.
\end{definition}

\begin{assumption}[Initial Failures]  % 5.6
  We usually generate initial failures from traces using the following function which generates 
  refusal sets according to the assumption. 
  @{const "stable_failures_traces_conformant.SFsemSigma.trace_state_set_to_failure_set"}
\end{assumption}

\begin{definition}[Removal of Former Terminal Failures]  % 5.19
  The former terminal failures are removed by @{const SClean}.
\end{definition}

\begin{definition}[Stable Failures Semantics of CUC]  % 5.20
  The sequential part of the traces semantics is defined in @{const stable_failures.SFsemSigma.sfsem}
  The concurrent part of the traces semantics is defined in @{const stable_failures_parallel.sfsem\<^sub>p}.
\end{definition}


\begin{theorem}[Correspondence Between Operational Characterization and Stable Failures Semantics]  % 5.2
  \label{correspondence_stable_failures}
  The correspondence between the operational semantics and the stable failures semantics is captured by theorems,
  one for each direction.
  @{thm[display] stable_failures.SFsemSigma.sfsem_implies_multistep}
  Observe that due to the implicit definition of stable states and refusal sets,
  the direction from multistep to failures is split in two theorems:
  One for terminating executions
  @{thm[display] stable_failures.SFsemSigma.Exec_implies_sfsem}
  and one for non-terminating executions
  @{thm[display] stable_failures.SFsemSigma.multistep_trace_implies_comm_failure}
\end{theorem}

*}


subsection {* Compatibility to CSP *} (* 5.3.5 *)

text{*  

We did not show T1 and T2 explicitly.

T1: If we assume that the initial traces are empty, we can use @{thm denotational.densem_super_id_element} to follow
that the traces semantics also contains a trace-state pair with the empty trace.

T2: We have shown that multistep is prefix closed in @{thm[display] small_step.liSet.multistep_prefix_closed}
Using the correspondence between multistep and the traces semantics, we conclude that the traces semantics
is also prefix closed. 

The properties SF1 to SF4 are linked to in the source file of this document: first for the sequential case, then for the concurrent case.

% @{thm stable_failures_traces_conformant.SFsemSigma.SF1}
% @{thm stable_failures_traces_conformant.SFsemSigma.SF2}
% @{thm stable_failures_traces_conformant.SFsemSigma.SF3}
% @{thm stable_failures_traces_conformant.SFsemSigma.SF4}
% 
% @{thm stable_failures_traces_conformance_parallel.SF1_parallel}
% @{thm stable_failures_traces_conformance_parallel.SF2_parallel}
% @{thm stable_failures_traces_conformance_parallel.SF3_parallel}
% @{thm stable_failures_traces_conformance_parallel.SF4_parallel}

*}


section {* Hoare Calculus *}  (* 5.4 *)

text{*

\begin{definition}[Hoare Tripel for CUC]  % 5.21
  The Hoare triple is defined in @{const hoare.SFsemSigma.hoare_valid}.
\end{definition}

\begin{definition}[Hoare Calculus for CUC]  % 5.22
  The Hoare calculus itself is defined in @{const hoare.SFsemSigma.hoare}.
\end{definition}

\begin{theorem}[Soundness of Our Hoare Calculus]  % 5.3
  We have proven the soundness of our Hoare Calculus in @{thm[display] hoare.SFsemSigma.hsem_soundness}
\end{theorem}

*}


section {* Relating CSP and CUC*}

text{*

\begin{lemma}[Traces Imply Stable Failures]  % 5.1
  When we only consider divergence free CUC programs, then all traces appear also in the stable failures semantics.
  @{thm[display] stable_failures_traces_conformant.SFsemSigma.tr_of_T_will_be_in_SF}
  The lemma talks about successors of starting states, as we allow arbitrary traces for initial trace-state pairs.
\end{lemma}

\begin{theorem}[Traces Refinement Implies Stable Failures Refinement for CUC]  % 5.4
  Again, this theorem only holds for divergence free programs.
  @{thm[display] stable_failures_traces_conformant.SFsemSigma.sfsem_no_divergencies_imp_sf_refinement_implies_traces_refinement}
\end{theorem}

\begin{assumption}[Divergence Freedom of CUC Programs]  % 5.7
  We assume divergence freedom directly in lemmas and theorems (see the theorems linked above).
  For our (quite simple) examples we used static code analysis, to ensure that every instruction will either be followed by
  a comm instruction (and thus produce a visible event) or will lead to termination.
  The definitions and lemmas can be found in @{theory static_code_analysis}. 
\end{assumption}

*}

subsection {* Example *} (* 5.5.1 *)
text{*
\label{example}

Reminder: For the formalization of CSP, we use the CSP Prover library~@{cite "csp-prover"} by Yoshinao Isobe and Markus Roggenbach.

We have formalized the example in the theory file @{theory concurrent_buffer_example}.
The major difference in the structure is that we first show that the code fulfills the sufficient property,
and afterwards that the sufficient property implies the specification. 
In contrast to the dissertation, in the formalization we need many more fine grained definitions and lemmas.

The CUC program is defined in @{const concurrent_buffer_example.SingleBuffer.LIS}.
The CSP specification is defined in the assumptions of the lemmas that use it, e.g.
@{thm[display] concurrent_buffer_example.SingleBuffer.Conn_Imp_Spec}
It is defined in fixpoint notation
@{term[display] "(([[$spec]]Ff [[PN]]Ffix)::(nat \<times> nat) domF) = [[(% x . (TIn, x)) ? k -> (TOut, k) -> $spec]]Ff [[PN]]Ffix"}
First, we give the name of the process (spec), and then apply a fixpoint to all process names
(PN; to handle possible mutual recursion).
After the ``='' sign, we see the almost CSP like syntax.
In CSP syntax, the process would look like the following:
$Spec = in?k \rightarrow out!k \rightarrow Spec$

The sufficient property is defined via its two disjuncts, F-empty and F-full, which are named
F-even and F-odd in the Isabelle/HOL formalization. The sufficient property is caled ``connecting property'' or ``conn''.
% @{const concurrent_buffer_example.SingleBuffer.TFeven}
% @{const concurrent_buffer_example.SingleBuffer.TFodd}

The sufficient property implies the CSP specification
@{thm[display] concurrent_buffer_example.SingleBuffer.Conn_Imp_Spec}

The sufficient property holds for all stable failures of the CUC program:
@{thm[display] concurrent_buffer_example.SingleBuffer.TSeq123_Imp_Conn}

The conclusion that the CUC program refines the CSP process
@{thm[display] concurrent_buffer_example.SingleBuffer.Spec_SF_Refinement_Impl}

We have proven the properties about the single buffer inside a locale @{locale concurrent_buffer_example.SingleBuffer}.
We can now instantiate this locale for each homogeneous component to reason about their parallel composition. 
The refinement of the concurrent versions is shown in @{thm[display] concurrent_buffer_example.Buffer_SF_Parallel_refinement}.

*}

(*<*)
end --"theory"
(*>*)