section {* Parallel Buffer example *}

text {* First, a locale will be introduced, where all theorems for the sequential code are contained.
Afterwards, via the composeable refinement property, we can show, that the parrallel compositions
also form a refinement.  *}

theory concurrent_buffer_example
imports stable_failures_parallel hoare cuc_csp_refinement
begin
  
text {* First, we introduce the buffer's local storage, and the event conversion function.
The event conversion function should be the same for all components.
*}
  
record R = free :: bool buffer :: nat
definition "EvtCV = id"

subsection {* Sequential Buffer @{text locale} *}

text {* The arguments for this locale are the channels for input and output. They have to be unequal. *}

locale SingleBuffer = fixes TIn :: nat and TOut :: nat assumes TInNeqTOut: "TIn \<noteq> TOut"
begin
  
paragraph {* Definition of the code *}
  
definition "\<Sigma> \<equiv> {(T, (x::nat)) . T = TIn \<or> T = TOut }"

definition "TDoF \<equiv> (\<lambda>(s::R). {s\<lparr>free := True\<rparr>})"
definition "TLine1 \<equiv> 1::do TDoF"
definition "TEf \<equiv> (\<lambda>s. {(I,(x::nat)) . free s = True \<and> I = TIn} \<union> {(out, (x::nat)) . free s = False \<and> x = buffer s \<and> out = TOut})"
definition "TF \<equiv> (\<lambda>s (event :: nat \<times> nat). 
  if (fst event = TIn) then (s\<lparr> buffer := snd event, free := False\<rparr>) else
  (s\<lparr> free := True \<rparr>))"
definition "TLine2 \<equiv> 2::comm TEf TF"
definition "TLine3 \<equiv> 3::cbr (\<lambda> s. True) 2 2"
definition "TSeq23 \<equiv> TLine2 \<oplus> TLine3"
definition "TSeq123 \<equiv> TLine1 \<oplus> TSeq23"
definition "LIS = ({(1, do TDoF), (2, comm TEf TF), (3, cbr (\<lambda> s . True) 2 2)} :: (R, nat \<times> nat) labeled_instruction_set)"
lemmas whole_code_unfold = TDoF_def TLine1_def TEf_def TF_def TLine2_def TLine3_def TSeq23_def TSeq123_def LIS_def
text {* This program fulfills the preconditions of the SFsemSigma locale. *}

interpretation Impl: SFsemSigma LIS "\<Sigma>"
  unfolding LIS_def
proof qed auto

text {* @{term fhelper} takes a list of @{term nat}s and converts it into an enriched list, where
each original element @{term x} is now represented by two successive tuples in the form of 
@{term "[(TIn, x), (TOut, x)]"}. *}

primrec fhelper :: "nat list \<Rightarrow> (nat \<times> nat) list" where
  "fhelper [] = []" |
  "fhelper (x # xs) = [(TIn, x), (TOut, x)] @ fhelper xs"

lemma fhelper_unfold_r:
  "(fhelper ns) @ [(TIn, x), (TOut, x)] = fhelper (ns @ [x])"
proof (induct ns, auto) qed

paragraph {* Definition of the connecting property sets \emph{even} *}

text {* For SF semantics, even length. *}
definition "TFeven \<equiv> {(c, X) .
  c \<in> { xs . \<exists> nats . xs = fhelper nats } \<and>
  (X \<subseteq> (\<Sigma> - {(tin, (x::nat)) . tin = TIn}))
}"
text {* For T semantics, even length. *}
definition "TTReven \<equiv> { xs . \<exists> nats . xs = fhelper nats }"

lemma eveneq: 
  "\<exists> X .((c, X) \<in> TFeven) = (c \<in> TTReven)"
  unfolding TFeven_def TTReven_def
  by auto

lemma evenimp: 
  "((c, X) \<in> TFeven) \<Longrightarrow> (c \<in> TTReven)"
  unfolding TFeven_def TTReven_def
  by auto

text {* The last element in an even trace is an output (sanity check). *}

lemma evenlastout:
  assumes a: "s @ [e] \<in> TTReven"
  shows "fst e = TOut"
proof  -
  obtain ns :: "nat list" where
    ao1: "s @ [e] = fhelper ns" using a TTReven_def mem_Collect_eq 
    by auto
  hence a1: "length ns > 0"
    using Nil_is_append_conv fhelper.simps(1) length_greater_0_conv not_Cons_self2
    by auto
  obtain ns' :: "nat list" and n :: nat where
    ao2: "ns = ns' @ [n]"
    using a1
    by (metis append_butlast_last_id length_greater_0_conv)
  with ao1 fhelper_unfold_r[of ns' n] have "fhelper ns' @ [(TIn, n), (TOut, n)] = s@[e]"
    by simp
  hence "\<exists> y . e = (TOut, y)"
    by auto
  thus "fst e = TOut"
    by auto
qed

text {* The empty trace is even. *}

lemma emptyeven: "X \<subseteq> (\<Sigma> - {(tin, (x::nat)) . tin = TIn}) \<Longrightarrow> ([], X) \<in> TFeven"
  unfolding TFeven_def
  by (metis (mono_tags, lifting) case_prodI fhelper.simps(1) mem_Collect_eq)

paragraph {* Definition of the connecting property sets \emph{odd} *}

text {* For SF semantics, odd length. *}
definition "TFodd \<equiv> {(c, X) . \<exists> y .
  c \<in> { xs . \<exists> nats . xs = (fhelper nats) @ [(TIn, y)]} \<and>
  (X \<subseteq> (\<Sigma> - {(tout, y') . y' = y \<and> tout = TOut}))
}"
text {* For T semantics, odd length. *}
definition "TTRodd \<equiv> { xs . \<exists> y nats . xs = (fhelper nats) @ [(TIn, y)]}"
lemma oddeq: 
  "\<exists> X .((c, X) \<in> TFodd) = (c \<in> TTRodd)"
  unfolding TFodd_def TTRodd_def
  by auto

lemma oddimp: 
  "((c, X) \<in> TFodd) \<Longrightarrow> (c \<in> TTRodd)"
  unfolding TFodd_def TTRodd_def
  by auto

text {* The last element in an odd trace is an input (sanity check). *}

lemma oddlastin:
  assumes a: "s @ [e] \<in> TTRodd"
  shows "fst e = TIn" using a
proof -
  have "\<exists>n ns. s @ [e] = fhelper ns @ [(TIn, n)]"
    using TTRodd_def assms by blast
  then show ?thesis
    by fastforce
qed

paragraph {* Connecting the odd and even sets *}

lemma evenappend:
  assumes a: "s \<in> TTReven"
  assumes b: "fst e = TIn"
  shows "s @ [e] \<in> TTRodd"
using a
unfolding TTReven_def TTRodd_def
proof
  from a obtain ns :: "nat list" where
    "s = fhelper (ns)" unfolding TTReven_def
    by force
  then have "\<exists>n ns. s @ [e] = fhelper ns @ [(TIn, n)]"
    by (metis (no_types) b eq_fst_iff)
  then show "s @ [e] \<in> {ps. \<exists>n ns. ps = fhelper ns @ [(TIn, n)]}"
    by blast
qed

lemma oddappend:
  assumes a: "s \<in> TTRodd"
  assumes b: "fst e = TOut"
  assumes c: "snd (last s) = snd e"
  shows "s @ [e] \<in> TTReven"
unfolding TTReven_def TTRodd_def
proof -
  from a obtain ns :: "nat list" and y :: nat where
    ao1: "s = fhelper (ns) @ [(TIn, y)]" unfolding TTRodd_def
    by force
  from ao1 b c have a1: "e = (TOut, y)"
    by (metis fst_conv last_snoc snd_conv surj_pair)
  with ao1 have "(s @ [e]) = (fhelper (ns) @ [(TIn, y), e])" by auto
  with a1 fhelper_unfold_r have "(s@[e]) = fhelper (ns @ [y])" by auto
  thus "s @ [e] \<in> {xs. \<exists>nats. xs = fhelper nats}" by auto
qed

lemma oddDblTail:
  assumes "(a # b # tr, X) \<in> TFodd"
  shows "(tr, X) \<in> TFodd"
proof -
  obtain nats y where
    abtr: "(a # b # tr) = fhelper nats @ [(TIn, y)]"
    using assms unfolding TFodd_def
    by blast
  have natsneq: "nats \<noteq> []"
    using abtr by auto
  then obtain n nats' where nats: "nats = n # nats'"
    using list.exhaust by blast
  hence "tr = fhelper nats' @ [(TIn, y)]"
    using abtr[unfolded nats fhelper.simps]
    by auto
  thus ?thesis
    using assms
    unfolding TFodd_def
    by fastforce
qed

lemma evenDblTail:
  assumes "(a # b # tr, X) \<in> TFeven"
  shows "(tr, X) \<in> TFeven"
proof -
  obtain nats where
    abtr: "(a # b # tr) = fhelper nats"
    using assms unfolding TFeven_def
    by blast
  have natsneq: "nats \<noteq> []"
    using abtr by auto
  then obtain n nats' where nats: "nats = n # nats'"
    using list.exhaust by blast
  hence "tr = fhelper nats'"
    using abtr[unfolded nats fhelper.simps]
    by auto
  thus ?thesis
    using assms
    unfolding TFeven_def
    by fast
qed

lemma first_two_are_in_out:
  assumes "(a # b # tr, X) \<in> TFeven \<or> (a # b # tr, X) \<in> TFodd"
  shows "\<exists> k . a = (TIn, k) \<and> b = (TOut, k)"
proof -
  obtain nats y where
    tr: "(a # b # tr) = fhelper nats \<or> (a # b # tr) = fhelper nats @ [(TIn, y)]"
    using assms unfolding TFeven_def TFodd_def
    by blast
  hence natsneq: "nats \<noteq> []" by auto
  then obtain k nats' where nats: "nats = k # nats'"
    using list.exhaust by blast
  show ?thesis
    using tr[unfolded nats fhelper.simps]
    by auto
qed

paragraph {* Definition of the starting and connecting property and invariants. *}

definition "TPre s \<equiv> NCPC (fst (snd s)) = 1 \<and> (fst s) = [] \<and> N (fst (snd s))"
lemma TPreEqPreStdSFseq: "PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q s = TPre s"
  unfolding PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q_def TPre_def by auto
definition "TI23 s \<equiv> (NCPC (fst (snd s)) = 2 \<or> NCPC (fst (snd s)) = 3) \<and> 
                    (((fst s, (snd (snd s))) \<in> TFeven \<and> free (NCRS (fst (snd s))) = True) \<or>
                      ((fst s, (snd (snd s))) \<in> TFodd \<and> free (NCRS (fst (snd s))) = False \<and>
                        (\<exists> x . buffer (NCRS (fst (snd s))) = x \<and> last (fst s) = (TIn, x))))"
definition "TInv s \<equiv> TPre s \<or> TI23 s"
definition "Conn x = (x \<in> TFeven \<or> x \<in> TFodd)"

definition "NSORB s \<equiv> (N (fst (snd s)) \<and> ((NCPC (fst (snd s))) \<in> {2,3}) \<and> 
                      ((free (NCRS (fst (snd s))) = True \<and> fst s \<in> TTReven) \<or> 
                       (free (NCRS (fst (snd s))) = False \<and> fst s \<in> TTRodd \<and> (\<exists> x . buffer (NCRS (fst (snd s))) = x \<and> last (fst s) = (TIn, x)))))"
definition "L1Q \<equiv> (\<lambda> s . TPre s \<or> TI23 s \<or> NSORB s) "
definition "L1G s \<equiv> N (fst (snd s)) \<and> NCPC (fst (snd s)) = 1"
definition "L1P1 s \<equiv> \<not> L1G s \<longrightarrow> L1Q s"
definition "L1P1inner s \<equiv> \<forall> (u :: R NCstate) . (N u \<and> NCPC u = Suc 1 \<and> NCRS u \<in> TDoF (NCRS (fst (snd s)))) \<longrightarrow>
              (\<forall> Y \<subseteq> \<Sigma> . L1Q (fst s, u, Y))"
definition "L1P2 s \<equiv> L1G s \<longrightarrow> L1P1inner s"
definition "L1P \<equiv> (\<lambda> s . L1P1 s \<and> L1P2 s)"
definition "L1ConsP2 \<equiv> (\<lambda> s . (L1Q s) \<and> NCPC (fst (snd s)) = 1 \<and> N (fst (snd s)))"
definition "L2Q \<equiv> (\<lambda> s . TI23 s \<or> NSORB s) "
definition "L2G s \<equiv> N (fst (snd s)) \<and> NCPC (fst (snd s)) = 2"
definition "L2P1 s \<equiv> \<not> L2G s \<longrightarrow> L2Q s"
definition "L2P1inner s \<equiv> ((\<forall> Y \<subseteq> \<Sigma> - TEf (NCRS (fst (snd s))) . L2Q(fst s, NCC (fst (snd s)), Y)) \<and> 
              (\<forall> (u :: R NCstate) e . N u \<and> e \<in> TEf (NCRS (fst (snd s))) \<and> NCPC u = Suc 2 \<and> NCRS u = TF (NCRS (fst (snd s))) e
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . L2Q((fst s)@[e], u, Y))))"
definition "L2P2 s \<equiv> L2G s \<longrightarrow> L2P1inner s"
definition "L2P \<equiv> (\<lambda> s . L2P1 s \<and> L2P2 s)"
definition "L2ConsP2 \<equiv> (\<lambda> s . (L2Q s) \<and> NCPC (fst (snd s)) = 2 \<and> N (fst (snd s)))"
definition "L3Q \<equiv> (\<lambda> s . TI23 s \<or> NSORB s) "
definition "L3G s \<equiv> N (fst (snd s)) \<and> NCPC (fst (snd s)) = 3"
definition "L3P1 s \<equiv> \<not> L3G s \<longrightarrow> L3Q s"
definition "L3P1inner s \<equiv> (\<forall> u. N u \<and> NCRS (fst (snd s)) = NCRS u \<and>
            (((\<lambda> s. True) (NCRS (fst (snd s))) \<and> NCPC u = 2) \<or> (\<not> (\<lambda> s. True) (NCRS (fst (snd s))) \<and> NCPC u = 2))
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . L3Q(fst s, u, Y)))"
definition "L3P2 s \<equiv> L3G s \<longrightarrow> L3P1inner s"
definition "L3P \<equiv> (\<lambda> s . L3P1 s \<and> L3P2 s)"
definition "L3ConsP2 \<equiv> (\<lambda> s . (L3Q s) \<and> NCPC (fst (snd s)) = 3 \<and> N (fst (snd s)))"

paragraph {* Proving the invariants *}

lemma hdohelphelp1: "(\<lambda>(tr, t, X).
         (\<not> (N t \<and> NCPC t = l) \<longrightarrow> Q (tr, t, X)) \<and>
         (N t \<and> NCPC t = l \<longrightarrow>
          (\<forall>u. N u \<and>
               NCPC u = Suc l \<and> NCRS u \<in> f (NCRS t) \<longrightarrow>
               (\<forall>Y\<subseteq>\<Sigma>. Q (tr, u, Y))))) = (\<lambda>s.
         (\<not> (N (fst (snd s)) \<and> NCPC (fst (snd s)) = l) \<longrightarrow> Q (fst s, (fst (snd s)), snd (snd s))) \<and>
         (N (fst (snd s)) \<and> NCPC (fst (snd s)) = l \<longrightarrow>
          (\<forall>u. N u \<and>
               NCPC u = Suc l \<and> NCRS u \<in> f (NCRS (fst (snd s))) \<longrightarrow>
               (\<forall>Y\<subseteq>\<Sigma>. Q (fst s, u, Y)))))" by (simp add: case_prod_unfold)

lemma hdohelp2: "(Impl.hoare
     (\<lambda>(tr, t, X).
         (\<not> (N t \<and> NCPC t = l) \<longrightarrow> Q (tr, t, X)) \<and>
         (N t \<and> NCPC t = l \<longrightarrow>
          (\<forall>u. N u \<and>
               NCPC u = Suc l \<and> NCRS u \<in> f (NCRS t) \<longrightarrow>
               (\<forall>Y\<subseteq>\<Sigma>. Q (tr, u, Y)))))
     (l :: do f) Q) \<Longrightarrow> (Impl.hoare
     (\<lambda>s.
         (\<not> (N (fst (snd s)) \<and> NCPC (fst (snd s)) = l) \<longrightarrow> Q s) \<and>
         (N (fst (snd s)) \<and> NCPC (fst (snd s)) = l \<longrightarrow>
          (\<forall>u. N u \<and>
               NCPC u = Suc l \<and> NCRS u \<in> f (NCRS (fst (snd s))) \<longrightarrow>
               (\<forall>Y\<subseteq>\<Sigma>. Q (fst s, u, Y)))))
     (l :: do f) Q)" using hdohelphelp1[of l Q f] by auto

lemma hcommhelp1: "(\<lambda> (tr, t, X) .(\<not>(N t \<and> NCPC t = l) \<longrightarrow> Q(tr, t, X)) \<and> 
              (N t \<and> NCPC t = l \<longrightarrow> ((\<forall> Y \<subseteq> \<Sigma> - ef (NCRS t) . Q(tr, NCC t, Y)) \<and> 
              (\<forall> u e . N u \<and> e \<in> ef (NCRS t) \<and> NCPC u = Suc l \<and> NCRS u = f2 (NCRS t) e
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q(tr@[e], u, Y)))))) =
            (\<lambda> s .(\<not>(N (fst (snd s)) \<and> NCPC (fst (snd s)) = l) \<longrightarrow> Q s) \<and> 
              (N (fst (snd s)) \<and> NCPC (fst (snd s)) = l \<longrightarrow> ((\<forall> Y \<subseteq> \<Sigma> - ef (NCRS (fst (snd s))) . Q(fst s, NCC (fst (snd s)), Y)) \<and> 
              (\<forall> u e . N u \<and> e \<in> ef (NCRS (fst (snd s))) \<and> NCPC u = Suc l \<and> NCRS u = f2 (NCRS (fst (snd s))) e
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q((fst s)@[e], u, Y))))))" by (simp add: case_prod_unfold)

lemma hcommhelp2: "(Impl.hoare (\<lambda> (tr, t, X) .(\<not>(N t \<and> NCPC t = l) \<longrightarrow> Q(tr, t, X)) \<and> 
              (N t \<and> NCPC t = l \<longrightarrow> ((\<forall> Y \<subseteq> \<Sigma> - ef (NCRS t) . Q(tr, NCC t, Y)) \<and> 
              (\<forall> u e . N u \<and> e \<in> ef (NCRS t) \<and> NCPC u = Suc l \<and> NCRS u = f2 (NCRS t) e
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q(tr@[e], u, Y)))))) (one l (comm ef f2)) Q) =
            (Impl.hoare (\<lambda> s .(\<not>(N (fst (snd s)) \<and> NCPC (fst (snd s)) = l) \<longrightarrow> Q s) \<and> 
              (N (fst (snd s)) \<and> NCPC (fst (snd s)) = l \<longrightarrow> ((\<forall> Y \<subseteq> \<Sigma> - ef (NCRS (fst (snd s))) . Q(fst s, NCC (fst (snd s)), Y)) \<and> 
              (\<forall> u e . N u \<and> e \<in> ef (NCRS (fst (snd s))) \<and> NCPC u = Suc l \<and> NCRS u = f2 (NCRS (fst (snd s))) e
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q((fst s)@[e], u, Y)))))) (one l (comm ef f2)) Q)" using hcommhelp1[of l Q ef f2] by simp

lemma hcbrhelp1: "(\<lambda> (tr, t, X) .(\<not>(N t \<and> NCPC t = l) \<longrightarrow> Q(tr, t, X))
                      \<and>
                      ((N t \<and> NCPC t = l) \<longrightarrow> (\<forall> u. N u \<and> NCRS t = NCRS u \<and>
            ((b (NCRS t) \<and> NCPC u = m) \<or> (\<not> b (NCRS t) \<and> NCPC u = n))
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q(tr, u, Y))))) =
(\<lambda> s .(\<not>(N (fst (snd s)) \<and> NCPC (fst (snd s)) = l) \<longrightarrow> Q(fst s, (fst (snd s)), snd (snd s)))
                      \<and>
                      ((N (fst (snd s)) \<and> NCPC (fst (snd s)) = l) \<longrightarrow> (\<forall> u. N u \<and> NCRS (fst (snd s)) = NCRS u \<and>
            ((b (NCRS (fst (snd s))) \<and> NCPC u = m) \<or> (\<not> b (NCRS (fst (snd s))) \<and> NCPC u = n))
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q(fst s, u, Y)))))" by (simp add: case_prod_unfold)

lemma hcbrhelp2: "(Impl.hoare (\<lambda> (tr, t, X) .(\<not>(N t \<and> NCPC t = l) \<longrightarrow> Q(tr, t, X))
                      \<and>
                      ((N t \<and> NCPC t = l) \<longrightarrow> (\<forall> u. N u \<and> NCRS t = NCRS u \<and>
            ((b (NCRS t) \<and> NCPC u = m) \<or> (\<not> b (NCRS t) \<and> NCPC u = n))
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q(tr, u, Y))))) (one l (cbr b m n)) Q) =
(Impl.hoare (\<lambda> s .(\<not>(N (fst (snd s)) \<and> NCPC (fst (snd s)) = l) \<longrightarrow> Q(fst s, (fst (snd s)), snd (snd s)))
                      \<and>
                      ((N (fst (snd s)) \<and> NCPC (fst (snd s)) = l) \<longrightarrow> (\<forall> u. N u \<and> NCRS (fst (snd s)) = NCRS u \<and>
            ((b (NCRS (fst (snd s))) \<and> NCPC u = m) \<or> (\<not> b (NCRS (fst (snd s))) \<and> NCPC u = n))
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q(fst s, u, Y))))) (one l (cbr b m n)) Q)" using hcbrhelp1[of l Q b m n]
by simp

lemma l1c: "L1ConsP2 s \<Longrightarrow> L1P s"
  unfolding L1P_def L1ConsP2_def TPre_def L1P1_def L1P2_def L1Q_def
proof
 assume "(NCPC (fst (snd s)) = 1 \<and> fst s = [] \<and> N (fst (snd s)) \<or> TI23 s \<or> NSORB s) \<and> NCPC (fst (snd s)) = 1 \<and> N (fst (snd s))"
 thus "\<not> L1G s \<longrightarrow> NCPC (fst (snd s)) = 1 \<and> fst s = [] \<and> N (fst (snd s)) \<or> TI23 s \<or> NSORB s"
  by auto
next
  assume aa1: "(NCPC (fst (snd s)) = 1 \<and> fst s = [] \<and> N (fst (snd s)) \<or> TI23 s \<or> NSORB s) \<and> NCPC (fst (snd s)) = 1 \<and> N (fst (snd s))"
  from aa1 have a1: "fst s = [] \<and> NCPC (fst (snd s)) = 1 \<and> N (fst (snd s))"
    unfolding TI23_def NSORB_def by auto
  have "L1P1inner s" unfolding L1P1inner_def L1Q_def TTReven_def NSORB_def 
  proof (rule allI, rule impI, rule allI, rule impI, rule disjI2, rule disjI2)
    fix t Y
    assume
      ba1: "N t \<and>
            NCPC t = Suc 1 \<and>
            NCRS t \<in> TDoF (NCRS (fst (snd s)))" and
      ba2: "Y \<subseteq> \<Sigma>"
    from ba1 have "N (fst (snd (fst s, t, Y)))"
      by auto
    moreover
    from ba1 have "NCPC (fst (snd (fst s, t, Y))) = 2"
      by auto
    moreover
    from ba1 have "free (NCRS (fst (snd (fst s, t, Y)))) = True"
      unfolding TDoF_def
      by auto
    moreover
    from a1 have "fst (fst s, t, Y) \<in> {xs. \<exists>nats. xs = fhelper nats}"
      using emptyeven TTReven_def evenimp fst_conv subset_eq
      by (metis (no_types, lifting))
    ultimately
    show "N (fst (snd (fst s, t, Y))) \<and>
       NCPC (fst (snd (fst s, t, Y))) \<in> {2, 3} \<and>
       (free (NCRS (fst (snd (fst s, t, Y)))) = True \<and>
        fst (fst s, t, Y)
        \<in> {xs. \<exists>nats. xs = fhelper nats} \<or>
        free (NCRS (fst (snd (fst s, t, Y)))) = False \<and>
        fst (fst s, t, Y) \<in> TTRodd \<and>
        (\<exists>x. buffer (NCRS (fst (snd (fst s, t, Y)))) = x \<and>
             last (fst (fst s, t, Y)) = (TIn, x)))"
      by auto
  qed
  thus "L1G s \<longrightarrow> L1P1inner s" by simp
qed

lemma l2c: "L2ConsP2 s \<Longrightarrow> L2P s"
  unfolding L2P_def L2ConsP2_def TPre_def L2P1_def L2P2_def L2Q_def
proof
  assume aa1: "(TI23 s \<or> NSORB s) \<and> NCPC (fst (snd s)) = 2 \<and> N (fst (snd s))"
  thus "\<not> L2G s \<longrightarrow> TI23 s \<or> NSORB s"
    by auto
next
  assume aa1: "(TI23 s \<or> NSORB s) \<and> NCPC (fst (snd s)) = 2 \<and> N (fst (snd s))"
  from aa1 have a1: "L2G s" unfolding L2G_def
    by simp
  from aa1 have a2: "(NCPC (fst (snd s)) = 2 \<and> N (fst (snd s)) \<and> (fst s, snd (snd s)) \<in> TFeven \<and>
    free (NCRS (fst (snd s))) = True \<or>
    (fst s, snd (snd s)) \<in> TFodd \<and>
    free (NCRS (fst (snd s))) = False \<and>
    (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
         last (fst s) = (TIn,x))) \<or>
   ((free (NCRS (fst (snd s))) = True \<and> fst s \<in> TTReven \<or> free (NCRS (fst (snd s))) = False \<and> fst s \<in> TTRodd \<and>
    (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
         last (fst s) = (TIn, x))) \<and>
  NCPC (fst (snd s)) = 2 \<and> N (fst (snd s)))" (is "(?A s \<or> ?B s) \<or> ?C s") unfolding TI23_def NSORB_def
    by auto

  hence "L2P1inner s" unfolding L2P1inner_def
  proof
    assume ba1: "?A s \<or> ?B s"
    hence "(\<forall>Y\<subseteq>\<Sigma> - TEf (NCRS (fst (snd s))).
        L2Q (fst s, NCC (fst (snd s)), Y))"
    proof
      assume ca1: "?A s"
      show "\<forall>Y\<subseteq>\<Sigma> - TEf (NCRS (fst (snd s))). L2Q (fst s, NCC (fst (snd s)), Y)" unfolding TEf_def L2Q_def TI23_def
      proof (rule allI, rule impI, rule disjI1)
        fix Y
        assume da1: "Y \<subseteq> \<Sigma> -
             ({(I, x).
               free (NCRS (fst (snd s))) = True \<and>
               I = TIn} \<union>
              {(out, x).
               free (NCRS (fst (snd s))) = False \<and>
               x = buffer (NCRS (fst (snd s))) \<and>
               out = TOut})"
        have dh1: "fst (fst s, NCC (fst (snd s)), Y) = fst s"
          by simp
        have dh2: "Y \<subseteq> \<Sigma> - {(tin, x). tin = TIn}" using da1 ca1
          by auto
        have dh3: "snd (snd (fst s, NCC (fst (snd s)), Y)) = Y"
          by simp
        from aa1 have "NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 2" using "NCstate.exhaust" "NCC.simps" "NCPC.simps"
          by (metis (no_types, hide_lams) fst_conv snd_conv)
        moreover
        from ca1 have "free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = True"
          using NCRSh1 NCRSh2
          by metis
        moreover
        from ca1 have "(fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFeven"
          using dh1 dh2 dh3
          unfolding TFeven_def
          by simp
        ultimately
        show "(NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 2 \<or> NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 3) \<and> ((fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFeven \<and> free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = True \<or> (fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFodd \<and> free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = False \<and> (\<exists>x. buffer (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = x \<and> last (fst (fst s, NCC (fst (snd s)), Y)) = (TIn, x)))"
          by simp 
      qed
    next
      assume
        ca1: "(fst s, snd (snd s)) \<in> TFodd \<and>
               free (NCRS (fst (snd s))) = False \<and>
               (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and> last (fst s) = (TIn, x))"
      show "\<forall>Y\<subseteq>\<Sigma> - TEf (NCRS (fst (snd s))). L2Q (fst s, NCC (fst (snd s)), Y)" unfolding  TEf_def L2Q_def TI23_def
      proof (rule allI, rule impI, rule disjI1)
        fix Y
        assume da1: "Y \<subseteq> \<Sigma> -
             ({(I, x).
               free (NCRS (fst (snd s))) = True \<and>
               I = TIn} \<union>
              {(out, x).
               free (NCRS (fst (snd s))) = False \<and>
               x = buffer (NCRS (fst (snd s))) \<and>
               out = TOut})"
        obtain x :: nat where do1: "buffer (NCRS (fst (snd s))) = x" by auto
        
        have dh0: "{(out, x). x = buffer (NCRS (fst (snd s))) \<and> out = TOut} =
          {(out, x'). x' = x \<and> out = TOut}" using do1 by simp
        have dh1: "fst (fst s, NCC (fst (snd s)), Y) = fst s"
          by simp
        have dh2: "Y \<subseteq> \<Sigma> - {(out, x). x = buffer (NCRS (fst (snd s))) \<and> out = TOut}" using da1 ca1
          by auto
        hence dh22: "Y \<subseteq> \<Sigma> - {(out, x'). x' = x \<and> out = TOut}" using dh0
          by simp
        from ca1 have dh5: "\<exists>y. fst s \<in> {xs. \<exists>nats. xs = fhelper nats @ [(TIn, y)]}" unfolding TFodd_def
          by blast
        obtain y :: nat where do2: "fst s \<in> {xs. \<exists>nats. xs = fhelper nats @ [(TIn, y)]}" using dh5
          by blast
        have dh3: "snd (snd (fst s, NCC (fst (snd s)), Y)) = Y" by simp
        have dh4: "last (fst s) = (TIn, x)" using do1 ca1 by simp
        from do1 do2 dh4 have dh6: "y = x" using Pair_inject last_snoc mem_Collect_eq by auto
        with do2 have dh7: "fst s \<in> {xs. \<exists>nats. xs = fhelper nats @ [(TIn, x)]}"
          by simp
        from aa1 have "NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 2" using "NCstate.exhaust" "NCC.simps" "NCPC.simps"
          by (metis (no_types, hide_lams) fst_conv snd_conv)
        moreover
        from ca1 have "free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = False"
          using NCRSh1 NCRSh2
          by metis
        moreover
        have d1:"(fst s, Y) \<in> TFodd"
          using dh22 dh7
          unfolding TFodd_def
          using DiffE DiffI case_prodI case_prod_beta' mem_Collect_eq subsetCE subsetI by blast
        moreover
        from d1 have "(fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFodd"
          using dh1 dh2 dh3
          unfolding TFodd_def
          by (simp)
        moreover
        from ca1 have "(\<exists>x. buffer (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = x \<and>
               last (fst (fst s, NCC (fst (snd s)), Y)) = (TIn,x))"
          using dh1 dh3
          by (metis NCRSh2 fst_conv snd_conv)
        ultimately
        show "(NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 2 \<or> NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 3) \<and> ((fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFeven \<and> free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = True \<or> (fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFodd \<and> free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = False \<and> (\<exists>x. buffer (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = x \<and> last (fst (fst s, NCC (fst (snd s)), Y)) = (TIn, x)))"
          by simp
      qed
    qed
    moreover
    from ba1 have "(\<forall>u e. N u \<and> e \<in> TEf (NCRS (fst (snd s))) \<and> NCPC u = Suc 2 \<and> NCRS u = TF (NCRS (fst (snd s))) e \<longrightarrow> (\<forall>Y\<subseteq>\<Sigma>. L2Q (fst s @ [e], u, Y)))"
    proof 
      assume ca1: "NCPC (fst (snd s)) = 2 \<and>
              N (fst (snd s)) \<and>
              (fst s, snd (snd s)) \<in> TFeven \<and>
              free (NCRS (fst (snd s))) = True"
      show "\<forall>u e. N u \<and> e \<in> TEf (NCRS (fst (snd s))) \<and> NCPC u = Suc 2 \<and> NCRS u = TF (NCRS (fst (snd s))) e \<longrightarrow> (\<forall>Y\<subseteq>\<Sigma>. L2Q (fst s @ [e], u, Y))"
        unfolding L2Q_def NSORB_def
      proof (rule allI, rule allI, rule impI, rule allI, rule impI, rule disjI2)
        fix u e Y
        assume 
          da1: "N u \<and>
                e \<in> TEf (NCRS (fst (snd s))) \<and>
                NCPC u = Suc 2 \<and>
                NCRS u = TF (NCRS (fst (snd s))) e" and
          da2: "Y \<subseteq> \<Sigma>"
        from da1 have "N (fst (snd (fst s @ [e], u, Y)))" by auto
        moreover
        from da1 have "NCPC (fst (snd (fst s @ [e], u, Y))) \<in> {1, 2, 3}" by auto
        moreover
        {
          from ca1 have "fst s \<in> TTReven" using evenimp by blast
          hence "fst s @ [e] \<in> TTRodd"
            using evenappend[of "fst s" e] da1 ca1 Product_Type.Collect_case_prodD Un_iff
            unfolding TEf_def
            by auto
        }
        moreover
        {
          have "fst e = TIn"
          proof -
            have "\<exists>n ns. fst s @ [e] = fhelper ns @ [(TIn, n)]"
              using TTRodd_def calculation(3) by blast
            then show ?thesis
              by force
          qed
        }
        moreover
        {
          with da1 have "NCRS (u) =  NCRS (fst (snd s))\<lparr>buffer := snd e, free := False\<rparr>"
            unfolding TF_def TEf_def
            by simp
          hence "free (NCRS (u)) = False"
            unfolding TF_def TEf_def
            by simp
          hence "free (NCRS (fst (snd (fst s @ [e], u, Y)))) = False"
            by simp
        }
        moreover
        {
          from da1 calculation(4) have "NCRS u = NCRS (fst (snd s))\<lparr>buffer := snd e, free := False\<rparr>"
            unfolding TF_def
            by simp
          hence "buffer (NCRS u) = snd e"
            by auto
          hence "\<exists> x . buffer (NCRS u) = x \<and> snd e = x"
            by simp
          with calculation(4) have "(\<exists>x. buffer (NCRS (fst (snd (fst s @ [e], u, Y)))) = x \<and> last (fst (fst s @ [e], u, Y)) = (TIn, x))"
            by (metis fst_conv last_snoc prod.collapse snd_conv)
        }
        ultimately
        show "N (fst (snd (fst s @ [e], u, Y))) \<and>
              NCPC (fst (snd (fst s @ [e], u, Y))) \<in> {2, 3} \<and>
              (free (NCRS (fst (snd (fst s @ [e], u, Y)))) = True \<and> fst (fst s @ [e], u, Y) \<in> TTReven \<or>
               free (NCRS (fst (snd (fst s @ [e], u, Y)))) = False \<and> fst (fst s @ [e], u, Y) \<in> TTRodd \<and>
        (\<exists>x. buffer (NCRS (fst (snd (fst s @ [e], u, Y)))) = x \<and> last (fst (fst s @ [e], u, Y)) = (TIn, x)))"
          by (simp add: da1)
      qed
    next
      assume ca1: "(fst s, snd (snd s)) \<in> TFodd \<and>
              free (NCRS (fst (snd s))) = False \<and>
              (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
              last (fst s) = (TIn, x))"
      show "\<forall>u e. N u \<and>
        e \<in> TEf (NCRS (fst (snd s))) \<and>
        NCPC u = Suc 2 \<and>
        NCRS u = TF (NCRS (fst (snd s))) e \<longrightarrow>
        (\<forall>Y\<subseteq>\<Sigma>. L2Q (fst s @ [e], u, Y))"
        unfolding L2Q_def NSORB_def
      proof (rule allI, rule allI, rule impI, rule allI, rule impI, rule disjI2)
        fix u e Y
        assume 
          da1: "N u \<and>
                e \<in> TEf (NCRS (fst (snd s))) \<and>
                NCPC u = Suc 2 \<and>
                NCRS u = TF (NCRS (fst (snd s))) e" and
          da2: "Y \<subseteq> \<Sigma>"
        from da1 have "N (fst (snd (fst s @ [e], u, Y)))" by auto
        moreover
        from da1 have "NCPC (fst (snd (fst s @ [e], u, Y))) \<in> {1, 2, 3}" by auto
        moreover
        {
          from ca1 have "fst s \<in> TTRodd" using oddimp by blast
          hence "fst s @ [e] \<in> TTReven"
            using oddappend[of "fst s" e] da1 ca1 Product_Type.Collect_case_prodD Un_iff
            unfolding TEf_def
            by auto
        }
        moreover
        {
          from calculation(3) evenlastout have "fst e = TOut"
            by simp
          hence "fst e \<noteq> TIn"
            using TInNeqTOut
            by (simp)
          with da1 have "NCRS (u) =  NCRS (fst (snd s))\<lparr>free := True\<rparr>"
            unfolding TF_def TEf_def
            by simp
          hence "free (NCRS (u)) = True"
            unfolding TF_def TEf_def
            by simp
          hence "free (NCRS (fst (snd (fst s @ [e], u, Y)))) = True"
            by simp
        }
        ultimately
        show "N (fst (snd (fst s @ [e], u, Y))) \<and>
              NCPC (fst (snd (fst s @ [e], u, Y))) \<in> {2, 3} \<and>
              (free (NCRS (fst (snd (fst s @ [e], u, Y)))) = True \<and> fst (fst s @ [e], u, Y) \<in> TTReven \<or>
               free (NCRS (fst (snd (fst s @ [e], u, Y)))) = False \<and> fst (fst s @ [e], u, Y) \<in> TTRodd \<and>
        (\<exists>x. buffer (NCRS (fst (snd (fst s @ [e], u, Y)))) = x \<and> last (fst (fst s @ [e], u, Y)) = (TIn, x)))"
          by (simp add: da1)
      qed
    qed
    ultimately
    show "(\<forall>Y\<subseteq>\<Sigma> - TEf (NCRS (fst (snd s))).
        L2Q (fst s, NCC (fst (snd s)), Y)) \<and>
    (\<forall>u e. N u \<and>
           e \<in> TEf (NCRS (fst (snd s))) \<and>
           NCPC u = Suc 2 \<and>
           NCRS u = TF (NCRS (fst (snd s))) e \<longrightarrow>
           (\<forall>Y\<subseteq>\<Sigma>. L2Q (fst s @ [e], u, Y)))" by auto
  next
    assume "(free (NCRS (fst (snd s))) = True \<and> fst s \<in> TTReven \<or> free (NCRS (fst (snd s))) = False \<and> fst s \<in> TTRodd \<and>
     (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
          last (fst s) = (TIn, x))) \<and> NCPC (fst (snd s)) = 2 \<and> N (fst (snd s))"
    hence b1:
      "(free (NCRS (fst (snd s))) = True \<and> fst s \<in> TTReven \<and> NCPC (fst (snd s)) = 2 \<and> N (fst (snd s))) \<or>
       (free (NCRS (fst (snd s))) = False \<and> fst s \<in> TTRodd  \<and>
     (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
          last (fst s) = (TIn, x)) \<and> NCPC (fst (snd s)) = 2 \<and> N (fst (snd s)))" 
      (is "?A s \<or> ?B s")
      by simp
    from b1 have "(\<forall>Y\<subseteq>\<Sigma> - TEf (NCRS (fst (snd s))). L2Q (fst s, NCC (fst (snd s)), Y))"
    proof
      assume ca1: "?A s"
      show "\<forall>Y\<subseteq>\<Sigma> - TEf (NCRS (fst (snd s))). L2Q (fst s, NCC (fst (snd s)), Y)" unfolding TEf_def L2Q_def TI23_def
      proof (rule allI, rule impI, rule disjI1)
        fix Y
        assume da1: "Y \<subseteq> \<Sigma> -
             ({(I, x).
               free (NCRS (fst (snd s))) = True \<and>
               I = TIn} \<union>
              {(out, x).
               free (NCRS (fst (snd s))) = False \<and>
               x = buffer (NCRS (fst (snd s))) \<and>
               out = TOut})"
        have dh1: "fst (fst s, NCC (fst (snd s)), Y) = fst s"
          by simp
        have dh2: "Y \<subseteq> \<Sigma> - {(tin, x). tin = TIn}" using da1 ca1
          by auto
        have dh3: "snd (snd (fst s, NCC (fst (snd s)), Y)) = Y"
          by simp
        from aa1 have "NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 2" using "NCstate.exhaust" "NCC.simps" "NCPC.simps"
          by (metis (no_types, hide_lams) fst_conv snd_conv)
        moreover
        from ca1 have "free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = True"
          using NCRSh1 NCRSh2
          by metis
        moreover
        {
        from ca1 have "(fst s, Y) \<in> TFeven"
          using dh1 dh2 dh3
          unfolding TFeven_def
          using CollectI TTReven_def case_prodI by auto
        hence "(fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFeven"
          by simp
        }
        ultimately
        show "(NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 2 \<or> NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 3) \<and> ((fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFeven \<and> free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = True \<or> (fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFodd \<and> free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = False \<and> (\<exists>x. buffer (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = x \<and> last (fst (fst s, NCC (fst (snd s)), Y)) = (TIn, x)))"
          by simp 
      qed
    next
      assume
        ca1: "free (NCRS (fst (snd s))) = False \<and>
              fst s \<in> TTRodd \<and>
              (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
                   last (fst s) = (TIn, x)) \<and>
              NCPC (fst (snd s)) = 2 \<and> N (fst (snd s))"
      show "\<forall>Y\<subseteq>\<Sigma> - TEf (NCRS (fst (snd s))). L2Q (fst s, NCC (fst (snd s)), Y)" unfolding  TEf_def L2Q_def TI23_def
      proof (rule allI, rule impI, rule disjI1)
        fix Y
        assume da1: "Y \<subseteq> \<Sigma> -
             ({(I, x).
               free (NCRS (fst (snd s))) = True \<and>
               I = TIn} \<union>
              {(out, x).
               free (NCRS (fst (snd s))) = False \<and>
               x = buffer (NCRS (fst (snd s))) \<and>
               out = TOut})"
        obtain x :: nat where do1: "buffer (NCRS (fst (snd s))) = x" by auto
        
        have dh0: "{(out, x). x = buffer (NCRS (fst (snd s))) \<and> out = TOut} =
          {(out, x'). x' = x \<and> out = TOut}" using do1 by simp
        have dh1: "fst (fst s, NCC (fst (snd s)), Y) = fst s"
          by simp
        have dh2: "Y \<subseteq> \<Sigma> - {(out, x). x = buffer (NCRS (fst (snd s))) \<and> out = TOut}" using da1 ca1
          by auto
        hence dh22: "Y \<subseteq> \<Sigma> - {(out, x'). x' = x \<and> out = TOut}" using dh0
          by simp
        from ca1 have dh5: "\<exists>y. fst s \<in> {xs. \<exists>nats. xs = fhelper nats @ [(TIn, y)]}" unfolding TTRodd_def
          by blast
        obtain y :: nat where do2: "fst s \<in> {xs. \<exists>nats. xs = fhelper nats @ [(TIn, y)]}" using dh5
          by blast
        have dh3: "snd (snd (fst s, NCC (fst (snd s)), Y)) = Y" by simp
        have dh4: "last (fst s) = (TIn, y)" using do2 by auto
        from do1 do2 dh4 ca1 have dh6: "y = x" using Pair_inject last_snoc mem_Collect_eq by auto
        with do2 have dh7: "fst s \<in> {xs. \<exists>nats. xs = fhelper nats @ [(TIn, x)]}"
          by simp
        from aa1 have "NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 2" using "NCstate.exhaust" "NCC.simps" "NCPC.simps"
          by (metis (no_types, hide_lams) fst_conv snd_conv)
        moreover
        from ca1 have "free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = False"
          using NCRSh1 NCRSh2
          by metis
        moreover
        have d1:"(fst s, Y) \<in> TFodd"
          using dh22 dh7
          unfolding TFodd_def
          using DiffE DiffI case_prodI case_prod_beta' mem_Collect_eq subsetCE subsetI by blast
        moreover
        from d1 have "(fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFodd"
          using dh1 dh2 dh3
          unfolding TFodd_def
          by (simp)
        moreover
        from ca1 have "(\<exists>x. buffer (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = x \<and>
               last (fst (fst s, NCC (fst (snd s)), Y)) = (TIn,x))"
          using dh1 dh3
          by (metis NCRSh2 fst_conv snd_conv)
        ultimately
        show "(NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 2 \<or> NCPC (fst (snd (fst s, NCC (fst (snd s)), Y))) = 3) \<and> ((fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFeven \<and> free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = True \<or> (fst (fst s, NCC (fst (snd s)), Y), snd (snd (fst s, NCC (fst (snd s)), Y))) \<in> TFodd \<and> free (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = False \<and> (\<exists>x. buffer (NCRS (fst (snd (fst s, NCC (fst (snd s)), Y)))) = x \<and> last (fst (fst s, NCC (fst (snd s)), Y)) = (TIn, x)))"
          by simp
      qed
    qed
    moreover
    from b1 have "(\<forall>u e. N u \<and> e \<in> TEf (NCRS (fst (snd s))) \<and> NCPC u = Suc 2 \<and> NCRS u = TF (NCRS (fst (snd s))) e \<longrightarrow> (\<forall>Y\<subseteq>\<Sigma>. L2Q (fst s @ [e], u, Y)))"
    proof 
      assume ca1: "?A s"
      show "\<forall>u e. N u \<and> e \<in> TEf (NCRS (fst (snd s))) \<and> NCPC u = Suc 2 \<and> NCRS u = TF (NCRS (fst (snd s))) e \<longrightarrow> (\<forall>Y\<subseteq>\<Sigma>. L2Q (fst s @ [e], u, Y))"
        unfolding L2Q_def NSORB_def
      proof (rule allI, rule allI, rule impI, rule allI, rule impI, rule disjI2)
        fix u e Y
        assume 
          da1: "N u \<and>
                e \<in> TEf (NCRS (fst (snd s))) \<and>
                NCPC u = Suc 2 \<and>
                NCRS u = TF (NCRS (fst (snd s))) e" and
          da2: "Y \<subseteq> \<Sigma>"
        from da1 have "N (fst (snd (fst s @ [e], u, Y)))" by auto
        moreover
        from da1 have "NCPC (fst (snd (fst s @ [e], u, Y))) \<in> {1, 2, 3}" by auto
        moreover
        {
          from ca1 have "fst s \<in> TTReven" using evenimp by blast
          hence "fst s @ [e] \<in> TTRodd"
            using evenappend[of "fst s" e] da1 ca1 Product_Type.Collect_case_prodD Un_iff
            unfolding TEf_def
            by auto
        }
        moreover
        {
          have "fst e = TIn"
          proof -
            have "\<exists>n ns. fst s @ [e] = fhelper ns @ [(TIn, n)]"
              using TTRodd_def calculation(3) by blast
            then show ?thesis
              by force
          qed
        }
        moreover
        {
          with da1 have "NCRS (u) =  NCRS (fst (snd s))\<lparr>buffer := snd e, free := False\<rparr>"
            unfolding TF_def TEf_def
            by simp
          hence "free (NCRS (u)) = False"
            unfolding TF_def TEf_def
            by simp
          hence "free (NCRS (fst (snd (fst s @ [e], u, Y)))) = False"
            by simp
        }
        moreover
        {
          from da1 calculation(4) have "NCRS u = NCRS (fst (snd s))\<lparr>buffer := snd e, free := False\<rparr>"
            unfolding TF_def
            by simp
          hence "buffer (NCRS u) = snd e"
            by auto
          hence "\<exists> x . buffer (NCRS u) = x \<and> snd e = x"
            by simp
          with calculation(4) have "(\<exists>x. buffer (NCRS (fst (snd (fst s @ [e], u, Y)))) = x \<and> last (fst (fst s @ [e], u, Y)) = (TIn, x))"
            by (metis fst_conv last_snoc prod.collapse snd_conv)
        }
        ultimately
        show "N (fst (snd (fst s @ [e], u, Y))) \<and>
              NCPC (fst (snd (fst s @ [e], u, Y))) \<in> {2, 3} \<and>
              (free (NCRS (fst (snd (fst s @ [e], u, Y)))) = True \<and> fst (fst s @ [e], u, Y) \<in> TTReven \<or>
               free (NCRS (fst (snd (fst s @ [e], u, Y)))) = False \<and> fst (fst s @ [e], u, Y) \<in> TTRodd \<and>
        (\<exists>x. buffer (NCRS (fst (snd (fst s @ [e], u, Y)))) = x \<and> last (fst (fst s @ [e], u, Y)) = (TIn, x)))"
          by (simp add: da1)
      qed
    next
      assume ca1: "?B s"
      show "\<forall>u e. N u \<and>
        e \<in> TEf (NCRS (fst (snd s))) \<and>
        NCPC u = Suc 2 \<and>
        NCRS u = TF (NCRS (fst (snd s))) e \<longrightarrow>
        (\<forall>Y\<subseteq>\<Sigma>. L2Q (fst s @ [e], u, Y))"
        unfolding L2Q_def NSORB_def
      proof (rule allI, rule allI, rule impI, rule allI, rule impI, rule disjI2)
        fix u e Y
        assume 
          da1: "N u \<and>
                e \<in> TEf (NCRS (fst (snd s))) \<and>
                NCPC u = Suc 2 \<and>
                NCRS u = TF (NCRS (fst (snd s))) e" and
          da2: "Y \<subseteq> \<Sigma>"
        from da1 have "N (fst (snd (fst s @ [e], u, Y)))" by auto
        moreover
        from da1 have "NCPC (fst (snd (fst s @ [e], u, Y))) \<in> {1, 2, 3}" by auto
        moreover
        {
          from ca1 have "fst s \<in> TTRodd" using oddimp by blast
          hence "fst s @ [e] \<in> TTReven"
            using oddappend[of "fst s" e] da1 ca1 Product_Type.Collect_case_prodD Un_iff
            unfolding TEf_def
            by auto
        }
        moreover
        {
          from calculation(3) evenlastout have "fst e = TOut"
            by simp
          hence "fst e \<noteq> TIn"
            using TInNeqTOut
            by (simp)
          with da1 have "NCRS (u) =  NCRS (fst (snd s))\<lparr>free := True\<rparr>"
            unfolding TF_def TEf_def
            by simp
          hence "free (NCRS (u)) = True"
            unfolding TF_def TEf_def
            by simp
          hence "free (NCRS (fst (snd (fst s @ [e], u, Y)))) = True"
            by simp
        }
        ultimately
        show "N (fst (snd (fst s @ [e], u, Y))) \<and>
              NCPC (fst (snd (fst s @ [e], u, Y))) \<in> {2, 3} \<and>
              (free (NCRS (fst (snd (fst s @ [e], u, Y)))) = True \<and> fst (fst s @ [e], u, Y) \<in> TTReven \<or>
               free (NCRS (fst (snd (fst s @ [e], u, Y)))) = False \<and> fst (fst s @ [e], u, Y) \<in> TTRodd \<and>
        (\<exists>x. buffer (NCRS (fst (snd (fst s @ [e], u, Y)))) = x \<and> last (fst (fst s @ [e], u, Y)) = (TIn, x)))"
          by (simp add: da1)
      qed
    qed
    ultimately
    show "(\<forall>Y\<subseteq>\<Sigma> - TEf (NCRS (fst (snd s))).
        L2Q (fst s, NCC (fst (snd s)), Y)) \<and>
    (\<forall>u e. N u \<and>
           e \<in> TEf (NCRS (fst (snd s))) \<and>
           NCPC u = Suc 2 \<and>
           NCRS u = TF (NCRS (fst (snd s))) e \<longrightarrow>
           (\<forall>Y\<subseteq>\<Sigma>. L2Q (fst s @ [e], u, Y)))" by simp
  qed
  with a1 show  "L2G s \<longrightarrow> L2P1inner s" by simp
qed

lemma l3c: "L3ConsP2 s \<Longrightarrow> L3P s"
  unfolding L3P_def L3ConsP2_def TPre_def L3P1_def L3P2_def L3Q_def
proof
  show "(TI23 s \<or> NSORB s) \<and> NCPC (fst (snd s)) = 3 \<and> N (fst (snd s)) \<Longrightarrow> \<not> L3G s \<longrightarrow> TI23 s \<or> NSORB s"
    by simp
next
  assume aa1: "(TI23 s \<or> NSORB s) \<and> NCPC (fst (snd s)) = 3 \<and> N (fst (snd s))"
  hence "(TI23 s \<or> NSORB s)" by simp
  hence "L3P1inner s"
    unfolding TI23_def NSORB_def L3P1inner_def
  proof
    assume "((NCPC (fst (snd s)) = 2 \<or> NCPC (fst (snd s)) = 3) \<and>
     ((fst s, snd (snd s)) \<in> TFeven \<and>
      free (NCRS (fst (snd s))) = True \<or>
      (fst s, snd (snd s)) \<in> TFodd \<and>
      free (NCRS (fst (snd s))) = False \<and>
      (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
           last (fst s) = (TIn, x))))"
    hence b1: "(fst s, snd (snd s)) \<in> TFeven \<and>
      free (NCRS (fst (snd s))) = True \<or>
      (fst s, snd (snd s)) \<in> TFodd \<and>
      free (NCRS (fst (snd s))) = False \<and>
      (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
           last (fst s) = (TIn, x))" (is "?A s \<or> ?B s")
      by simp
    thus "\<forall>u. N u \<and> NCRS (fst (snd s)) = NCRS u \<and> (True \<and> NCPC u = 2 \<or> \<not> True \<and> NCPC u = 2) \<longrightarrow> (\<forall>Y\<subseteq>\<Sigma>. L3Q (fst s, u, Y))"
    proof
      assume ca1: "?A s"
      show "\<forall>u. N u \<and> NCRS (fst (snd s)) = NCRS u \<and> (True \<and> NCPC u = 2 \<or> \<not> True \<and> NCPC u = 2) \<longrightarrow> (\<forall>Y\<subseteq>\<Sigma>. L3Q (fst s, u, Y))"
        unfolding L3Q_def NSORB_def 
        using ca1 evenimp by auto
    next
      assume ca1: "?B s"
      show "\<forall>u. N u \<and> NCRS (fst (snd s)) = NCRS u \<and> (True \<and> NCPC u = 2 \<or> \<not> True \<and> NCPC u = 2) \<longrightarrow> (\<forall>Y\<subseteq>\<Sigma>. L3Q (fst s, u, Y))"
        unfolding L3Q_def NSORB_def
        using ca1 oddimp by auto
    qed
  next
    assume ba1: "N (fst (snd s)) \<and> NCPC (fst (snd s)) \<in> {2, 3} \<and>
      (free (NCRS (fst (snd s))) = True  \<and> fst s \<in> TTReven \<or>
       free (NCRS (fst (snd s))) = False \<and> fst s \<in> TTRodd \<and> (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
          last (fst s) = (TIn, x)))"
    hence b1: "free (NCRS (fst (snd s))) = True  \<and> fst s \<in> TTReven \<or>
       free (NCRS (fst (snd s))) = False \<and> fst s \<in> TTRodd \<and> (\<exists>x. buffer (NCRS (fst (snd s))) = x \<and>
          last (fst s) = (TIn, x))" (is "?A s \<or> ?B s")
      by simp
    thus "\<forall>u. N u \<and>
        NCRS (fst (snd s)) = NCRS u \<and>
        (True \<and> NCPC u = 2 \<or> \<not> True \<and> NCPC u = 2) \<longrightarrow>
        (\<forall>Y\<subseteq>\<Sigma>. L3Q (fst s, u, Y))" (is "?C s")
    proof
      assume ca1: "?A s"
      show "?C s"
       by (metis L3Q_def NSORB_def ca1 fst_conv insertI1 snd_conv)
    next
      assume ca1: "?B s"
      show "?C s"
      proof -
        have f1: "\<not> free (NCRS (fst (snd s))) \<and> fst s \<in> TTRodd \<and> (\<exists>n. buffer (NCRS (fst (snd s))) = n \<and> last (fst s) = (TIn, n))"
          using ca1 by force
        obtain nn where
          "(\<exists>v0. (N v0 \<and> NCRS (fst (snd s)) = NCRS v0 \<and> NCPC v0 = 2) \<and> (\<exists>v1\<subseteq>\<Sigma>. \<not> L3Q (fst s, v0, v1))) = ((N nn \<and> NCRS (fst (snd s)) = NCRS nn \<and> NCPC nn = 2) \<and> (\<exists>v1\<subseteq>\<Sigma>. \<not> L3Q (fst s, nn, v1)))"
          by (metis (no_types))
        then obtain rr :: "(nat \<times> nat) set" where
          f2: "(\<exists>n. (N n \<and> NCRS (fst (snd s)) = NCRS n \<and> NCPC n = 2) \<and> (\<exists>r\<subseteq>\<Sigma>. \<not> L3Q (fst s, n, r))) = ((N nn \<and> NCRS (fst (snd s)) = NCRS nn \<and> NCPC nn = 2) \<and> rr \<subseteq> \<Sigma> \<and> \<not> L3Q (fst s, nn, rr))"
          by auto
        have "NCRS (fst (snd (fst s, nn, rr))) = NCRS (fst (snd s)) \<longrightarrow> \<not> N nn \<or> NCRS (fst (snd s)) \<noteq> NCRS nn \<or> NCPC nn \<noteq> 2 \<or> \<not> rr \<subseteq> \<Sigma> \<or> L3Q (fst s, nn, rr)"
          using f1 by (simp add: L3Q_def NSORB_def)
        then have "\<not> N nn \<or> NCRS (fst (snd s)) \<noteq> NCRS nn \<or> NCPC nn \<noteq> 2 \<or> \<not> rr \<subseteq> \<Sigma> \<or> L3Q (fst s, nn, rr)"
          by fastforce
        then show ?thesis
          using f2 by blast
       qed 
     qed
  qed
  thus "L3G s \<longrightarrow> L3P1inner s" ..
qed

lemma hoareeq:
  assumes
    "P1 = P2" and
    "c1 = c2" and
    "Q1 = Q2"
  shows "Impl.hoare P1 c1 Q1 = Impl.hoare P2 c2 Q2"
  using assms by simp

lemma "L1Q s \<Longrightarrow> L1P1 s" unfolding L1P1_def by blast

lemma conjimphelp: "\<lbrakk>A; B; (A \<Longrightarrow> B \<Longrightarrow> C)\<rbrakk> \<Longrightarrow> C" by simp

lemma TSeq123FulfillsTPreAndTInv:
  "Impl.hoare TPre TSeq123 TInv"
proof -
  have l1: "Impl.hoare L1P TLine1 L1Q" unfolding L1P_def L1P1_def L1P2_def L1P1inner_def TLine1_def L1G_def
    using "Impl.hdo"[of 1 L1Q] hdohelp2[of 1 L1Q] by simp
  have l2: "Impl.hoare L2P TLine2 L2Q" unfolding L2P_def L2P1_def L2P2_def L2P1inner_def TLine2_def L2G_def
    using "Impl.hcomm"[of 2 L2Q] hcommhelp2[of 2 L2Q] by simp
  have l3: "Impl.hoare L3P TLine3 L3Q" unfolding L3P_def L3P1_def L3P2_def L3P1inner_def TLine3_def L3G_def
    using "Impl.hcbr"[of 3 L3Q "(\<lambda>s. True)" 2 2] hcbrhelp2[of 3 L3Q "(\<lambda>s. True)" 2 2] by simp

  from l1 l1c Impl.hcons have l1': "Impl.hoare L1ConsP2 TLine1 L1Q" by blast
  from l2 l2c Impl.hcons have l2': "Impl.hoare L2ConsP2 TLine2 L2Q" by blast
  from l3 l3c Impl.hcons have l3': "Impl.hoare L3ConsP2 TLine3 L3Q" by blast

  have l2ql3qeq: "L2Q = L3Q"
    unfolding L2Q_def L3Q_def ..
    
  {
    {
      {
        have hlpr: "(\<lambda>(tr, t, X). (TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> NCPC t \<in> dom (2 :: comm TEf TF) \<and> N t) = (\<lambda>s. (TI23 s \<or> NSORB s) \<and> NCPC (fst (snd s)) = 2 \<and> N (fst (snd s)))"
          using case_prod_beta' unfolding sc_inj.simps indom_def by fastforce
        from l2' have "Impl.hoare (\<lambda>(tr, t, X). (TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (NCPC t \<in> dom TLine2) \<and> N t) TLine2 (\<lambda>s. TI23 s \<or> NSORB s)"
          unfolding  TLine2_def L2ConsP2_def L2Q_def hlpr  
          by blast
      }
      moreover
      {
        have hlpr: "(\<lambda>(tr, t, X). (TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> NCPC t \<in> dom (3 :: cbr (\<lambda>s. True) 2 2) \<and> N t) = (\<lambda>s. (TI23 s \<or> NSORB s) \<and> NCPC (fst (snd s)) = 3 \<and> N (fst (snd s)))"
          using case_prod_beta' unfolding sc_inj.simps indom_def by fastforce
        from l3' L3Q_def "Impl.hcons" have "Impl.hoare (\<lambda>(tr, t, X). (TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (NCPC t \<in> dom TLine3) \<and> N t) TLine3 (\<lambda>s. TI23 s \<or> NSORB s)"
          unfolding L3ConsP2_def L3Q_def TLine3_def hlpr
          by presburger
      }
      ultimately
      have "Impl.hoare (\<lambda>s. TI23 s \<or> NSORB s) (TLine2 \<oplus> TLine3)
        (\<lambda>(tr, t, X). (TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (N t \<longrightarrow> \<not> (NCPC t \<in> dom (TLine2 \<oplus> TLine3))))"
        using Impl.hseq[of "(\<lambda> s . TI23 s \<or> NSORB s)" TLine2 TLine3]
        unfolding TLine2_def TLine3_def by force
      moreover
      {
        have "\<And> s . (\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s) \<and> (NCPC (fst (snd s)) \<in> dom (TLine2 \<oplus> TLine3)) \<and> N (fst (snd s))) s \<Longrightarrow> (\<lambda>s. TI23 s \<or> NSORB s) s"
          unfolding TLine2_def TLine3_def TPre_def
          by auto
        hence "(\<And>q. (TPre q \<or> TI23 q \<or> NSORB q) \<and>
          (NCPC (fst (snd q)) \<in> dom (TLine2 \<oplus> TLine3)) \<and>
          N (fst (snd q)) \<Longrightarrow>
          TI23 q \<or> NSORB q)"
        by simp
      }
      moreover  
      {
        have "\<And> s . (\<lambda>(tr, t, X). (TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (N t \<longrightarrow> \<not> (NCPC t \<in> dom (TLine2 \<oplus> TLine3)))) s \<Longrightarrow>
          (\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s)) s"
          unfolding TLine2_def TLine3_def TPre_def
          by auto
        hence "(\<And>q. case q of
          (tr, t, X) \<Rightarrow>
            (TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and>
            (N t \<longrightarrow>
             \<not> (NCPC t \<in> dom (TLine2 \<oplus> TLine3))) \<Longrightarrow>
          TPre q \<or> TI23 q \<or> NSORB q)"
          by simp
      }
      ultimately
      have "Impl.hoare (\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s) \<and> (NCPC (fst (snd s)) \<in> dom (TLine2 \<oplus> TLine3)) \<and> N (fst (snd s))) (TLine2 \<oplus> TLine3)
        (\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s))"
        using "Impl.hcons"[of 
          "(\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s) \<and> (NCPC (fst (snd s)) \<in> dom (TLine2 \<oplus> TLine3)) \<and> N (fst (snd s)))"
          "(\<lambda>s. TI23 s \<or> NSORB s)"
          "(TLine2 \<oplus> TLine3)"
          "(\<lambda>(tr, t, X). (TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (N t \<longrightarrow> \<not> (NCPC t \<in> dom (TLine2 \<oplus> TLine3))))"
          "(\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s))"
]        unfolding TLine2_def TLine3_def

        by blast
      moreover
      have "\<And> s . (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (NCPC t \<in> dom (TLine2 \<oplus> TLine3)) \<and> N t) (fst s, fst (snd s), snd (snd s)) \<Longrightarrow> (\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s) \<and> (NCPC (fst (snd s)) \<in> dom (TLine2 \<oplus> TLine3)) \<and> N (fst (snd s))) s"
        unfolding TLine2_def TLine3_def
        by auto
      moreover
        have "\<And> s . (\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s)) s \<Longrightarrow> (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X))) (fst s, fst (snd s), snd (snd s))" 
        by auto
      ultimately
      have "Impl.hoare (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (NCPC t \<in> dom (TLine2 \<oplus> TLine3)) \<and> N t) (TLine2 \<oplus> TLine3)
        (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)))"
        using Impl.hcons[of 
          "(\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (NCPC t \<in> dom (TLine2 \<oplus> TLine3)) \<and> N t)"
          "(\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s) \<and> (NCPC (fst (snd s)) \<in> dom (TLine2 \<oplus> TLine3)) \<and> N (fst (snd s)))"
          "(TLine2 \<oplus> TLine3)"
          "(\<lambda>s. (TPre s \<or> TI23 s \<or> NSORB s))"
          "(\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)))"
          ]
        unfolding TLine2_def TLine3_def
        by force
    }
    moreover
    {
      from l1' have "Impl.hoare
     (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and>
           NCPC t = 1 \<and> N t)
     TLine1 (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)))"
        unfolding L1ConsP2_def L1Q_def
        by (simp add: case_prod_beta')
      hence "Impl.hoare
     (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and>
           ((NCPC t) \<in> dom TLine1) \<and> N t)
     TLine1 (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)))"
        unfolding TLine1_def
        by auto
    }
    ultimately
    have "Impl.hoare (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X))) TSeq123 (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (N t \<longrightarrow> \<not>(NCPC t \<in> dom TSeq123)))" 
      using Impl.hseq
        [of
          "(\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)))"
          "TLine1" "TLine2 \<oplus> TLine3"
          ]
          case_prod_beta' case_prodI
      unfolding L1ConsP2_def L1Q_def TLine1_def TLine2_def TLine3_def TSeq123_def TSeq23_def
      by auto
   }
   moreover
   {
     have "\<And> s . (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and> (N t \<longrightarrow> \<not>(NCPC t \<in> dom TSeq123))) (fst s, fst (snd s), snd (snd s)) \<Longrightarrow> TInv s"
      unfolding TInv_def NSORB_def TSeq123_def TLine1_def TSeq23_def TLine2_def TLine3_def
      by auto
   }
   moreover
   {
     have "\<And> s . TPre s \<Longrightarrow> (\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X))) (fst s, fst (snd s), snd (snd s))"
      by simp
   }
   ultimately
   show ?thesis
     using Impl.hcons[of TPre 
       "(\<lambda>(tr, t, X). (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)))"
       TSeq123
       "(\<lambda>(tr, t, X).
         (TPre (tr, t, X) \<or> TI23 (tr, t, X) \<or> NSORB (tr, t, X)) \<and>
         (N t \<longrightarrow> \<not> (NCPC t \<in> dom TSeq123)))"
       TInv
       ]
     unfolding TInv_def TSeq123_def TLine1_def TSeq23_def TLine2_def TLine3_def
     by auto
qed

lemma InvImpliedConn:
  fixes tr s X
  assumes
    a: "TInv (tr, s, X)" and
    b: "\<not> N s"
  shows "Conn (tr, X)"
using assms
unfolding TInv_def TPre_def TI23_def
unfolding Conn_def
using fst_conv snd_conv by auto

text {* Now, this is the big end. Each state in the SF semantics, whose starting state fulfilled
the precondition, now also fulfills the connecting property @{term Conn}. *}

lemma TSeq123_Imp_Conn:
  fixes tr s X
  assumes "(tr, s, X) \<in> Impl.sfsem TSeq123 {x . TPre x}"
  shows "Conn (tr, X)"
proof -
  have preHasPc1: "\<forall> tr s X . TPre (tr, s, X) \<longrightarrow> NCPC s = 1"
    unfolding TPre_def snd_conv fst_conv
    by simp
  have oneInDom: "1 \<in> dom TSeq123"
    unfolding TSeq123_def TLine1_def
    by auto
  have noPathOutside: "noPathOutside TSeq123"
    unfolding TSeq123_def TLine1_def TSeq23_def TLine2_def TLine3_def noPathOutside_def
    by auto
  have notNs: "\<not> N s"
    using Impl.CannotHaveNormalOutput[of TSeq123 "{x . TPre x}"]
    using noPathOutside assms oneInDom preHasPc1
    unfolding TSeq123_def TLine1_def TSeq23_def TLine2_def TLine3_def
    by fastforce
  show ?thesis
    using InvImpliedConn[OF _ notNs, of tr X]
      Impl.hsem_soundness[OF TSeq123FulfillsTPreAndTInv, unfolded "Impl.hoare_valid_def"]
      assms
    by blast
qed

paragraph {* Connecting CUC and CSP *}

text {* When we have some @{term "(tr, X)"} pair in the CSP SF semantics for the buffer program which,
as a next step, would communicate the value out, we can add an additional reading step in the beginning
to the CSP program and prepend this to the trace, accordingly.
Due to the recursive definition of the CSP program, this is the same as instead leaving the first step
of the program away.
*}

lemma trace_prepend_in:
  fixes X :: "(nat \<times> nat) set"
  assumes "(CV\<^sub>t\<^sub>r tr EvtCV, CV\<^sub>X X EvtCV) :f ((sndF ([[(tout, k) -> $spec]]Ff [[PN]]Ffix)))"
    and unfold: "[[$spec]]Ff [[PN]]Ffix = [[Pair tin ? (k::nat) -> (tout, k) -> $spec]]Ff [[PN]]Ffix"
  shows "(CV\<^sub>t\<^sub>r ((tin, k) # tr) EvtCV, CV\<^sub>X X EvtCV) :f sndF ([[$spec]]Ff [[PN]]Ffix::(nat \<times> nat) domF)"
proof -
  have base: "(CV\<^sub>t\<^sub>r tr EvtCV, CV\<^sub>X X EvtCV) :f {f. (\<exists>X. f = (<>, X) \<and> Ev (tout, inv (Pair tin) (tin, k)) \<notin> X) \<or>
                                  (\<exists>s X. f = (<Ev (tout, inv (Pair tin) (tin, k))> ^^^ s, X) \<and> (s, X) :f sndF ([[PN]]Ffix spec))}f"
  proof -
    have f1: "(CV\<^sub>t\<^sub>r tr EvtCV, CV\<^sub>X X EvtCV) :f failures ((tout, k) -> $spec) [[PN]]Ffix"
      using assms unfolding EvtCV_def by simp
    have "\<And>n na. (n::nat) = inv (Pair (na::nat)) (na, n)"
      by (metis (no_types) f_inv_into_f rangeI snd_conv)
    then show ?thesis
      using f1 by (simp add: failures.simps(15) failures.simps(4))
  qed
  have range: "(tin, k) \<in> range (Pair tin)" by simp
  have hlpr: "\<And> P t . t : {x . P x} \<Longrightarrow> HC_F2 {x . P x} \<Longrightarrow> t :f {x . P x}f"
    by (simp add: CollectF_open_memF setF_def memberF)
  have G1: "(CV\<^sub>t\<^sub>r ((tin, k) # tr) EvtCV, CV\<^sub>X X EvtCV)
        : ({f. (\<exists>Y. f = (<>, Y) \<and> Ev ` range (Pair tin) \<inter> Y = {}) \<or>
           (\<exists>a s Y. f = (<Ev a> ^^^ s, Y) \<and>
                    (s, Y) :f {f. (\<exists>X. f = (<>, X) \<and> Ev (tout, inv (Pair tin) a) \<notin> X) \<or>
                                  (\<exists>s X. f = (<Ev (tout, inv (Pair tin) a)> ^^^ s, X) \<and> (s, X) :f sndF ([[PN]]Ffix spec))}f \<and>
                    a \<in> range (Pair tin))}::(nat \<times> nat) failure set)"
    unfolding "CUCtoCSPtrace.simps"
    using base range unfolding EvtCV_def by fastforce
  moreover have G2: "HC_F2 ({f. (\<exists>Y. f = (<>, Y) \<and> Ev ` range (Pair tin) \<inter> Y = {}) \<or>
           (\<exists>a s Y. f = (<Ev a> ^^^ s, Y) \<and>
                    (s, Y) :f {f. (\<exists>X. f = (<>, X) \<and> Ev (tout, inv (Pair tin) a) \<notin> X) \<or>
                                  (\<exists>s X. f = (<Ev (tout, inv (Pair tin) a)> ^^^ s, X) \<and> (s, X) :f sndF ([[PN]]Ffix spec))}f \<and>
                    a \<in> range (Pair tin))}::(nat \<times> nat) failure set)"
    using Ext_pre_choice_setF setF_def by fast
  show "(CV\<^sub>t\<^sub>r ((tin, k) # tr) EvtCV, CV\<^sub>X X EvtCV) :f sndF ([[$spec]]Ff [[PN]]Ffix::(nat \<times> nat) domF)"
    unfolding unfold
    unfolding unfoldFtofailues
    unfolding Rec_prefix_def
    unfolding failures.simps
    using hlpr[OF G1 G2] .
qed

text {* When we have some @{term "(tr, X)"} pair in the CSP SF semantics for the buffer program which,
as a next step, would read some value, we can add an additional out-communication step in the beginning
to the CSP program and prepend this to the trace, accordingly.
Due to the recursive definition of the CSP program, this is the same as instead leaving the first step
of the program away.
*}

lemma trace_prepend_out:
  fixes X :: "(nat \<times> nat) set"
  assumes "(CV\<^sub>t\<^sub>r tr EvtCV, CV\<^sub>X X EvtCV) :f ((sndF ([[$spec]]Ff [[PN]]Ffix)))"
    and unfold: "[[$spec]]Ff [[PN]]Ffix = [[Pair tin ? (k::nat) -> (tout, k) -> $spec]]Ff [[PN]]Ffix"
  shows "(CV\<^sub>t\<^sub>r ((tout, k) # tr) EvtCV, CV\<^sub>X X EvtCV) :f sndF ([[(tout, k) -> $spec]]Ff [[PN]]Ffix)"
proof -
  have base: "(CV\<^sub>t\<^sub>r tr EvtCV, CV\<^sub>X X EvtCV) :f sndF ([[PN]]Ffix (spec))"
    using assms(1) in_failures_Proc_name by force
  have "(CV\<^sub>t\<^sub>r ((tout, k) # tr) EvtCV, CV\<^sub>X X EvtCV)
        : {f. (\<exists>X. f = (<>, X) \<and> Ev (tout, k) \<notin> X) \<or> (\<exists>s X. f = (<Ev (tout, k)> ^^^ s, X) \<and> (s, X) :f sndF ([[PN]]Ffix spec))}"
    unfolding "CUCtoCSPtrace.simps"
    using base unfolding EvtCV_def by fastforce
  moreover have "HC_F2 {f. (\<exists>X. f = (<>, X) \<and> Ev (tout, k) \<notin> X) \<or> (\<exists>s X. f = (<Ev (tout, k)> ^^^ s, X) \<and> (s, X) :f sndF ([[PN]]Ffix spec))}"
    using Act_prefix_setF setF_def
    by force
  ultimately show ?thesis
    unfolding unfold
    unfolding unfoldFtofailues
    unfolding failures.simps
    by (simp add: CollectF_open_memF setF_def)
qed

text {* If a @{term "(tr, X)"} pair fulfill @{term Conn}, then they are also contained in the CSP SF semantics. *}

lemma Conn_Imp_Spec:
  fixes X
  assumes
    conn: "Conn (tr, X)" and
    unfold: "(([[$spec]]Ff [[PN]]Ffix)::(nat \<times> nat) domF) = [[(% x . (TIn, x)) ? k -> (TOut, k) -> $spec]]Ff [[PN]]Ffix"
  shows
    "(CV\<^sub>t\<^sub>r tr EvtCV, CV\<^sub>X X EvtCV) :f sndF ([[$spec]]Ff [[PN]]Ffix)"
using conn
proof (induct tr rule: length_induct, clarsimp)
  fix xs
  assume
    IH: "\<forall>ys. length ys < length xs \<longrightarrow> Conn (ys, X) \<longrightarrow> (CV\<^sub>t\<^sub>r ys EvtCV, CV\<^sub>X X EvtCV) :f failures ($spec) [[PN]]Ffix" and
    conn: "Conn (xs, X)"
  have hc_f2: "HC_F2 {f. (\<exists>Y. f = (<>, Y) \<and> Ev ` range (Pair TIn) \<inter> Y = {}) \<or>
           (\<exists>a s Y. f = (<Ev a> ^^^ s, Y) \<and>
                    (s, Y) :f {f. (\<exists>X. f = (<>, X) \<and> Ev (TOut, inv (Pair TIn) a) \<notin> X) \<or>
                                  (\<exists>s X. f = (<Ev (TOut, inv (Pair TIn) a)> ^^^ s, X) \<and> (s, X) :f sndF ([[PN]]Ffix spec))}f \<and>
                    a \<in> range (Pair TIn))}"
  proof -
    have "\<And>P f. {p. (\<exists>E. p = (<>, E) \<and> Ev ` P \<inter> E = {}) \<or> (\<exists>pa t E. p = (<Ev (pa::nat \<times> 'b)> ^^^ t, E) \<and> (t, E) :f f pa \<and> pa \<in> P)} \<in> Collect HC_F2"
      using Ext_pre_choice_setF setF_def by blast
    then show ?thesis
      by fastforce
  qed
  have funfold: "failures ($spec) [[PN]]Ffix = failures ((Pair TIn ? k -> (TOut, k) -> $spec)::('a, nat \<times> nat) proc) [[PN]]Ffix"
    by (metis sndF_semFf unfold)
  consider (even) "(xs, X) \<in> TFeven" | (odd) "(xs, X) \<in> TFodd"
    using conn[unfolded Conn_def] by blast
  thus "(CV\<^sub>t\<^sub>r xs EvtCV, CV\<^sub>X X EvtCV) :f failures ($spec) [[PN]]Ffix"
  proof (cases)
    case even
    then obtain nats where xs: "xs = fhelper nats"
      unfolding TFeven_def
      by blast
    have Xrestr: "X \<subseteq> \<Sigma> - {(tin, x). tin = TIn}"
      using even unfolding TFeven_def by simp
    have "length xs \<noteq> 1"
      using xs
    proof (induct nats, auto) qed
    then consider (zero) "length xs = 0" | (many) "length xs > 1"
      by linarith
    thus ?thesis
    proof (cases)
      case zero
      hence xs: "xs = []" by simp
      have "Ev ` range (Pair TIn) \<inter> {x. \<exists>x'. x' \<in> X \<and> x = Ev x'} = {}"
        using Xrestr
        by blast
      hence "(<>, {x. \<exists>x'. x' \<in> X \<and> x = Ev x'})
          : {f. (\<exists>Y. f = (<>, Y) \<and> Ev ` range (Pair TIn) \<inter> Y = {}) \<or>
           (\<exists>a s Y. f = (<Ev a> ^^^ s, Y) \<and>
                    (s, Y) :f {f. (\<exists>X. f = (<>, X) \<and> Ev (TOut, inv (Pair TIn) a) \<notin> X) \<or>
                                  (\<exists>s X. f = (<Ev (TOut, inv (Pair TIn) a)> ^^^ s, X) \<and> (s, X) :f sndF ([[PN]]Ffix spec))}f \<and>
                    a \<in> range (Pair TIn))}"
        using Xrestr
        by blast
      thus ?thesis
        unfolding xs
        unfolding funfold
        unfolding Rec_prefix_def
        unfolding "failures.simps"
        unfolding "CUCtoCSPrefusal_def" "CUCtoCSPtrace.simps"
        using hc_f2
        using CollectF_open_memF setF_def
        unfolding EvtCV_def
        by fastforce
    next
      case many
      then obtain a b xs' where xs: "xs = a # b # xs'"
        by (metis One_nat_def \<open>length xs \<noteq> 1\<close> add.right_neutral add_Suc_right list.size(3) list.size(4) neq_Nil_conv not_one_less_zero)
      obtain k where a: "a = (TIn, k)" and b: "b = (TOut, k)"
        using first_two_are_in_out[OF conn[unfolded xs Conn_def]]
        by blast
      have xs'InEven: "(xs', X) \<in> TFeven"
        using evenDblTail [OF even[unfolded xs]] .
      show "(CV\<^sub>t\<^sub>r xs EvtCV, CV\<^sub>X X EvtCV) :f failures ($spec) [[PN]]Ffix"
        unfolding xs a b
      proof -
        have "length xs' < length xs"
          by (simp add: xs)
        then show "(CV\<^sub>t\<^sub>r ((TIn, k) # (TOut, k) # xs') EvtCV, CV\<^sub>X X EvtCV) :f failures ($spec) [[PN]]Ffix"
          by (metis Conn_def IH assms(2) semFf_def sndF_semFf trace_prepend_in trace_prepend_out xs'InEven)
      qed
    qed
  next
    case odd
    then obtain nats y where xs: "xs = fhelper nats @ [(TIn, y)]"
      unfolding TFodd_def
      by blast
    have Xrestr: "X \<subseteq> \<Sigma> - {(tout, y'). y' = y \<and> tout = TOut}"
      using odd xs unfolding TFodd_def by simp
    have "length xs \<noteq> 0 \<and> length xs \<noteq> 2"
      using xs
    proof (induct nats, auto) qed
    then consider (one) "length xs = 1" | (many) "length xs > 2"
      by linarith
    thus ?thesis
    proof (cases)
      case one
      then obtain a where xs: "xs = [a]" and a: "a = (TIn, y)"
        by (simp add: xs)
      have hlpr: "\<And> a . (inv (Pair TIn) (TIn, a)) = a"
        by (meson f_inv_into_f prod.inject rangeI)
      have range: "(TIn, y) \<in> range (Pair TIn)" by simp
      have "(<>, {x. \<exists>x'. x' \<in> X \<and> x = Ev x'}) : {f. (\<exists>X. f = (<>, X) \<and> Ev (TOut, inv (Pair TIn) a) \<notin> X) \<or>
                                  (\<exists>s X. f = (<Ev (TOut, inv (Pair TIn) a)> ^^^ s, X) \<and> (s, X) :f sndF ([[PN]]Ffix spec))}"
        using Xrestr unfolding a hlpr by fastforce
      hence "(CV\<^sub>t\<^sub>r [] EvtCV, CV\<^sub>X X EvtCV) :f failures ((TOut, y) ->  $spec) [[PN]]Ffix"
        unfolding a hlpr
        unfolding funfold
        unfolding Rec_prefix_def
        unfolding "failures.simps"
        unfolding "CUCtoCSPrefusal_def" "CUCtoCSPtrace.simps"
      by (simp add: Act_prefix_setF CollectF_open_memF EvtCV_def)
      thus ?thesis
        unfolding xs a
        unfolding funfold
        unfolding Rec_prefix_def
        unfolding "failures.simps"
        unfolding "CUCtoCSPrefusal_def" "CUCtoCSPtrace.simps"
        using hc_f2 hlpr[of y] range CollectF_open_memF setF_def
        unfolding EvtCV_def
        by fastforce
    next
      case many
      then obtain a b xs' where xs: "xs = a # b # xs'"
        by (metis One_nat_def \<open>length xs \<noteq> 0 \<and> length xs \<noteq> 2\<close> add.right_neutral add_Suc_right list.size(3) list.size(4) not_numeral_less_one remdups_adj.cases)
      obtain k where a: "a = (TIn, k)" and b: "b = (TOut, k)"
        using first_two_are_in_out[OF conn[unfolded xs Conn_def]]
        by blast
      have xs'InOdd: "(xs', X) \<in> TFodd"
        using oddDblTail [OF odd[unfolded xs]] .
      show "(CV\<^sub>t\<^sub>r xs EvtCV, CV\<^sub>X X EvtCV) :f failures ($spec) [[PN]]Ffix"
        unfolding xs a b
        using trace_prepend_in
      proof -
        have "length xs' < length xs"
          by (simp add: xs)
        then show "(CV\<^sub>t\<^sub>r ((TIn, k) # (TOut, k) # xs') EvtCV, CV\<^sub>X X EvtCV) :f failures ($spec) [[PN]]Ffix"
          by (metis Conn_def IH sndF_semFf trace_prepend_in trace_prepend_out unfold xs'InOdd)
      qed
    qed
  qed
qed

text{* We have a CUC CSP SF refinement, still using @{term Conn}. *}

theorem Spec_SF_Refinement_Impl_With_Conn:
  assumes
    unfold: "([[$spec]]Ff [[PN]]Ffix::(nat \<times> nat) domF) = [[(% x . (TIn, x)) ? k -> (TOut, k) -> $spec]]Ff [[PN]]Ffix"
  shows "Impl.spec_impl_refinement_sf_seq_pre_set ($spec) PN (CUCConnToCSPConn Conn (inv EvtCV)) EvtCV TSeq123"
unfolding Impl.spec_impl_refinement_sf_seq_pre_set_def
proof(rule, rule, rule, rule, rule)
  fix tr s X
  assume aa1: "(tr, s, X) \<in> Impl.sfsem TSeq123 (Collect PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q)"
  have bij: "bij EvtCV" unfolding EvtCV_def by simp
  have eq: "{x. TPre x} = Collect PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q"
    using TPreEqPreStdSFseq by blast
  from TSeq123_Imp_Conn aa1 have Conn: "Conn (tr, X)"
    by (simp add: eq)
  thus "CUCConnToCSPConn Conn (inv EvtCV) (CV\<^sub>t\<^sub>r tr EvtCV, CV\<^sub>X X EvtCV)"
    unfolding CUCConnToCSPConn_eq[OF bij] .
next
  have bij: "bij EvtCV" unfolding EvtCV_def by simp
  from CUCConnToCSPConn_imp[OF bij] Conn_Imp_Spec[OF _ unfold]
  show "\<forall>tr X. CUCConnToCSPConn Conn (inv EvtCV) (tr, X) \<longrightarrow> (tr, X) :f sndF ([[$spec]]Ff [[PN]]Ffix)"
    by metis
qed

text{* We have a CUC CSP SF refinement, without @{term Conn}. *}

theorem Spec_SF_Refinement_Impl:
  assumes
    unfold: "([[$spec]]Ff [[PN]]Ffix::(nat \<times> nat) domF) = [[(% x . (TIn, x)) ? k -> (TOut, k) -> $spec]]Ff [[PN]]Ffix"
  shows "Impl.spec_impl_refinement_sf_seq_no_conn_pre_set ($spec) PN EvtCV TSeq123"
unfolding Impl.spec_impl_refinement_sf_seq_no_conn_pre_set_def
proof(rule, rule, rule, rule)
  fix tr s X
  assume aa1: "(tr, s, X) \<in> Impl.sfsem TSeq123 (Collect PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q)"
  have eq: "{x. TPre x} = Collect PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q"
    using TPreEqPreStdSFseq by blast
  from TSeq123_Imp_Conn[unfolded eq, OF aa1] Conn_Imp_Spec[OF _ unfold]
    show "(CV\<^sub>t\<^sub>r tr EvtCV, CV\<^sub>X X EvtCV) :f sndF ([[$spec]]Ff [[PN]]Ffix)" by blast
qed
    
end
  
subsection {* Parallel instantiations *}

paragraph {* Defining the CSP Process *}

datatype PN = Spec1 | Spec2 | Spec12

text {* Defining the channel IDs used for communication. *}

definition ftin: "TIn::nat \<equiv> 0"
definition ftmid: "TMid::nat \<equiv> 1"
definition ftout: "TOut::nat \<equiv> 2"
type_synonym CSP_Evt = "nat \<times> nat"

interpretation Impl1: SingleBuffer TIn TMid
proof
  show "TIn \<noteq> TMid"
  unfolding ftin ftmid by simp
qed

interpretation Impl2: SingleBuffer TMid TOut
proof
  show "TMid \<noteq> TOut"
  unfolding ftmid ftout by simp
qed

text {* This function is the core CSP process definition. *}

primrec	
  PNdef :: "PN => (PN, CSP_Evt) proc"
where
--"Defined as @{term Rec_Prefix}. The k will be put into the function before the question mark,
     and then unpacked when passed to Spec."
 "PNdef   (Spec1) = (% x . (TIn, x)) ? k -> (TMid, k) -> $Spec1"
|"PNdef   (Spec2) = (% x . (TMid, x)) ? k -> (TOut, k) -> $Spec2"
--"To make it guarded, we have to replace the firs occurrence of @{term Spec1} and @{term Spec2} with the unfolded form."
--"Without the guarded property, the whole lemma machenery wouldn't work, thus we should accept it ..."
|"PNdef  (Spec12) = ((% x . (TIn, x)) ? k -> (TMid, k) -> $Spec1) |[{(mid, x) . x \<in> \<nat> \<and> mid = TMid}, {(mid,x) . x \<in> \<nat> \<and> mid = TMid}]| ((% x . (TMid, x)) ? k -> (TOut, k) -> $Spec2)"

text {* As it is done in the CPS Prover example, we have to overload this function. *}

overloading Set_PNdef == 
  "PNfun :: (PN, CSP_Evt) pnfun"
begin
  definition "PNfun == PNdef"
end
  
declare Set_PNdef_def [simp]

lemma guardedfun_PN[simp]:
      "guardedfun PNdef"
by (simp add: guardedfun_def, rule allI, induct_tac p, simp_all)

overloading FPmode == "FPmode :: fpmode"
begin
  definition "FPmode == CMSmode"
end

declare FPmode_def [simp]

definition "\<Sigma>12 \<equiv> {(T, (x::nat)) . T = TIn \<or> T = TMid \<or> T = TOut}"
definition "CS = {(mid, x::nat) . mid = TMid}"
definition "impl1 = (seqn (
                    Impl1.TSeq123,
                    Impl1.\<Sigma>)
                )"
definition "impl2 = (seqn (
                    Impl2.TSeq123,
                    Impl2.\<Sigma>)
                )"
definition "impl12 = (parn impl1 (CS, CS) impl2)"

text {* We used a generic version of the following as assumption in the CUC buffer locale.
We now have to prove, that we really can unfold the CSP buffer. *}

lemma Spec1Unfold: "([[$Spec1]]F::(nat \<times> nat) domF) = [[Pair TIn ? k -> (TMid, k) -> $Spec1]]F"
proof -
  have eqCMS: "FPmode = CMSmode"
    unfolding FPmode_def
    ..
  have t: "[[$Spec1]]F = (UFP [[PNdef]]Ffun Spec1)"
    using semF_UFP_cms[OF _ guardedfun_PN eqCMS, of Spec1]
    by simp
  have funeqdef: "PNfun = PNdef"
    using Set_PNdef_def
    by simp
  have hasUFP: "[[PNdef]]Ffun hasUFP"
    by (simp add: semF_hasUFP_cms)
  hence hasUFP': "[[(PNfun::(PN \<Rightarrow> (PN, nat \<times> nat) proc))]]Ffun hasUFP"
    unfolding funeqdef[symmetric]
    .
  obtain UFIXP where UFIXP: "UFIXP isUFP [[PNdef]]Ffun"
    using UFP_is hasUFP by blast
  have UFIXP_unfold: "UFIXP = [[PNdef]]Ffun UFIXP"
    by (meson UFIXP isUFP_def)
  hence UFIXP_BuffIn_unfold: "UFIXP Spec1 = [[PNdef]]Ffun UFIXP Spec1"
    by auto
  have "UFIXP = UFP [[PNdef]]Ffun"
    using UFIXP isUFP_UFP by fastforce
  hence UFIXP': "UFIXP = (THE x. x isUFP [[PNdef]]Ffun)"
    by (simp add: UFP_def hasUFP)
  show ?thesis
    unfolding t
    unfolding UFP_def if_P[OF hasUFP]
    unfolding UFIXP'[symmetric]
    unfolding UFIXP_BuffIn_unfold
    unfolding semFfun_def
    unfolding "PNdef.simps"
    by (simp add: MF_def \<open>UFIXP = UFP [[PNdef]]Ffun\<close> semF_def semFfix_def)
qed

lemma Spec1Unfold': "([[$Spec1]]Ff [[PNdef]]Ffix::(nat \<times> nat) domF) = [[Pair TIn ? k -> (TMid, k) -> $Spec1]]Ff [[PNdef]]Ffix"
  using Spec1Unfold
  by (simp add: MF_def Set_PNdef_def semF_def)

lemma Spec2Unfold:
  shows "([[$Spec2]]F::(nat \<times> nat) domF) = [[Pair TMid ? k -> (TOut, k) -> $Spec2]]F"
proof -
  have eqCMS: "FPmode = CMSmode"
    unfolding FPmode_def
    ..
  have t: "[[$Spec2]]F = (UFP [[PNdef]]Ffun Spec2)"
    using semF_UFP_cms[OF _ guardedfun_PN eqCMS, of Spec2]
    by simp
  have funeqdef: "PNfun = PNdef"
    using Set_PNdef_def
    by simp
  have hasUFP: "[[PNdef]]Ffun hasUFP"
    by (simp add: semF_hasUFP_cms)
  hence hasUFP': "[[(PNfun::(PN \<Rightarrow> (PN, nat \<times> nat) proc))]]Ffun hasUFP"
    unfolding funeqdef[symmetric]
    .
  obtain UFIXP where UFIXP: "UFIXP isUFP [[PNdef]]Ffun"
    using UFP_is hasUFP by blast
  have UFIXP_unfold: "UFIXP = [[PNdef]]Ffun UFIXP"
    by (meson UFIXP isUFP_def)
  hence UFIXP_BuffMid_unfold: "UFIXP Spec2 = [[PNdef]]Ffun UFIXP Spec2"
    by auto
  have "UFIXP = UFP [[PNdef]]Ffun"
    using UFIXP isUFP_UFP by fastforce
  hence UFIXP': "UFIXP = (THE x. x isUFP [[PNdef]]Ffun)"
    by (simp add: UFP_def hasUFP)
  show ?thesis
    unfolding t
    unfolding UFP_def if_P[OF hasUFP]
    unfolding UFIXP'[symmetric]
    unfolding UFIXP_BuffMid_unfold
    unfolding semFfun_def
    unfolding "PNdef.simps"
    by (simp add: MF_def \<open>UFIXP = UFP [[PNdef]]Ffun\<close> semF_def semFfix_def)
qed
                           
lemma Spec2Unfold': "([[$Spec2]]Ff [[PNdef]]Ffix::(nat \<times> nat) domF) = [[Pair TMid ? k -> (TOut, k) -> $Spec2]]Ff [[PNdef]]Ffix"
  using Spec2Unfold
  by (simp add: MF_def semF_def)

text {* The big final. *}

theorem Buffer_SF_Parallel_refinement:
  "($Spec1 |[CS,CS]| $Spec2) PNdef <=(EvtCV)=F impl12"
proof -
  have wff1: "wff\<^sub>p impl1"
    unfolding impl1_def wff\<^sub>p_def
      Impl1.whole_code_unfold
      "sc_wellformed.simps"
      definitions.sc_inj.simps
      definitions.dom.simps
      SFsemSigma_def liSet_def
      by fastforce
  have wff2: "wff\<^sub>p impl2"
    unfolding impl2_def wff\<^sub>p_def
      Impl2.whole_code_unfold
      "sc_wellformed.simps"
      definitions.sc_inj.simps
      definitions.dom.simps
      SFsemSigma_def liSet_def
      by fastforce
  have inj: "inj EvtCV" unfolding EvtCV_def by simp
  have ref_1: "($Spec1) PNdef <=(EvtCV)=F impl1"
    using
      seq_refinement_imp_par_refinement_no_conn_pre_set
      Impl1.Spec_SF_Refinement_Impl Spec1Unfold' wff1 impl1_def
    by metis
  have ref_2: "($Spec2) PNdef <=(EvtCV)=F impl2"
    using
      seq_refinement_imp_par_refinement_no_conn_pre_set
      Impl2.Spec_SF_Refinement_Impl Spec2Unfold' wff2 impl2_def
    by metis
  show ?thesis
    using
      refinement_parallel_composition_without_conn_Pre_set'[
        OF ref_1 ref_2 wff1 wff2 inj, of CS CS
      ]
    unfolding impl12_def EvtCV_def
    by auto
qed
  
end
