section{*Syntax and State Definitions*}

theory definitions imports Main 
begin

text{*Definitions of a polymorphic state and labeled sets of instructions.*}

type_synonym label = nat

text{*A state consists of variables (or registers), represented by R and the current program counter.*}

record ('a) state = R  ::'a   PC ::label


text{* Introduce events which populate the traces. *}

type_synonym 'e event = "'e"
type_synonym 'e trace = "'e event list"
type_synonym ('s, 'e) trace_state = "'e trace \<times> 's state"

text{* Introduce convenient wrapper datatypes. *}
type_synonym 's do_fun = "'s \<Rightarrow> 's set"
type_synonym 's cbr_fun = "'s \<Rightarrow> bool"
type_synonym ('s, 'e) event_fun = "'s \<Rightarrow> 'e event set"
type_synonym ('s, 'e) state_fun = "'s \<Rightarrow> 'e event \<Rightarrow> 's"


text{* The theory of CUC consists of three basic instructions: The nondeterministic do, a noncomputing branch and a nondeterministic communication primitive. *}
datatype ('s, 'e)instr = do "('s do_fun)"
                   | cbr "('s cbr_fun)" label label
                   | comm "('s, 'e) event_fun" "('s, 'e) state_fun"

type_synonym ('s, 'e)labeled_instruction_set = "(label \<times> ('s, 'e)instr) set"

definition indom :: "label \<Rightarrow> ('s, 'e)labeled_instruction_set \<Rightarrow> bool" (*<*)(infix "\<in>\<^sub>l\<^sub>i\<^sub>s" 55)(*>*) where
  "x \<in>\<^sub>l\<^sub>i\<^sub>s lis = (\<exists>y. (x,y) \<in> lis)"

text{* The label has to be in some of its components. *}

lemma indom_splits:
"pc \<in>\<^sub>l\<^sub>i\<^sub>s (c1 \<union> c2) \<Longrightarrow> pc \<in>\<^sub>l\<^sub>i\<^sub>s c1 \<or> pc \<in>\<^sub>l\<^sub>i\<^sub>s c2"
apply(simp add: indom_def)
apply clarify
apply(rule_tac x="y" in exI)
apply blast
done

lemma not_indom_splits: 
"\<not> pc \<in>\<^sub>l\<^sub>i\<^sub>s (c1 \<union> c2) \<Longrightarrow> \<not> pc \<in>\<^sub>l\<^sub>i\<^sub>s c1 \<and> \<not> pc \<in>\<^sub>l\<^sub>i\<^sub>s c2"
by(simp add: indom_def)


text{*The following datatype formalizes structured code and is used to obtain a compositional
big-step semantics or denotational semantics.*}

datatype 'c structured_code_base = 
  one label "'c" ("_ :: _" (*<*) [60, 60] 10(*>*)) | 
  seq "'c structured_code_base" "'c structured_code_base" (infixr "\<oplus>" 10)

type_synonym ('s, 'e)structured_code = "(('s, 'e) instr) structured_code_base"

text {* Transforming the tree structure of the code into a set, losing the structural information. *}
primrec sc_inj :: " ('s, 'e)structured_code \<Rightarrow> ('s, 'e)labeled_instruction_set"  ("\<union>\<^sub>s\<^sub>c") where
  one:  "\<union>\<^sub>s\<^sub>c (one l ins) = {(l, ins)}"  |
  seq:  "\<union>\<^sub>s\<^sub>c (seq sc1 sc2) = (\<union>\<^sub>s\<^sub>c sc1) \<union> (\<union>\<^sub>s\<^sub>c sc2)"
  
primrec dom :: "('s, 'e)structured_code \<Rightarrow> (nat set)" (*("\<union>\<^sub>l")*) where
   one:  "dom (one l ins) = {l}"  |
   seq:  "dom (seq sc1 sc2) = (dom sc1) \<union> (dom sc2)"
   
(*notation (xsymbols) dom ("\<union>\<^sub>l")*)

text {* @{term dom} is like @{term sc_inj} without the instructions. *}
lemma dom_sc_inj_eq: "dom c = { l . \<exists> i . (l, i) \<in> \<union>\<^sub>s\<^sub>c c}"
proof (induct c, auto) qed

lemma dom_eq_indom: "l \<in> (dom c) = l \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c c)"
by (simp add: dom_sc_inj_eq indom_def)

text {* In a code tree, there always has to be some instruction. This is true as we do not have a "Nil" constructor for the tree. *}
lemma sc_inj_not_empty: "\<union>\<^sub>s\<^sub>c c \<noteq> {}"
proof (induct c, auto) qed

lemma dom_not_empty: "dom c \<noteq> {}"
proof (induct c, auto) qed

lemma sc_inj_dom_singleton:
  assumes "\<union>\<^sub>s\<^sub>c c = {(l, i)}"
  shows "dom c = {l}"
  using assms dom_sc_inj_eq
by fast

lemma indom_sc_subset_imp_indom_dom_subset: 
  assumes "\<union>\<^sub>s\<^sub>c c1 \<subseteq> \<union>\<^sub>s\<^sub>c c2"
  shows "definitions.dom c1 \<subseteq> definitions.dom c2"
  using assms dom_sc_inj_eq by blast
  
lemma indom_sc_eq_imp_indom_dom_eq:
  assumes "\<union>\<^sub>s\<^sub>c c1 = \<union>\<^sub>s\<^sub>c c2"
  shows "definitions.dom c1 = definitions.dom c2"
using indom_sc_subset_imp_indom_dom_subset assms
by blast

lemma not_in_sc_inj_imp_not_in_dom:
  assumes "\<forall>i . (l, i) \<notin> \<union>\<^sub>s\<^sub>c c"
  shows "l \<notin> dom c"
using assms dom_sc_inj_eq by blast

lemma in_sc_inj_imp_in_dom:
  assumes "(l, i) \<in> \<union>\<^sub>s\<^sub>c c"
  shows "l \<in> dom c"
using assms dom_sc_inj_eq by blast

text {* Wellformedness means: Each label appears at most once in the whole code. So, for any label @{term l} in code @{term "c1 \<oplus> c2"}, we want it to appear either in @{term c1} or in @{term c2}, but not in both. This means, no label will appear in both subcodes. *}
primrec sc_wellformed :: "('s, 'e)structured_code \<Rightarrow> bool"  ("wff\<^sub>s\<^sub>c") where
   one:  "wff\<^sub>s\<^sub>c (one l instr) = True"  |
   seq:  "wff\<^sub>s\<^sub>c (seq sc1 sc2) = (((wff\<^sub>s\<^sub>c sc1) \<and> (wff\<^sub>s\<^sub>c sc2)) \<and> ((dom sc1 \<inter> dom sc2) = {}))"

lemma sc_wellformed_xor:
  assumes "wff\<^sub>s\<^sub>c (c1 \<oplus> c2)"
  shows "l \<notin> dom c1 \<or> l \<notin> dom c2"
  using assms by auto

text {* Then we now, that for a wellformed @{term "c1 \<oplus> c2"}, if their lis does not overlap, their domains will not, too. *}

lemma sc_inj_intersection_empty_imp_dom_intersection_empty:
  assumes "\<union>\<^sub>s\<^sub>c c1 \<inter> \<union>\<^sub>s\<^sub>c c2 = {}" and "wff\<^sub>s\<^sub>c (c1 \<oplus> c2)"
  shows "dom c1 \<inter> dom c2 = {}"
using assms
proof (induct c1, auto) qed

text {* A wellformed seq can never only consist of one instruction. *}

lemma sc_inj_num:
  assumes "wff\<^sub>s\<^sub>c (c1 \<oplus> c2)"
  shows "\<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<noteq> {a}"
proof (rule ccontr, simp)
  assume aa1: "\<union>\<^sub>s\<^sub>c c1 \<union> \<union>\<^sub>s\<^sub>c c2 = {a}"
  obtain l i where
    ao1: "a = (l, i)"
    by fastforce
  hence a1: "{(l, i)} = \<union>\<^sub>s\<^sub>c c1 \<and> {(l, i)} = \<union>\<^sub>s\<^sub>c c2"
    unfolding sc_inj.seq
    using aa1 sc_inj_not_empty[of c1] sc_inj_not_empty[of c2]
    by (simp add: Un_singleton_iff)
  hence BANG: "definitions.dom c1 = {l} \<and> definitions.dom c2 = {l}"
    using sc_inj_dom_singleton
    by metis
  show False
    using BANG assms
    by auto
qed

text {* Two @{term "(label \<times> instr)"} pairs with the same label will always share to the same instruction, if the code is wellformed.
  Note: The opposite is not true, as wellformedness also requires to have each label appearing at most once. *}
lemma wff_code_implies_unique_labels:
  assumes 
    wff_c: "wff\<^sub>s\<^sub>c c" and
    i1InC: "(l, i1) \<in> \<union>\<^sub>s\<^sub>c c" and
    i2InC: "(l, i2) \<in> \<union>\<^sub>s\<^sub>c c"
  shows
    "i1 = i2"
using assms
proof (induct c)
  fix c1 c2
  assume
    wff_c: " wff\<^sub>s\<^sub>c (c1 \<oplus> c2)" and
    aa2: "(l, i1) \<in> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)" and
    aa3: "(l, i2) \<in> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)" and
    aa4: "wff\<^sub>s\<^sub>c c1 \<Longrightarrow>
        (l, i1) \<in> \<union>\<^sub>s\<^sub>c c1 \<Longrightarrow>
        (l, i2) \<in> \<union>\<^sub>s\<^sub>c c1 \<Longrightarrow> i1 = i2" and
    aa5: "wff\<^sub>s\<^sub>c c2 \<Longrightarrow>
        (l, i1) \<in> \<union>\<^sub>s\<^sub>c c2 \<Longrightarrow>
        (l, i2) \<in> \<union>\<^sub>s\<^sub>c c2 \<Longrightarrow> i1 = i2"
  have wff_c1c2: "wff\<^sub>s\<^sub>c c1 \<and> wff\<^sub>s\<^sub>c c2"
    using wff_c
    by auto
  have "(l, i1) \<in> \<union>\<^sub>s\<^sub>c c1 \<and> (l, i2) \<in> \<union>\<^sub>s\<^sub>c c1 \<or> (l, i1) \<in> \<union>\<^sub>s\<^sub>c c2 \<and> (l, i2) \<in> \<union>\<^sub>s\<^sub>c c2"
    using sc_wellformed_xor[OF wff_c, of l] dom_sc_inj_eq[of "c1 \<oplus> c2"] aa2 aa3 
    using UnE in_sc_inj_imp_in_dom unfolding sc_inj.seq  by fastforce
  thus ?thesis
    using aa2 aa3 wff_c1c2 aa4 aa5 by fast
qed auto

text{* Reconstruction of a single instruction. *}

lemma sc_inj_singleton_equal:
  assumes a: "\<union>\<^sub>s\<^sub>c c \<subseteq> {(l, i)}" and b: "wff\<^sub>s\<^sub>c c"
  shows "c = one l i"
by (metis a b prod.inject sc_inj.one sc_inj_not_empty sc_inj_num singletonD singletonI structured_code_base.exhaust subset_singletonD)

text {* For two code structures, reconstruct the wellformedness property for their sequential composition. *}
lemma wff_reconstruction:
  assumes 
    a: "wff\<^sub>s\<^sub>c B" and
    b: "\<union>\<^sub>s\<^sub>c l \<subset> \<union>\<^sub>s\<^sub>c B" and
    c: "\<union>\<^sub>s\<^sub>c r \<subset> \<union>\<^sub>s\<^sub>c B" and
    d: "\<union>\<^sub>s\<^sub>c l \<inter> \<union>\<^sub>s\<^sub>c r = {}" and
    e: "wff\<^sub>s\<^sub>c l" and
    f: "wff\<^sub>s\<^sub>c r"
  shows "wff\<^sub>s\<^sub>c (l \<oplus> r)"
proof (rule ccontr)
  assume aa1: "\<not> wff\<^sub>s\<^sub>c (l \<oplus> r)"
  hence domInter: "\<not> dom l \<inter> dom r = {}"
    unfolding sc_wellformed.seq
    using e f
    by auto
  obtain lbl :: label where
    ao1: "lbl \<in> dom l \<inter> dom r"
    using domInter
    by blast
  obtain i1 where
    ao2: "(lbl, i1) \<in> \<union>\<^sub>s\<^sub>c l"
    using not_in_sc_inj_imp_not_in_dom ao1
    by blast
  obtain i2 where
    ao3: "(lbl, i2) \<in> \<union>\<^sub>s\<^sub>c r"
    using not_in_sc_inj_imp_not_in_dom ao1
    by blast
  from ao2 have a1: "(lbl, i1) \<in> \<union>\<^sub>s\<^sub>c B" using b
    by blast
  from ao3 have a2: "(lbl, i2) \<in> \<union>\<^sub>s\<^sub>c B" using c
    by blast
  have "i1 = i2" using wff_code_implies_unique_labels[OF a a1 a2] .
  with ao2 ao3 d show False
    by blast
qed

paragraph {* Shorthands for conjunction extraction *}

lemma cj2_1: "A \<and> B \<Longrightarrow> A" by simp
lemma cj2_2: "A \<and> B \<Longrightarrow> B" by simp
lemma cj3_1: "A \<and> B \<and> C \<Longrightarrow> A" by simp
lemma cj3_2: "A \<and> B \<and> C \<Longrightarrow> B" by simp
lemma cj3_3: "A \<and> B \<and> C \<Longrightarrow> C" by simp
lemma cj4_1: "A \<and> B \<and> C \<and> D \<Longrightarrow> A" by simp
lemma cj4_2: "A \<and> B \<and> C \<and> D \<Longrightarrow> B" by simp
lemma cj4_3: "A \<and> B \<and> C \<and> D \<Longrightarrow> C" by simp
lemma cj4_4: "A \<and> B \<and> C \<and> D \<Longrightarrow> D" by simp
lemma cj5_1: "A \<and> B \<and> C \<and> D \<and> E \<Longrightarrow> A" by simp
lemma cj5_2: "A \<and> B \<and> C \<and> D \<and> E \<Longrightarrow> B" by simp
lemma cj5_3: "A \<and> B \<and> C \<and> D \<and> E \<Longrightarrow> C" by simp
lemma cj5_4: "A \<and> B \<and> C \<and> D \<and> E \<Longrightarrow> D" by simp
lemma cj5_5: "A \<and> B \<and> C \<and> D \<and> E \<Longrightarrow> E" by simp
lemma cj6_1: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<Longrightarrow> A" by simp
lemma cj6_2: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<Longrightarrow> B" by simp
lemma cj6_3: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<Longrightarrow> C" by simp
lemma cj6_4: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<Longrightarrow> D" by simp
lemma cj6_5: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<Longrightarrow> E" by simp
lemma cj6_6: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<Longrightarrow> F" by simp
lemma cj7_1: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<Longrightarrow> A" by simp
lemma cj7_2: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<Longrightarrow> B" by simp
lemma cj7_3: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<Longrightarrow> C" by simp
lemma cj7_4: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<Longrightarrow> D" by simp
lemma cj7_5: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<Longrightarrow> E" by simp
lemma cj7_6: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<Longrightarrow> F" by simp
lemma cj7_7: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<Longrightarrow> G" by simp
lemma cj12_1: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> A" by simp
lemma cj12_2: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> B" by simp
lemma cj12_3: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> C" by simp
lemma cj12_4: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> D" by simp
lemma cj12_5: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> E" by simp
lemma cj12_6: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> F" by simp
lemma cj12_7: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> G" by simp
lemma cj12_8: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> H" by simp
lemma cj12_9: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> I" by simp
lemma cj12_10: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> J" by simp
lemma cj12_11: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> K" by simp
lemma cj12_12: "A \<and> B \<and> C \<and> D \<and> E \<and> F \<and> G \<and> H \<and> I \<and> J \<and> K \<and> L \<Longrightarrow> L" by simp
end