section {* Hoare Calculus for SF semantics *}

theory hoare
imports stable_failures
begin

context SFsemSigma
begin
  
paragraph {* Hoare triple validity definitions *}

type_synonym 'b assn = "'b \<Rightarrow> bool"
definition
hoare_valid_orig :: "('s, 'e) failure assn \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> ('s, 'e) failure assn \<Rightarrow> bool" ("\<Turnstile>* {(1_)}/ (_)/ {(1_)}" 50) where
"\<Turnstile>* {P}c{Q} = (\<forall> s t. P s \<and> (t \<in> (sfsem c {s}))  \<longrightarrow>  Q t)"

text {* This formulation reasons about having a set of starting state. *}

definition
hoare_valid_orig_set :: "('s, 'e) failure assn \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> ('s, 'e) failure assn \<Rightarrow> bool" ("\<Turnstile>** {(1_)}/ (_)/ {(1_)}" 50) where
"\<Turnstile>** {P}c{Q} = (\<forall> S t. (\<forall> s \<in> S . P s) \<and> (t \<in> (sfsem c S))  \<longrightarrow>  Q t)"

text {* This is the exact formulation, as it is used in the paper. *}

definition
hoare_valid_orig_paper :: "('s, 'e) failure assn \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> ('s, 'e) failure assn \<Rightarrow> bool" ("\<Turnstile>*** {(1_)}/ (_)/ {(1_)}" 50) where
"\<Turnstile>*** {P}c{Q} = (\<forall> s . (P s \<longrightarrow> (\<forall> t \<in> (sfsem c {s}) . Q t)))"

text {* This formulation is a little bit more convenient to use, as the failure states are already split up into the triple. *}

definition
hoare_valid :: "('s, 'e) failure assn \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> ('s, 'e) failure assn \<Rightarrow> bool" ("\<Turnstile> {(1_)}/ (_)/ {(1_)}" 50) where
"\<Turnstile> {P}c{Q} = (\<forall> t_tr t_t t_X S . ((\<forall> (s_tr, s_t, s_X) \<in> S .  P (s_tr, s_t, s_X)) \<and> ((t_tr, t_t, t_X) \<in> (sfsem c S))) \<longrightarrow>  Q (t_tr, t_t, t_X))"

text {* We want to use the convenient version, but ultimately proof the paper version. Thus we have to show, that
proving the convenient version also implies proving it for the paper version. *}

lemma hoare_valid_imp_hoare_valid_orig:
  assumes "\<Turnstile> {P}c{Q}"
  shows "\<Turnstile>** {P}c{Q}"
  using assms unfolding hoare_valid_def hoare_valid_orig_set_def Ball_def
by auto

lemma hoare_valid_orig_imp_hoare_valid_orig_paper:
  assumes "\<Turnstile>** {P}c{Q}"
  shows "\<Turnstile>*** {P}c{Q}"
  using assms unfolding hoare_valid_orig_paper_def hoare_valid_orig_set_def Ball_def
by blast

lemma hoare_valid_imp_hoare_valid_orig_paper:
  assumes "\<Turnstile> {P}c{Q}"
  shows "\<Turnstile>*** {P}c{Q}"
  using assms hoare_valid_imp_hoare_valid_orig hoare_valid_orig_imp_hoare_valid_orig_paper
by blast
  
paragraph {* Hoare rules *}

text {* The hoare definition as it is introduced in the paper. *}

inductive hoare :: "(('s, 'e) failure) assn \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> (('s, 'e) failure) assn \<Rightarrow> bool"
where
hdo: "hoare (\<lambda> (tr, t, X) .(\<not>(N t \<and> NCPC t = l) \<longrightarrow> Q(tr, t, X))
                      \<and>
                      ((N t \<and> NCPC t = l) \<longrightarrow> (\<forall> u. N u \<and> NCPC u = Suc l \<and> NCRS u \<in> f (NCRS t) 
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q(tr, u, Y))))) (one l (do f)) Q" |
hcbr: "hoare (\<lambda> (tr, t, X) .(\<not>(N t \<and> NCPC t = l) \<longrightarrow> Q(tr, t, X))
                      \<and>
                      ((N t \<and> NCPC t = l) \<longrightarrow> (\<forall> u. N u \<and> NCRS t = NCRS u \<and>
            ((b (NCRS t) \<and> NCPC u = m) \<or> (\<not> b (NCRS t) \<and> NCPC u = n))
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q(tr, u, Y))))) (one l (cbr b m n)) Q" |
hcomm: "hoare (\<lambda> (tr, t, X) .(\<not>(N t \<and> NCPC t = l) \<longrightarrow> Q(tr, t, X)) \<and> 
              (N t \<and> NCPC t = l \<longrightarrow> ((\<forall> Y \<subseteq> \<Sigma> - ef (NCRS t) . Q(tr, NCC t, Y)) \<and> 
              (\<forall> u e . N u \<and> e \<in> ef (NCRS t) \<and> NCPC u = Suc l \<and> NCRS u = f2 (NCRS t) e
            \<longrightarrow> (\<forall> Y \<subseteq> \<Sigma> . Q(tr@[e], u, Y)))))) (one l (comm ef f2)) Q" |
hseq: "\<lbrakk>hoare (\<lambda> (tr, t, X). I(tr, t, X) \<and> ((NCPC t) \<in> dom code\<^sub>1) \<and> N t) code\<^sub>1 (I); hoare (\<lambda> (tr, t, X). I(tr, t, X) \<and> ((NCPC t) \<in> dom code\<^sub>2) \<and> N t) code\<^sub>2 (I)\<rbrakk> \<Longrightarrow>
        hoare I (code\<^sub>1 \<oplus> code\<^sub>2) (\<lambda>(tr,t,X). I(tr,t,X) \<and> (N t \<longrightarrow> \<not> ((NCPC t) \<in> dom (code\<^sub>1 \<oplus> code\<^sub>2))))" |
hcons: "\<lbrakk>\<And> q . P q \<Longrightarrow> P' q ; hoare P' code Q'; \<And> q. Q' q \<Longrightarrow> Q q\<rbrakk> \<Longrightarrow> hoare P code Q"

inductive_cases ri_hdo[elim!]: "hoare P (one l (do f)) Q"
inductive_cases ri_hcbr[elim!]: "hoare P (one l (cbr b m n)) Q"
inductive_cases ri_hcomm[elim!]: "hoare P (one l (comm ef f2)) Q"
inductive_cases ri_hseq[elim]: "hoare I (code\<^sub>1 \<oplus> code\<^sub>2) (\<lambda>(tr,t,X). I(tr,t,X) \<and> (N t \<longrightarrow> \<not> ((NCPC t) \<in> dom (code\<^sub>1 \<oplus> code\<^sub>2))))"

text {* Extends @{text Set.UnE} to three sets. Will be used below. *}
  
lemma UnE2 [elim!]: "c \<in> A \<union> B \<union> C \<Longrightarrow> (c \<in> A \<Longrightarrow> P) \<Longrightarrow> (c \<in> B \<Longrightarrow> P) \<Longrightarrow> (c \<in> C \<Longrightarrow> P) \<Longrightarrow> P"
  by auto

paragraph {* Soundness *}

theorem hsem_soundness:
  fixes P Q :: "(('s, 'e) failure) assn" and
        code :: "('s, 'e) structured_code"
  assumes "hoare P code Q"
  shows "hoare_valid P code Q"
  using assms
proof (induct rule: hoare.induct)
  case (hdo l Q f)
  show ?case (is "\<Turnstile> {\<lambda> a . case a of (tr, t, X) \<Rightarrow> ?A tr t X} (one l (do f)) {Q}") unfolding hoare_valid_def
  proof (rule allI, rule allI, rule allI, rule allI, rule impI)
    fix 
      S :: "('s, 'e) failure set" and
      t_tr :: "'e trace" and
      t_t :: "'s NCstate" and
      t_X :: "'e event set"
    assume b1: "(\<forall>(s_tr, s_t, s_X)\<in>S. case (s_tr, s_t, s_X) of (tr, t, X) \<Rightarrow> ?A tr t X) \<and> (t_tr, t_t, t_X) \<in> \<lbrakk>one l (do f)\<rbrakk>S"
    hence b2: "(t_tr, t_t, t_X) \<in> \<lbrakk>one l (do f)\<rbrakk>S"
      by simp
    --"Split set S, such that we can reason for both conditions separately."
    have "S =
        {(tr, t, X) . (tr, t, X) \<in> S \<and>   (NCPC t = l \<and> N t) } \<union>
        {(tr, t, X) . (tr, t, X) \<in> S \<and> \<not> (NCPC t = l \<and> N t) }"
      by blast
    then obtain S'1 S'2 :: "('s, 'e) failure set" where
      e10: "S = S'1 \<union> S'2 \<and>
        S'1 = {(tr, t, X) . (tr, t, X) \<in> S \<and>   (NCPC t = l \<and> N t) } \<and>
        S'2 = {(tr, t, X) . (tr, t, X) \<in> S \<and> \<not> (NCPC t = l \<and> N t) }"
        by auto
    hence e20: "S = S'1 \<union> S'2" ..
    from e10 have e30: "\<forall> (tr, t, X) \<in> S'2 . \<not> ( NCPC t = l \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
      by blast
    from e10 have e40: "\<forall> (tr, t, X) \<in> S'1 . ( NCPC t = l \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
      by blast
    from b2 e20 e30 e40 sfsem_remove_nondom_set have "(t_tr, t_t, t_X) \<in> \<lbrakk>one l (do f)\<rbrakk>S'1 \<union> S'2"
      by auto
    moreover
    {
      from b1 e20 have "(\<forall>(s_tr, s_t, s_X)\<in>S'2. case (s_tr, s_t, s_X) of (tr, t, X) \<Rightarrow> ?A tr t X)" by simp
      with e30 have "\<forall>(s_tr, s_t, s_X)\<in>S'2. Q (s_tr, s_t, s_X)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
      by auto
    }
    moreover
    {
      from b1 e20 have "(\<forall>(s_tr, s_t, s_X)\<in>S'1.
        case (s_tr, s_t, s_X) of
        (tr, t, X) \<Rightarrow>
          (\<not> (N t \<and> NCPC t = l) \<longrightarrow>
           Q (tr, t, X)) \<and>
          (N t \<and> NCPC t = l \<longrightarrow>
           (\<forall>u. N u \<and>
                NCPC u = Suc l \<and>
                NCRS u \<in> f (NCRS t) \<longrightarrow>
                (\<forall>Y\<subseteq>\<Sigma>. Q (tr, u, Y)))))" by simp
       with e40 have c2: "\<forall>(s_tr, s_t, s_X)\<in>S'1. (\<forall>u. N u \<and>
                NCPC u = Suc l \<and>
                NCRS u \<in> f (NCRS s_t) \<longrightarrow>
                (\<forall>Y\<subseteq>\<Sigma>. Q (s_tr, u, Y)))" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
                by auto
        from e40 have c3: "S'1 -\<^sub>s\<^sub>f {l} = {}" unfolding SClean_def
          by auto
        from c3 have "\<forall> (tr, t, X) \<in> sfsem (one l (do f)) S'1 . \<exists> (s_tr, s_t, s_X) \<in> S'1 .
          tr = s_tr \<and> N t \<and> NCPC t = Suc l \<and> NCRS t \<in> f (NCRS s_t) \<and> X \<subseteq> \<Sigma>"
          by auto
        with c2 have "\<forall>(tr, t, X) \<in> sfsem (one l (do f)) S'1 . Q (tr, t, X)"
          using Pair_inject case_prodE case_prodI2 by fastforce
          (* using split_def by fastforce *)
    }
    ultimately
    show "Q (t_tr, t_t, t_X)" by auto
  qed
next
  case (hcbr l Q b m n)
  show ?case (is "\<Turnstile> {\<lambda> a . case a of (tr, t, X) \<Rightarrow> ?A tr t X} (one l (cbr b m n)) {Q}") unfolding hoare_valid_def
  proof (rule allI, rule allI, rule allI, rule allI, rule impI)
    fix 
      S :: "('s, 'e) failure set" and
      t_tr :: "'e trace" and
      t_t :: "'s NCstate" and
      t_X :: "'e event set"
    (* fix s t :: "('s, 'e) failure" *)
    assume b1: "(\<forall>(s_tr, s_t, s_X)\<in>S. case (s_tr, s_t, s_X) of (tr, t, X) \<Rightarrow> ?A tr t X) \<and> (t_tr, t_t, t_X) \<in> \<lbrakk>one l (cbr b m n)\<rbrakk>S"
    hence b2: "(t_tr, t_t, t_X) \<in> \<lbrakk>one l (cbr b m n)\<rbrakk>S"
      by simp
    have "S =
        {(tr, t, X) . (tr, t, X) \<in> S \<and>   (NCPC t = l \<and> N t) } \<union>
        {(tr, t, X) . (tr, t, X) \<in> S \<and> \<not> (NCPC t = l \<and> N t) }"
      by blast
    then obtain S'1 S'2 :: "('s, 'e) failure set" where
      e10: "S = S'1 \<union> S'2 \<and>
        S'1 = {(tr, t, X) . (tr, t, X) \<in> S \<and>   (NCPC t = l \<and> N t) } \<and>
        S'2 = {(tr, t, X) . (tr, t, X) \<in> S \<and> \<not> (NCPC t = l \<and> N t) }"
        by auto
    hence e20: "S = S'1 \<union> S'2" ..
    from e10 have e30: "\<forall> (tr, t, X) \<in> S'2 . \<not> ( NCPC t = l \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
      by blast
    from e10 have e40: "\<forall> (tr, t, X) \<in> S'1 . ( NCPC t = l \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
      by blast
    from b2 e20 e30 e40 sfsem_remove_nondom_set have "(t_tr, t_t, t_X) \<in> \<lbrakk>one l (cbr b m n)\<rbrakk>S'1 \<union> S'2"
      by auto
    moreover
    {
      from b1 e20 have "(\<forall>(s_tr, s_t, s_X)\<in>S'2. case (s_tr, s_t, s_X) of (tr, t, X) \<Rightarrow> ?A tr t X)" by simp
      with e30 have "\<forall>(s_tr, s_t, s_X)\<in>S'2. Q (s_tr, s_t, s_X)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
      by auto
    }
    moreover
    {
      from b1 e20 have "(\<forall>(s_tr, s_t, s_X)\<in>S'1.
        case (s_tr, s_t, s_X) of
        (tr, t, X) \<Rightarrow>
          (\<not> (N t \<and> NCPC t = l) \<longrightarrow>
           Q (tr, t, X)) \<and>
          (N t \<and> NCPC t = l \<longrightarrow>
           (\<forall>u. N u \<and>
                NCRS t = NCRS u \<and>
                (b (NCRS t) \<and> NCPC u = m \<or>
                 \<not> b (NCRS t) \<and>
                 NCPC u = n) \<longrightarrow>
                (\<forall>Y\<subseteq>\<Sigma>. Q (tr, u, Y)))))" by simp
       with e40 have c2: "\<forall>(s_tr, s_t, s_X)\<in>S'1. (\<forall>u. N u \<and>
                NCRS s_t = NCRS u \<and>
                (b (NCRS s_t) \<and> NCPC u = m \<or>
                 \<not> b (NCRS s_t) \<and>
                 NCPC u = n) \<longrightarrow>
                (\<forall>Y\<subseteq>\<Sigma>. Q (s_tr, u, Y)))" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
                by auto
        from e40 have c3: "S'1 -\<^sub>s\<^sub>f {l} = {}" unfolding SClean_def
          by auto
        from c3 have "\<forall> (tr, t, X) \<in> sfsem (one l (cbr b m n)) S'1 . \<exists> (s_tr, s_t, s_X) \<in> S'1 .
          tr = s_tr \<and> N t \<and> NCPC s_t = l \<and> NCRS t = NCRS s_t \<and> (b (NCRS s_t) \<and> NCPC t = m \<or>
                 \<not> b (NCRS s_t) \<and>
                 NCPC t = n) \<and> X \<subseteq> \<Sigma>"
          by auto
        with c2 have "\<forall>(tr, t, X) \<in> sfsem (one l (cbr b m n)) S'1 . Q (tr, t, X)"
          using Pair_inject case_prodE case_prodI2 by fastforce
    }
    ultimately
    show "Q (t_tr, t_t, t_X)" by auto
  qed
next
  case (hcomm l Q ef f2)
  show ?case (is "\<Turnstile> {\<lambda> a . case a of (tr, t, X) \<Rightarrow> ?A tr t X} (one l (comm ef f2)) {Q}") unfolding hoare_valid_def
  proof (rule allI, rule allI, rule allI, rule allI, rule impI)
     fix 
      S :: "('s, 'e) failure set" and
      t_tr :: "'e trace" and
      t_t :: "'s NCstate" and
      t_X :: "'e event set"
    (* fix s t :: "('s, 'e) failure" *)
    assume b1: "(\<forall>(s_tr, s_t, s_X)\<in>S. case (s_tr, s_t, s_X) of (tr, t, X) \<Rightarrow> ?A tr t X) \<and> (t_tr, t_t, t_X) \<in> \<lbrakk>one l (comm ef f2)\<rbrakk>S"
    hence b2: "(t_tr, t_t, t_X) \<in> \<lbrakk>one l (comm ef f2)\<rbrakk>S"
      by simp
    have "S =
        {(tr, t, X) . (tr, t, X) \<in> S \<and>   (NCPC t = l \<and> N t) } \<union>
        {(tr, t, X) . (tr, t, X) \<in> S \<and> \<not> (NCPC t = l \<and> N t) }"
      by blast
    then obtain S'1 S'2 :: "('s, 'e) failure set" where
      e10: "S = S'1 \<union> S'2 \<and>
        S'1 = {(tr, t, X) . (tr, t, X) \<in> S \<and>   (NCPC t = l \<and> N t) } \<and>
        S'2 = {(tr, t, X) . (tr, t, X) \<in> S \<and> \<not> (NCPC t = l \<and> N t) }"
        by auto
    hence e20: "S = S'1 \<union> S'2" ..
    from e10 have e30: "\<forall> (tr, t, X) \<in> S'2 . \<not> ( NCPC t = l \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
      by blast
    from e10 have e40: "\<forall> (tr, t, X) \<in> S'1 . ( NCPC t = l \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
      by blast
    from b2 e20 e30 e40 sfsem_remove_nondom_set have "(t_tr, t_t, t_X) \<in> \<lbrakk>one l (comm ef f2)\<rbrakk>S'1 \<union> S'2"
      by fastforce
    moreover
    {
      from b1 e20 have "(\<forall>(s_tr, s_t, s_X)\<in>S'2. case (s_tr, s_t, s_X) of (tr, t, X) \<Rightarrow> ?A tr t X)" by simp
      with e30 have "\<forall>(s_tr, s_t, s_X)\<in>S'2. Q (s_tr, s_t, s_X)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
      by auto
    }
    moreover
    {
      from b1 e20 have "(\<forall>(s_tr, s_t, s_X)\<in>S'1.
        case (s_tr, s_t, s_X) of
        (tr, t, X) \<Rightarrow>
          (\<not> (N t \<and> NCPC t = l) \<longrightarrow>
           Q (tr, t, X)) \<and>
          (N t \<and> NCPC t = l \<longrightarrow>
           (\<forall>Y\<subseteq>\<Sigma> - ef (NCRS t). Q (tr, NCC t,
              Y)) \<and>
           (\<forall>u e.
               N u \<and>
               e \<in> ef (NCRS t) \<and>
               NCPC u = Suc l \<and>
               NCRS u = f2 (NCRS t) e \<longrightarrow>
               (\<forall>Y\<subseteq>\<Sigma>. Q (tr @ [e], u, Y)))))"
         by simp
       with e40 have c2: "\<forall>(s_tr, s_t, s_X)\<in>S'1. (\<forall>Y\<subseteq>\<Sigma> - ef (NCRS s_t). Q (s_tr, NCC s_t,
              Y)) \<and>
           (\<forall>u e.
               N u \<and>
               e \<in> ef (NCRS s_t) \<and>
               NCPC u = Suc l \<and>
               NCRS u = f2 (NCRS s_t) e \<longrightarrow>
               (\<forall>Y\<subseteq>\<Sigma>. Q (s_tr @ [e], u, Y)))" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
          by auto
        from e40 have c3: "S'1 -\<^sub>s\<^sub>f {l} = {}" unfolding SClean_def
          by auto
        from c3 sfcomm[of l ef f2 S'1] have "sfsem (one l (comm ef f2)) S'1 =
          {(tr, t, X). \<exists>u X'. (tr, u, X') \<in> S'1 \<and> NCPC u = l \<and> N u \<and> X \<subseteq> \<Sigma> - ef (NCRS u) \<and> t = NCC u} \<union>
          {(tr', t, X). \<exists>u X' tr ev. (tr, u, X') \<in> S'1 \<and> NCPC u = l \<and> N u \<and> N t \<and> ev \<in> ef (NCRS u) \<and>
           NCPC t = Suc l \<and> NCRS t = f2 (NCRS u) ev \<and> tr' = tr @ [ev] \<and> X \<subseteq> \<Sigma>}"
          by auto
        with c3 have c4_0: "\<forall> (tr, t, X) \<in> sfsem (one l (comm ef f2)) S'1 . \<exists> (s_tr, s_t, s_X) \<in> S'1 .
       ((tr = s_tr \<and> X \<subseteq> \<Sigma> - ef (NCRS s_t) \<and> t = NCC s_t) \<or>
       (\<exists> ev . NCPC s_t = l \<and> N s_t \<and> ev \<in> (ef (NCRS s_t)) \<and> N t \<and> NCPC t = Suc l \<and> NCRS t = (f2 (NCRS s_t) ev) \<and> tr = s_tr@[ev] \<and> X \<subseteq> \<Sigma>))"
          by auto
        with c2 have "\<forall>(tr, t, X) \<in> sfsem (one l (comm ef f2)) S'1 . Q (tr, t, X)"
          using Pair_inject case_prodE case_prodI2
          by fastforce
    }
    ultimately
    show "Q (t_tr, t_t, t_X)" by blast
  qed
next
  fix
    I :: "('s, 'e) failure assn" and 
    c1 c2 :: "('s, 'e) structured_code"
  assume
    a1: "hoare
        (\<lambda>(tr, t, X).
            I (tr, t, X) \<and> (NCPC t \<in> dom c1) \<and> N t)
        c1 I" and
    a2: "\<Turnstile> {\<lambda>a. case a of
               (tr, t, X) \<Rightarrow>
                 I (tr, t, X) \<and>
                 (NCPC t \<in> dom c1) \<and> N t}
       c1 {I}" and
    a3: "hoare
        (\<lambda>(tr, t, X).
            I (tr, t, X) \<and> (NCPC t \<in> dom c2) \<and> N t)
        c2 I" and
    a4: "\<Turnstile> {\<lambda>a. case a of
               (tr, t, X) \<Rightarrow>
                 I (tr, t, X) \<and>
                 (NCPC t \<in> dom c2) \<and> N t}
       c2 {I}"
  show "\<Turnstile> {I} c1 \<oplus> c2
       {\<lambda>a. case a of
             (tr, t, X) \<Rightarrow>
               I (tr, t, X) \<and>
               (N t \<longrightarrow>
                \<not> (NCPC t \<in> dom (c1 \<oplus> c2)))}" (is "\<Turnstile> {I} c1 \<oplus> c2 {\<lambda>a. case a of (tr, t, X) \<Rightarrow> ?A tr t X}") unfolding hoare_valid_def
  proof (rule allI, rule allI, rule allI, rule allI, rule impI, rule case_prodI, rule case_prodI)
    fix 
        S :: "('s, 'e) failure set" and
        t_tr :: "'e trace" and
        t_t :: "'s NCstate" and
        t_X :: "'e event set"
    assume b1: "(\<forall>(s_tr, s_t, s_X)\<in>S. I (s_tr, s_t, s_X)) \<and>
       (t_tr, t_t, t_X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
    hence b2: "(t_tr, t_t, t_X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S" ..
    from b1 have b1_2: "(\<forall>(s_tr, s_t, s_X)\<in>S. I (s_tr, s_t, s_X))" ..
    show "I (t_tr, t_t, t_X) \<and>
    (N t_t \<longrightarrow> \<not> (NCPC t_t \<in> dom (c1 \<oplus> c2)))"
    proof (rule conjI)
      show "N t_t \<longrightarrow> \<not> (NCPC t_t \<in> dom (c1 \<oplus> c2))" using b1 sfseq dom_eq_indom unfolding SClean_def
        by blast
    next
      show "I (t_tr, t_t, t_X)" using b2 b1_2 unfolding sfseq
      proof (induction arbitrary: S rule: fixp_induct_sf)
        fix 
          S' :: "('s, 'e) failure set" and
          x :: "('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set"
        assume
          d1: "(\<And>S .(t_tr, t_t, t_X) \<in> (x S) -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2) \<Longrightarrow>
                     \<forall>(s_tr, s_t, s_X)\<in>S. I (s_tr, s_t, s_X) \<Longrightarrow>
                I (t_tr, t_t, t_X))" and
          d2: "(t_tr, t_t, t_X) \<in> (extends (c1 \<oplus> c2) x S') -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)" and
          d2_2: "\<forall>(s_tr, s_t, s_X)\<in>S'. I (s_tr, s_t, s_X)"
        with extends[of c1 c2 x] have d3:
          "(t_tr, t_t, t_X) \<in> (
            x (\<lbrakk>c1\<rbrakk>S') \<union> x (\<lbrakk>c2\<rbrakk>S') \<union> S'
            ) -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)"
          by auto
        hence d3: "(t_tr, t_t, t_X) \<in> (
            (x (\<lbrakk>c1\<rbrakk>S')-\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)) \<union> (x (\<lbrakk>c2\<rbrakk>S')-\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)) \<union> (S'
            -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)))"
          by simp
        show "I (t_tr, t_t, t_X)" using d3
        proof
          assume ea1: "(t_tr, t_t, t_X) \<in> (x (\<lbrakk>c1\<rbrakk>S')-\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2))"
          with d1[of "\<lbrakk>c1\<rbrakk>S'"] have ea2: "\<forall>(s_tr, s_t, s_X)\<in>\<lbrakk>c1\<rbrakk>S'. I (s_tr, s_t, s_X) \<Longrightarrow> I (t_tr, t_t, t_X)"
            by auto
          have "S' =
              {(tr, t, X) . (tr, t, X) \<in> S' \<and>   (NCPC t \<in> dom c1 \<and> N t) } \<union>
              {(tr, t, X) . (tr, t, X) \<in> S' \<and> \<not> (NCPC t \<in> dom c1 \<and> N t) }"
            by blast
          then obtain S'1 S'2 :: "('s, 'e) failure set" where
            e10: "S' = S'1 \<union> S'2 \<and>
              S'1 = {(tr, t, X) . (tr, t, X) \<in> S' \<and>   (NCPC t \<in> dom c1 \<and> N t) } \<and>
              S'2 = {(tr, t, X) . (tr, t, X) \<in> S' \<and> \<not> (NCPC t \<in> dom c1 \<and> N t) }"
              by auto
          hence e20: "S' = S'1 \<union> S'2" ..
          from e10 have e30: "\<forall> (tr, t, X) \<in> S'2 . \<not> ( NCPC t \<in> dom c1 \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
            by blast
          from e10 have e40: "\<forall> (tr, t, X) \<in> S'1 . ( NCPC t \<in> dom c1 \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
            by blast
          from ea2 e10 e30 sfsem_remove_nondom_set have e50: "\<forall>(s_tr, s_t, s_X)\<in>(\<lbrakk>c1\<rbrakk>S'1 \<union> S'2). I (s_tr, s_t, s_X) \<Longrightarrow> I (t_tr, t_t, t_X)"
            by blast
          have "\<forall>(s_tr, s_t, s_X)\<in>(S'2). I (s_tr, s_t, s_X)" using d2_2 e20
            by blast
          moreover
          {
            have e60: "\<forall>(s_tr, s_t, s_X)\<in>(S'1). I (s_tr, s_t, s_X)" using d2_2 e20
              by blast
            with e40 have e70: "\<forall>(s_tr, s_t, s_X)\<in>(S'1). I (s_tr, s_t, s_X) \<and>  (NCPC s_t \<in> dom c1) \<and> N s_t" using dom_eq_indom
              by blast
            with a2 
            have t2: "\<forall>(s_tr, s_t, s_X)\<in>(\<lbrakk>c1\<rbrakk>S'1). I (s_tr, s_t, s_X)" unfolding hoare_valid_def using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
              by auto
          }
          ultimately
          show ?thesis using e50
            by blast
        next
          assume ea1: "(t_tr, t_t, t_X) \<in> (x (\<lbrakk>c2\<rbrakk>S')-\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2))"
          with d1[of "\<lbrakk>c2\<rbrakk>S'"] have ea2: "\<forall>(s_tr, s_t, s_X)\<in>\<lbrakk>c2\<rbrakk>S'. I (s_tr, s_t, s_X) \<Longrightarrow> I (t_tr, t_t, t_X)"
            by auto
          have "S' =
              {(tr, t, X) . (tr, t, X) \<in> S' \<and>   (NCPC t \<in> dom c2 \<and> N t) } \<union>
              {(tr, t, X) . (tr, t, X) \<in> S' \<and> \<not> (NCPC t \<in> dom c2 \<and> N t) }"
            by blast
          then obtain S'1 S'2 :: "('s, 'e) failure set" where
            e10: "S' = S'1 \<union> S'2 \<and>
              S'1 = {(tr, t, X) . (tr, t, X) \<in> S' \<and>   (NCPC t \<in> dom c2 \<and> N t) } \<and>
              S'2 = {(tr, t, X) . (tr, t, X) \<in> S' \<and> \<not> (NCPC t \<in> dom c2 \<and> N t) }"
              by auto
          hence e20: "S' = S'1 \<union> S'2" ..
          from e10 have e30: "\<forall> (tr, t, X) \<in> S'2 . \<not> ( NCPC t \<in> dom c2 \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
            by blast
          from e10 have e40: "\<forall> (tr, t, X) \<in> S'1 . ( NCPC t \<in> dom c2 \<and> N t)" using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
            by blast
          from ea2 e10 e30 sfsem_remove_nondom_set have e50: "\<forall>(s_tr, s_t, s_X)\<in>(\<lbrakk>c2\<rbrakk>S'1 \<union> S'2). I (s_tr, s_t, s_X) \<Longrightarrow> I (t_tr, t_t, t_X)"
            by blast
          have "\<forall>(s_tr, s_t, s_X)\<in>(S'2). I (s_tr, s_t, s_X)" using d2_2 e20
            by blast
          moreover
          {
            have e60: "\<forall>(s_tr, s_t, s_X)\<in>(S'1). I (s_tr, s_t, s_X)" using d2_2 e20
              by blast
            with e40 have e70: "\<forall>(s_tr, s_t, s_X)\<in>(S'1). I (s_tr, s_t, s_X) \<and>  (NCPC s_t \<in> dom c2) \<and> N s_t" using dom_eq_indom
              by blast
            with a4 
            have t2: "\<forall>(s_tr, s_t, s_X)\<in>(\<lbrakk>c2\<rbrakk>S'1). I (s_tr, s_t, s_X)" unfolding hoare_valid_def using BNF_Def.Collect_case_prodD case_prodD case_prodI prod.sel(2) surj_pair split_def
              by auto
          }
          ultimately
          show ?thesis using e50
            by blast
        next
          assume "(t_tr, t_t, t_X) \<in> (S' -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2))"
          with d2_2 show ?thesis unfolding hoare_valid_def
            by simp
        qed
      qed (auto simp add: ccpo.admissible_def)
    qed
  qed
next
  --"Rule of consequence"
  fix
    P P' Q' Q:: "(('s, 'e) failure) assn" and
    c :: "('s, 'e) structured_code"
  assume 
    a1: "(\<And> q. P q \<Longrightarrow> P' q)" and
    a2: "hoare P' c Q'" and
    a3: "\<Turnstile> {P'} c {Q'}" and
    a4: "(\<And> q. Q' q \<Longrightarrow> Q q)"
  show "\<Turnstile> {P} c {Q}" unfolding hoare_valid_def
  proof (rule allI, rule allI, rule allI, rule allI, rule impI)
    fix 
      S :: "('s, 'e) failure set" and
      t_tr :: "'e trace" and
      t_t :: "'s NCstate" and
      t_X :: "'e event set"
    assume
      b1: "(\<forall>(s_tr, s_t, s_X)\<in>S. P (s_tr, s_t, s_X)) \<and> (t_tr, t_t, t_X) \<in> \<lbrakk>c\<rbrakk>S"
      from b1 a1 have p': "(\<forall>(s_tr, s_t, s_X)\<in>S. P' (s_tr, s_t, s_X))" by simp
      from a3 b1 have q': "Q' (t_tr, t_t, t_X)" using p' unfolding hoare_valid_def by blast
      from q' a4 show q: "Q (t_tr, t_t, t_X)" by simp
   qed
qed

end --"context liSet"

end
