section{*Smallstep semantics*}

theory small_step imports definitions small_step_non_locale
begin

text{* The semantics is defined with respect to a set of labeled instructions. This is
formalized using the concept of a @{text locale}. *}

locale liSet=fixes lis :: "('s, 'e)labeled_instruction_set"
 assumes lis_finite: "finite lis"
    and  lis_wellformed: "\<forall>l i1 i2. (l, i1) \<in> lis \<and> (l, i2) \<in> lis \<longrightarrow> i1 = i2"
begin

text {* There can't be two different instructions in the same line of code. *}

lemma sameinstruction: "\<lbrakk>(l, i1) \<in> lis; (l, i2) \<in> lis\<rbrakk> \<Longrightarrow> i1 = i2 "
by (simp add: lis_wellformed)

lemma indomSC_element:
  assumes "l \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstr)"
          "\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis"
          "(l, ins) \<in> lis"
  shows   "(l, ins) \<in> \<union>\<^sub>s\<^sub>c strInstr"
  using assms sameinstruction  indom_def[of l "( \<union>\<^sub>s\<^sub>c strInstr)"]
  by blast

text {* @{term Smallstep} is the semantic of a single line of code. We simply reuse the definition of the
non-locale variant. *}

definition Smallstep :: "(('s, 'e) trace_state \<times> ('s, 'e) trace_state) set"
where
  Stos   :"Smallstep \<equiv> smallstep(lis)"

text {* A @{term Smallstep} implies, that some instructions was executed. *}

lemma Smallstep_rev:
  assumes "((trs, s), (trt, t)) \<in> Smallstep"
  shows "(\<exists>l f x. (l, do f) \<in> lis \<and> x \<in> f (R s) \<and> PC s = l \<and> t = s\<lparr> R:=x, PC:=PC s + 1\<rparr>) \<and> trs=trt
         \<or>
         (\<exists>l b m n. (l,cbr b m n) \<in> lis \<and> PC s = l \<and> b (R s) \<and> t = s\<lparr>PC := m\<rparr>) \<and> trs=trt
         \<or>
         (\<exists>l b m n. (l,cbr b m n) \<in> lis \<and> PC s = l \<and> \<not> (b (R s)) \<and> t = s\<lparr>PC := n\<rparr>) \<and> trs=trt
         \<or>
         (\<exists>l ef f e.  (l, comm ef f) \<in> lis \<and> e \<in> ef (R s) \<and> PC s = l \<and> 
                             trt = trs@[e] \<and> t = s\<lparr> R := f (R s) e, PC:=PC s + 1\<rparr>)" (is "?P")
using Stos smallstep.cases[of trs s trt t lis] smallstep_rev assms
proof -
  from assms Stos have "((trs, s), (trt, t)) \<in> smallstep lis" by blast
  with smallstep_rev[of trs s trt t lis] show ?P by blast
qed

text{* We formalize two relations that model @{term mulitstep} executions and prove properties
about them. Those are mainly used to establish the relation between the smallstep
and the bigstep semantics.*}

text{* single Stuck executions for @{term Smallstep}*}
inductive_set Exec  :: "(('s, 'e) trace_state \<times> ('s, 'e) labeled_instruction_set \<times> ('s, 'e) trace_state) set"
where
  Exseq    : "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ( (trs, s),  (trs', s')) \<in> Smallstep \<and> 
               ((trs', s'), instr_set , (trt, t)) \<in> Exec\<rbrakk> 
              \<Longrightarrow> ((trs, s),instr_set, (trt, t)) \<in> Exec" 
|
  Exterm   : "\<lbrakk>instr_set \<subseteq> lis;  \<not>(indom (PC s)) instr_set\<rbrakk>
              \<Longrightarrow> ((tr, s),instr_set, (tr, s)) \<in> Exec"

text {* This is shorthand for going from a single @{term smallstep} to @{term Exec}. *}
lemma Exsing: "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ((trs, s), (trt, t)) \<in> Smallstep \<and> 
               \<not>(indom (PC t) instr_set)\<rbrakk>
              \<Longrightarrow> ((trs, s),instr_set, (trt, t)) \<in> Exec"
by (meson Exseq Exterm)
              
inductive_set multistep  :: "(('s, 'e) trace_state \<times> ('s, 'e) labeled_instruction_set \<times> ('s, 'e) trace_state) set"
where
  Mseq    : "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ( (trs, s),  (trs', s')) \<in> Smallstep \<and> 
               ((trs', s'), instr_set , (trt, t)) \<in> multistep\<rbrakk> 
              \<Longrightarrow> ((trs, s),instr_set, (trt, t)) \<in> multistep" 
|
  Mterm   : "\<lbrakk>instr_set \<subseteq> lis\<rbrakk>
              \<Longrightarrow> ((tr, s),instr_set, (tr, s)) \<in> multistep"
              
text {* This is shorthand for going from a single @{term smallstep} to @{term multistep} *}
lemma Msing: "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ((trs, s), (trt, t)) \<in> Smallstep\<rbrakk>
              \<Longrightarrow> ((trs, s),instr_set, (trt, t)) \<in> multistep"
by (meson Mseq Mterm)

text {* Restricting the number of smallsteps allows for easier reasoning, but is otherwise equivalent to the k-less original. *}

inductive_set Exec_k  :: "(('s, 'e) trace_state \<times> ('s, 'e)labeled_instruction_set \<times> ('s, 'e) trace_state \<times> nat) set"
where
  --"If we reach s' from s in one step and t from s' in k steps, we can reach t from s in k + 1 steps." 
  Exseq_k    : "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ( (trs, s),  (trs', s')) \<in> Smallstep \<and> 
               ((trs', s'), instr_set , (trt, t), k) \<in> Exec_k\<rbrakk> 
              \<Longrightarrow> ((trs, s),instr_set, (trt, t), Suc k) \<in> Exec_k" 
|
  Exterm_k   : "\<lbrakk>instr_set \<subseteq> lis;  \<not>(indom (PC s)) instr_set\<rbrakk>
              \<Longrightarrow> ((tr, s),instr_set, (tr, s), 0) \<in> Exec_k"

text {* This is shorthand for going from a single @{term smallstep} to @{term Exec_k}. *}
lemma Exsing_k   : "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ((trs, s), (trt, t)) \<in> Smallstep \<and> 
               \<not>(indom (PC t) instr_set)\<rbrakk>
              \<Longrightarrow> ((trs, s),instr_set, (trt, t), Suc 0) \<in> Exec_k"
by (meson Exseq_k Exterm_k)

inductive_set multistep_k  :: "(('s, 'e) trace_state \<times> ('s, 'e)labeled_instruction_set \<times> ('s, 'e) trace_state \<times> nat) set"
where
  --"If we reach s' from s in one step and t from s' in k steps, we can reach t from s in k + 1 steps."
  Mseq_k    : "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ( (trs, s),  (trs', s')) \<in> Smallstep \<and> 
               ((trs', s'), instr_set , (trt, t), k) \<in> multistep_k\<rbrakk> 
              \<Longrightarrow> ((trs, s),instr_set, (trt, t), Suc k) \<in> multistep_k" 
|
  Mterm_k   : "\<lbrakk>instr_set \<subseteq> lis\<rbrakk>
              \<Longrightarrow> ((tr, s),instr_set, (tr, s), 0) \<in> multistep_k"

text {* This is shorthand for going from a single @{term smallstep} to @{term multistep_k}. *}
lemma Msing_k: "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ((trs, s), (trt, t)) \<in> Smallstep\<rbrakk>
              \<Longrightarrow> ((trs, s),instr_set, (trt, t), Suc 0) \<in> multistep_k"
by (meson Mseq_k Mterm_k)

lemma Exec_lastPC: 
  assumes "((trs, s), c, (trt, t)) \<in> Exec"
  shows   "\<not> (PC t) \<in>\<^sub>l\<^sub>i\<^sub>s c"
  using assms
proof (induction rule: Exec.induct) qed auto

text {* The "out of @{term lis}" property for the final state in @{term Exec} truly is the only enrichment to @{term multistep} - otherwise they are equal. *}

lemma notIndomIsStuck: "(((trs, s),instr_set, (trt, t)) \<in> multistep \<and> \<not>(indom (PC t)) instr_set)
                         = (((trs, s),instr_set, (trt, t)) \<in> Exec)"
apply(rule iffI)
 apply(erule conjE)
 apply(subgoal_tac "\<not> indom (PC t) instr_set \<longrightarrow> ((trs, s), instr_set, trt, t) \<in> Exec")
  apply simp
 apply(erule multistep.induct)
   using liSet.Exec.Exseq liSet_axioms apply blast
 apply (simp add: Exec.Exterm)
apply(frule Exec_lastPC)
apply(rule conjI)
 apply(erule Exec.induct)
  apply auto[]
 using liSet.multistep.Mseq liSet_axioms apply blast
apply (simp add: multistep.Mterm)
apply assumption
done

lemma Exec_k_0_same: "((trs, s), c, (trt, t), 0) \<in> Exec_k \<Longrightarrow> t = s \<and> trs = trt \<and> \<not> indom (PC s) c"
by(erule Exec_k.cases, auto)

lemma multistep_k_0_same: "((trs, s), c, (trt, t), 0) \<in> multistep_k \<Longrightarrow> t = s \<and> trs = trt"
by(erule multistep_k.cases, auto)

lemma Exec_k_indom: "((trs, s), c, (trt, t), k) \<in> Exec_k \<Longrightarrow> k > 0 \<Longrightarrow> indom (PC s) c"
apply(erule Exec_k.cases, auto)
 apply(simp add: indom_def)
 apply(rule_tac x=ins in exI, simp)
done

lemma multistep_k_indom: "((trs, s), c, (trt, t), k) \<in> multistep_k \<Longrightarrow> k > 0 \<Longrightarrow> indom (PC s) c"
apply(erule multistep_k.cases, auto)
 apply(simp add: indom_def)
 apply(rule_tac x=ins in exI, simp)
done

lemma Exec_k_lastPC: "((trs, s), c, (trt, t), k) \<in> Exec_k \<Longrightarrow> \<not>  indom (PC t) c"
by(erule Exec_k.induct, auto)

text {* Having executed @{term "Suc k"} steps means we can extract a @{term Smallstep} off the front and
@{term k} steps remain. *}

lemma Exec_k_split: "((trs, s), c, (trt, t), Suc k) \<in> Exec_k \<Longrightarrow> 
\<exists>s' trs'. ((trs, s), (trs', s')) \<in> Smallstep \<and> ((trs', s'), c, (trt, t), k) \<in> Exec_k \<and> (k > 0 \<longrightarrow> indom (PC (s')) c)
      \<and> (k = 0 \<longrightarrow> \<not> indom (PC (s')) c)"
apply(erule Exec_k.cases, auto)
apply(rule_tac x=s' in exI)
apply(rule_tac x=trs' in exI)
apply simp
apply(rule conjI)
 apply(rule impI)
 apply(rule_tac s=s' in Exec_k_indom, simp, simp)
apply(rule impI)
apply simp
apply(drule Exec_k_0_same, simp)
done

text {* The same extraction applies to the bare @{term multistep} semantics. *}

lemma multistep_k_split: "((trs, s), c, (trt, t), Suc k) \<in> multistep_k \<Longrightarrow> 
\<exists>s' trs'. ((trs, s), (trs', s')) \<in> Smallstep \<and> ((trs', s'), c, (trt, t), k) \<in> multistep_k \<and> (trs' = trs \<or> (\<exists> a . trs' = trs @ [a]))"
apply (erule multistep_k.cases, auto, auto simp add: smallstep_head_growth Stos)
using smallstep_head_growth unfolding Stos by fast

text {* The operational semantics with and without @{term k} can be transformed into their counterpart. *}

lemma Exec_implies_Exec_k: " ((trs, s), code, (trs',t)) \<in> Exec \<Longrightarrow> \<exists> k. ((trs, s), code, (trs',t), k) \<in> Exec_k"
apply(erule Exec.induct)
  apply clarsimp
 apply (metis Exec_k.Exseq_k)
by (metis Exec_k.Exterm_k)

lemma multistep_implies_multistep_k: " ((trs, s), code, (trs',t)) \<in> multistep \<Longrightarrow> \<exists> k. ((trs, s), code, (trs',t), k) \<in> multistep_k"
apply(erule multistep.induct)
  apply clarsimp
 apply (metis multistep_k.Mseq_k)
apply (metis multistep_k.Mterm_k)
done

lemma Exec_k_implies_Exec: "((s,trs), code, (t,trs'), k) \<in> Exec_k \<Longrightarrow> ((s,trs), code, (t,trs')) \<in> Exec"
apply(erule Exec_k.induct)
  using Exec.Exseq apply blast
by (metis Exec.Exterm)

lemma multistep_k_implies_multistep: "((s,trs), code, (t,trs'), k) \<in> multistep_k \<Longrightarrow> ((s,trs), code, (t,trs')) \<in> multistep"
apply(erule multistep_k.induct)
  using multistep.Mseq apply blast
by (metis Mterm)

lemma Exec_k_suc: "((trs, s), c, (trt, t), Suc k) \<in> Exec_k \<Longrightarrow> indom (PC (s)) c"
apply(erule Exec_k.cases)
  apply clarsimp
 apply(simp add: indom_def)
 apply(rule_tac x="ins" in exI, blast+)
done

lemma multistep_k_suc: "((trs, s), c, (trt, t), Suc k) \<in> multistep_k \<Longrightarrow> indom (PC (s)) c"
apply(erule multistep_k.cases)
  apply clarsimp
 apply(simp add: indom_def)
 apply(rule_tac x="ins" in exI, blast+)
done

lemma Exec_k_one: "((trs, s), instrSet, (trt, t), Suc 0) \<in> Exec_k \<Longrightarrow> \<exists>l ins.
   (l, ins) \<in> instrSet \<and>
   PC (s) = l \<and>
   ((trs, s), (trt, t)) \<in> Smallstep \<and> \<not> indom (PC (t)) instrSet \<and> instrSet \<subseteq> lis"
apply(erule Exec_k.cases)
  using Exec_k_0_same apply blast+
done

lemma multistep_k_one: "((trs, s), instrSet, (trt, t), Suc 0) \<in> multistep_k \<Longrightarrow> \<exists>l ins.
   (l, ins) \<in> instrSet \<and>
   PC (s) = l \<and>
   ((trs, s), (trt, t)) \<in> Smallstep \<and> instrSet \<subseteq> lis"
apply(erule multistep_k.cases)
  using multistep_k_0_same apply blast+
done

text {* We can extend the instruction set with arbitrary instructions from lis and still find the way through
it from s to t.*}

lemma multistep_extend_lis:
  assumes "(s, c, t) \<in> multistep" and "c' \<subseteq> lis"
  shows "(s, c \<union> c', t) \<in> multistep"
proof -
  obtain trs ss trt st where
    s: "s = (trs, ss)" and
    t: "t = (trt, st)"
    by fastforce
  show ?thesis
    using assms unfolding s t
  proof (induct)
    fix instr_set l ins s trs trs' s' trt t
    assume
      subs: "instr_set \<subseteq> lis" and
      prec: "(l, ins) \<in> instr_set \<and> PC s = l \<and> ((trs, s), trs', s') \<in> Smallstep \<and> ((trs', s'), instr_set, trt, t) \<in> multistep \<and> (c' \<subseteq> lis \<longrightarrow> ((trs', s'), instr_set \<union> c', trt, t) \<in> multistep)" and
      c'subs: "c' \<subseteq> lis"
    have subs: "instr_set \<union> c' \<subseteq> lis"
      using c'subs subs by simp
    show "((trs, s), instr_set \<union> c', trt, t) \<in> multistep"
      using Mseq[OF subs] prec c'subs by fast
  next
    fix instr_set tr s
    assume
      subs: "instr_set \<subseteq> lis" and
      c'subs: "c' \<subseteq> lis"
    have subs: "instr_set \<union> c' \<subseteq> lis"
      using c'subs subs by simp
    show "((tr, s), instr_set \<union> c', tr, s) \<in> multistep"
      using Mterm[OF subs] by simp
  qed
qed

text {* We can combine two multisteps. *}

lemma multistep_seq:
  assumes
    step1: "(s, c, t) \<in> multistep" and
    step2: "(t, c', u) \<in> multistep"
  shows "(s, c \<union> c', u) \<in> multistep"
proof -
  obtain trs ss trt st tru su where
    s: "s = (trs, ss)" and
    t: "t = (trt, st)" and
    u: "u = (tru, su)"
    by fastforce
  show ?thesis
    using assms
    unfolding s t u
  proof (induct)
    fix instr_set l ins s trs trs' s' trt t
    assume
      subs: "instr_set \<subseteq> lis" and
      prec: "(l, ins) \<in> instr_set \<and> PC s = l \<and> ((trs, s), trs', s') \<in> Smallstep \<and> ((trs', s'), instr_set, trt, t) \<in> multistep \<and> (((trt, t), c', tru, su) \<in> multistep \<longrightarrow> ((trs', s'), instr_set \<union> c', tru, su) \<in> multistep)" and
      ms: "((trt, t), c', tru, su) \<in> multistep"
    hence "instr_set \<union> c' \<subseteq> lis"
      using local.multistep.induct by simp
    thus "((trs, s), instr_set \<union> c', tru, su) \<in> multistep"
      using ms prec Mseq by blast
  next
    fix instr_set tr s
    assume
      subs: "instr_set \<subseteq> lis" and
      ms: "((tr, s), c', tru, su) \<in> multistep"
    hence "instr_set \<union> c' \<subseteq> lis"
      using local.multistep.induct by simp
    thus "((tr, s), instr_set \<union> c', tru, su) \<in> multistep"
      by (simp add: ms multistep_extend_lis sup_commute) 
  qed
qed

lemma subsetlemma: "a \<in> A \<union> Be \<Longrightarrow> A  \<subseteq> lis \<Longrightarrow> Be \<subseteq> lis \<Longrightarrow> a \<in> lis"
by auto

text {* Some kind of opposite to the previous lemma. We can split any execution in some unstructured code
  into a prefix, which runs solely in some subset, and the remaining portion of the calculation in the big part. 
*}

lemma OnescinseqMultistepCombinedBoth:
  assumes a: "((trs, s), \<union>\<^sub>s\<^sub>c strInstr, (trt, t), k) \<in> multistep_k"
      and b: "\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis"
      and c: "\<union>\<^sub>s\<^sub>c strInstrSub \<subseteq> \<union>\<^sub>s\<^sub>c strInstr"
      and d: "(PC (s)) \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub)" and
      kneqzero: "k \<noteq> 0"
  shows "\<exists> (n::nat) (m::nat) trs' s' . (k = n + m \<and> (n > 0) \<and> ((trs, s) , \<union>\<^sub>s\<^sub>c strInstrSub , (trs', s'), n) \<in> multistep_k 
                                       \<and> ((trs', s') , \<union>\<^sub>s\<^sub>c strInstr , (trt, t) , m) \<in> multistep_k
                                       \<and> ((\<union>\<^sub>s\<^sub>c strInstrSub \<subset> \<union>\<^sub>s\<^sub>c strInstr \<and> m > 0) \<longrightarrow> \<not> ((PC (s')) \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub))))"
using a d kneqzero
proof (induction k arbitrary: trs s trt t  rule: nat_induct )
  case 0 thus ?case by blast
next
  case (Suc k)
  show ?case
  proof (cases rule: multistep_k.cases[of trs s "\<union>\<^sub>s\<^sub>c strInstr" trt t "Suc k"], auto)
    from Suc.prems show "((trs, s),  \<union>\<^sub>s\<^sub>c strInstr, (trt, t), Suc k) \<in> multistep_k" by blast
  next
    fix l ins trs' s'
    assume
      subs: "\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis" and
      inC: "(PC s, ins) \<in>  \<union>\<^sub>s\<^sub>c strInstr" and
      ss: "((trs, s), trs', s') \<in> Smallstep" and
      ms_k: "((trs', s'),  \<union>\<^sub>s\<^sub>c strInstr, (trt, t), k) \<in> multistep_k"
    show "\<exists>n m. Suc k = n + m \<and> 0 < n \<and> (\<exists>trs' s'. ((trs, s), \<union>\<^sub>s\<^sub>c strInstrSub, (trs', s'), n) \<in> multistep_k \<and> ((trs', s'), \<union>\<^sub>s\<^sub>c strInstr, (trt, t), m) \<in> multistep_k \<and> (\<union>\<^sub>s\<^sub>c strInstrSub \<subset> \<union>\<^sub>s\<^sub>c strInstr \<and> 0 < m \<longrightarrow> \<not> PC s' \<in>\<^sub>l\<^sub>i\<^sub>s \<union>\<^sub>s\<^sub>c strInstrSub))"
    proof (cases "k = 0")
      assume keq0: "k = 0"
      from ss Msing_k have "((trs, s),  \<union>\<^sub>s\<^sub>c strInstrSub, (trs', s'), Suc 0) \<in> multistep_k"
        by (metis (no_types, lifting) Suc.prems(2) c indom_def subs subset_trans)
      thus ?thesis using ss keq0 ms_k by force 
    next
      assume kneq0: "k \<noteq> 0"
      show ?thesis
      proof (cases "PC s' \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub)")
        assume c1: "(PC s' \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub))"
        from Suc.IH [OF ms_k c1 kneq0] obtain trs'' s'' n m
        where o1: "k = n + m \<and>
         0 < n \<and>
         ((trs', s'),  \<union>\<^sub>s\<^sub>c strInstrSub, (trs'', s''), n) \<in> multistep_k \<and>
         ((trs'', s''),  \<union>\<^sub>s\<^sub>c strInstr, (trt, t), m) \<in> multistep_k \<and>
         (\<union>\<^sub>s\<^sub>c strInstrSub \<subset> \<union>\<^sub>s\<^sub>c strInstr \<and> 0 < m \<longrightarrow> \<not> (PC s'' \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub)))" by blast
        hence sk : "Suc k = Suc n + m" by arith
        from o1 ss b Mseq_k have "((trs, s),  \<union>\<^sub>s\<^sub>c strInstrSub, (trs'', s''), Suc n) \<in> multistep_k"
        proof -
          have "\<union>\<^sub>s\<^sub>c strInstrSub \<subseteq> lis"
            using subs c by auto
          then show ?thesis
            by (meson Mseq_k Suc.prems(2) ss indom_def o1)
        qed
        with o1 sk show ?thesis by blast
      next
        assume c1: "\<not> (PC s' \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub))"
        from ss Msing_k have "((trs, s),  \<union>\<^sub>s\<^sub>c strInstrSub, (trs', s'), Suc 0) \<in> multistep_k"
          by (metis (no_types, lifting) Suc.prems(2) c indom_def subs subset_trans)
        with inC ss ms_k c1 show ?thesis by force
      qed
    qed
  qed
qed

lemma OnescinseqCombinedBoth:
  assumes a: "((trs, s), \<union>\<^sub>s\<^sub>c strInstr, (trt, t), k) \<in> Exec_k"
      and b: "\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis"
      and c: "\<union>\<^sub>s\<^sub>c strInstrSub \<subseteq> \<union>\<^sub>s\<^sub>c strInstr"
      and d: "(PC (s)) \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub)"
  shows "\<exists> (n::nat) (m::nat) trs' s' . (k = n + m \<and> (n > 0) \<and> ((trs, s) , \<union>\<^sub>s\<^sub>c strInstrSub , (trs', s'), n) \<in> Exec_k 
                                       \<and> ((trs', s') , \<union>\<^sub>s\<^sub>c strInstr , (trt, t) , m) \<in> Exec_k
                                       \<and> \<not> ((PC (s')) \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub)))"
using a d
proof (induction k arbitrary: trs s trt t  rule: nat_induct )
  case 0
  with Exec_k_0_same have "\<not> indom (PC s) (\<union>\<^sub>s\<^sub>c strInstr)" by blast
  with c not_indom_splits sup.order_iff have "\<not> (PC s \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub))" by metis
  with 0 show ?case by auto
next
  case (Suc k)
  show ?case
  proof (cases rule: Exec_k.cases[of trs s "\<union>\<^sub>s\<^sub>c strInstr" trt t "Suc k"])
    from Suc.prems show "((trs, s),  \<union>\<^sub>s\<^sub>c strInstr, (trt, t), Suc k) \<in> Exec_k" by blast
  next
    fix l ins trs' s'
    assume
      b2: "\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis" and
      b3: "
            (l, ins) \<in>  \<union>\<^sub>s\<^sub>c strInstr \<and>
            PC s = l \<and>
            ((trs, s), trs', s') \<in> Smallstep \<and>
            ((trs', s'),  \<union>\<^sub>s\<^sub>c strInstr, (trt, t), k)
            \<in> Exec_k"
    show ?case
    proof (cases "PC s' \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub)")
      assume c1: "(PC s' \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub))"
      with Suc.IH b3 obtain trs'' s'' n m
      where o1: "k = n + m \<and>
       0 < n \<and>
       ((trs', s'),  \<union>\<^sub>s\<^sub>c strInstrSub, (trs'', s''), n) \<in> Exec_k \<and>
       ((trs'', s''),  \<union>\<^sub>s\<^sub>c strInstr, (trt, t),
        m)
       \<in> Exec_k \<and>
       \<not> (PC s'' \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub))" by blast
      hence sk : "Suc k = Suc n + m" by arith
      from o1 b3 b Exseq_k have "((trs, s),  \<union>\<^sub>s\<^sub>c strInstrSub, (trs'', s''), Suc n) \<in> Exec_k"
      proof -
        have "\<union>\<^sub>s\<^sub>c strInstrSub \<subseteq> lis"
          using b2 c by auto
        then show ?thesis
          by (meson Exec_k.Exseq_k Suc.prems(2) b3 indom_def o1)
      qed
      with o1 sk show ?case by blast
    next
      assume c1: "\<not> (PC s' \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstrSub))"
      with b3 Exsing_k have "((trs, s),  \<union>\<^sub>s\<^sub>c strInstrSub, (trs', s'), Suc 0) \<in> Exec_k"
        by (metis (no_types, hide_lams) Suc.prems(2) b2 c indom_def subset_trans)
      with b3 c1 show ?case by force
    qed
  qed
qed

text {* We can append the smallstep also on the other end, and grow the execution more intuitively. *}

lemma multistep_seq_other_end_k:
  assumes
    a: "instr_set \<subseteq> lis" and
    b: "(l, ins) \<in> instr_set" and
    c: "PC s' = l" and (* in case of s' = s, it would not be guaranteed otherwise. *)
    d: "((trs', s'),  (trt, t)) \<in> Smallstep" and
    e: "((trs, s), instr_set , (trs', s'), k) \<in> multistep_k" 
  shows "((trs, s),instr_set, (trt, t), Suc k) \<in> multistep_k"
using e b c d
proof (induct)
    fix instr_set la insa s trs trs' s' trta ta k
    assume
      aa1: "instr_set \<subseteq> lis" and
      aa2: "(la, insa) \<in> instr_set \<and> PC s = la \<and> ((trs, s), trs', s') \<in> Smallstep \<and>
       ((trs', s'), instr_set, (trta, ta), k) \<in> multistep_k \<and>
       
       ((l, ins) \<in> instr_set \<longrightarrow>PC ta = l \<longrightarrow> ((trta, ta), trt, t) \<in> Smallstep \<longrightarrow>
        ((trs', s'), instr_set, (trt, t), Suc k) \<in> multistep_k)" and
      aa3: "PC ta = l" and
      aa4: "((trta, ta), trt, t) \<in> Smallstep" and
      aa5: "instr_set \<subseteq> lis" and
      aa6: "(l, ins) \<in> instr_set"
    have a1: "((trs', s'), instr_set, (trt, t), Suc k) \<in> multistep_k"
      using aa2 aa3 aa4 aa5 aa6
      by blast
    have a2: "((trs, s), trs', s') \<in> Smallstep"
      using aa2 by fast
    (* have a3: "((trs', s'), instr_set, (trta, ta), k) \<in> multistep_k"
      using aa2 by fast *)
    from Mseq_k[OF aa1, of la insa s trs trs' s' trt t "Suc k"]
    show "((trs, s), instr_set, (trt, t), Suc (Suc k)) \<in> multistep_k"
      using a1 a2 aa2[THEN cj2_1] aa2[THEN cj3_2]
      by blast
  next
    show "\<And>instr_set tr s.
       instr_set \<subseteq> lis \<Longrightarrow>
       PC s = l \<Longrightarrow>
       ((tr, s), trt, t) \<in> Smallstep \<Longrightarrow>
       instr_set \<subseteq> lis \<Longrightarrow>
       (l, ins) \<in> instr_set \<Longrightarrow>
       ((tr, s), instr_set, (trt, t), Suc 0)
       \<in> multistep_k"
    by (simp add: Msing_k)
qed

lemma multistep_seq_other_end:
  assumes
    a: "instr_set \<subseteq> lis" and
    b: "(l, ins) \<in> instr_set" and
    c: "PC s' = l" and (* in case of s' = s, it would not be guaranteed otherwise. *)
    d: "((trs', s'),  (trt, t)) \<in> Smallstep" and
    e: "((trs, s), instr_set , (trs', s')) \<in> multistep" 
  shows "((trs, s),instr_set, (trt, t)) \<in> multistep"
using multistep_seq_other_end_k[OF a b c d, of trs s]
  multistep_implies_multistep_k[OF e] multistep_k_implies_multistep[of trs s instr_set trt t]
by fast

text {* For any calculation, we can extract each smallstep as we want. *}

lemma multistep_k_split2:
  assumes
    a: "((tr', s'), instrSet, (tr, s), k) \<in> multistep_k" and
    b: "instrSet \<subseteq> lis" and
    c: "Suc (m + n) = k"
  shows "\<exists> tra' sa' tra sa .
          ((tr', s'), instrSet, (tra', sa'), m) \<in> multistep_k
          \<and> ((tra', sa'), (tra, sa)) \<in> Smallstep
          \<and> ((tra, sa), instrSet, (tr, s), n) \<in> multistep_k \<and>
          indom (PC sa') instrSet \<and> ((n > 0 \<or> indom (PC s) instrSet) \<longrightarrow> indom (PC sa) instrSet) 
          "
using a c
proof(induct m arbitrary: tr' s' k)
  fix tr' s' k
  assume 
    aa1: "((tr', s'), instrSet, (tr, s), k) \<in> multistep_k" and
    aa2: "Suc (0 + n) = k"
  obtain trs' ss' where
    ao1: "((tr', s'), trs', ss') \<in> Smallstep" and
    ao2: "((trs', ss'), instrSet, (tr, s), n) \<in> multistep_k"
    using multistep_k_split[OF aa1[unfolded aa2[symmetric, simplified]]]
    by blast
  moreover
  have "((tr', s'), instrSet, (tr', s'), 0) \<in> multistep_k"
    by (simp add: b multistep_k.Mterm_k)
  moreover
  have "indom (PC s') instrSet"
    using aa1 aa2 multistep_k_suc by blast
  moreover
  have "(n > 0 \<or> indom (PC s) instrSet) \<longrightarrow> indom (PC ss') instrSet"
    using ao2 multistep_k_0_same multistep_k_indom by blast
  ultimately
  show "\<exists>tra' sa' tra sa.
          ((tr', s'), instrSet, (tra', sa'), 0) \<in> multistep_k \<and>
          ((tra', sa'), tra, sa) \<in> Smallstep \<and> ((tra, sa), instrSet, (tr, s), n) \<in> multistep_k
           \<and> (indom (PC sa') instrSet) \<and> ((n > 0 \<or> indom (PC s) instrSet) \<longrightarrow> indom (PC sa) instrSet)"
    by blast
next
  fix m tr' s' k
  assume
    aa1: "\<And>tr' s' k.
           ((tr', s'), instrSet, (tr, s), k) \<in> multistep_k \<Longrightarrow>
           Suc (m + n) = k \<Longrightarrow>
           \<exists>tra' sa' tra sa.
              ((tr', s'), instrSet, (tra', sa'), m) \<in> multistep_k \<and>
              ((tra', sa'), tra, sa) \<in> Smallstep \<and>
              ((tra, sa), instrSet, (tr, s), n) \<in> multistep_k \<and>
              indom (PC sa') instrSet \<and>
              (0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa) instrSet)" and
    aa2: "((tr', s'), instrSet, (tr, s), k) \<in> multistep_k" and
    aa3: "Suc (Suc m + n) = k"
  obtain tru' su' where
    ao1: "((tr', s'), tru', su') \<in> Smallstep" and
    ao2: "((tru', su'), instrSet, (tr, s), Suc m + n) \<in> multistep_k"
    using multistep_k_split[OF aa2[unfolded aa3[symmetric]]]
    by blast
  obtain tra' sa' tra sa where
    ao3: "((tru', su'), instrSet, (tra', sa'), m) \<in> multistep_k" and
    ao4: "((tra', sa'), tra, sa) \<in> Smallstep" and
    ao5: "((tra, sa), instrSet, (tr, s), n) \<in> multistep_k" and
    ao10: "indom (PC sa') instrSet" and
    ao11: "0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa) instrSet"
    using aa1[OF ao2]
    by auto
  obtain ins where
    ao7: "(PC s', ins) \<in> instrSet"
  proof -
    assume "\<And>ins. (PC s', ins) \<in> instrSet \<Longrightarrow> thesis"
    then show ?thesis
      by (meson aa2[unfolded aa3[symmetric]] indom_def multistep_k_suc)
  qed
  have a1: "((tr', s'), instrSet, (tra', sa'), Suc m) \<in> multistep_k"
    using Mseq_k[OF b, of "PC s'" ins s' tr' tru' su' tra' sa' m]
      ao7 ao3 ao1
    by blast
  thus "\<exists>tra' sa' tra sa.
          ((tr', s'), instrSet, (tra', sa'), Suc m) \<in> multistep_k \<and>
          ((tra', sa'), tra, sa) \<in> Smallstep \<and> ((tra, sa), instrSet, (tr, s), n) \<in> multistep_k \<and>
          (indom (PC sa') instrSet) \<and> (0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa) instrSet)"
    using ao4 ao5 ao10 ao11
    by blast 
qed

text {* The trace in multistep can only grow at most by k, if we do k steps. *}

lemma multistep_k_growing_trace:
  assumes
    a: "((tr', s'), instrSet, (tr, s), k) \<in> multistep_k"
  shows "\<exists> r . k \<ge> length r \<and> tr = tr' @ r"
using a
proof (induct k)
  fix instr_set l ins s trs trs' s' trt t k
  assume
    aa1: "instr_set \<subseteq> lis" and
    aa2: "(l, ins) \<in> instr_set \<and>
       PC s = l \<and>
       ((trs, s), trs', s') \<in> Smallstep \<and>
       ((trs', s'), instr_set, (trt, t), k) \<in> multistep_k \<and>
       (\<exists>r. length r \<le> k \<and> trt = trs' @ r)"
  obtain r where
    ao1: "length r \<le> k" and
    ao2: "trt = trs' @ r"
    using aa2
    by blast
  show "\<exists>r. length r \<le> Suc k \<and> trt = trs @ r"
    using smallstep_head_growth[OF aa2[THEN cj4_3, unfolded Stos]]
      ao1 ao2
    by force
qed blast

lemma multistep_growing_trace:
  assumes "((tr', s'), instrSet, (tr, s)) \<in> multistep"
  shows "\<exists> r . tr = tr' @ r"
using multistep_implies_multistep_k[OF assms] multistep_k_growing_trace
by blast
  
text {* We also can "remove" calculations to get to the point, where we have added some event to the trace.
This is kind of backstepping. *}

lemma multistep_k_prefix_closed':
  assumes
    a: "instrSet \<subseteq> lis" and
    b: "((tr, s'), instrSet, (tr @ r @ [a] @ l, s), k) \<in> multistep_k"
  shows "\<exists> sa sa' m n . Suc (m + n) = k \<and> (* This cannot be put into the assumptions *)
    ((tr, s'), instrSet, (tr @ r, sa), m) \<in> multistep_k \<and>
    ((tr @ r, sa), (tr @ r @ [a], sa')) \<in> Smallstep \<and> 
    ((tr @ r @ [a], sa'), instrSet, (tr @ r @ [a] @ l, s), n) \<in> multistep_k \<and>
    indom (PC sa) instrSet \<and> (0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa') instrSet)"
using b
proof (induct k arbitrary: s s' tr l r a rule: nat.induct)
  fix nat s l r a s' tr
  assume
    aa1: "\<And>s s' tr l r a. ((tr, s'), instrSet, (tr @ r @ [a] @ l, s), nat) \<in> multistep_k \<Longrightarrow>
               \<exists>sa sa' m n.
                  Suc (m + n) = nat \<and>
                  ((tr, s'), instrSet, (tr @ r, sa), m) \<in> multistep_k \<and>
                  ((tr @ r, sa), tr @ r @ [a], sa') \<in> Smallstep \<and>
                  ((tr @ r @ [a], sa'), instrSet, (tr @ r @ [a] @ l, s), n) \<in> multistep_k \<and>
                  indom (PC sa) instrSet \<and> (0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa') instrSet)" and
    aa2: "((tr, s'), instrSet, (tr @ r @ [a] @ l, s), Suc nat) \<in> multistep_k"
  obtain tru su tru' su' where
    ao1: "((tr, s'), instrSet, (tru, su), nat) \<in> multistep_k" and
    ao2: "((tru, su), tru', su') \<in> Smallstep" and
    ao3: "((tru', su'), instrSet, (tr @ r @ [a] @ l, s), 0) \<in> multistep_k" and
    ao4: "indom (PC su) instrSet" and
    ao5: "0 < (0::nat) \<or> indom (PC s) instrSet \<longrightarrow> indom (PC su') instrSet"
    using multistep_k_split2[OF aa2 a, of nat 0]
    unfolding add_0_right
    by blast
  obtain ins where ao4: "(PC su, ins) \<in> instrSet"
    using ao4 unfolding indom_def by blast
  obtain ins' where ao6: "(PC s', ins') \<in> instrSet"
    using multistep_k_indom[OF aa2 Nat.zero_less_Suc]
    unfolding indom_def
    by blast
  have
    a1: "tru' = tr @ r @ [a] @ l \<and> su' = s"
    using multistep_k_0_same[OF ao3]
    by meson
  let ?G = "\<exists>sa sa' m n.
          Suc (m + n) = Suc nat \<and>
          ((tr, s'), instrSet, (tr @ r, sa), m) \<in> multistep_k \<and>
          ((tr @ r, sa), tr @ r @ [a], sa') \<in> Smallstep \<and> ((tr @ r @ [a], sa'), instrSet, (tr @ r @ [a] @ l, s), n) \<in> multistep_k \<and>
          indom (PC sa) instrSet \<and> (0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa') instrSet)"
  {
    assume nat0: "nat = 0"
    have b1: "tru = tr \<and> su = s'"
      using multistep_k_0_same[OF ao1[unfolded nat0]]
      by fast
    have b2: "((tr, s'), tr @ r @ [a] @ l, s) \<in> Smallstep"
      using ao2[unfolded b1[THEN cj2_1] b1[THEN cj2_2] a1[THEN cj2_1] a1[THEN cj2_2]]
      by blast
    have b3: "r = [] \<and> l = []"
      using smallstep_head_growth2[OF b2[unfolded Stos]]
      .
    have ?G
      unfolding b3[THEN cj2_1] b3[THEN cj2_2] append_Nil append_Nil2
      using b2 nat0 Mterm_k[OF a]
      using \<open>l = []\<close> \<open>r = []\<close> add_Suc_right append_Nil append_self_conv length_Cons list.size(3) list.size(4)
      by (metis (no_types, hide_lams) ao6 indom_def not_less0)
      (* by auto *)
  }
  moreover
  {
    assume ba1: "l = []" and ba2: "nat \<noteq> 0"
    obtain nat' where nat': "nat = Suc nat'"
      using ba2
      using not0_implies_Suc by blast 
    have "tru = tr @ r @ [a] \<or> tru = tr @ r"
      using a1[unfolded ba1]
      using smallstep_head_growth[OF ao2[unfolded Stos]]
      by fastforce
    moreover
    {
      assume ca1: "tru = tr @ r @ [a]"
      hence c1: "tru = tr @ r @ [a] @ []"
        by simp
      obtain sa sa' m n where
        co1: "Suc (m + n) = nat" and
        co2: "((tr, s'), instrSet, (tr @ r, sa), m) \<in> multistep_k" and
        co3: "((tr @ r, sa), tr @ r @ [a], sa') \<in> Smallstep" and
        co4: "((tr @ r @ [a], sa'), instrSet, (tr @ r @ [a] @ [], su), n) \<in> multistep_k" and
        co5: "indom (PC sa) instrSet" and
        co6: "(0 < n \<or> indom (PC su) instrSet \<longrightarrow> indom (PC sa') instrSet)"
        using aa1[OF ao1[unfolded c1]]
        by blast
      from multistep_seq_other_end_k[OF a ao4 _ ao2[unfolded c1] co4]
        have "((tr @ r @ [a], sa'), instrSet, (tru', su'), Suc n) \<in> multistep_k"
        by blast
      moreover
      have "0 < Suc n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa') instrSet"
        using calculation multistep_k_suc by blast
      ultimately
      have ?G
        using a1 add_Suc_right[of m n] co1 co2 co3 co5 by blast
    }
    moreover
    {
      assume ca1: "tru = tr @ r"
      obtain trx sx where
        co1: "((tr, s'), trx, sx) \<in> Smallstep" and
        co2: "((trx, sx), instrSet, (tr @ r, su), nat') \<in> multistep_k"
        using multistep_k_split[OF ao1[unfolded nat'], unfolded ca1]
        by blast
      obtain lx where co3: "trx = tr @ lx"
        using co1 Smallstep_rev append_Nil2 by blast
      obtain rx where co4: "tr @ lx @ rx = tr @ r"
        using append_assoc co3 multistep_k_growing_trace[OF co2]
        by fastforce
      have c1: "((trx, sx), instrSet, (trx @ rx @ [a] @ [], s), nat) \<in> multistep_k"
        using multistep_seq_other_end_k[OF a ao4 _ ao2 co2[unfolded ca1[symmetric]]]
        unfolding a1[THEN cj2_1] a1[THEN cj2_2] ba1 append_Nil2
        unfolding co3 nat'[symmetric]
        using co4[symmetric]
        by fastforce
      obtain m n sp sq where
        co5: "Suc (m + n) = nat" and
        co6: "((trx, sx), instrSet, (trx @ rx, sp), m) \<in> multistep_k" and
        co7: "((trx @ rx, sp), trx @ rx @ [a], sq) \<in> Smallstep" and
        co8: "((trx @ rx @ [a], sq), instrSet, (trx @ rx @ [a] @ [], s), n) \<in> multistep_k" and
        co9: "indom (PC sp) instrSet" and
        co10: "0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sq) instrSet"
        using aa1[OF c1]
        by fast
      have c2: "((tr, s'), instrSet, (trx @ rx, sp), Suc m) \<in> multistep_k"
        using Mseq_k[OF a, of "PC s'" ins' s' tr trx sx "trx @ rx" sp m]
          co1 a1 co6 ao6
        by blast
      have ?G
        using c2 co7 co8 co5 co4 co3 co9 co10
        using ba1 add_Suc[of m n]
        by force
    }
    ultimately
    have "?G"
      by fast
  }
  moreover
  {
    assume ba1: "l \<noteq> []" and ba2: "nat \<noteq> 0"
    obtain l1 l2 where
      bo1: "((tr @ r @ [a] @ l1, su), tr @ r @ [a] @ l1 @ l2, su') \<in> Smallstep" and
      bo2: "l = l1 @ l2" and
      bo3: "tru = tr @ r @ [a] @ l1"
      using smallstep_head_growth[OF ao2[unfolded a1[THEN cj2_1] Stos]]
        smallstep_rev[OF ao2[unfolded a1[THEN cj2_1] Stos]]
    proof -
      assume c1: "\<And>l1 l2. \<lbrakk>((tr @ r @ [a] @ l1, su), tr @ r @ [a] @ l1 @ l2, su') \<in> Smallstep; l = l1 @ l2; tru = tr @ r @ [a] @ l1\<rbrakk> \<Longrightarrow> thesis"
      have "\<And>p ps psa. (p # ps) @ psa \<noteq> []"
        by blast
      then have f2: "tr @ r @ [a] @ butlast l = tru \<or> tr @ r @ [a] @ l = tru \<or> r @ [a] @ l = []"
        using ao2[unfolded a1[THEN cj2_1]]        
        by (metis (no_types) Stos smallstep_head_growth append_Nil2 ba1 butlast.simps(2) butlast_append)
      have "r @ [a] @ l \<noteq> []"
        by force
      then show ?thesis
        using f2 c1 ao2[unfolded a1[THEN cj2_1]]
        by (metis (no_types) append_Nil2 ba1 butlast.simps(2) butlast_append list.distinct(1) rev_exhaust)
    qed
    obtain sp sq m n where
      bo4: "Suc (m + n) = nat" and
      bo5: "((tr, s'), instrSet, (tr @ r, sp), m) \<in> multistep_k" and
      bo6: "((tr @ r, sp), tr @ r @ [a], sq) \<in> Smallstep" and
      bo7: "((tr @ r @ [a], sq), instrSet, (tr @ r @ [a] @ l1, su), n) \<in> multistep_k" and
      bo8: "indom (PC sp) instrSet" and
      bo9: "(0 < n \<or> indom (PC su) instrSet \<longrightarrow> indom (PC sq) instrSet)"
      using aa1[OF ao1[unfolded bo3]]
      by blast
    have b1: "((tr @ r @ [a], sq), instrSet, (tru', su'), Suc n) \<in> multistep_k"
      using multistep_seq_other_end_k[OF a ao4 _ ao2 bo7[unfolded bo3[symmetric]]]
      by blast
    have ?G
      using b1 bo4 bo5 bo6 bo8 bo9
      unfolding a1[THEN cj2_1] a1[THEN cj2_2]
      using add_Suc_right multistep_k_suc by blast
  }
  ultimately
  show ?G
    by blast
next
  show "\<And>s s' tr l r a.
       ((tr, s'), instrSet, (tr @ r @ [a] @ l, s), 0) \<in> multistep_k \<Longrightarrow>
       \<exists>sa sa' m n.
          Suc (m + n) = 0 \<and>
          ((tr, s'), instrSet, (tr @ r, sa), m) \<in> multistep_k \<and>
          ((tr @ r, sa), tr @ r @ [a], sa') \<in> Smallstep \<and>
          ((tr @ r @ [a], sa'), instrSet, (tr @ r @ [a] @ l, s), n) \<in> multistep_k \<and>
          indom (PC sa) instrSet \<and> (0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa') instrSet)"
    using multistep_k_0_same
    by fast
qed

lemma multistep_prefix_closed':
  assumes
    subs: "instrSet \<subseteq> lis" and
    ms: "((tr, s'), instrSet, (tr @ r @ [a] @ l, s)) \<in> multistep"
  shows "\<exists> sa sa' .
    ((tr, s'), instrSet, (tr @ r, sa)) \<in> multistep \<and>
    ((tr @ r, sa), (tr @ r @ [a], sa')) \<in> Smallstep \<and> 
    ((tr @ r @ [a], sa'), instrSet, (tr @ r @ [a] @ l, s)) \<in> multistep \<and>
    indom (PC sa) instrSet \<and> (s \<noteq> sa' \<or> l \<noteq> [] \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa') instrSet)"
proof -
  obtain k where ms_k: "((tr, s'), instrSet, (tr @ r @ [a] @ l, s), k) \<in> multistep_k"
    using ms multistep_implies_multistep_k by blast
  obtain sa sa' m n where
    k: "k = Suc (m + n)" and
    ms_m: "((tr, s'), instrSet, (tr @ r, sa), m) \<in> multistep_k" and
    ss: "((tr @ r, sa), tr @ r @ [a], sa') \<in> Smallstep" and
    ms_n: "((tr @ r @ [a], sa'), instrSet, (tr @ r @ [a] @ l, s), n) \<in> multistep_k" and
    pc_in: "PC sa \<in>\<^sub>l\<^sub>i\<^sub>s instrSet" and
    imp_pc: "(0 < n \<or> PC s \<in>\<^sub>l\<^sub>i\<^sub>s instrSet \<longrightarrow> PC sa' \<in>\<^sub>l\<^sub>i\<^sub>s instrSet)"
    using multistep_k_prefix_closed'[OF subs ms_k] by blast
  moreover have "s \<noteq> sa' \<longrightarrow> indom (PC sa') instrSet"
    using ms_n multistep_k_0_same gr0I imp_pc by auto
  moreover have "l \<noteq> [] \<longrightarrow> indom (PC sa') instrSet"
    using ms_n multistep_k_0_same imp_pc append_self_conv same_append_eq by fastforce
  moreover have "indom (PC s) instrSet \<longrightarrow> indom (PC sa') instrSet"
    using imp_pc by fastforce
  ultimately show ?thesis using multistep_k_implies_multistep by blast
qed

text {* A more restricted formulation of the above:
  For any trace being prefix of tr and having tr' as suffix, we can extract intermediate
  states, where this trace appears, thus being prefix closed.
*}

lemma multistep_k_prefix_closed:
  assumes
    a: "instrSet \<subseteq> lis" and
    b: "((tr', s'), instrSet, (tr' @ r @ l, s), k) \<in> multistep_k"
  shows "
    (\<exists> sa m n . m + n = k \<and>
    ((tr', s'), instrSet, (tr' @ r, sa), m) \<in> multistep_k \<and>
    ((tr' @ r, sa), instrSet, (tr' @ r @ l, s), n) \<in> multistep_k \<and>
    (0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa) instrSet))"
proof -
  let ?G = "\<exists>sa m n. m + n = k \<and>
                ((tr', s'), instrSet, (tr' @ r, sa), m) \<in> multistep_k \<and>
                ((tr' @ r, sa), instrSet, (tr' @ r @ l, s), n) \<in> multistep_k \<and>
                (0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa) instrSet)"
  {
    assume ba1: "l = [] \<and> r = []"
    have ?G
      using Mterm_k[OF a, of "tr' @ l @ r" s] b
      unfolding ba1[THEN cj2_1] ba1[THEN cj2_2] append_Nil append_Nil2
      by force
  }
  moreover
  {
    assume ba1: "l \<noteq> []"
    obtain a l' where bo1: "l = [a] @ l'"
      using ba1
      by (metis append_Cons append_Nil neq_Nil_conv)
    have tmp: "((tr', s'), instrSet, (tr' @ r @ [a] @ l', s), k) \<in> multistep_k"
      using b[unfolded bo1]
      by fastforce
    obtain sa sa' m n where
      bo2: "Suc (m + n) = k" and
      bo3: "((tr', s'), instrSet, (tr' @ r, sa), m) \<in> multistep_k" and
      bo4: "((tr' @ r, sa), tr' @ r @ [a], sa') \<in> Smallstep" and
      bo5: "((tr' @ r @ [a], sa'), instrSet, (tr' @ r @ [a] @ l', s), n) \<in> multistep_k" and
      bo6: "indom (PC sa) instrSet" and
      bo7: "0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa') instrSet"
      using multistep_k_prefix_closed'[OF a tmp]
      by blast
    obtain insa where bo8: "(PC sa, insa) \<in> instrSet"
      using bo6 unfolding indom_def
      by blast
    have b1: "((tr' @ r, sa), instrSet, (tr' @ r @ l, s), Suc n) \<in> multistep_k"
      using Mseq_k[OF a, of "PC sa" insa sa "tr' @ r" "tr' @ r @ [a]" sa' "tr' @ r @ [a] @ l'" s n]
        bo8 bo4 bo5
      unfolding bo1[symmetric]
      by fast
    obtain n' where bo9: "n' = Suc n" by blast
    have b2: "m + n' = k" using bo9 bo2 by simp
    have b1': "((tr' @ r, sa), instrSet, (tr' @ r @ l, s), n') \<in> multistep_k"
      using b1[unfolded bo9[symmetric]] .
    have ?G
      using 
        b1' b2
        bo7 bo3 bo6
      by blast
  }
  moreover
  {
    assume ba1: "r \<noteq> []"
    obtain a r' where bo1: "r = r' @ [a]"
      using ba1
      by (metis append_butlast_last_id)
    have tmp: "((tr', s'), instrSet, (tr' @ r' @ [a] @ l, s), k) \<in> multistep_k"
      using b[unfolded bo1]
      by fastforce
    obtain sa sa' m n where
      bo2: "Suc (m + n) = k" and
      bo3: "((tr', s'), instrSet, (tr' @ r', sa), m) \<in> multistep_k" and
      bo4: "((tr' @ r', sa), tr' @ r' @ [a], sa') \<in> Smallstep" and
      bo5: "((tr' @ r' @ [a], sa'), instrSet, (tr' @ r' @ [a] @ l, s), n) \<in> multistep_k" and
      bo6: "indom (PC sa) instrSet" and
      bo7: "0 < n \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa') instrSet"
      using multistep_k_prefix_closed'[OF a tmp]
      by blast
    obtain insa where bo8: "(PC sa, insa) \<in> instrSet"
      using bo6 unfolding indom_def
      by blast
    obtain m' where bo9: "m' = Suc m" by blast
    have b1: "((tr', s'), instrSet, (tr' @ r, sa'), m') \<in> multistep_k"
      using multistep_seq_other_end_k[OF a bo8 _ bo4 bo3]
      unfolding bo1[symmetric] bo9[symmetric]
      by fast
    have b2: "m' + n = k" using bo9 bo2 by simp
    have ?G
      using b1 b2 bo7 bo5 bo1[symmetric]
      by fastforce
  }
  ultimately
  show ?G
    by fast
qed

lemma multistep_prefix_closed:
  assumes
    subs: "instrSet \<subseteq> lis" and
    ms: "((tr', s'), instrSet, (tr' @ r @ l, s)) \<in> multistep"
  shows " \<exists> sa .
    (((tr', s'), instrSet, (tr' @ r, sa)) \<in> multistep \<and>
    ((tr' @ r, sa), instrSet, (tr' @ r @ l, s)) \<in> multistep \<and>
    (s \<noteq> sa \<or> l \<noteq> [] \<or> indom (PC s) instrSet \<longrightarrow> indom (PC sa) instrSet))"
proof -
  obtain k where ms_k: "((tr', s'), instrSet, (tr' @ r @ l, s), k) \<in> multistep_k"
    using ms multistep_implies_multistep_k by blast
  obtain sa m n where
    k: "k = m + n" and
    ms_m: "((tr', s'), instrSet, (tr' @ r, sa), m) \<in> multistep_k" and
    ms_n: "((tr' @ r, sa), instrSet, (tr' @ r @ l, s), n) \<in> multistep_k" and
    imp_pc: "(0 < n \<or> PC s \<in>\<^sub>l\<^sub>i\<^sub>s instrSet \<longrightarrow> PC sa \<in>\<^sub>l\<^sub>i\<^sub>s instrSet)"
    using multistep_k_prefix_closed[OF subs ms_k] by blast
  moreover have "s \<noteq> sa \<longrightarrow> indom (PC sa) instrSet"
    using ms_n multistep_k_0_same gr0I imp_pc by auto
  moreover have "l \<noteq> [] \<longrightarrow> indom (PC sa) instrSet"
    using ms_n multistep_k_0_same imp_pc append_self_conv same_append_eq by fastforce
  moreover have "indom (PC s) instrSet \<longrightarrow> indom (PC sa) instrSet"
    using imp_pc by fastforce
  ultimately show ?thesis using multistep_k_implies_multistep by blast
qed

lemma Smallstep_prefix_exchange:
  assumes "((tr, sa'), tr @ new, s) \<in> Smallstep"
  shows "((tr', sa'), tr' @ new, s) \<in> Smallstep"
using smallstep_prefix_exchange assms
unfolding Stos by fast

lemma multistep_k_prefix_exchange:
  assumes
    subs: "instrSet \<subseteq> lis" and
    ms_k: "((tr, s'), instrSet, (tr @ new, s), k) \<in> multistep_k"
  shows "((tr', s'), instrSet, (tr' @ new, s), k) \<in> multistep_k"
using ms_k
proof (induct k arbitrary: new s rule: nat_induct)
  show "\<And>new s. ((tr, s'), instrSet, (tr @ new, s), 0) \<in> multistep_k \<Longrightarrow> ((tr', s'), instrSet, (tr' @ new, s), 0) \<in> multistep_k"
    using multistep_k_0_same Mterm_k[OF subs] by fastforce
next
  fix n new s
  assume
    IH: "\<And>new s. ((tr, s'), instrSet, (tr @ new, s), n) \<in> multistep_k \<Longrightarrow> ((tr', s'), instrSet, (tr' @ new, s), n) \<in> multistep_k" and
    ms_sucn: "((tr, s'), instrSet, (tr @ new, s), Suc n) \<in> multistep_k"
  obtain tra' sa' where
    ms_n: "((tr, s'), instrSet, (tra', sa'), n) \<in> multistep_k" and
    ss: "((tra', sa'), tr @ new, s) \<in> Smallstep" and
    pcsa': "PC sa' \<in>\<^sub>l\<^sub>i\<^sub>s instrSet" and
    pcimp: "(PC s \<in>\<^sub>l\<^sub>i\<^sub>s instrSet \<longrightarrow> PC s \<in>\<^sub>l\<^sub>i\<^sub>s instrSet)"
    using multistep_k_split2[OF ms_sucn subs, of n 0, simplified]
      multistep_k_0_same
    by blast
  obtain new' where tra': "tra' = tr @ new'" using ms_n multistep_k_growing_trace
    by blast
  have ms_n: "((tr', s'), instrSet, (tr' @ new', sa'), n) \<in> multistep_k"
    using IH [OF ms_n[unfolded tra']] .
  have new: "new' = new \<or> (\<exists> e . new = new' @ [e])" using ss[unfolded tra']
    using Smallstep_rev by force
  have ss: "((tr' @ new', sa'), tr' @ new, s) \<in> Smallstep"
    using new
  proof
    assume eq: "new' = new"
    have ss: "((tr @ new, sa'), (tr @ new) @ [], s) \<in> Smallstep"
      using ss[unfolded tra' eq] by simp
    show ?thesis
      using Smallstep_prefix_exchange[OF ss, of "tr' @ new"]
      unfolding append_Nil2 eq .
  next
    assume neq: "\<exists>e. new = new' @ [e]"
    then obtain e where new: "new = new' @ [e]" by blast
    have ss: "((tr @ new', sa'), (tr @ new') @ [e], s) \<in> Smallstep"
      using ss[unfolded tra' new] by simp
    show ?thesis
      using Smallstep_prefix_exchange[OF ss, of "tr' @ new'"]
      unfolding append_Nil2 new by simp
  qed
  show "((tr', s'), instrSet, (tr' @ new, s), Suc n) \<in> multistep_k"
    using subs pcsa' ss tra' unfolding indom_def
    using multistep_seq_other_end_k[OF subs _  _ ss[unfolded tra'] ms_n[unfolded tra']] by fast
qed

lemma multistep_prefix_exchange:
  assumes
    subs: "instrSet \<subseteq> lis" and
    ms_k: "((tr, s'), instrSet, (tr @ new, s)) \<in> multistep"
  shows "((tr', s'), instrSet, (tr' @ new, s)) \<in> multistep"
using multistep_k_prefix_exchange multistep_implies_multistep_k multistep_k_implies_multistep assms
by meson

end --"locale liSet"

end