theory small_step_non_locale
  imports Main definitions
begin

section {* Smallstep and Multistep definitions *}

text {* In this part, the smallstep and multistep relations will be introduced. Also,
an enriched version of multistep is introduced: Exec. This is as multistep, but the final
state points out of the instruction set, thus we cannot execute a further instruction with
this state. *}

text {* The raw definition of the @{term smallstep} relation. This relates with the execution of a single instruction.
This instruction will be searched out of the applied @{term lis}, according to the current states' program counter. *}

inductive_set smallstep :: "('s, 'e) labeled_instruction_set \<Rightarrow> (('s, 'e) trace_state \<times> ('s, 'e) trace_state) set"
for lis
where
  --"Execute a nondeterministic function, assign any of the results to the new state and increment the program counter."
  sdof   :"\<lbrakk>(l, do f) \<in> lis; x \<in> f (R s); PC s = l; t = s\<lparr>R:=x, PC:=PC s + 1\<rparr>\<rbrakk>
           \<Longrightarrow> ((tr, s), (tr, t)) \<in> smallstep(lis)" 
|
  --"Branch to m, if the condition is fulfilled."
  scbrt  :"\<lbrakk>(l, cbr b m n) \<in> lis; PC s = l; b (R s);  t=s\<lparr>PC := m\<rparr>\<rbrakk>
           \<Longrightarrow> ((tr,s),(tr,t)) \<in> smallstep(lis)"
|
  --"Branch to n, if the condition is not fulfilled"
  scbrf  :"\<lbrakk>(l, cbr b m n) \<in> lis; PC s = l; \<not> (b (R s));  t=s\<lparr>PC := n\<rparr>\<rbrakk>
           \<Longrightarrow> ((tr,s),(tr,t)) \<in> smallstep(lis)"
|
  -- "Choose nondeterministically one of the possible communication events, extend the trace,
      update the current state with the deterministic function and increment the program counter."
  scomm  :"\<lbrakk> (l, comm ef f) \<in> lis; e \<in> ef (R s); PC s = l; 
             trt = trs @ [e]; t = s\<lparr>R := f (R s) e, PC:=PC s + 1\<rparr>\<rbrakk>
           \<Longrightarrow> ((trs, s), (trt, t)) \<in> smallstep(lis)"
           
text {* Reverse smallstep for easier reasoning. *}

lemma smallstep_rev: "((trs, s), (trt, t)) \<in> smallstep lis \<Longrightarrow>

(\<exists>l f x. (l, do f) \<in> lis \<and> x \<in> f (R s) \<and> PC s = l \<and> 
                    t = s\<lparr>R:=x, PC:=PC s + 1\<rparr>) \<and> trs=trt
\<or>
(\<exists>l b m n. (l,cbr b m n) \<in> lis \<and> PC s = l \<and> b (R s) \<and> t = s\<lparr>PC := m\<rparr>) \<and> trs=trt
\<or>
(\<exists>l b m n. (l,cbr b m n) \<in> lis \<and> PC s = l \<and> \<not> (b (R s)) \<and> t = s\<lparr>PC := n\<rparr>) \<and> trs=trt
\<or>
(\<exists>l ef f e.  (l, comm ef f) \<in> lis \<and> e \<in> ef (R s) \<and> PC s = l \<and> 
                    trt = trs@[e] \<and> t = s\<lparr>R := f (R s) e, PC:=PC s + 1\<rparr>)
"
proof (erule smallstep.cases, fast+) qed

text {* Exec is a relation containing all @{term multistep} executions, whose final state's program counter is outside of lis. *}

inductive_set exec  :: "('s, 'e) labeled_instruction_set \<Rightarrow> (('s, 'e) trace_state \<times> ('s, 'e)labeled_instruction_set \<times> ('s, 'e) trace_state) set"
for lis
where
  --"Prepend the smallstep."
  exseq    : "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ((trs, s),  (trs', s')) \<in> smallstep(lis) \<and> 
               ((trs', s'), instr_set , (trt, t)) \<in> exec(lis)\<rbrakk> 
              \<Longrightarrow> ((trs, s),instr_set, (trt, t)) \<in> exec(lis)" 
|  
  --"All pairs outside of the instruction set are in exec."
  exterm   : "\<lbrakk>instr_set \<subseteq> lis;  \<not>(indom (PC s)) instr_set\<rbrakk>
              \<Longrightarrow> ((tr, s) ,instr_set, (tr, s)) \<in> exec(lis)"

text {* This is shorthand for going from a single @{term smallstep} to @{term exec}. *}
lemma exsing: "\<lbrakk>instr_set \<subseteq> lis; (l, ins) \<in> instr_set \<and> PC s = l \<and> ((trs, s), (trt, t)) \<in> smallstep(lis) \<and> 
               \<not>(indom (PC t) instr_set)\<rbrakk>
              \<Longrightarrow> ((trs, s),instr_set, (trt, t)) \<in> exec(lis)" 
by (meson exec.exseq exec.exterm)

text {* For each @{term smallstep}, the trace is expanded by at most one event. *}
lemma smallstep_head_growth: "((tr', s'), (tr, s)) \<in> smallstep(lis) \<Longrightarrow> (tr' = tr) \<or> (\<exists> a . (tr = tr' @ [a]))"
  using smallstep_rev
  by blast

lemma smallstep_head_growth2:
  assumes
    a: "((tr, s'), tr @ r @ [a] @ l, s) \<in> smallstep(lis)"
  shows "r = [] \<and> l = []"
proof -
  have f1: "\<forall>as asa. (([]::'a list) = as @ asa) = (as = [] \<and> asa = [])"
    by (meson Nil_is_append_conv)
  have f2: "[] \<noteq> [a] @ l"
    by blast
  obtain aa :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a" where
    f3: "\<forall>x2 x4. (\<exists>v5. x2 = x4 @ [v5]) = (x2 = x4 @ [aa x2 x4])"
    by moura
  have "tr @ r @ [a] @ l \<noteq> tr"
    by blast
  then have f4: "[aa (tr @ r @ [a] @ l) tr] = r @ [a] @ l"
    using f3 by (metis (no_types) append_eq_append_conv assms smallstep_head_growth)
  then have "r = []"
    using f2 f1 by (metis list.sel(3) tl_append2)
  then show ?thesis
    using f4 by simp
qed

lemma smallstep_prefix_exchange:
  assumes "((tr, sa'), tr @ new, s) \<in> smallstep lis"
  shows "((tr', sa'), tr' @ new, s) \<in> smallstep lis"
using smallstep_rev[OF assms]
proof auto
  show "\<And>f x. new = [] \<Longrightarrow> (PC sa', do f) \<in> lis \<Longrightarrow> x \<in> f (R sa') \<Longrightarrow> s = sa'\<lparr>R := x, PC := Suc (PC sa')\<rparr> \<Longrightarrow> ((tr', sa'), tr', sa'\<lparr>R := x, PC := Suc (PC sa')\<rparr>) \<in> smallstep lis"
    by (simp add: sdof)
  show "\<And>b m n. new = [] \<Longrightarrow> (PC sa', cbr b m n) \<in> lis \<Longrightarrow> b (R sa') \<Longrightarrow> s = sa'\<lparr>PC := m\<rparr> \<Longrightarrow> ((tr', sa'), tr', sa'\<lparr>PC := m\<rparr>) \<in> smallstep lis"
    by (simp add: scbrt)
  show "\<And>b m n. new = [] \<Longrightarrow> (PC sa', cbr b m n) \<in> lis \<Longrightarrow> \<not> b (R sa') \<Longrightarrow> s = sa'\<lparr>PC := n\<rparr> \<Longrightarrow> ((tr', sa'), tr', sa'\<lparr>PC := n\<rparr>) \<in> smallstep lis"
    by (simp add: scbrf)
  show "\<And>ef f e. (PC sa', comm ef f) \<in> lis \<Longrightarrow> e \<in> ef (R sa') \<Longrightarrow> s = sa'\<lparr>R := f (R sa') e, PC := Suc (PC sa')\<rparr> \<Longrightarrow> new = [e] \<Longrightarrow> ((tr', sa'), tr' @ [e], sa'\<lparr>R := f (R sa') e, PC := Suc (PC sa')\<rparr>) \<in> smallstep lis"
    by (simp add: scomm)
qed

lemma smallstep_lis_subset:
  assumes
    op: "((tr, s), (tr', s')) \<in> smallstep big"
  shows "\<exists> i . (PC s, i) \<in> big \<and> ((tr, s), (tr', s')) \<in> smallstep {(PC s, i)}"
using smallstep_rev[OF op]
proof -
  obtain nn :: nat and AA :: "'b \<Rightarrow> 'a set" and bb :: "'b \<Rightarrow> 'a \<Rightarrow> 'b" and aa :: 'a where
      f1: "(\<exists>v0 v1 v2 v3. (v0, comm v1 v2) \<in> big \<and> v3 \<in> v1 (R s) \<and> PC s = v0 \<and> tr' = tr @ [v3] \<and> s' = s\<lparr>R := v2 (R s) v3, PC := PC s + 1\<rparr>) = ((nn, comm AA bb) \<in> big \<and> aa \<in> AA (R s) \<and> PC s = nn \<and> tr' = tr @ [aa] \<and> s' = s\<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr>)"
    by blast
  obtain nna :: nat and bba :: "'b \<Rightarrow> bool" and nnb :: nat and nnc :: nat where
    f2: "(\<exists>v0 v1 v2 v3. (v0, cbr v1 v2 v3) \<in> big \<and> PC s = v0 \<and> \<not> v1 (R s) \<and> s' = s\<lparr>PC := v3\<rparr>) = ((nna, cbr bba nnb nnc) \<in> big \<and> PC s = nna \<and> \<not> bba (R s) \<and> s' = s\<lparr>PC := nnc\<rparr>)"
    by blast
  obtain nnd :: nat and bbb :: "'b \<Rightarrow> bool" and nne :: nat and nnf :: nat where
    f3: "(\<exists>v0 v1 v2 v3. (v0, cbr v1 v2 v3) \<in> big \<and> PC s = v0 \<and> v1 (R s) \<and> s' = s\<lparr>PC := v2\<rparr>) = ((nnd, cbr bbb nne nnf) \<in> big \<and> PC s = nnd \<and> bbb (R s) \<and> s' = s\<lparr>PC := nne\<rparr>)"
    by blast
  obtain nng :: nat and BB :: "'b \<Rightarrow> 'b set" and bbc :: 'b where
    "(\<exists>v0 v1 v2. (v0, do v1) \<in> big \<and> v2 \<in> v1 (R s) \<and> PC s = v0 \<and> s' = s\<lparr>R := v2, PC := PC s + 1\<rparr>) = ((nng, do BB) \<in> big \<and> bbc \<in> BB (R s) \<and> PC s = nng \<and> s' = s\<lparr>R := bbc, PC := PC s + 1\<rparr>)"
    by blast
  then have f4: "(nng, do BB) \<in> big \<and> bbc \<in> BB (R s) \<and> PC s = nng \<and> s' = s\<lparr>R := bbc, PC := PC s + 1\<rparr> \<and> tr = tr' \<or> (nnd, cbr bbb nne nnf) \<in> big \<and> PC s = nnd \<and> bbb (R s) \<and> s' = s\<lparr>PC := nne\<rparr> \<and> tr = tr' \<or> (nna, cbr bba nnb nnc) \<in> big \<and> PC s = nna \<and> \<not> bba (R s) \<and> s' = s\<lparr>PC := nnc\<rparr> \<and> tr = tr' \<or> (nn, comm AA bb) \<in> big \<and> aa \<in> AA (R s) \<and> PC s = nn \<and> tr' = tr @ [aa] \<and> s' = s\<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr>"
    using f3 f2 f1 \<open>(\<exists>l f x. (l, do f) \<in> big \<and> x \<in> f (R s) \<and> PC s = l \<and> s' = s\<lparr>R := x, PC := PC s + 1\<rparr>) \<and> tr = tr' \<or> (\<exists>l b m n. (l, cbr b m n) \<in> big \<and> PC s = l \<and> b (R s) \<and> s' = s\<lparr>PC := m\<rparr>) \<and> tr = tr' \<or> (\<exists>l b m n. (l, cbr b m n) \<in> big \<and> PC s = l \<and> \<not> b (R s) \<and> s' = s\<lparr>PC := n\<rparr>) \<and> tr = tr' \<or> (\<exists>l ef f e. (l, comm ef f) \<in> big \<and> e \<in> ef (R s) \<and> PC s = l \<and> tr' = tr @ [e] \<and> s' = s\<lparr>R := f (R s) e, PC := PC s + 1\<rparr>)\<close> by presburger
  have f5: "\<forall>n p na nb P s sa as. ((n, cbr p na nb) \<notin> P \<or> PC s \<noteq> n \<or> p (R s::'b) \<or> sa \<noteq> s\<lparr>PC := nb\<rparr>) \<or> ((as::'a list, s), as, sa) \<in> smallstep P"
    using smallstep.scbrf by blast
  have f6: "(nn, comm AA bb) \<notin> {(PC s, comm AA bb)} \<or> aa \<notin> AA (R \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>) \<or> PC \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> \<noteq> nn \<or> s\<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr> \<noteq> \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> \<lparr>R := bb (R \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>) aa, PC := PC \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> + 1\<rparr> \<or> ((tr, \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>), tr @ [aa], s \<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr>) \<in> smallstep {(PC s, comm AA bb)}"
    by (simp add: smallstep.scomm)
  { assume "s\<lparr>R := bbc, PC := PC s + 1\<rparr> = \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> \<lparr>R := bbc, PC := PC \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> + 1\<rparr>"
    { assume "((tr', \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>), tr', s \<lparr>R := bbc, PC := PC s + 1\<rparr>) \<in> smallstep {(PC s, do BB)}"
      then have "((tr', \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>), tr', s \<lparr>R := bbc, PC := PC s + 1\<rparr>) = ((tr, s), tr', s') \<longrightarrow> ((tr, s), tr', s') \<in> smallstep {(PC s, do BB)}"
        by metis
      then have "(tr', \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>) = (tr, s) \<longrightarrow> ((nng, do BB) \<notin> big \<or> bbc \<notin> BB (R s) \<or> PC s \<noteq> nng \<or> s' \<noteq> s\<lparr>R := bbc, PC := PC s + 1\<rparr> \<or> tr \<noteq> tr') \<or> (\<exists>i. (PC s, i) \<in> big \<and> ((tr, s), tr', s') \<in> smallstep {(PC s, i)})"
        by blast }
    then have "(tr', \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>) = (tr, s) \<longrightarrow> ((nng, do BB) \<notin> big \<or> bbc \<notin> BB (R s) \<or> PC s \<noteq> nng \<or> s' \<noteq> s\<lparr>R := bbc, PC := PC s + 1\<rparr> \<or> tr \<noteq> tr') \<or> (\<exists>i. (PC s, i) \<in> big \<and> ((tr, s), tr', s') \<in> smallstep {(PC s, i)})"
      by (metis (no_types) insertI1 smallstep.sdof) }
  then have "(tr', \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>) = (tr, s) \<longrightarrow> ((nng, do BB) \<notin> big \<or> bbc \<notin> BB (R s) \<or> PC s \<noteq> nng \<or> s' \<noteq> s\<lparr>R := bbc, PC := PC s + 1\<rparr> \<or> tr \<noteq> tr') \<or> (\<exists>i. (PC s, i) \<in> big \<and> ((tr, s), tr', s') \<in> smallstep {(PC s, i)})"
    by fastforce
  then have "(\<exists>i. (PC s, i) \<in> big \<and> ((tr, s), tr', s') \<in> smallstep {(PC s, i)}) \<or> (nn, comm AA bb) \<in> big \<and> aa \<in> AA (R s) \<and> PC s = nn \<and> tr' = tr @ [aa] \<and> s' = s\<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr> \<or> (nna, cbr bba nnb nnc) \<in> big \<and> PC s = nna \<and> \<not> bba (R s) \<and> s' = s\<lparr>PC := nnc\<rparr> \<and> tr = tr'"
    using f4 by (metis (no_types) insertI1 smallstep.scbrt surjective)
  moreover
  { assume "(nn, comm AA bb) \<in> big \<and> aa \<in> AA (R s) \<and> PC s = nn \<and> tr' = tr @ [aa] \<and> s' = s\<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr>"
    then have "((nn, comm AA bb) \<in> big \<and> aa \<in> AA (R s) \<and> PC s = nn \<and> tr' = tr @ [aa] \<and> s' = s\<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr>) \<and> ((tr, \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>), tr @ [aa], s \<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr>) \<in> smallstep {(PC s, comm AA bb)}"
      using f6 by force
    then have ?thesis
      by (metis surjective) }
  ultimately show ?thesis
    using f5 by blast
qed


lemma smallstep_lis_subset':
  assumes
    op: "((tr, s), (tr', s')) \<in> smallstep big" and
    subs: "small \<subseteq> big" and
    PCs: "(PC s, i) \<in> small" and
    wff: "\<forall>l i1 i2. (l, i1) \<in> big \<and> (l, i2) \<in> big \<longrightarrow> i1 = i2"
  shows "((tr, s), (tr', s')) \<in> smallstep small"
using smallstep_rev[OF op]
using smallstep.simps
using subs PCs wff
proof -
  obtain nn :: nat and AA :: "'b \<Rightarrow> 'a set" and bb :: "'b \<Rightarrow> 'a \<Rightarrow> 'b" and aa :: 'a where
    f1: "(\<exists>v0 v1 v2 v3. (v0, comm v1 v2) \<in> big \<and> v3 \<in> v1 (R s) \<and> PC s = v0 \<and> tr' = tr @ [v3] \<and> s' = s\<lparr>R := v2 (R s) v3, PC := PC s + 1\<rparr>) = ((nn, comm AA bb) \<in> big \<and> aa \<in> AA (R s) \<and> PC s = nn \<and> tr' = tr @ [aa] \<and> s' = s\<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr>)"
    by blast
  obtain nna :: nat and bba :: "'b \<Rightarrow> bool" and nnb :: nat and nnc :: nat where
    f2: "(\<exists>v0 v1 v2 v3. (v0, cbr v1 v2 v3) \<in> big \<and> PC s = v0 \<and> \<not> v1 (R s) \<and> s' = s\<lparr>PC := v3\<rparr>) = ((nna, cbr bba nnb nnc) \<in> big \<and> PC s = nna \<and> \<not> bba (R s) \<and> s' = s\<lparr>PC := nnc\<rparr>)"
    by blast
  obtain nnd :: nat and bbb :: "'b \<Rightarrow> bool" and nne :: nat and nnf :: nat where
    f3: "(\<exists>v0 v1 v2 v3. (v0, cbr v1 v2 v3) \<in> big \<and> PC s = v0 \<and> v1 (R s) \<and> s' = s\<lparr>PC := v2\<rparr>) = ((nnd, cbr bbb nne nnf) \<in> big \<and> PC s = nnd \<and> bbb (R s) \<and> s' = s\<lparr>PC := nne\<rparr>)"
    by blast
  obtain nng :: nat and BB :: "'b \<Rightarrow> 'b set" and bbc :: 'b where
    "(\<exists>v0 v1 v2. (v0, do v1) \<in> big \<and> v2 \<in> v1 (R s) \<and> PC s = v0 \<and> s' = s\<lparr>R := v2, PC := PC s + 1\<rparr>) = ((nng, do BB) \<in> big \<and> bbc \<in> BB (R s) \<and> PC s = nng \<and> s' = s\<lparr>R := bbc, PC := PC s + 1\<rparr>)"
    by blast
  then have f4: "(nng, do BB) \<in> big \<and> bbc \<in> BB (R s) \<and> PC s = nng \<and> s' = s\<lparr>R := bbc, PC := PC s + 1\<rparr> \<and> tr = tr' \<or> (nnd, cbr bbb nne nnf) \<in> big \<and> PC s = nnd \<and> bbb (R s) \<and> s' = s\<lparr>PC := nne\<rparr> \<and> tr = tr' \<or> (nna, cbr bba nnb nnc) \<in> big \<and> PC s = nna \<and> \<not> bba (R s) \<and> s' = s\<lparr>PC := nnc\<rparr> \<and> tr = tr' \<or> (nn, comm AA bb) \<in> big \<and> aa \<in> AA (R s) \<and> PC s = nn \<and> tr' = tr @ [aa] \<and> s' = s\<lparr>R := bb (R s) aa, PC := PC s + 1\<rparr>"
    using f3 f2 f1 \<open>(\<exists>l f x. (l, do f) \<in> big \<and> x \<in> f (R s) \<and> PC s = l \<and> s' = s\<lparr>R := x, PC := PC s + 1\<rparr>) \<and> tr = tr' \<or> (\<exists>l b m n. (l, cbr b m n) \<in> big \<and> PC s = l \<and> b (R s) \<and> s' = s\<lparr>PC := m\<rparr>) \<and> tr = tr' \<or> (\<exists>l b m n. (l, cbr b m n) \<in> big \<and> PC s = l \<and> \<not> b (R s) \<and> s' = s\<lparr>PC := n\<rparr>) \<and> tr = tr' \<or> (\<exists>l ef f e. (l, comm ef f) \<in> big \<and> e \<in> ef (R s) \<and> PC s = l \<and> tr' = tr @ [e] \<and> s' = s\<lparr>R := f (R s) e, PC := PC s + 1\<rparr>)\<close> by presburger
  have f5: "(PC s, i) \<in> big"
    using PCs assms(2) by blast
  moreover
  { assume "s\<lparr>R := bbc, PC := PC s + 1\<rparr> = \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> \<lparr>R := bbc, PC := PC \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> + 1\<rparr>"
    { assume "((tr', \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>), tr', s \<lparr>R := bbc, PC := PC s + 1\<rparr>) \<in> smallstep small"
      then have "((tr', \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>), tr', s \<lparr>R := bbc, PC := PC s + 1\<rparr>) = ((tr, s), tr', s') \<longrightarrow> ((tr, s), tr', s') \<in> smallstep small"
        by metis
      then have "(tr', \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>) = (tr, s) \<longrightarrow> ((nng, do BB) \<notin> big \<or> bbc \<notin> BB (R s) \<or> PC s \<noteq> nng \<or> s' \<noteq> s\<lparr>R := bbc, PC := PC s + 1\<rparr> \<or> tr \<noteq> tr') \<or> ((tr, s), tr', s') \<in> smallstep small"
        by blast }
    then have "(tr', \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr>) = (tr, s) \<longrightarrow> ((nng, do BB) \<notin> big \<or> bbc \<notin> BB (R s) \<or> PC s \<noteq> nng \<or> s' \<noteq> s\<lparr>R := bbc, PC := PC s + 1\<rparr> \<or> tr \<noteq> tr') \<or> ((tr, s), tr', s') \<in> smallstep small"
      using f5 by (metis PCs assms(4) smallstep.intros(1)) }
  moreover
  { assume "s\<lparr>R := bbc, PC := PC s + 1\<rparr> \<noteq> \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> \<lparr>R := bbc, PC := PC \<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> + 1\<rparr>"
    moreover
    { assume "\<lparr>R = R s, PC = PC s, \<dots> = more s\<rparr> \<lparr>R := bbc\<rparr> \<noteq> \<lparr>R = bbc, PC = nng, \<dots> = more s\<rparr>"
      then have "s \<noteq> \<lparr>R = R s, PC = nng, \<dots> = more s\<rparr>"
        by (metis (full_types) surjective update_convs(1)) }
    ultimately have "(nng, do BB) \<notin> big \<or> bbc \<notin> BB (R s) \<or> PC s \<noteq> nng \<or> s' \<noteq> s\<lparr>R := bbc, PC := PC s + 1\<rparr> \<or> tr \<noteq> tr'"
      by simp }
  ultimately show ?thesis
    using f4 by (metis (no_types) PCs assms(4) smallstep.intros(2) smallstep.intros(3) smallstep.intros(4) surjective)
qed

end