section {* Parallel SF semantics *}

text{* This section introduces parallel concepts of the stable failure semantics and ultimately exercises it on a parallel buffer example. *}

theory "stable_failures_parallel"
imports stable_failures parallel_base trace_restriction
begin

subsection {* Preparations *}

paragraph {* Datatypes *}

text {* First we need to introduce datatypes to handle the tree structure of parallel code.  *}

type_synonym ('s)parallelState = "'s NCstate ATree"

type_synonym ('s, 'e) parallelFailure = "'e definitions.trace \<times> 's parallelState \<times> 'e definitions.event set"

text {* A slightly modified induction scheme with already expanded variables. *}

paragraph {* Functions describing state and code characteristics *}

text {* For every sequential state, we want it to be normal. *}
definition N\<^sub>p :: "'s parallelState \<Rightarrow> bool" where
"N\<^sub>p ss = (leaf1_property (\<lambda> s . N s) op \<and> ss)"

lemma N\<^sub>p_inherit: "N\<^sub>p (par sl sr) = (N\<^sub>p sl \<and> N\<^sub>p sr)"
  using leaf1_property_inherit[of N\<^sub>p N] N\<^sub>p_def by blast

text {* True, iff all states have that one label. *}
definition AllPC\<^sub>S\<^sub>F :: "'s parallelState \<Rightarrow> nat \<Rightarrow> bool" where
"AllPC\<^sub>S\<^sub>F ss lbl = leaf1_property (\<lambda> s . NCPC s = lbl) op \<and> ss"

lemma AllPC\<^sub>S\<^sub>F_inherit: "AllPC\<^sub>S\<^sub>F (par sl sr) lbl = (AllPC\<^sub>S\<^sub>F sl lbl \<and> AllPC\<^sub>S\<^sub>F sr lbl)"
  using leaf1_property_inherit AllPC\<^sub>S\<^sub>F_def by blast

paragraph {* Transformations between the two parallel state worlds. *}

fun parallelFailureToFailure :: "('s, 'e) parallelFailure \<Rightarrow> ('s, 'e) failure set" where
"parallelFailureToFailure (tr, seq s, X) = {(tr, s, X)}" |
"parallelFailureToFailure (tr, par _ _, X) = {}"

definition parallelFailureSetToFailureSet :: "('s, 'e) parallelFailure set \<Rightarrow> ('s, 'e) failure set" where
"parallelFailureSetToFailureSet S = {x . \<exists> x' . x' \<in> S \<and> x \<in> parallelFailureToFailure x'}"

lemma parallelFailureToFailure_ex_seq: "(tr, s, X) \<in> parallelFailureToFailure (tr', s', X') \<Longrightarrow> (tr, seq s, X) = (tr', s', X')"
  by (metis ATree.exhaust empty_iff fst_conv insert_iff parallelFailureToFailure.simps(1) parallelFailureToFailure.simps(2) snd_conv)

lemma parallelFailureSingletonToFailureSingleton:
 "parallelFailureSetToFailureSet {(tr, s, X)} \<noteq> {} \<Longrightarrow> \<exists> s' . {(tr, s', X)} = parallelFailureSetToFailureSet {(tr, s, X)} \<and> s = seq s'"
  unfolding parallelFailureSetToFailureSet_def
proof (induct s, auto) qed

definition failureSetToParallelFailureSet :: "('s, 'e) failure set \<Rightarrow> ('s, 'e) parallelFailure set" where
"failureSetToParallelFailureSet S = {(tr, s, X) . \<exists> s' . (tr, s', X) \<in> S \<and> s = seq s'}"

subsection {* The parallel semantics *}

text{* Note: This definition already practically restricts the interfaces: A parent interface is only
   allowed to be a subset of the children interfaces' union, as otherwise the trace could
   contain potentially too many events. (in theory one could still write it, though) *}
fun sfsem\<^sub>p :: "('s, 'e) structured_code_parallel \<Rightarrow>
              ('s, 'e) parallelFailure set \<Rightarrow>
              ('s, 'e) parallelFailure set"(*<*)("\<lbrakk>_\<rbrakk>\<^sub>p\<^sub>S\<^sub>F(_)")(*>*) where
sfsemp_par: "(\<lbrakk>parn c1 (\<alpha>, \<beta>) c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F (S)) = {(tr, t12, X) .
  \<exists> X1 X2 X' t1 t2 s1 s2 s12. ([], s12, X') \<in> S \<and> N\<^sub>p s1 \<and> N\<^sub>p s2 \<and>
  X \<inter> (\<alpha> \<union> \<beta>) = (X1 \<inter> \<alpha>) \<union> (X2 \<inter> \<beta>) \<and>
  (tr \<restriction> \<alpha>, t1, X1) \<in> sfsem\<^sub>p c1 {([], s1, X')} \<and>
  (tr \<restriction> \<beta>, t2, X2) \<in> sfsem\<^sub>p c2 {([], s2, X')} \<and>
  set tr \<subseteq> (\<alpha> \<union> \<beta>) \<and>
  t12 = par t1 t2 \<and>
  s12 = par s1 s2
}" |
sfsemp_seq: "(\<lbrakk>seqn (c, \<Sigma>)\<rbrakk>\<^sub>p\<^sub>S\<^sub>F(S)) = failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> c (parallelFailureSetToFailureSet S))"

lemma sfsemp_seq_unfold: "((tr, seq s, X) \<in> (\<lbrakk>seqn (c, \<Sigma>)\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', seq s', X')})) = ((tr, s, X) \<in> (SFsemSigma.sfsem \<Sigma> c {(tr',  s', X')}))"
  unfolding sfsemp_seq failureSetToParallelFailureSet_def parallelFailureSetToFailureSet_def
  using parallelFailureToFailure.simps
  by simp


text {* We can add arbitrary many other states to the besides the known predecessor state. *}

lemma sfsemp_monotone_single:
  fixes
    X X' :: "'e definitions.event set" and
    s s' :: "'s parallelState"
  assumes
    a: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F({(tr', s', X')})" and
    b: "(tr', s', X') \<in> S" and
    wff: "wff\<^sub>p c"
  shows "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F(S)"
using a b wff
proof (induct c rule: structured_code_parallel_induct, simp_all add: wff\<^sub>p_def)
  fix \<Sigma> :: "'e definitions.event set" and x :: "('s, 'e) structured_code"
  assume
    aa1: "(tr, s, X) \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> x (parallelFailureSetToFailureSet {(tr', s', X')}))" and
    aa2: "(tr', s', X') \<in> S" and
    wff: "wff\<^sub>s\<^sub>c x \<and> liSet (\<union>\<^sub>s\<^sub>c x)"
  have SFsemSigma: "SFsemSigma (\<union>\<^sub>s\<^sub>c x)"
    using wff unfolding SFsemSigma_def by simp
  have subs: "parallelFailureSetToFailureSet {(tr', s', X')} \<subseteq> parallelFailureSetToFailureSet S"
    unfolding parallelFailureSetToFailureSet_def
    using aa2
    by blast
  moreover have "\<forall> S T . S \<subseteq> T \<longrightarrow> failureSetToParallelFailureSet S \<subseteq> failureSetToParallelFailureSet T"
    unfolding failureSetToParallelFailureSet_def
    by fast
  ultimately show "(tr, s, X) \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> x (parallelFailureSetToFailureSet S))"
    using aa1 SFsemSigma.sfsem_monotone[OF SFsemSigma subs, of \<Sigma> x]
    by blast
qed blast

lemma sfsemp_predecessor:
  fixes
    X X' :: "'e definitions.event set" and
    s s' :: "'s parallelState"
  assumes
    a: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F(S)" and
    wff: "wff\<^sub>p c"
  shows "\<exists> tr' s' X' . (tr', s', X') \<in> S \<and> (tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F({(tr', s', X')}) \<and> treen_mirror s' c \<and> treen_mirror s c \<and> tree_mirror s s'"
using a wff
proof (induct c arbitrary: tr s X S rule: structured_code_parallel_induct, simp_all add: wff\<^sub>p_def)
  fix x \<Sigma> tr s X S
  assume
    inside: "(tr, s, X) \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> x (parallelFailureSetToFailureSet S))" and
    wff: "wff\<^sub>s\<^sub>c x \<and> liSet (\<union>\<^sub>s\<^sub>c x)"
  have SFsemSigma: "SFsemSigma (\<union>\<^sub>s\<^sub>c x)"
    using wff unfolding SFsemSigma_def by simp
  obtain s' where
    inside: "(tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> x {x. \<exists>x'. x' \<in> S \<and> x \<in> parallelFailureToFailure x'}" and
    s: "s = ATree.seq s'"
    using inside
    unfolding failureSetToParallelFailureSet_def parallelFailureSetToFailureSet_def
    by fast
  obtain trp sp Xp where
    inside_single: "(tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> x {(trp, sp, Xp)}" and
    inS: "(trp, sp, Xp) \<in> {x. \<exists>x'. x' \<in> S \<and> x \<in> parallelFailureToFailure x'}"
    using SFsemSigma.sfsem_predecessor[OF SFsemSigma inside] by blast
  then obtain trs ss Xs where
    p1: "(trp, sp, Xp) \<in> {x. \<exists>x'. x' \<in> {(trs, ss, Xs)} \<and> x \<in> parallelFailureToFailure x'}" and
    pInS: "(trs, ss, Xs) \<in> S"
    using inS by auto
  have inside: "(tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> x {x. \<exists>x'. x' \<in> {(trs, ss, Xs)} \<and> x \<in> parallelFailureToFailure x'}"
    using p1 inside_single
    by (meson SFsemSigma.sfsem_monotone_single SFsemSigma)
  have "\<exists>s. ss = ATree.seq s"
    using p1 parallelFailureToFailure.simps
  proof -
    have "(trp, sp, Xp) \<in> parallelFailureSetToFailureSet {(trs, ss, Xs)}"
      using p1 parallelFailureSetToFailureSet_def by blast
    then show ?thesis
      by (metis (no_types) empty_iff parallelFailureSingletonToFailureSingleton)
  qed
  thus "\<exists>tr' s' X'. (tr', s', X') \<in> S \<and>
          (tr, s, X) \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> x (parallelFailureSetToFailureSet {(tr', s', X')})) \<and> (\<exists>s. s' = ATree.seq s) \<and> (\<exists>s'. s = ATree.seq s') \<and> tree_mirror s s'"
    using inside pInS p1 s
    unfolding failureSetToParallelFailureSet_def parallelFailureSetToFailureSet_def
    by auto
next
  fix cl \<alpha> \<beta> cr tra sa Xa Sa
  assume
    IH1: "\<And>tr s X S. (tr, s, X) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS \<Longrightarrow> \<exists>tr' s' X'. (tr', s', X') \<in> S \<and> (tr, s, X) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s', X')} \<and> treen_mirror s' cl \<and> treen_mirror s cl \<and> tree_mirror s s'" and
    IH2: "\<And>tr s X S. (tr, s, X) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS \<Longrightarrow> \<exists>tr' s' X'. (tr', s', X') \<in> S \<and> (tr, s, X) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s', X')} \<and> treen_mirror s' cr \<and> treen_mirror s cr \<and> tree_mirror s s'" and
    base: "\<exists>X1 X2 X' t1 t2 s1 s2.
          ([], par s1 s2, X') \<in> Sa \<and>
          N\<^sub>p s1 \<and>
          N\<^sub>p s2 \<and>
          Xa \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<and>
          (tra \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')} \<and> (tra \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')} \<and> set tra \<subseteq> \<alpha> \<union> \<beta> \<and> sa = par t1 t2"
  obtain X1 X2 X' t1 t2 s1 s2 where
    "([], par s1 s2, X') \<in> Sa" and
    N1: "N\<^sub>p s1" and
    N2: "N\<^sub>p s2" and
    refusal: "Xa \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
    ex_l: "(tra \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}" and
    ex_r: "(tra \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}" and
    settra: "set tra \<subseteq> \<alpha> \<union> \<beta>" and
    sa: "sa = par t1 t2"
    using base by blast
  moreover
  have "treen_mirror (par s1 s2) (parn cl (\<alpha>, \<beta>) cr) \<and> treen_mirror (sa) (parn cl (\<alpha>, \<beta>) cr) \<and> tree_mirror (sa) (par s1 s2)"
    using IH1[OF ex_l, simplified] IH2[OF ex_r, simplified] sa
    by auto
  ultimately
  show "\<exists>s' X'. ([], s', X') \<in> Sa \<and>
               (\<exists>X1 X2 t1 t2 s1.
                   N\<^sub>p s1 \<and>
                   (\<exists>s2. N\<^sub>p s2 \<and>
                         Xa \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<and>
                         (tra \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')} \<and>
                         (tra \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')} \<and> set tra \<subseteq> \<alpha> \<union> \<beta> \<and> sa = par t1 t2 \<and> s' = par s1 s2)) \<and>
               treen_mirror s' (parn cl (\<alpha>, \<beta>) cr) \<and> treen_mirror sa (parn cl (\<alpha>, \<beta>) cr) \<and> tree_mirror sa s'"
    by blast
qed

lemma sfsemp_monotone:
  assumes subs: "S \<subseteq> T" and wff: "wff\<^sub>p c"
  shows "\<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS \<subseteq> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>FT"
proof
  fix x
  assume inside: "x \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS"
  obtain tr s X where x: "x = (tr, s, X)" using prod_cases3 by blast
  obtain tr' s' X' where
    ex: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s', X')}" and
    inS: "(tr', s', X') \<in> S"
    using sfsemp_predecessor[OF inside[unfolded x] wff]
    by blast
  have inT: "(tr', s', X') \<in> T" using subs inS by blast
  thus "x \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>FT"
    using sfsemp_monotone_single ex wff
    unfolding x
    by blast
qed

lemma sfsemp_starting_set_can_contain_malformed_states:
  fixes
    c :: "('s, 'e) structured_code_parallel"
  assumes
    wff: "wff\<^sub>p c"
  shows "\<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. Pre x' \<and> treen_mirror (fst (snd x')) c} = \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. Pre x'}"
proof -
  {
    fix tr s X
    assume inside: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. Pre x' \<and> treen_mirror (fst (snd x')) c}"
    hence "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. Pre x'}"
      using sfsemp_monotone[of "{x'. Pre x' \<and> treen_mirror (fst (snd x')) c}" "{x'. Pre x'}" c] wff
      by blast
  } moreover {
    fix tr s X
    assume inside: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. Pre x'}"
    obtain tr' s' X' where
      ex: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s', X')}" and
      inS: "(tr', s', X') \<in> {x'. Pre x'}" and
      tm: "treen_mirror s' c"
      using sfsemp_predecessor[OF inside wff]
      by blast
    hence start: "(tr', s', X') \<in> {x'. Pre x' \<and> treen_mirror (fst (snd x')) c}"
      by simp
    have "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. Pre x' \<and> treen_mirror (fst (snd x')) c}"
      using sfsemp_monotone_single[OF ex start wff] .
  } ultimately show ?thesis by fast
qed

lemma sfsemp_general_sigma_exchange:
  fixes
    X X' :: "'e definitions.event set" and
    s s' :: "'s parallelState"
  assumes
    a: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F({(tr', s', X')})" and
    wff: "wff\<^sub>p c"
  shows "\<exists> X . (tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>S\<^sub>F({(tr', s', X's)})"
using assms
proof (induct c arbitrary: tr s s' X rule: structured_code_parallel_induct, simp_all add: wff\<^sub>p_def, auto)
  fix x :: "('s, 'e) structured_code" and \<Sigma> tr s s' X
  assume
    inside: "(tr, s, X) \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> x (parallelFailureSetToFailureSet {(tr', s', X')}))" and
    wff: "wff\<^sub>s\<^sub>c x" and
    ax: "liSet (\<union>\<^sub>s\<^sub>c x)"
  have SFsemSigma: "SFsemSigma (\<union>\<^sub>s\<^sub>c x)"
    using ax unfolding SFsemSigma_def by simp
  obtain ss where
    inside: "(tr, ss, X) \<in> SFsemSigma.sfsem \<Sigma> x {x. \<exists>x'. x' \<in> {(tr', s', X')} \<and> x \<in> parallelFailureToFailure x'}" and
    s: "s = ATree.seq ss"
    using inside
    unfolding failureSetToParallelFailureSet_def parallelFailureSetToFailureSet_def
    by fast
  obtain ss' where
    s': "s' = seq ss'" and
    pred: "{(tr', ss', X')} = {x. \<exists>x'. x' \<in> {(tr', s', X')} \<and> x \<in> parallelFailureToFailure x'}"
    using parallelFailureSingletonToFailureSingleton[of tr' s' X']
    using SFsemSigma.sfsem_predecessor[OF SFsemSigma inside]
    unfolding parallelFailureSetToFailureSet_def
    by auto
  have exch: "\<exists> X . (tr, ss, X) \<in> SFsemSigma.sfsem \<Sigma> x {(tr', ss', X's)}"
    using SFsemSigma wff inside[unfolded pred[symmetric]]
      SFsemSigma.sfsem_general_sigma_exchange by fast
  have pred's: "{(tr', ss', X's)} = {x. \<exists>x'. x' \<in> {(tr', s', X's)} \<and> x \<in> parallelFailureToFailure x'}"
    using pred s' "parallelFailureToFailure.simps" by fastforce
  have "\<exists> X . (tr, ss, X) \<in> SFsemSigma.sfsem \<Sigma> x (parallelFailureSetToFailureSet {(tr', s', X's)})"
    using exch[unfolded pred's parallelFailureSetToFailureSet_def[symmetric]] .
  thus "\<exists> X . (tr, s, X) \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> x (parallelFailureSetToFailureSet {(tr', s', X's)}))"
    unfolding failureSetToParallelFailureSet_def s by blast
next
  fix cl cr :: "('s, 'e) structured_code_parallel" and \<alpha> \<beta> t1 t2 s1 s2 tr X1 X2 s s' X
  assume
    IH1: "\<And>tr s s' X. (tr, s, X) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s', X')} \<Longrightarrow> \<exists>X. (tr, s, X) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s', X's)}" and
    IH2: "\<And>tr s s' X. (tr, s, X) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s', X')} \<Longrightarrow> \<exists>X. (tr, s, X) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s', X's)}" and
    tr': "tr' = []" and
    wff_l: "leafn1_property (\<lambda>(c, ab). wff\<^sub>s\<^sub>c c \<and> liSet (\<union>\<^sub>s\<^sub>c c)) op \<and> cl" and
    wff_r: "leafn1_property (\<lambda>(c, ab). wff\<^sub>s\<^sub>c c \<and> liSet (\<union>\<^sub>s\<^sub>c c)) op \<and> cr" and
    Ns1: "N\<^sub>p s1" and
    Ns2: "N\<^sub>p s2" and
    rs: "X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
    ex_l: "(tr \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}" and
    ex_r: "(tr \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}" and
    tr: "set tr \<subseteq> \<alpha> \<union> \<beta>"
  obtain X1s where i1: "(tr \<restriction> \<alpha>, t1, X1s) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X's)}"
    using IH1[OF ex_l] by blast
  obtain X2s where i2: "(tr \<restriction> \<beta>, t2, X2s) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X's)}"
    using IH2[OF ex_r] by blast
  obtain X12s where X12s: "X12s = X1s \<inter> \<alpha> \<union> X2s \<inter> \<beta>" by blast
  have "X12s \<inter> (\<alpha> \<union> \<beta>) = X1s \<inter> \<alpha> \<union> X2s \<inter> \<beta>" unfolding X12s by blast
  thus "\<exists>X X1 X2. X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<and> (tr \<restriction> \<alpha>, t1, X1) \<in> (\<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X's)}) \<and> (tr \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X's)}"
    using i1 i2 by blast
qed

subsection {* Refusal set manipulation *}

text {* When there exist multiple sets of refused elements, then they are also refused all together.
Note note that this lemma is not used afterwards. However, it is a nice to have sanity check.
*}
lemma sfsemp_refusal_set_Union:
  assumes a: "\<forall> a \<in> XA . (tr, s, a) \<in> sfsem\<^sub>p code {(tr', s', X')}" and
      not_empty: "XA \<noteq> {}" and
      d: "wff\<^sub>p code"
  shows "(tr, s, \<Union>XA) \<in> sfsem\<^sub>p code {(tr', s', X')}"
  using assms
proof (induct code arbitrary: tr s tr' s' X' XA, clarsimp simp add: wff\<^sub>p_inherit wff\<^sub>p_def)
  case (parn c1 n c2)
  obtain \<alpha> \<beta> where n: "n = (\<alpha>, \<beta>)" by fastforce
  have unf: "\<forall>a\<in>XA.
          tr' = [] \<and> (\<exists>X1 X2 t1 t2 s1 s2.
          N\<^sub>p s1 \<and> N\<^sub>p s2 \<and> (a) \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<and>
          (tr \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>c1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')} \<and>
          (tr \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')} \<and>
          set tr \<subseteq> \<alpha> \<union> \<beta> \<and> s = par t1 t2 \<and> s' = par s1 s2)"
    using parn(3)[unfolded n sfsemp_par, simplified] by simp
  then obtain s1 s2 t1 t2 where
    s: "s = par t1 t2" and
    s': "s' = par s1 s2" and
    tr: "set tr \<subseteq> \<alpha> \<union> \<beta>" and
    tr': "tr' = []" and
    Ns1: "N\<^sub>p s1" and Ns2: "N\<^sub>p s2"
    using parn(3-5) by fastforce
  have children: "\<forall>a\<in>XA. \<exists>X1 X2 . 
          (a) \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<and>
          (tr \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>c1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')} \<and>
          (tr \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}"
    using s s' unf by fastforce
  define X1A :: "'a set set" where X1A: "X1A = { X1 . \<exists> a X2 . a \<in> XA \<and>
          (a) \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<and>
          (tr \<restriction> \<alpha>, t1, X1) \<in> (\<lbrakk>c1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}) \<and>
          (tr \<restriction> \<beta>, t2, X2) \<in> (\<lbrakk>c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')})}"
  have eq: "\<And> s . {} \<union> s = s" by simp
  have g1: "\<forall>a\<in>X1A. (tr \<restriction> \<alpha>, t1, a) \<in> \<lbrakk>c1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}" unfolding eq using X1A children
    by blast
  have X1A_not_empty: "X1A \<noteq> {}" using X1A parn(4)
    using Collect_empty_eq children by auto
  have g1: "(tr \<restriction> \<alpha>, t1, \<Union>X1A) \<in> \<lbrakk>c1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}"
    using parn(1)[OF g1 X1A_not_empty] parn(5) wff\<^sub>p_inherit n unfolding eq by blast
  define X2A :: "'a set set" where X2A: "X2A = { X2 . \<exists> a X1 . a \<in> XA \<and>
          (a) \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<and>
          (tr \<restriction> \<alpha>, t1, X1) \<in> (\<lbrakk>c1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}) \<and>
          (tr \<restriction> \<beta>, t2, X2) \<in> (\<lbrakk>c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')})}"
  have eq: "\<And> s . {} \<union> s = s" by simp
  have g2: "\<forall>a\<in>X2A. (tr \<restriction> \<beta>, t2, a) \<in> \<lbrakk>c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}" unfolding eq using X2A children
    by blast
  have X2A_not_empty: "X2A \<noteq> {}" using X2A parn(4)
    using Collect_empty_eq children by auto
  have g2: "(tr \<restriction> \<beta>, t2, \<Union>X2A) \<in> \<lbrakk>c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}"
    using parn(2)[OF g2 X2A_not_empty] parn(5) wff\<^sub>p_inherit n unfolding eq by blast
  have refsets: "(\<Union>XA) \<inter> (\<alpha> \<union> \<beta>) = \<Union>X1A \<inter> \<alpha> \<union> \<Union>X2A \<inter> \<beta>"
  proof -
    {
      fix e
      assume e: "e \<in> (\<Union>XA) \<inter> (\<alpha> \<union> \<beta>)"
      have "e \<notin> \<Union>X1A \<inter> \<alpha> \<longrightarrow> e \<in> \<Union>X2A \<inter> \<beta>"
      proof 
        assume notInX1A: "e \<notin> \<Union>X1A \<inter> \<alpha>"
        obtain xa where xa: "xa \<in> XA" and e: "e \<in> xa \<inter> (\<alpha> \<union> \<beta>)" using parn(4) e by blast
        obtain X1 X2 where
          refsets: "(xa) \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
          sf1: "(tr \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>c1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}" and
          sf2: "(tr \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}"
          using children xa by meson
        have X1inX1A: "X1 \<in> X1A" unfolding X1A using refsets sf1 sf2 xa by blast
        have X2inX2A: "X2 \<in> X2A" unfolding X2A using refsets sf1 sf2 xa by blast
        have "e \<in> X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" using e refsets by blast
        thus "e \<in> \<Union>X2A \<inter> \<beta>"
          using notInX1A X1inX1A X2inX2A
          by blast
      qed
      hence "e \<in> \<Union>X1A \<inter> \<alpha> \<union> \<Union>X2A \<inter> \<beta>" by blast
    } moreover {
      fix e
      assume e: "e \<in> \<Union>X1A \<inter> \<alpha> \<union> \<Union>X2A \<inter> \<beta>"
      then consider (inX1A) "e \<in> \<Union>X1A \<inter> \<alpha>" | (inX2A) "e \<in> \<Union>X2A \<inter> \<beta>" by blast
      hence "e \<in> (\<Union>XA) \<inter> (\<alpha> \<union> \<beta>)"
      proof cases
        case inX1A
        then obtain xa X1 X2 where
          xa: "xa \<in> XA" and
          refsets: "(xa) \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
          sf1: "(tr \<restriction> \<alpha>, t1, X1) \<in> (\<lbrakk>c1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')})" and
          sf2: "(tr \<restriction> \<beta>, t2, X2) \<in> (\<lbrakk>c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')})" and
          e': "e \<in> X1 \<inter> \<alpha>"
          unfolding X1A by blast
        have X1inX1A: "X1 \<in> X1A" unfolding X1A using refsets sf1 sf2 xa by blast
        thus ?thesis
          using e' refsets xa by blast
      next
        case inX2A
        then obtain xa X1 X2 where
          xa: "xa \<in> XA" and
          refsets: "(xa) \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
          sf1: "(tr \<restriction> \<alpha>, t1, X1) \<in> (\<lbrakk>c1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')})" and
          sf2: "(tr \<restriction> \<beta>, t2, X2) \<in> (\<lbrakk>c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')})" and
          e': "e \<in> X2 \<inter> \<beta>"
          unfolding X2A by blast
        have X1inX1A: "X2 \<in> X2A" unfolding X2A using refsets sf1 sf2 xa by blast
        thus ?thesis
          using e' refsets xa by blast
      qed
    } ultimately show ?thesis by blast
  qed
  show "(tr, s, \<Union>XA) \<in> \<lbrakk>parn c1 n c2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s', X')}"
    unfolding n sfsemp_par s' s
    using tr' refsets g1 g2 Ns1 Ns2 tr
    by blast
next
  fix \<Sigma> tr s X tr' s' X' XA and c :: "('s, 'e) structured_code"
  assume
    sf: "\<forall>a\<in>XA. (tr, s, a)
                 \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> c (parallelFailureSetToFailureSet {(tr', s', X')}))" and
    not_empty: "XA \<noteq> {}" and
    wff: "wff\<^sub>s\<^sub>c c" and
    ax: "liSet (\<union>\<^sub>s\<^sub>c c)"
  have ax: "SFsemSigma (\<union>\<^sub>s\<^sub>c c)" using ax SFsemSigma_def by blast
  obtain ss where s: "s = ATree.seq ss"
  proof -
    assume th: "\<And>ss. s = ATree.seq ss \<Longrightarrow> thesis"
    { assume "\<not> (\<exists> ss . s = seq ss)"
      then obtain s1 s2 where s: "s = par s1 s2" using ATree.exhaust by metis
      obtain xa where xa: "xa \<in> XA" using not_empty by blast
      hence "(tr, s, xa) \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> c (parallelFailureSetToFailureSet {(tr', s', X')}))"
        using sf by blast
      hence False using s
          unfolding failureSetToParallelFailureSet_def by fast
    } thus ?thesis using th by blast
  qed
  have insf: "\<forall>a\<in>XA. (tr, ss, a) \<in> SFsemSigma.sfsem \<Sigma> c (parallelFailureSetToFailureSet {(tr', s', X')})"
    using sf not_empty unfolding failureSetToParallelFailureSet_def s by fast
  obtain s's where eq: "{(tr', s's, X')} = parallelFailureSetToFailureSet {(tr', s', X')}" and s': "s' = seq s's"
    using insf SFsemSigma.sfsem_empty[OF ax] not_empty
    unfolding parallelFailureSetToFailureSet_def
    by (metis (mono_tags) empty_iff equals0I parallelFailureSetToFailureSet_def parallelFailureSingletonToFailureSingleton)
  show "(tr, s, \<Union>XA) \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> c (parallelFailureSetToFailureSet {(tr', s', X')}))"
    using SFsemSigma.sfsem_refusal_set_Union[OF ax insf[unfolded eq[symmetric]] not_empty wff]
    unfolding eq  failureSetToParallelFailureSet_def s
    by blast
qed

text {* We can intersect the start and resulting refusal set by some other common set. *}

lemma sfsemp_refusal_set_intersection:
  assumes
    in_sf_x: "(tr, s, X) \<in> sfsem\<^sub>p code {(tr', s', Xs)}" and
    wff: "wff\<^sub>p code"
  shows "(tr, s, X \<inter> A) \<in> sfsem\<^sub>p code {(tr', s', Xs \<inter> A)}"
    using assms
proof (induct code arbitrary: tr s X s')
  case (seqn x)
  obtain ss ss' where
    s: "s = seq ss" and
    s': "s' = seq ss'"
    using sfsemp_predecessor[OF seqn(1,2)]
    unfolding treen_mirror_unfold_seqr
    by fast
  obtain c \<Sigma> where x: "x = (c, \<Sigma>)" by fastforce
  have ex: "(tr, ss, X) \<in> SFsemSigma.sfsem \<Sigma> c {(tr', ss', Xs)}"
    using seqn(1) unfolding sfsemp_seq_unfold s s' x .
  have sf_axiom: "SFsemSigma (\<union>\<^sub>s\<^sub>c c)"
    using seqn(2)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] SFsemSigma_def by auto
  have wff: "wff\<^sub>s\<^sub>c c"
    using seqn(2)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] by simp
  have set_eq1: "X - (UNIV - A) = X \<inter> A" by auto
  have set_eq2: "Xs - (UNIV - A) = Xs \<inter> A" by auto
  have "(tr, ss, X \<inter> A) \<in> SFsemSigma.sfsem \<Sigma> c {(tr', ss', Xs \<inter> A)}"
    unfolding x sfsemp_seq s s'
    using SFsemSigma.sfsem_refusal_substraction[OF sf_axiom ex wff, of "UNIV - A"]
    unfolding set_eq1 set_eq2 .
  thus ?case
    unfolding sfsemp_seq_unfold x s' s
    using SFsemSigma.sfsem_predecessor2 SFsemSigma.sfsem_sigma_irrelevant_single sf_axiom wff by fastforce
next
  case (parn code1 \<alpha>\<beta> code2)
  obtain \<alpha> \<beta> where \<alpha>\<beta>: "\<alpha>\<beta> = (\<alpha>, \<beta>)" by fastforce
  have wff1: "wff\<^sub>p code1" using parn(4) \<alpha>\<beta> wff\<^sub>p_inherit by blast
  have wff2: "wff\<^sub>p code2" using parn(4) \<alpha>\<beta> wff\<^sub>p_inherit by blast
  obtain s1 s2 s1' s2' where
    s: "s = par s1 s2" and
    s': "s' = par s1' s2'"
    using parn(3)[unfolded \<alpha>\<beta>, simplified]
    by blast
  moreover obtain X1 X2 where
    inS: "([], par s1' s2', Xs) \<in> {(tr', s', Xs)}" and
    Ns1: "N\<^sub>p s1'" and Ns2: "N\<^sub>p s2'" and
    X: "X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
    sf1: "(tr \<restriction> \<alpha>, s1, X1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s1', Xs)}" and
    sf2: "(tr \<restriction> \<beta>, s2, X2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s2', Xs)}" and
    tr: "set tr \<subseteq> \<alpha> \<union> \<beta>"
    using parn(3)[unfolded \<alpha>\<beta>, simplified] s s'
    by blast
  moreover have ex1: "(tr \<restriction> \<alpha>, s1, X1 \<inter> A) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s1', Xs \<inter> A)}"
    using parn(1)[OF sf1 wff1] .
  moreover have ex2: "(tr \<restriction> \<beta>, s2, X2 \<inter> A) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s2', Xs \<inter> A)}"
    using parn(2)[OF sf2 wff2] .
  moreover have "(X \<inter> A) \<inter> (\<alpha> \<union> \<beta>) = (X1 \<inter> A) \<inter> \<alpha> \<union> (X2 \<inter> A) \<inter> \<beta>"
    using X by blast
  ultimately show ?case
    unfolding \<alpha>\<beta> by auto
qed

text {* A reformulation of the above one. *}

lemma sfsemp_refusal_set_subtraction:
  fixes R
  assumes
    ex: "(tr, s, X) \<in> sfsem\<^sub>p code {(tr', s', X')}" and
    wff: "wff\<^sub>p code"
  shows "(tr, s, X - R) \<in> sfsem\<^sub>p code  {(tr', s', X' - R)}"
proof -
  have eq1: "X \<inter> (UNIV - R) = X - R" by auto
  have eq2: "X' \<inter> (UNIV - R) = X' - R" by auto
  show ?thesis
    using sfsemp_refusal_set_intersection[OF ex wff, of "UNIV - R"]
    unfolding eq1 eq2 .
qed



text {*
 If we extend the start refusal set @{term Xs} with some set @{term X'}, the original resulting
 refusal set @{term X} has to be extended, too, by some unknown subset of @{term X'}. *}

lemma sfsemp_extend_start_refusal_set:
  assumes
    in_sf_x: "(tr, s, X) \<in> sfsem\<^sub>p code {(tr', s', Xs)}" and
    Xs_subs_\<Sigma>: "Xs \<subseteq> \<Sigma>" and
    subs: "X \<union> X' \<subseteq> \<Sigma>" and
    wff: "wff\<^sub>p code" and
    wff_sigma: "wff_p_sigma code \<Sigma>"
  shows "\<exists> XR . X \<subseteq> XR \<and> XR \<subseteq> X \<union> X' \<and> (tr, s, XR) \<in> sfsem\<^sub>p code {(tr', s', Xs \<union> X')}"
  using in_sf_x subs wff wff_sigma 
proof (induct code arbitrary: tr s X s')
  case (seqn x)
  obtain ss ss' where
    s: "s = seq ss" and
    s': "s' = seq ss'"
    using sfsemp_predecessor[OF seqn(1,3)]
    unfolding treen_mirror_unfold_seqr
    by fast
  obtain c \<Sigma> where x: "x = (c, \<Sigma>)" by fastforce
  have ex: "(tr, ss, X) \<in> SFsemSigma.sfsem \<Sigma> c {(tr', ss', Xs)}"
    using seqn(1) unfolding sfsemp_seq_unfold s s' x .
  have sf_axiom: "SFsemSigma (\<union>\<^sub>s\<^sub>c c)"
    using seqn(3)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] SFsemSigma_def by auto
  have wff: "wff\<^sub>s\<^sub>c c"
    using seqn(3)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] by simp
  consider
    (is_fresh) "SFsemSigma.sfsem_fresh \<Sigma> (tr, ss, X) c {(tr', ss', Xs)}" |
    (not_fresh) "\<not> SFsemSigma.sfsem_fresh \<Sigma> (tr, ss, X) c {(tr', ss', Xs)}"
    by blast
  hence "(tr, ss, X) \<in> SFsemSigma.sfsem \<Sigma> c {(tr', ss', Xs \<union> X')} \<or> (tr, ss, X \<union> X') \<in> SFsemSigma.sfsem \<Sigma> c {(tr', ss', Xs \<union> X')}"
  proof cases
    case is_fresh
    then show ?thesis
      using SFsemSigma.sfsem_sigma_irrelevant_single[
          OF sf_axiom
        ]
      unfolding SFsemSigma.sfsem_fresh[OF sf_axiom]
      using wff by blast
  next
    case not_fresh
    hence "(tr, ss, X) = (tr', ss', Xs)"
      using sf_axiom wff
      using SFsemSigma.sfsem_freshI1 ex by blast
    then show ?thesis
      using sf_axiom wff ex
      using SFsemSigma.pc_does_not_point_to_single SFsemSigma.sfsem_sigma_irrelevant_single by fastforce
  qed
  then show ?case
    unfolding sfsemp_seq_unfold[symmetric] x s' s
    by auto
next
  case (parn code1 \<alpha>\<beta> code2)
  obtain \<alpha> \<beta> where \<alpha>\<beta>: "\<alpha>\<beta> = (\<alpha>, \<beta>)" by fastforce
  have wff1: "wff\<^sub>p code1" using parn(5) \<alpha>\<beta> wff\<^sub>p_inherit by blast
  have wff2: "wff\<^sub>p code2" using parn(5) \<alpha>\<beta> wff\<^sub>p_inherit by blast
  have wffS1: "wff_p_sigma code1 \<Sigma>" using parn(6) \<alpha>\<beta> wff_p_sigma_inherit by blast
  have wffS2: "wff_p_sigma code2 \<Sigma>" using parn(6) \<alpha>\<beta> wff_p_sigma_inherit by blast
  obtain s1 s2 s1' s2' where
    s: "s = par s1 s2" and
    s': "s' = par s1' s2'"
    using parn(3)[unfolded \<alpha>\<beta>, simplified]
    by blast
  obtain X1 X2 where
    inS: "([], par s1' s2', Xs) \<in> {(tr', s', Xs)}" and
    Ns1: "N\<^sub>p s1'" and Ns2: "N\<^sub>p s2'" and X: "X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
    sf1: "(tr \<restriction> \<alpha>, s1, X1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s1', Xs)}" and
    sf2: "(tr \<restriction> \<beta>, s2, X2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s2', Xs)}" and
    tr: "set tr \<subseteq> \<alpha> \<union> \<beta>"
    using parn(3)[unfolded \<alpha>\<beta>, simplified] s s' by blast
  have sf1: "(tr \<restriction> \<alpha>, s1, X1 \<inter> \<Sigma>) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s1', Xs)}"
    using sfsemp_refusal_set_intersection[OF _ wff, of _]
    by (metis Xs_subs_\<Sigma> inf.orderE sf1 sfsemp_refusal_set_intersection wff1)
  have sf2: "(tr \<restriction> \<beta>, s2, X2 \<inter> \<Sigma>) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s2', Xs)}"
    using sfsemp_refusal_set_intersection[OF _ wff, of _]
    by (metis Xs_subs_\<Sigma> inf.orderE sf2 sfsemp_refusal_set_intersection wff2)
  have "\<exists>XR. (X1 \<inter> \<Sigma>) \<subseteq> XR \<and> XR \<subseteq> (X1 \<inter> \<Sigma>) \<union> X' \<and> (tr \<restriction> \<alpha>, s1, XR) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s1', Xs \<union> X')}"
  proof (rule parn(1)[OF sf1 _ wff1 wffS1])
    show "X1 \<inter> \<Sigma> \<union> X' \<subseteq> \<Sigma>" using parn(4) by simp
  qed
  then obtain XR1 where
    XR1_bounds: "(X1 \<inter> \<Sigma>) \<subseteq> XR1 \<and> XR1 \<subseteq> (X1 \<inter> \<Sigma>) \<union> X'" and
    sf1XR1: "(tr \<restriction> \<alpha>, s1, XR1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s1', Xs \<union> X')}"
    by blast
  have "\<exists>XR. (X2 \<inter> \<Sigma>) \<subseteq> XR \<and> XR \<subseteq> (X2 \<inter> \<Sigma>) \<union> X' \<and> (tr \<restriction> \<beta>, s2, XR) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s2', Xs \<union> X')}"
  proof (rule parn(2)[OF sf2 _ wff2 wffS2])
    show "X2 \<inter> \<Sigma> \<union> X' \<subseteq> \<Sigma>" using parn(4) by simp
  qed
  then obtain XR2 where
    XR2_bounds: "(X2 \<inter> \<Sigma>) \<subseteq> XR2 \<and> XR2 \<subseteq> (X2 \<inter> \<Sigma>) \<union> X'" and
    sf2XR2: "(tr \<restriction> \<beta>, s2, XR2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s2', Xs \<union> X')}"
    by blast
  have interface_bounded: "\<alpha> \<subseteq> \<Sigma> \<and> \<beta> \<subseteq> \<Sigma>"
    using parn(6)[unfolded \<alpha>\<beta> wff_p_sigma_def node1_property.simps]
    by fast
  define XR where XR: "XR = X \<union> (XR1 \<inter> \<alpha>) \<union> (XR2 \<inter> \<beta>)"
  have "XR \<inter> (\<alpha> \<union> \<beta>) = XR1 \<inter> \<alpha> \<union> XR2 \<inter> \<beta>"
    unfolding XR 
    using XR1_bounds XR2_bounds X interface_bounded parn(4)
    by blast
  moreover have "X \<subseteq> XR \<and> XR \<subseteq> X \<union> X'"
    unfolding XR 
    using XR1_bounds XR2_bounds X interface_bounded parn(4)
    by blast
  ultimately show ?case
    unfolding \<alpha>\<beta> sfsemp_par s s'
    using sf1XR1 sf2XR2 Ns1 Ns2 inS tr
  proof -
    have "(tr, par s1 s2, XR) \<in> {(as, a, A). \<exists>Aa Ab Ac aa ab ac ad ae. ([], ae, Ac) \<in> {(tr', par s1' s2', Xs \<union> X')} \<and> N\<^sub>p ac \<and> N\<^sub>p ad \<and> A \<inter> (\<alpha> \<union> \<beta>) = Aa \<inter> \<alpha> \<union> Ab \<inter> \<beta> \<and> (as \<restriction> \<alpha>, aa, Aa) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], ac, Ac)} \<and> (as \<restriction> \<beta>, ab, Ab) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], ad, Ac)} \<and> set as \<subseteq> \<alpha> \<union> \<beta> \<and> a = par aa ab \<and> ae = par ac ad}"
      using Ns1 Ns2 \<open>XR \<inter> (\<alpha> \<union> \<beta>) = XR1 \<inter> \<alpha> \<union> XR2 \<inter> \<beta>\<close> inS sf1XR1 sf2XR2 tr by blast
    then show "\<exists>A. X \<subseteq> A \<and> A \<subseteq> X \<union> X' \<and> (tr, par s1 s2, A) \<in> {(as, a, A). \<exists>Aa Ab Ac aa ab ac ad ae. ([], ae, Ac) \<in> {(tr', par s1' s2', Xs \<union> X')} \<and> N\<^sub>p ac \<and> N\<^sub>p ad \<and> A \<inter> (\<alpha> \<union> \<beta>) = Aa \<inter> \<alpha> \<union> Ab \<inter> \<beta> \<and> (as \<restriction> \<alpha>, aa, Aa) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], ac, Ac)} \<and> (as \<restriction> \<beta>, ab, Ab) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], ad, Ac)} \<and> set as \<subseteq> \<alpha> \<union> \<beta> \<and> a = par aa ab \<and> ae = par ac ad}"
      by (meson \<open>X \<subseteq> XR \<and> XR \<subseteq> X \<union> X'\<close>)
  qed
qed

text {* If we know of two start / result refusal set pairs, we can merge them accordingly. *}

lemma sfsemp_refusal_set_merge:
  assumes
    in_sf_x: "(tr, s, X1) \<in> sfsem\<^sub>p code {([], s', Xs1)}" and
    in_sf_x: "(tr, s, X2) \<in> sfsem\<^sub>p code {([], s', Xs2)}" and
    wff: "wff\<^sub>p code"
  shows "(tr, s, X1 \<union> X2) \<in> sfsem\<^sub>p code {([], s', Xs1 \<union> Xs2)}"
    using assms
proof (induct code arbitrary: tr s X1 X2 s')
  case (seqn x)
  obtain ss ss' where
    s: "s = seq ss" and
    s': "s' = seq ss'"
    using sfsemp_predecessor[OF seqn(1,3)]
    unfolding treen_mirror_unfold_seqr
    by fast
  obtain c \<Sigma> where x: "x = (c, \<Sigma>)" by fastforce
  have ex1: "(tr, ss, X1) \<in> SFsemSigma.sfsem \<Sigma> c {([], ss', Xs1)}"
    using seqn(1) unfolding sfsemp_seq_unfold s s' x .
  have ex2: "(tr, ss, X2) \<in> SFsemSigma.sfsem \<Sigma> c {([], ss', Xs2)}"
    using seqn(2) unfolding sfsemp_seq_unfold s s' x .
  have sf_axiom: "SFsemSigma (\<union>\<^sub>s\<^sub>c c)"
    using seqn(3)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] SFsemSigma_def by auto
  have wff: "wff\<^sub>s\<^sub>c c"
    using seqn(3)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] by simp
  consider
    (is_fresh) "SFsemSigma.sfsem_fresh \<Sigma> (tr, ss, X1) c {([], ss', Xs1)}" |
    (not_fresh) "\<not> SFsemSigma.sfsem_fresh \<Sigma> (tr, ss, X1) c {([], ss', Xs1)}"
    by blast
  thus ?case
    unfolding sfsemp_seq_unfold x s' s
  proof cases
    case is_fresh
    have nice_to_have: "NCPC ss' \<in> definitions.dom c \<and> N ss' \<and> (tr, ss, X1) \<in> SFsemSigma.sfsem \<Sigma> c {([], ss', Xs1)}"
      using is_fresh[unfolded SFsemSigma.sfsem_fresh[OF sf_axiom], simplified] .
    consider
      (self_cbr) "(tr, ss, Xs1) \<in> {([], ss', Xs1)}" |
      (good_instruction) "(tr, ss, Xs1) \<notin> {([], ss', Xs1)}"
      by blast
    hence "(tr, ss, X1 \<union> X2) \<in> SFsemSigma.sfsem \<Sigma> c {([], ss', Xs1)}"
    proof cases
      case self_cbr
      show ?thesis
        using SFsemSigma.sfsem_not_outside_means_cbr[OF sf_axiom ex1]
        using SFsemSigma.sfsem_not_outside_means_cbr[OF sf_axiom ex2]
        using self_cbr nice_to_have by fastforce
    next
      case good_instruction
      hence "\<forall> X . (tr, ss, X) \<notin> {([], ss', Xs1)}"
        by blast
      hence "X1 \<union> X2 \<subseteq> \<Sigma>"
        using SFsemSigma.sfsem_refusal_subset_of_sigma[OF sf_axiom ex1]
        using SFsemSigma.sfsem_refusal_subset_of_sigma[OF sf_axiom ex2]
        by blast
      then show ?thesis
        by (metis (no_types, lifting) SFsemSigma.sfsem_greatest_refusal SFsemSigma.sfsem_sigma_irrelevant_single Un_subset_iff \<open>\<forall>X. (tr, ss, X) \<notin> {([], ss', Xs1)}\<close> ex2 nice_to_have sf_axiom wff)
    qed
    thus "(tr, ss, X1 \<union> X2) \<in> SFsemSigma.sfsem \<Sigma> c {([], ss', Xs1 \<union> Xs2)}"
      using SFsemSigma.sfsem_sigma_irrelevant_single nice_to_have sf_axiom wff by blast
  next
    case not_fresh
    then show "(tr, ss, X1 \<union> X2) \<in> SFsemSigma.sfsem \<Sigma> c {([], ss', Xs1 \<union> Xs2)}"
      by (metis SFsemSigma.pc_does_not_point_to_single SFsemSigma.sfsem_fresh SFsemSigma.sfsem_predecessor2 ex1 ex2 sf_axiom singletonI)
  qed
next
  case (parn code1 \<alpha>\<beta> code2)
  obtain \<alpha> \<beta> where \<alpha>\<beta>: "\<alpha>\<beta> = (\<alpha>, \<beta>)" by fastforce
  have wff1: "wff\<^sub>p code1" using parn(5) \<alpha>\<beta> wff\<^sub>p_inherit by blast
  have wff2: "wff\<^sub>p code2" using parn(5) \<alpha>\<beta> wff\<^sub>p_inherit by blast
  obtain s1 s2 s1' s2' where
    s: "s = par s1 s2" and
    s': "s' = par s1' s2'"
    using parn(3)[unfolded \<alpha>\<beta>, simplified]
    by blast
  moreover obtain X11 X12 X21 X22 where
    inS: "([], par s1' s2', Xs1) \<in> {([], s', Xs1)}" and
    inS: "([], par s1' s2', Xs2) \<in> {([], s', Xs2)}" and
    Ns1: "N\<^sub>p s1'" and Ns2: "N\<^sub>p s2'" and
    X1: "X1 \<inter> (\<alpha> \<union> \<beta>) = X11 \<inter> \<alpha> \<union> X12 \<inter> \<beta>" and
    X2: "X2 \<inter> (\<alpha> \<union> \<beta>) = X21 \<inter> \<alpha> \<union> X22 \<inter> \<beta>" and
    sf11: "(tr \<restriction> \<alpha>, s1, X11) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1', Xs1)}" and
    sf12: "(tr \<restriction> \<beta>, s2, X12) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2', Xs1)}" and
    sf21: "(tr \<restriction> \<alpha>, s1, X21) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1', Xs2)}" and
    sf22: "(tr \<restriction> \<beta>, s2, X22) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2', Xs2)}" and
    tr: "set tr \<subseteq> \<alpha> \<union> \<beta>"
    using parn(3)[unfolded \<alpha>\<beta>, simplified] parn(4)[unfolded \<alpha>\<beta>, simplified] s s'
    by blast
  moreover have ex1: "(tr \<restriction> \<alpha>, s1, X11 \<union> X21) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1', Xs1 \<union> Xs2)}"
    using parn(1)[OF sf11 sf21 wff1] .
  moreover have ex2: "(tr \<restriction> \<beta>, s2, X12 \<union> X22) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2', Xs1 \<union> Xs2)}"
    using parn(2)[OF sf12 sf22 wff2] .
  moreover have "(X1 \<union> X2) \<inter> (\<alpha> \<union> \<beta>) = (X11 \<union> X21) \<inter> \<alpha> \<union> (X12 \<union> X22) \<inter> \<beta>"
    using X1 X2 by blast
  ultimately show ?case
    unfolding \<alpha>\<beta> by auto
qed

text {* The start refusal set @{term Xs} can be replaced by the resulting refusal set @{term X}. *}

lemma sfsemp_replace_start_refusal_set_by_result_refusal_set:
  assumes
    ex: "(tr, s, X) \<in> \<lbrakk>code\<rbrakk>\<^sub>p\<^sub>S\<^sub>F {(tr', s', Xs)}" and
    X_subs_\<Sigma>: "X \<subseteq> \<Sigma>" and
    wff: "wff\<^sub>p code" and
    wff_p_sigma: "wff_p_sigma code \<Sigma>"
  shows "(tr, s, X) \<in> \<lbrakk>code\<rbrakk>\<^sub>p\<^sub>S\<^sub>F {(tr',  s', X)}"
proof -
  have *: "(tr, s, X) \<in> \<lbrakk>code\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{(tr', s', Xs \<inter> X)}"
    using sfsemp_refusal_set_intersection[OF ex wff, of X, simplified]
    by simp
  have hlpr: "Xs \<inter> X \<union> X = X"
    by blast
  show ?thesis
    using sfsemp_extend_start_refusal_set[OF * _ _ wff wff_p_sigma, of X, simplified]
    using X_subs_\<Sigma>
    unfolding hlpr
    by fastforce
qed

subsection {* Connecting Property and Refinement *}

text {* In the following definitions, we always need to carry the process name function as it is used in the
CSP Prover. For its documentation, refer to the CSP Prover package.
In general, we have two basic concepts of the CUC-CSP refinement. One with the user definable connecting property,
the other one without a connecting property. Also, both for SF semantics and T semantics.
A further split is due to the starting property, which in general can be generic. There is also a predefined precondition,
which fixes the starting program counter to be 1, the state to be a normal state, and the trace to be empty. The overall idea is,
that each code probably starts the execution at some program counter, which we simply fixed arbitrarily. Also, the trace should always
be empty in the beginning. Further assumptions concerning the state (i.e. initial values of the registers) can be made
explicit via some initializing instruction(s). Hence, this fixed starting property should be sufficient. Still, we want
to praise genericity and thus allow for arbitrary starting properties.
A third parameter, being required in any case, is the conversion function to relate the CUC event type to the CSP event type.
*}

paragraph {* Preconditions *}

text {* The precondition for start states - we want them all to start at a welldefined pc, be normal
and have an empty trace. *}

definition "PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q x = (case x of (tr, s, _) \<Rightarrow> NCPC s = 1 \<and> N s \<and> tr = [])"
definition "PreStd\<^sub>S\<^sub>F x = (case x of (tr, s, _) \<Rightarrow> AllPC\<^sub>S\<^sub>F s 1 \<and> N\<^sub>p s \<and> tr = [])"
lemma PreStd_prod: "PreStd\<^sub>S\<^sub>F x \<Longrightarrow> x = (tr, s, X) \<Longrightarrow> AllPC\<^sub>S\<^sub>F s 1 \<and> N\<^sub>p s \<and> tr = []"
  using PreStd\<^sub>S\<^sub>F_def by fastforce

lemma PreStd_conv: "parallelFailureSetToFailureSet (Collect PreStd\<^sub>S\<^sub>F) = Collect PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q"
  unfolding parallelFailureSetToFailureSet_def PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q_def PreStd\<^sub>S\<^sub>F_def
    AllPC\<^sub>S\<^sub>F_def 
proof -
  define S1 :: "('a list \<times> 'b NCstate \<times> 'a set) set" where
    S1: "S1 = {x. \<exists>x'. x' \<in> {x. case x of (tr, s, xa) \<Rightarrow> leaf1_property (\<lambda>s. NCPC s = 1) op \<and> s \<and> N\<^sub>p s \<and> tr = []} \<and> x \<in> parallelFailureToFailure x'}"
  define S2 :: "('a list \<times> 'b NCstate \<times> 'a set) set" where
    S2: "S2 = {x. case x of (tr, s, xa) \<Rightarrow> NCPC s = 1 \<and> N s \<and> tr = []}"
  {
    fix tr s X
    assume inS1: "(tr, s, X) \<in> S1"
    obtain ss X' where
      lp: "leaf1_property (\<lambda>s. NCPC s = Suc 0) op \<and> ss" and
      Na: "N\<^sub>p ss" and
      conv: "(tr, s, X) \<in> parallelFailureToFailure ([], ss, X')"
      using inS1[unfolded S1, simplified] by blast
    have ss:"ss = seq s" using conv
      by (metis Pair_inject parallelFailureToFailure_ex_seq)
    have "(tr, s, X) \<in> S2"
      using lp Na conv
      unfolding S2 ss leaf1_property.simps N\<^sub>p_def parallelFailureToFailure.simps
      by fastforce
  } moreover {
    fix tr s X
    assume inS2: "(tr, s, X) \<in> S2"
    have conds: "NCPC s = Suc 0 \<and> N s \<and> tr = []"
      using inS2[unfolded S2, simplified] .
    hence "(tr, s, X) \<in> parallelFailureToFailure ([], seq s, X)"
      unfolding parallelFailureToFailure.simps by simp
    moreover
    have "([], seq s, X) \<in> {(tr, s, xa). leaf1_property (\<lambda>s. NCPC s = 1) op \<and> s \<and> N\<^sub>p s \<and> tr = []}"
      using conds leaf1_property.simps
      by (simp add: N\<^sub>p_def)
    ultimately
    have "(tr, s, X) \<in> S1"
      unfolding S1
      by blast      
  } ultimately show "S1 = S2" by fast    
qed  

text {* Converts a predicate over a non-parallel stable failure to a parallel stable failure, considering only the cases where
the state is a single state.*}

definition "CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre = (\<lambda> x . case x of (tr, seq s', X) \<Rightarrow> Pre (tr, s', X) | _ \<Rightarrow> False)"

fun parallelState_is_NCstate :: "('s) parallelState \<Rightarrow> bool" where
"parallelState_is_NCstate (seq _) = True" |
"parallelState_is_NCstate (par _ _) = False"

text {* If we have an NCstate, then it has the same height as another tree of size 1 (being just a single node). *}

lemma parallelState_eq: "parallelState_is_NCstate s = treen_mirror s (seqn (a, b, c))"
  unfolding treen_mirror
  by (metis leafn_property.simps(1) leafn_property_seql parallelState_is_NCstate.elims(2) parallelState_is_NCstate.simps(1))

text {* Unfolding the short definition into a longer one. *}

lemma convert_pre: "(CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q) = (\<lambda> x . PreStd\<^sub>S\<^sub>F x \<and> parallelState_is_NCstate (fst (snd x)))"
  unfolding PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q_def PreStd\<^sub>S\<^sub>F_def CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F_def
proof (auto, rule ext, rule iffI)
  fix x :: "'a list \<times> 'b NCstate ATree \<times> 'c"
  assume aa1: "case x of (tr, seq s', X) \<Rightarrow> NCPC s' = 1 \<and> N s' \<and> tr = []
         | (tr, par parallelState1 parallelState2, X) \<Rightarrow> False"
  obtain tr ss' X where x: "x = (tr, ss', X)" using prod_cases3 by blast
  then consider (s) "\<exists> s' . ss' = seq s'" | (p) "\<exists> cl cr . ss' = par cl cr"
    by (meson parallelState_is_NCstate.cases)
  thus "(case x of (tr, s, xa) \<Rightarrow> AllPC\<^sub>S\<^sub>F s (Suc 0) \<and> N\<^sub>p s \<and> tr = []) \<and> parallelState_is_NCstate (fst (snd x))"
  proof cases
    case s
    then obtain s' where ss': "ss' = seq s'" by blast
    hence "NCPC s' = Suc 0 \<and> N s' \<and> tr = []" using aa1[unfolded x ss', simplified] by simp
    hence "AllPC\<^sub>S\<^sub>F (seq s') (Suc 0) \<and> N\<^sub>p (seq s') \<and> tr = [] \<and> parallelState_is_NCstate (fst (snd x))"
      unfolding x ss' fst_conv snd_conv leaf1_property.simps "parallelState_is_NCstate.simps" "AllPC\<^sub>S\<^sub>F_def" N\<^sub>p_def by simp
    thus "(case x of (tr, s, xa) \<Rightarrow> AllPC\<^sub>S\<^sub>F s (Suc 0) \<and> N\<^sub>p s \<and> tr = []) \<and> parallelState_is_NCstate (fst (snd x))"
      unfolding x ss'
      by simp
  next
    case p
    with aa1 x have False by fastforce
    thus ?thesis by simp
  qed
next
  fix x :: "'a list \<times> 'b NCstate ATree \<times> 'c"
  assume aa1: "(case x of (tr, s, xa) \<Rightarrow> AllPC\<^sub>S\<^sub>F s (Suc 0) \<and> N\<^sub>p s \<and> tr = []) \<and> parallelState_is_NCstate (fst (snd x))"
  then obtain tr s X where x: "x = (tr, s, X)" by blast
  with aa1 have G: "AllPC\<^sub>S\<^sub>F s (Suc 0) \<and> N\<^sub>p s \<and> tr = [] \<and> parallelState_is_NCstate (fst (snd x))"
    by fast
  then obtain s' where s: "s = seq s'"
    unfolding fst_conv snd_conv x using "parallelState_is_NCstate.elims"(2)
    by metis
  have "NCPC s' = 1 \<and> N s' \<and> tr = []"
    using G
    unfolding s "AllPC\<^sub>S\<^sub>F_def" N\<^sub>p_def by simp
  thus "case x of (tr, seq s', X) \<Rightarrow> NCPC s' = 1 \<and> N s' \<and> tr = []
         | (tr, par parallelState1  parallelState2, X) \<Rightarrow> False"
    unfolding x s
    by auto
qed

lemma convert_pre2: "(CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q) = (\<lambda> x . PreStd\<^sub>S\<^sub>F x \<and> treen_mirror (fst (snd x)) (seqn (a, b, c)))"
  by (meson convert_pre parallelState_eq)

lemma convert_pre3: "Collect (CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q) = { x . PreStd\<^sub>S\<^sub>F x \<and> treen_mirror (fst (snd x)) (seqn (a, b, c))}"
  by (meson convert_pre parallelState_eq)

lemma seq_sf_refinement_set_eq:
  "{(tr, s, X) . \<exists> s' . (tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> Impl (Collect Pre) \<and> s = seq s'} =
   \<lbrakk>seqn (Impl, \<Sigma>)\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre x'}"
proof -
  have eq: "Collect Pre = (parallelFailureSetToFailureSet {x'. CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre x'})"
    unfolding parallelFailureSetToFailureSet_def CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F_def
  proof (auto)
    fix a aa b
    assume ca1: "Pre (a, aa, b)"
    hence "(case seq aa of seq aa \<Rightarrow> Pre (a, aa, b) | par parallelState1 parallelState2 \<Rightarrow> False) \<and>
               treen_mirror (seq aa) (seqn (Impl, \<Sigma>)) \<and> (a, aa, b) \<in> parallelFailureToFailure (a, seq aa, b)"
      unfolding "parallelFailureToFailure.simps"
      by simp
    thus "\<exists>aaa aaaa ba.
               (case aaaa of seq s' \<Rightarrow> Pre (aaa, s', ba) | par parallelState1 parallelState2 \<Rightarrow> False) \<and>
               (a, aa, b) \<in> parallelFailureToFailure (aaa, aaaa, ba)"
      by blast
  next
    fix a aa b ab ac ba
    assume
      seq: "case ac of ATree.seq s' \<Rightarrow> Pre (ab, s', ba) | par ATree1 ATree2 \<Rightarrow> False" and
      ins: "(a, aa, b) \<in> parallelFailureToFailure (ab, ac, ba)"
    obtain s' where ac: "ac = seq s'" using seq
      by (metis empty_iff ins parallelFailureToFailure.simps(2) parallelState_is_NCstate.cases)
    have eq: "(a, aa, b) = (ab, s', ba)" using ins unfolding ac parallelFailureToFailure.simps
      by simp
    thus "Pre (a, aa, b)"
      using ac seq
      by simp
  qed
  show ?thesis
  proof (auto simp del: treen_mirror_unfold_seqr)
    fix a b s'
    assume aa1: "(a, s', b) \<in> SFsemSigma.sfsem \<Sigma> Impl (Collect Pre)"
    
    show "(a, seq s', b)
                \<in> failureSetToParallelFailureSet
                    (SFsemSigma.sfsem \<Sigma> Impl (parallelFailureSetToFailureSet {x'. CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre x'}))"
      using aa1[unfolded eq]
      unfolding failureSetToParallelFailureSet_def
      by (metis (mono_tags, lifting) case_prodI eq mem_Collect_eq)
  next
    fix a aa b
    assume aa1: "(a, aa, b) \<in> failureSetToParallelFailureSet
                    (SFsemSigma.sfsem \<Sigma> Impl (parallelFailureSetToFailureSet {x'. CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre x'}))"
    thus "\<exists>s'. (a, s', b) \<in> SFsemSigma.sfsem \<Sigma> Impl (Collect Pre) \<and> aa = seq s'"
      unfolding eq
      unfolding failureSetToParallelFailureSet_def
      by simp
  qed
qed

lemma seq_sf_refinement_set_eq_pre_set:
  "wff\<^sub>p (seqn (Impl, \<Sigma>)) \<Longrightarrow> {(tr, s, X) . \<exists> s' . (tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> Impl (Collect PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q) \<and> s = seq s'} =
   \<lbrakk>seqn (Impl, \<Sigma>)\<rbrakk>\<^sub>p\<^sub>S\<^sub>F(Collect PreStd\<^sub>S\<^sub>F )"
  by (simp add: PreStd_conv failureSetToParallelFailureSet_def)

end
