section {* Stable Failures - Traces conformance *}

theory "stable_failures_traces_conformant"
imports denotational "stable_failures"
begin

subsection {* Preparation *}

text{* One general problem between the two semantics is their different types of states. Therefor,
we need to introduce transformations. *}

definition failureSetToTraceStateSet :: "('s, 'e) failure set \<Rightarrow> ('s, 'e) trace_state set" where
  "failureSetToTraceStateSet S = {(tr, s) . \<exists> X s' . (tr, s', X) \<in> S \<and> s = NCS s'}"

context SFsemSigma
begin

definition trace_state_set_to_failure_set :: "('s, 'e) trace_state set \<Rightarrow> ('s, 'e) failure set" where
  "trace_state_set_to_failure_set S = {(tr, s, X) . \<exists> s' . (tr, s') \<in> S \<and> s = normal s' \<and> X \<subseteq> \<Sigma>}"

text {* A sanity check: Adding information to a @{term "trace_state set"} and removing those later,
gives the original @{term "trace_state set"} back. *}

lemma setConversionIdentity:
  "failureSetToTraceStateSet o trace_state_set_to_failure_set = id"
  unfolding comp_def failureSetToTraceStateSet_def trace_state_set_to_failure_set_def 
  by fastforce
  
text {* If a new state is created in the traces semantics, this new state also appears in the SF semantics. *}

lemma densem_new_state_is_sfsem_new_state:
  assumes "(tr, s) \<in> (densem (one l i) {(tr', s')}) - {(tr', s')}" and "X' \<subseteq> \<Sigma>"
  shows "(tr, normal s, X') \<in> (sfsem (one l i) {(tr', normal s', X')}) - {(tr', normal s', X')}"
using assms
proof (induct i, auto) qed

text {* A critical execution path in the traces semantics is also a critical execution path in the 
stable failures semantics. Criticality in this case assures, that the resulting state won't be erased
in the SF semantics. *}

lemma densem_critical_path_is_sfsem_path:
  assumes
    a: "densem_exec_path_is_critical (tr, s) lil S" and
    b: "X' \<subseteq> \<Sigma>"
  shows
    "(tr, normal s, X') \<in> sfsem_exec lil (trace_state_set_to_failure_set S)"
using a b
proof (induct lil arbitrary: tr s X')
  show "\<And>tr s X.
       densem_exec_path_is_critical (tr, s) [] S \<Longrightarrow>
       X \<subseteq> \<Sigma> \<Longrightarrow>
       (tr, normal s, X)
       \<in> sfsem_exec []
           (trace_state_set_to_failure_set S)"
    unfolding trace_state_set_to_failure_set_def densem_exec_path_is_critical.simps
      densem_exec_def "rev.simps" "densem_exec_b.simps" sfsem_exec_def "sfsem_exec_b.simps"
    by blast
next
  fix li :: "nat \<times> ('s, 'e) instr" and lil :: "(nat \<times> ('s, 'e) instr) list" and
    tr :: "'e event list" and s :: "'s state" and X :: "'e event set"
  assume
    aa1: "\<And>tr s X. densem_exec_path_is_critical (tr, s) lil S \<Longrightarrow>
           X \<subseteq> \<Sigma> \<Longrightarrow> (tr, normal s, X) \<in> sfsem_exec lil (trace_state_set_to_failure_set S)" and
    aa2: "densem_exec_path_is_critical (tr, s) (li # lil) S" and
    aa3: "X \<subseteq> \<Sigma>"
  obtain tr' s' where
    ao1: "densem_exec_path_is_critical (tr, s) [li] {(tr', s')} \<and>
         densem_exec_path_is_critical (tr', s') lil S"
    using aa2 densem_exec_critical_path_unfold[of "(tr, s)" "[li]" lil S]
    by auto
  have a1: "(tr', normal s', X) \<in> sfsem_exec lil (trace_state_set_to_failure_set S)"
    using aa1[of tr' s' X, OF ao1[THEN conjunct2], OF aa3] 
    .
  obtain l i where
    ao2: "li = (l, i)"
    by fastforce
  from ao1[THEN conjunct1, unfolded ao2 densem_exec_path_is_critical.simps]
    densem_exec_unfold[of l i "[]" "{(tr', s')}", unfolded densem_exec_def[of "[]", unfolded rev.simps(1), unfolded densem_exec_b.simps(1)]]
    have a2: "(tr, s) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr', s')} \<and> (tr', s') \<noteq> (tr, s)"
    by blast
  have a3: "densem_exec_path_is_critical (tr', s') lil S"
    using ao1[unfolded ao2 densem_exec_path_is_critical.simps]
    by simp
  hence "(tr, s) \<in> \<lbrakk>one l i\<rbrakk> {(tr', s')} - {(tr', s')}"
    using a2
    unfolding densem_exec_empty densem_exec_unfold
    by fastforce
  with densem_new_state_is_sfsem_new_state[of tr s l i tr' s' X] aa3 have
    "(tr, normal s, X) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr', normal s', X)} - {(tr', normal s', X)}"
    by fastforce
  hence "(tr, normal s, X) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr', normal s', X)}"
    by auto
  with a1 have "(tr, normal s, X) \<in> \<lbrakk>one l i\<rbrakk> sfsem_exec lil (trace_state_set_to_failure_set S)"
    using contra_subsetD insert_is_Un mk_disjoint_insert sfsem_monotone sup.cobounded1
    by (metis (no_types, hide_lams) )
  thus
       "(tr, normal s, X)
       \<in> sfsem_exec (li # lil)
           (trace_state_set_to_failure_set S)"
    unfolding ao2
    using sfsem_exec_unfold[of l i lil "(trace_state_set_to_failure_set S)"]
    by auto
qed

text {* Every execution path in the SF semantics is also an execution path in the traces semantics.
As in the traces semantics states are not erased, criticality is not necessary. *}

lemma sfsem_path_is_densem_path:
  assumes
    a: "(tr, s, X') \<in> sfsem_exec lil S"
  shows
    "(tr, NCS s) \<in> densem_exec lil (failureSetToTraceStateSet S)"
using assms
proof (induct lil arbitrary: tr s X')
  show "\<And> tr s X . (tr, s, X) \<in> sfsem_exec [] S \<Longrightarrow>
    (tr, NCS s) \<in> densem_exec [] (failureSetToTraceStateSet S)"
    unfolding sfsem_exec_empty densem_exec_empty
     failureSetToTraceStateSet_def by blast
next
  fix a :: "label \<times> ('s, 'e) instr" and lil :: "(label \<times> ('s, 'e) instr) list" and
    tr :: "'e event list" and s :: "'s NCstate" and X :: "'e event set"
  assume
    aa1: "\<And> tr s X . (tr, s, X) \<in> sfsem_exec lil S \<Longrightarrow>
        (tr, NCS s)
        \<in> densem_exec lil
            (failureSetToTraceStateSet S)" and
    aa2: "(tr, s, X) \<in> sfsem_exec (a # lil) S"
  obtain l i where
    ao1: "a = (l, i)"
    by fastforce
  obtain tr' s' X' where
    ao2: "(tr', s', X') \<in> sfsem_exec lil S \<and> (tr, s, X) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr', s', X')}"
    using sfsem_predecessor[OF aa2[unfolded ao1 sfsem_exec_unfold]]
    by blast
  have a0: "(tr', NCS s') \<in> densem_exec lil (failureSetToTraceStateSet S)"
    using aa1[OF ao2[THEN cj2_1]] .
  have a1: "(tr, s, X) \<in> {(tr', s', X')} -\<^sub>s\<^sub>f {l} \<Longrightarrow> (tr, NCS s) \<in> {(tr', NCS s')}"
    by auto
  have a2: "(tr, NCS s) \<in> \<lbrakk>(one l i)\<rbrakk>{(tr', NCS s')}"
  proof(cases i)
    fix dof :: "'s do_fun"
    assume ba1: "i = do dof"
    {
      assume ca1: "(tr, s, X) \<in> {(tr, t, X).
      \<exists>u X'a. (tr, u, X'a) \<in> {(tr', s', X')} \<and> NCPC u = l \<and> N u \<and> N t \<and>
         NCPC t = Suc l \<and> NCRS t \<in> dof (NCRS u) \<and> X \<subseteq> \<Sigma>}"
      hence "(tr, NCS s)
    \<in> {(tr, t). \<exists>s x. (tr, s) \<in> {(tr', NCS s')} \<and> x \<in> dof (R s) \<and> PC s = l \<and>
           t = s\<lparr>R := x, PC := PC s + 1\<rparr>}"
      proof(clarsimp)
        assume
          da1: "tr = tr'" and
          da2: "l = NCPC s'" and
          da3: "N s'" and
          da4: "N s" and
          da5: "NCPC s = Suc (NCPC s')" and
          da6: "NCRS s \<in> dof (NCRS s')" and
          da7: "X \<subseteq> \<Sigma>"
        have "PC (NCS s') = NCPC s'" using PCNCS by blast
        moreover
        have "R (NCS s) = NCRS s" using RNCS by blast
        moreover
        have "PC (NCS s) = Suc (PC (NCS s'))" using da5 PCNCS by metis
        moreover
        have "NCS s = NCS s' \<lparr>R := NCRS s, PC := Suc (PC (NCS s'))\<rparr>"
          using calculation(2-3)
          by (simp add: N.simps(1) NCPC.simps(1) NCRS.simps(1) NCstate.inject(1) NC_state_eq state.select_convs(1) state.select_convs(2) state.update_convs(1) state.update_convs(2))
        ultimately
        show "\<exists>x. x \<in> dof (R (NCS s')) \<and>
        PC (NCS s') = NCPC s' \<and>
        NCS s = NCS s'
        \<lparr>R := x, PC := Suc (PC (NCS s'))\<rparr>"
          using da6
          by (metis RNCS state.surjective state.update_convs(1))
      qed
    }
    thus "(tr, NCS s) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr', NCS s')}"
      using ao2[THEN cj2_2] a1
      unfolding ba1 sfdof ddof
      by fast
  next
    fix f :: "'s cbr_fun" and m n :: label
    assume ba1: "i = cbr f m n"
    {
      assume ca1: "(tr, s, X) \<in> {(tr, t, X).
        \<exists>u X'a. (tr, u, X'a) \<in> {(tr', s', X')} \<and> NCPC u = l \<and> N u \<and> N t \<and> NCRS u = NCRS t \<and> X \<subseteq> \<Sigma> \<and>
           (f (NCRS u) \<and> NCPC t = m \<or> \<not> f (NCRS u) \<and> NCPC t = n)}"
      hence "(tr, NCS s)
    \<in> {(tr, t). \<exists>s. (tr, s) \<in> {(tr', NCS s')} \<and>
            PC s = l \<and> (f (R s) \<and> t = s\<lparr>PC := m\<rparr> \<or> \<not> f (R s) \<and> t = s\<lparr>PC := n\<rparr>)}"
        (* using PCNCS *)
      proof (simp)
        assume da1: "tr = tr' \<and> NCPC s' = l \<and> N s' \<and> N s \<and> NCRS s' = NCRS s \<and> X \<subseteq> \<Sigma> \<and>
          (f (NCRS s') \<and> NCPC s = m \<or> \<not> f (NCRS s') \<and> NCPC s = n)"
        {
          assume "f (NCRS s') \<and> NCPC s = m"
          hence "f (R (NCS s')) \<and> NCS s = NCS s'\<lparr>PC := m\<rparr>"
            unfolding PCNCS[symmetric]
            by (simp add: RNCS da1)
        }
        moreover
        {
          assume "\<not> f (NCRS s') \<and> NCPC s = n"
          hence "\<not> f (R (NCS s')) \<and> NCS s = NCS s'\<lparr>PC := n\<rparr>"
            unfolding PCNCS[symmetric]
            by (simp add: RNCS da1)
        }
        ultimately
        show "PC (NCS s') = l \<and>
          (f (R (NCS s')) \<and> NCS s = NCS s'\<lparr>PC := m\<rparr> \<or>
          \<not> f (R (NCS s')) \<and> NCS s = NCS s'\<lparr>PC := n\<rparr>)"
          unfolding PCNCS
          using da1
          by fast
      qed
    }
    thus "(tr, NCS s) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr', NCS s')}"
      using ao2[THEN cj2_2] a1
      unfolding ba1 sfcbr dcbr
      by fast
  next
    fix evfun :: "('s, 'e) event_fun" and sfun :: "('s, 'e) state_fun"
    assume ba1: "i = comm evfun sfun"
    {
      assume ca1: "(tr, s, X)
    \<in> {(tr, t, X). \<exists>u X'a. (tr, u, X'a) \<in> {(tr', s', X')} \<and>
           NCPC u = l \<and> N u \<and> X \<subseteq> \<Sigma> - evfun (NCRS u) \<and> t = NCC u}"
      hence "(tr, NCS s) \<in> {(tr', NCS s')}"
        using NCS_NCC by auto
    }
    moreover
    {
      assume ca1: "(tr, s, X) \<in> {(tr'a, t, X).
        \<exists>u X'a tr ev.
           (tr, u, X'a) \<in> {(tr', s', X')} \<and>
           NCPC u = l \<and>
           N u \<and>
           N t \<and>
           ev \<in> evfun (NCRS u) \<and>
           NCPC t = Suc l \<and>
           NCRS t = sfun (NCRS u) ev \<and>
           tr'a = tr @ [ev] \<and> X \<subseteq> \<Sigma>}"
      hence "(tr, NCS s) \<in>
       {(tr'a, t).
        \<exists>tr s e.
           (tr, s) \<in> {(tr', NCS s')} \<and>
           e \<in> evfun (R s) \<and>
           PC s = l \<and>
           tr'a = tr @ [e] \<and>
           t = s
           \<lparr>R := sfun (R s) e, PC := PC s + 1\<rparr>}"
      proof(auto)
        show "\<And> a . a \<in> evfun (NCRS s') \<Longrightarrow> a \<in> evfun (R (NCS s'))"
          unfolding RNCS .
      next
        show "PC (NCS s') = NCPC s'"
          unfolding PCNCS ..
      next
        fix a :: "'e event"
        assume
          da1: "l = NCPC s'" and
          da2: "N s'" and
          da3: "N s" and
          da4: "a \<in> evfun (NCRS s')" and
          da5: "NCPC s = Suc (NCPC s')" and
          da6: "NCRS s = sfun (NCRS s') a" and
          da7: "X \<subseteq> \<Sigma>" and
          da8: "tr = tr' @ [a]"
        have "R (NCS s) = sfun (R (NCS s')) a"
          using da6 unfolding RNCS .
        moreover
        have "PC (NCS s) = Suc (PC (NCS s'))"
          using da5 unfolding PCNCS .
        ultimately
        show "NCS s = NCS s' \<lparr>R := sfun (R (NCS s')) a,
                              PC := Suc (PC (NCS s'))\<rparr>"
          by fastforce
      qed
    }
    ultimately
    show "(tr, NCS s) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr', NCS s')}"
      using ao2[THEN cj2_2] a1
      unfolding ba1 sfcomm dcomm
      by fast
  qed
  thus "(tr, NCS s) \<in> densem_exec (a # lil) (failureSetToTraceStateSet S)"
    using a0
    unfolding ao1 densem_exec_unfold
    using densem_subset_single by blast
qed

subsection {* Conformance proofs *}

paragraph {* SF1 *}

text {* Each state in the SF semantics also is in the traces semantics. *}

lemma SF1:
  fixes X
  assumes "(tr, s, X) \<in> sfsem code S"
  shows "(tr, NCS s) \<in> densem code (failureSetToTraceStateSet S)"
using sfsem_path_is_densem_path[of tr s X _ S]
  sfsem_exec1[OF assms] densem_exec2[of _ code tr "NCS s" "failureSetToTraceStateSet S"]
by fast

paragraph {* SF2 *}

text {* For each state in the SF semantics, one can arbitrarily shrink its refusal set and still find
it in the SF semantics. *}

lemma SF2:
  fixes X
  assumes exec_code: "(tr, s, X) \<in> sfsem code S"
      and X'subX: "X' \<subseteq> X"
      and base: "\<forall> tr s X X'. (tr, s, X) \<in> S \<and> X' \<subseteq> X \<longrightarrow> (tr, s, X') \<in> S"
      and wff: "wff\<^sub>s\<^sub>c code"
  shows "(tr, s, X') \<in> sfsem code S"
proof -
  consider (inc) "NCPC s \<in> dom code \<and> N s" | (ninc) "\<not> (NCPC s \<in> dom code \<and> N s)" by blast
  thus ?thesis
  proof cases
    case inc
    show ?thesis
      using sfsem_not_outside_means_cbr[OF exec_code inc]
        X'subX
        sfsem_monotone_single[of "(tr, s, _)" S "(tr, s, X')" code]
      by (meson subset_trans)
  next
    case ninc
    obtain tr' s' X'' where
      in'S: "(tr', s', X'') \<in> S" and
      exec_code': "(tr, s, X) \<in> sfsem code {(tr', s', X'')}" and
      may_eq: "tr = tr' \<and> s = s' \<and> X = X'' \<or> NCPC s' \<in> definitions.dom code \<and> N s'"
      using sfsem_predecessor[OF exec_code] by blast
    consider (neq) "\<forall>X. (tr, s, X) \<notin> {(tr', s', X'')}" | (eq) "\<exists> X. (tr, s, X) \<in> {(tr', s', X'')}" by blast
    thus ?thesis
    proof cases
      case neq
      thus ?thesis
        using X'subX exec_code'
        using sfsem_greatest_refusal[OF exec_code' neq wff]
        using sfsem_monotone_single[OF in'S, of "(tr, s, X')" code]
        by force
    next
      case eq
      thus ?thesis
        using X'subX base in'S may_eq ninc 
        using pc_does_not_point_to_single[OF ninc, of tr X' S]
        by blast
    qed
  qed
qed

paragraph {* SF3 *}

text {* A helper lemma first:
If a triple @{term "(tr, s, X)"} is in the SF semantics, but extending @{term X} with @{term X'} is not,
then one can find a single element @{term "a \<subseteq> X'"}, such that extending @{term X} with @{term "{a}"}
already is not in the SF semantics. It is used in a contradictory proof in SF3.
*}

lemma SF3h: 
  fixes X
  assumes
    a: "(tr, s, X) \<in> sfsem c S" and
    b: "(tr, s, X \<union> X') \<notin> sfsem c S" and
    d: "wff\<^sub>s\<^sub>c c" and
    e: "\<forall> Xo. (tr, s, Xo) \<notin> S" (* care outside for other cases *)
  shows "\<exists> x . x \<in> X' \<and> (tr, s, X \<union> {x}) \<notin> sfsem c S"
proof -
  obtain X\<^sub>g where
    ao1: "(tr, s, X\<^sub>g) \<in> \<lbrakk>c\<rbrakk>S \<and>
     (\<forall>X''. (tr, s, X'') \<in> \<lbrakk>c\<rbrakk>S = (X'' \<subseteq> X\<^sub>g)) \<and>
     (N s \<longrightarrow> X\<^sub>g = \<Sigma>) \<and>
     (\<not> N s \<longrightarrow>
      (\<exists>evf sf.
          (NCPC s, comm evf sf) \<in> \<union>\<^sub>s\<^sub>c c \<and>
          X\<^sub>g = \<Sigma> - evf (NCRS s)))"
    using sfsem_greatest_refusal[OF a e d]
    by presburger
  from ao1[THEN cj3_2] b have a1: "\<not>(X \<union> X' \<subseteq> X\<^sub>g)"
    by blast
  from ao1[THEN cj3_2] a have a2: "X \<subseteq> X\<^sub>g"
    by blast
  obtain x where
    ao2: "x \<in> X' \<and> \<not> (X \<union> {x} \<subseteq> X\<^sub>g)"
    using a1 a2
    by blast
  have "(tr, s, X \<union> {x}) \<notin> \<lbrakk>c\<rbrakk>S"
    using ao1[THEN cj3_2] ao2[THEN cj2_2]
    by presburger
  thus ?thesis
    using ao2[THEN cj2_1]
    by blast
qed

lemma SF3hlpr3:
  assumes
    fresh: "sfsem_fresh (tr, s, X) c S" and subs: "Xs \<subseteq> X" and wff: "wff\<^sub>s\<^sub>c c"
  shows "sfsem_fresh (tr, s, Xs) c S"
proof -
  obtain tr' s' X' where
    in_S: "(tr', s', X') \<in> S" and
    indom: "NCPC s' \<in> definitions.dom c" and
    N: "N s'" and
    sf: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}"
    using fresh[unfolded sfsem_fresh]
    by fast
  consider (eq) "(tr, s) = (tr', s')" | (neq) "(tr, s) \<noteq> (tr', s')" by blast
  thus ?thesis
  proof cases
    case neq
    hence tmp: "\<forall>X. (tr, s, X) \<notin> {(tr', s', X')}" by blast
    obtain Xg where
      sfg: "(tr, s, Xg) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}" and
      greatest: "(\<forall>X''. ((tr, s, X'') \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}) = (X'' \<subseteq> Xg))"
      using sfsem_greatest_refusal[OF sf tmp wff]
      by blast
    hence "(tr, s, Xs) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}" using sf subs
      by blast
    thus ?thesis
      unfolding sfsem_fresh
      using in_S indom N
      by blast
  next
    case eq
    with sfsem_not_outside_means_cbr[OF sf] indom N subs
      have "(tr, s, Xs) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}" by blast
    thus ?thesis
      unfolding sfsem_fresh
      using in_S indom N
      by blast
  qed
qed

lemma SF3hlpr4:
  fixes X
  assumes
    wff: "wff\<^sub>s\<^sub>c c" and
    ex: "(tr, s, X) \<in> sfsem c {(tr', s', X')}" and
    --"For the sequential case:"
    a_subs: "a \<in> \<Sigma>" and
    fresh: "sfsem_fresh (tr, s, X) c {(tr', s', X')}" and
    not_sf: "(tr, s, X \<union> {a}) \<notin> sfsem c {(tr', s', X')}"
  shows
    "\<exists> t . (tr @ [a], t) \<in> densem c {(tr', NCS s')}"
proof -
  have outdom: "\<not> (NCPC s \<in> definitions.dom c \<and> N s)"
    using sfsem_not_outside_means_cbr[of tr s X c "{(tr', s', X)}"]
      a_subs not_sf
    by (metis (no_types, lifting) Un_insert_right ex insert_subset sfsem_predecessor2 sfsem_sigma_irrelevant_single singletonD sup_bot.right_neutral wff)
  obtain lil where
    subs: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    ex_lil: "(tr, s, X) \<in> sfsem_exec lil {(tr', s', X')}" and
    crit: "sfsem_exec_path_is_critical (tr, s, X) lil {(tr', s', X')}"
    using sfsem_exec1_crit[OF ex]
    by auto
  hence not_ex_lil: "(tr, s, X \<union> {a}) \<notin> sfsem_exec lil {(tr', s', X')}"
    using sfsem_exec2[of lil c tr s "X \<union> {a}" "{(tr', s', X')}"] not_sf wff outdom
    by auto
  have "lil \<noteq> []" using ex_lil fresh[unfolded sfsem_fresh, simplified] outdom by force
  then obtain l i lil' where
    lil: "lil = (l, i) # lil'"
    by (metis neq_Nil_conv prod.collapse)
  obtain tru su Xu where
    ex_one: "(tr, s, X) \<in> \<lbrakk>one l i\<rbrakk>{(tru, su ,Xu)}" and
    ex_lil': "(tru, su, Xu) \<in> sfsem_exec lil' {(tr', s', X')}" and
    neq: "(tr, s, X) \<noteq> (tru, su, Xu)" and
    crit': "sfsem_exec_path_is_critical (tru, su, Xu) lil' {(tr', s', X')}"
    using crit[unfolded lil, simplified] sfsem_exec_is_critical_is_exec
    by blast
  have not_ex_one: "(tr, s, X \<union> {a}) \<notin> \<lbrakk>one l i\<rbrakk>{(tru, su, Xu)}"
    using ex_lil' lil not_ex_lil sfsem_exec_unfold sfsem_monotone_single by blast
  then obtain ef f where
    i: "i = comm ef f" and
    notN: "\<not> N s" and
    N: "N su" 
    using ex_one a_subs neq
  proof (induct i)
    case (comm ef f)
    thus ?case 
      by (metis (no_types, hide_lams) Un_insert_right empty_iff insert_iff insert_subset sfsem_N_has_all_X sfsem_predecessor2 sfsem_refusal_subset_of_sigma sup_bot.right_neutral)
  qed auto
  have tr_pc: "tr = tru \<and> NCPC su = l"
    using ex_one N notN
    unfolding i sfsem.simps SClean_def by fast
  have a_in_ef: "a \<in> ef (NCRS su)"
    using not_ex_one tr_pc N notN ex_one a_subs
    unfolding i sfsem.simps SClean_def by force
  obtain t :: "('s) state" where
    "PC t = Suc (NCPC su) \<and> R t = f (NCRS su) a"
    by (meson select_convs(1) select_convs(2))
  hence "N (normal t) \<and> NCPC (normal t) = Suc (NCPC su) \<and> NCRS (normal t) = f (NCRS su) a"
    by simp
  hence "(tr @ [a], normal t, \<Sigma>) \<in> \<lbrakk>one l i\<rbrakk>{(tru, su, Xu)}"
    using N tr_pc a_in_ef N notN
    unfolding i sfsem.simps SClean_def by auto
  hence ex: "(tr @ [a], normal t, \<Sigma>) \<in> sfsem_exec lil {(tr', s', X')}"
    using ex_lil' lil sfsem_exec_unfold sfsem_monotone_single
    by blast
  show ?thesis
    using sfsem_path_is_densem_path[OF ex, unfolded failureSetToTraceStateSet_def, simplified]
      densem_exec2 subs by blast
qed

text{* For each state in the SF semantics, if for that trace we cannot find the trace with an extended event @{term a} in the
trace semantics, we can extend the refusal set arbitrarily by @{term a} (and each other event we cannot find in the traces semantics)
and still find it in the SF semantics. *}

lemma SF3:
  fixes X
  assumes a: "(tr, s, X) \<in> sfsem c S"
      and base: "\<forall> tr s X X' . (tr, s, X) \<in> S \<and> X \<union> X' \<subseteq> \<Sigma> \<longrightarrow> (tr, s, X \<union> X') \<in> S"
      and b: "\<forall> t . \<forall> a \<in> X' . (tr@[a], t) \<notin> densem c (failureSetToTraceStateSet S)"
      and c: "X \<union> X' \<subseteq> \<Sigma>"
      and d: "wff\<^sub>s\<^sub>c c"
  shows "(tr, s, X \<union> X') \<in> sfsem c S"
proof (cases "NCPC s \<in> definitions.dom c \<and> N s")
  assume aa1: "NCPC s \<in> definitions.dom c \<and> N s"
  from sfsem_not_outside_means_cbr[OF a aa1]
  obtain X m n l f where
     ao1: "(tr, s, X) \<in> S \<and>
     c = (l :: cbr f m n) \<and>
     (f (NCRS s) \<longrightarrow> m = l) \<and>
     (\<not> f (NCRS s) \<longrightarrow> n = l)"
     by fastforce
  have "\<forall>X. (tr, s, X)
        \<notin> {(tr, s, X).
            (tr, s, X) \<in> S \<and>
            (NCPC s \<notin> {l} \<or> \<not> N s)}"
     using aa1[unfolded ao1[THEN conjunct2, THEN conjunct1]] by auto
  have "\<forall>X \<subseteq> \<Sigma> . (tr, s, X)
        \<in> {(tr, t, X). \<exists>u X'. (tr, u, X') \<in> S \<and>
               NCPC u = l \<and> N u \<and> N t \<and> NCRS u = NCRS t \<and> X \<subseteq> \<Sigma> \<and>
               (f (NCRS u) \<and> NCPC t = m \<or>
                \<not> f (NCRS u) \<and> NCPC t = n)}"
    using aa1[unfolded ao1[THEN conjunct2, THEN conjunct1]]
    using ao1 case_prodI by auto
  hence "\<forall> X \<subseteq> \<Sigma>. (tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
    unfolding ao1[THEN conjunct2, THEN conjunct1] sfcbr SClean_def
    by auto
  thus ?thesis
    using c
    by blast
next
  assume aa0: "\<not> (NCPC s \<in> definitions.dom c \<and> N s)"
  show "(tr, s, X \<union> X') \<in> \<lbrakk>c\<rbrakk>S"
  proof (rule ccontr)
    assume aa1: "(tr, s, X \<union> X') \<notin> \<lbrakk>c\<rbrakk>S"
    have X'noSubsetOfX: "\<not> X' \<subseteq> X"
      using aa1 a
      by (metis sup.orderE)
    have notInS: (* "\<forall> X \<subseteq> \<Sigma> .  *)"(tr, s, X) \<notin> S"
    proof -
      have ba1: "(tr, s, X \<union> X') \<notin> S"
        using pc_does_not_point_to_single[OF aa0 _] aa1
        by blast
      thus ?thesis using base c by blast
    qed
    have SF3h':  "\<exists> x . x \<in> X' \<and> (tr, s, X \<union> {x}) \<notin> sfsem c S"
    proof (cases "\<forall> X . (tr, s, X) \<notin> S")
      assume ba1: "\<forall>X. (tr, s, X) \<notin> S"
      show ?thesis using SF3h[OF a aa1 d ba1]
        .
    next
      assume ba1: "\<not> (\<forall>X. (tr, s, X) \<notin> S)"
      hence b1: "\<exists> X . (tr, s, X) \<in> S"
        by simp
      obtain Strs Sr where
        bo1: "S = Strs \<union> Sr \<and>
         Strs = {(tr', s', X\<^sub>0) . (tr', s', X\<^sub>0) \<in> S \<and> tr' = tr \<and> s' = s} \<and>
         Sr = {(tr', s', X\<^sub>0) . (tr', s', X\<^sub>0) \<in> S \<and> \<not> (tr' = tr \<and> s' = s)}"
        by fastforce
      obtain S1 S2 where
        bo2: "Strs = S1 \<union> S2 \<and>
         S1 = {(tr', s', X\<^sub>0) . (tr', s', X\<^sub>0) \<in> Strs \<and> X\<^sub>0 \<subseteq> X \<union> X'} \<and>
         S2 = {(tr', s', X\<^sub>0) . (tr', s', X\<^sub>0) \<in> Strs \<and> \<not> X\<^sub>0 \<subseteq> X \<union> X'}"
        using bo1
        by fastforce
      have b2: "\<forall> (tr', s', X) \<in> S1 \<union> S2 . tr' = tr \<and> s' = s"
        unfolding bo2[THEN cj3_2, unfolded bo1[THEN cj3_2]] bo2[THEN cj3_3, unfolded bo1[THEN cj3_2]]
        by fast
      hence b2': "\<forall> x \<in> S1 \<union> S2 . fst x = tr \<and> fst (snd x) = s"
        by (simp add: case_prod_unfold)
      have "\<forall> (tr', s', X) \<in> Sr . \<not> (tr' = tr \<and> s' = s)"
        unfolding bo1[THEN cj3_3]
        by fast
      hence b3: "\<forall>Xo. (tr, s, Xo) \<notin> Sr"
        by fast
      have S1empty: "S1 = {}"
      proof(rule ccontr, auto)
        fix trc sc Xc
        assume ca1: "(trc, sc, Xc) \<in> S1"
        hence c0: "trc = tr \<and> sc = s"
          using b2 by blast
        hence c1: "\<not> (NCPC sc \<in> definitions.dom c \<and> N sc)"
          using aa0 by blast
        have c2: "Xc \<subseteq> X \<union> X'"
          using bo2[THEN cj3_2] ca1
          by blast
        have c3: "(trc, sc, Xc) \<in> S"
          unfolding bo1[THEN cj2_1] bo2[THEN cj2_1]
          using ca1
          by simp
        have c4: "(tr, s, X \<union> X') \<in> S"
          using c3[unfolded c0[THEN cj2_1] c0[THEN cj2_2]] c2 base c
            sup.absorb_iff2[of Xc "X \<union> X'"]
          by metis 
        show False
          using pc_does_not_point_to_single[OF aa0 c4] aa1
          by blast
      qed
      have "\<forall> x \<in> Strs. \<not> (NCPC (fst (snd x)) \<in> definitions.dom c \<and> N (fst (snd x)))"
        using b2' aa0
        unfolding bo2[THEN cj2_1]
        by simp
      hence b4: "\<forall> (tr, s, X) \<in> Strs. \<not> (NCPC s \<in> definitions.dom c \<and> N s)"
        by fastforce
      have b5: "(tr, s, X) \<notin> sfsem c Strs"
        using notInS bo1[THEN cj2_1, unfolded bo2[THEN cj2_1]] 
        unfolding bo2[THEN cj2_1] pc_does_not_point_to[OF b4, unfolded bo2[THEN cj2_1]]
        by blast
      hence b6: "(tr, s, X) \<in> sfsem c Sr"
        using bo1[THEN cj2_1] a
        by (simp add: Un_iff sfsem_union_additive)
      have b7: "(tr, s, X \<union> X') \<notin> \<lbrakk>c\<rbrakk>Sr"
        using aa1[unfolded bo1[THEN cj2_1]]
        by (simp add: sfsem_union_additive)
      obtain x where
        bo3: "x \<in> X' \<and> (tr, s, X \<union> {x}) \<notin> \<lbrakk>c\<rbrakk>Sr"
        using SF3h[OF b6 b7 d b3]
        by blast
      have b8: "(tr, s, X \<union> {x}) \<notin> Strs"
        using S1empty bo3[THEN cj2_1]
        unfolding bo2[THEN cj2_1]
        unfolding bo2[THEN cj3_2] bo2[THEN cj3_3]
        by fast
      have "(tr, s, X \<union> {x}) \<notin> \<lbrakk>c\<rbrakk>S"
        using pc_does_not_point_to[OF b4,symmetric] b8 bo3[THEN cj2_2] bo1[THEN cj2_1]
        by (simp add: Un_iff sfsem_union_additive)
      thus ?thesis
        using bo3[THEN cj2_1]
        by blast
    qed
    obtain x where
      ao1: "(tr, s, X \<union> {x}) \<notin> \<lbrakk>c\<rbrakk>S \<and> x \<in> X'"
      using SF3h'
      by blast
    obtain lil\<^sub>o where
      ao2: "set lil\<^sub>o \<subseteq> \<union>\<^sub>s\<^sub>c c \<and> (tr, s, X) \<in> sfsem_exec lil\<^sub>o S"
      using sfsem_exec1[OF assms(1)]
      by auto
    obtain lil where 
      ao3: "set lil \<subseteq> set lil\<^sub>o \<and> length lil \<le> length lil\<^sub>o \<and>
       sfsem_exec_path_is_critical (tr, s, X) lil S"
      using sfsem_exec_criticalPath[OF ao2[THEN conjunct2]]
      by blast
    have a0: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<and> \<not> (NCPC s \<in> definitions.dom c \<and> N s) \<and> wff\<^sub>s\<^sub>c c"
      using ao3[THEN conjunct1] ao2[THEN conjunct1] a d aa0
      by auto
    hence a1: "(tr, s, X \<union> {x}) \<notin> sfsem_exec lil S"
      using ao1[THEN conjunct1] sfsem_exec2[of lil c tr s "X \<union> {x}" S]
      by auto
    have a2: "(tr, s, X) \<in> sfsem_exec lil S"
      using sfsem_exec_is_critical_is_exec[OF ao3[THEN conjunct2, THEN conjunct2]] .
    have a3: "X \<union> {x} \<subseteq> \<Sigma>"
      using ao1[THEN conjunct2] c
      by blast
    {
      assume ba1: "lil = []"
      have "(tr, s, X) \<in> S"
        using ao3[THEN conjunct2, THEN conjunct2, unfolded ba1 sfsem_exec_path_is_critical.simps]
        .
      hence False
        using notInS
        by simp
    }
    moreover
    {
      assume ba1: "lil \<noteq> []"
      obtain li lil' where
        bo1: "lil = li # lil'"
        using ba1
        by (meson last_in_set list.set_cases)
      obtain lbl i where
        bo2: "li = (lbl, i)"
        by fastforce
      have b1: "(tr, s, X) \<notin> sfsem_exec lil' S"
        using ao3[THEN cj3_3, unfolded bo1 bo2 sfsem_exec_path_is_critical.simps, THEN cj2_1]
        .
      have b2: "(tr, s, X) \<in> \<lbrakk>(lbl :: i)\<rbrakk>sfsem_exec lil' S"
        using a2[unfolded bo1 bo2 sfsem_exec_unfold]
        .
      have b3: "(tr, s, X \<union> {x}) \<notin> \<lbrakk>(lbl :: i)\<rbrakk>sfsem_exec lil' S"
        using a1[unfolded bo1 bo2 sfsem_exec_unfold]
        .
      have notN: "\<not> N s"
      proof (cases "NCPC s \<in> definitions.dom (lbl :: i)")
        assume co1: "NCPC s \<in> definitions.dom (lbl :: i)"
        have "dom (lbl :: i) \<subseteq> dom c"
          using bo2 bo1 a0[THEN cj2_1]
          using dom.one in_sc_inj_imp_in_dom list.set_intros(1) singletonD subset_eq
          by auto
        thus "\<not> N s" using co1 aa0
          by auto
      next
        assume "NCPC s \<notin> definitions.dom (lbl :: i)"
        thus "\<not> N s"
          using b1 b2 b3
          using a3
          using sfsem_N_has_all_X[of tr s X lbl i "sfsem_exec lil' S" "X \<union> {x}"]
          by fast
      qed
      obtain evfun sfun where
        bo3: "i = comm evfun sfun"
        using sfsem_C_from_comm[OF b2 b1 notN]
        by blast
      have b4: "(tr, s, X) \<in> {(tr, t, X).
      \<exists>u X'.
         (tr, u, X') \<in> sfsem_exec lil' S \<and>
         NCPC u = lbl \<and> N u \<and> X \<subseteq> \<Sigma> - evfun (NCRS u) \<and> t = NCC u}"
        using b2 notN b1
        unfolding bo3 sfcomm
        by fastforce
      obtain u X'' where
        bo4: "(tr, u, X'') \<in> sfsem_exec lil' S \<and>
         NCPC u = lbl \<and> N u \<and> X \<subseteq> \<Sigma> - evfun (NCRS u) \<and> s = NCC u"
        using b4
        by blast
      {
        assume ca1: "x \<notin> evfun (NCRS u)"
        hence "(tr, u, X'') \<in> sfsem_exec lil' S \<and>
         NCPC u = lbl \<and> N u \<and> X \<union> {x} \<subseteq> \<Sigma> - evfun (NCRS u) \<and> s = NCC u"
          using bo4 a3
          by auto
        hence "(tr, s, X \<union> {x}) \<in> {(tr, t, X).
      \<exists>u X'.
         (tr, u, X') \<in> sfsem_exec lil' S \<and>
         NCPC u = lbl \<and> N u \<and> X \<subseteq> \<Sigma> - evfun (NCRS u) \<and> t = NCC u}"
          by auto
        hence "(tr, s, X \<union> {x}) \<in> \<lbrakk>(lbl :: i)\<rbrakk>sfsem_exec lil' S"
          unfolding bo3
          by auto
        with b3 have False
          by auto
      }
      moreover
      {
        assume ca1: "x \<in> evfun (NCRS u)"
        obtain tr' lbl' rs' X''' where
         co1: "(tr, u, X'') \<in> sfsem_exec lil' S \<and>
             NCPC u = lbl \<and>
             N u \<and>
             x \<in> evfun (NCRS u) \<and>
             lbl' = Suc lbl \<and>
             tr' = tr @ [x] \<and>
             rs' = sfun (NCRS u) x \<and> X''' \<subseteq> \<Sigma>"
          using bo4 a3 ca1
          by blast
        obtain t where
          co2: "t = normal (\<lparr>R = rs', PC = lbl'\<rparr>)"
          by presburger
        from co2 have nt: "N t"
          by (simp add: N.simps(1))
        hence "(tr, u, X'') \<in> sfsem_exec lil' S \<and>
             NCPC u = lbl \<and>
             N u \<and>
             N t \<and>
             x \<in> evfun (NCRS u) \<and>
             NCPC t = Suc lbl \<and>
             NCRS t = sfun (NCRS u) x \<and>
             tr' = tr @ [x] \<and> X''' \<subseteq> \<Sigma>"
          using co1 co2
          by auto
        hence "\<exists> tr' t X.
             (tr, u, X'') \<in> sfsem_exec lil' S \<and>
             NCPC u = lbl \<and>
             N u \<and>
             N t \<and>
             x \<in> evfun (NCRS u) \<and>
             NCPC t = Suc lbl \<and>
             NCRS t = sfun (NCRS u) x \<and>
             tr' = tr @ [x] \<and> X \<subseteq> \<Sigma>"
          using bo4
           by auto
        hence "\<exists> t X . (tr@[x], t, X) \<in> \<lbrakk>(lbl :: i)\<rbrakk>sfsem_exec lil' S"
          unfolding bo3 sfcomm
          by blast
        hence "\<exists> t . (tr@[x], t) \<in> densem (one lbl i) (densem_exec lil' (failureSetToTraceStateSet S))"
          unfolding sfsem_exec_unfold[symmetric] densem_exec_unfold[symmetric]
          using sfsem_path_is_densem_path[of "tr @ [x]" _ _ "(lbl, i) # lil'" S]
          by fast
        hence "\<exists> t . (tr@[x], t) \<in> densem_exec ((lbl, i) # lil') (failureSetToTraceStateSet S)"
          unfolding densem_exec_unfold[symmetric] .
        hence "\<exists> t . (tr@[x], t) \<in> densem c (failureSetToTraceStateSet S)"
          unfolding bo2[symmetric] bo1[symmetric]
          using densem_exec2[OF a0[THEN cj2_1], of "tr @ [x]" _ "failureSetToTraceStateSet S"]
          by blast
        with b ao1[THEN cj2_2] have False
          by blast
      }
      ultimately
      have False by auto
    }
    ultimately show False
      by blast
  qed
qed

paragraph {* SF4 *}

text {* Each state in the trace semantics being outside of the code's domain can also be found in
the SF semantics. This is the opposite to SF1. *}

lemma SF4:
  assumes a: "(tr, s) \<in> densem c (S)"
      and b: "\<not> (PC s \<in> dom c)"
      and X'SUB: "X \<subseteq> \<Sigma>"
      and wff: "wff\<^sub>s\<^sub>c c"
  shows "(tr, normal s, X) \<in> sfsem c (trace_state_set_to_failure_set S)"
proof -
  obtain lil :: "(nat \<times> ('s, 'e) instr) list" where
    ao1: "(tr, s) \<in> densem_exec lil S \<and> set lil \<subseteq> \<union>\<^sub>s\<^sub>c c"
    using densem_exec1[of tr s c S, OF a]
    by blast
  obtain lil_crit :: "(nat \<times> ('s, 'e) instr) list" where
    ao2: "set lil_crit \<subseteq> set lil \<and> densem_exec_path_is_critical (tr, s) lil_crit S"
    using densem_exec_critical_path[of "(tr, s)" lil S, OF ao1[THEN conjunct1]]
    by blast
  from densem_critical_path_is_sfsem_path[OF ao2[THEN conjunct2], OF X'SUB]
    have a1: "(tr, normal s, X) \<in> sfsem_exec lil_crit (trace_state_set_to_failure_set S)"
    .
  from subset_trans[OF ao2[THEN conjunct1], OF ao1[THEN conjunct2]]
    have a2: "set lil_crit \<subseteq> \<union>\<^sub>s\<^sub>c c"
    .
  show "(tr, normal s, X) \<in> \<lbrakk>c\<rbrakk>(trace_state_set_to_failure_set S)"
    using wff a1 b dom_eq_indom[of "PC s" c] sfsem_exec2[OF a2, OF a1, unfolded NCPC.simps(1)]
    by blast
qed

text {* A reformulation of SF4. *}

lemma SF4':
  fixes X
  assumes a: "(tr, NCS s) \<in> densem c (S)"
      and b: "\<not> (NCPC s \<in> dom c)"
      and c: "N s"
      and X'SUB: "X \<subseteq> \<Sigma>"
      and wff: "wff\<^sub>s\<^sub>c c"
  shows "(tr, s, X) \<in> sfsem c (trace_state_set_to_failure_set S)"
proof -
  show ?thesis
    using SF4[OF a, unfolded PCNCS, OF b, OF X'SUB, OF wff] Nnormal[OF c]
    by simp
qed

paragraph {* SF refinement implies T refinement *}

text {* This section requires the static code analysis, where each execution path eventually executes a @{term comm} instruction
or leads out of the code's domain. Again, in the end this strong condition is weakened a little: We need only the condition, that
each possible starting state has a successor. This still allows for infinite loops without a comm, but suffices. *}

text {* We can use any refusal set in the start. *}

lemma sfsem_exec_transforming_start_is_possible:
  fixes X
  assumes
    a: "(tr @ l, s, X) \<in> sfsem_exec lil (trace_state_set_to_failure_set {(tr, s')})" and
    b: "l \<noteq> [] \<or> s \<noteq> normal s'" (* We really progressed, so the start is irrelevant *)
  shows "(tr @ l, s, X) \<in> sfsem_exec lil {(tr, normal s', X')}"
proof -
  obtain x where
    execX: "(tr @ l, s, X) \<in> sfsem_exec lil {x}" and
    xIn: "x \<in> (trace_state_set_to_failure_set {(tr, s')})"
    using a
    using sfsem_exec_pred2 by blast
  then obtain X'' where
    x: "x = (tr, normal s', X'')"
    using trace_state_set_to_failure_set_def
    by auto
  have fresh: "sfsem_exec_fresh (tr @ l, s, X) lil {x}"
    unfolding x
    using b append_self_conv execX fst_conv sfsem_exec_freshI1 singletonD snd_conv x
    by auto
  from sfsem_exec_fresh_X_irrelevant[OF execX[unfolded x] fresh[unfolded x], of X']
    show ?thesis .
qed

text {* If each starting state has a successor, then each trace in the traces semantis
can be found in the SF semantics. *}

lemma tr_of_T_will_be_in_SF:
  assumes
    --"Each starting state has a successor."
    a: "\<forall> tr s X . sfsem c {(tr, s, X)} \<noteq> {}" and
    wffc: "wff\<^sub>s\<^sub>c c" and
    tr_in_densem: "(tr, s) \<in> densem c (failureSetToTraceStateSet S)" and
    allInSAreNormal: "\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> N s" and
    csublis: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis"
  shows "\<exists> s X . (tr, s, X) \<in> sfsem  c S"
proof -
  obtain tr' s' where 
    bo1: "(tr, s) \<in> \<lbrakk>c\<rbrakk> {(tr', s')}" and
    bo2: "(tr', s') \<in> failureSetToTraceStateSet S"
  using tr_in_densem densem_predecessor by fastforce
  {
    assume tr: "tr = tr'"
    then obtain s'' X where x: "(tr, s'', X) \<in> S" and s': "s' = NCS s''"
      using bo2 allInSAreNormal
      unfolding failureSetToTraceStateSet_def
      by blast
    hence "N s''" using allInSAreNormal by blast
    hence "s'' = normal s'"
      using s' Nnormal by blast
    hence x: "(tr, normal s', X) \<in> S" using x by simp
    with sfsem_not_empty_has_same_trace_successor[of c tr "normal s'" X, OF _ csublis wffc] a
      have ?thesis
      using sfsem_monotone by blast
  }
  moreover
  {
    assume neq: "tr \<noteq> tr'"
    have b1: "((tr', s'), \<union>\<^sub>s\<^sub>c c, tr, s) \<in> multistep"
      using densem_implies_multistep[OF csublis bo1, simplified] .
    obtain k where
      k: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr, s), k) \<in> multistep_k"
      using multistep_implies_multistep_k[OF b1]
      by blast
    obtain l where l: "tr = tr' @ l"
      using k multistep_k_growing_trace by blast
    then obtain a l' where tr: "tr = tr' @ [] @ [a] @ l'"
      by (metis append_Nil append_Nil2 local.list_append_singleton neq neq_Nil_conv)
    have comminc: "\<exists> ef sf lbl . (lbl, comm ef sf) \<in> \<union>\<^sub>s\<^sub>c c"
    proof -
      obtain sa sa' m n where
        sucmn: "Suc (m + n) = k" and
        msm: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr', sa), m) \<in> multistep_k" and
        smstp: "((tr', sa), tr' @ [a], sa') \<in> Smallstep" and
        msn: "((tr' @ [a], sa'), \<union>\<^sub>s\<^sub>c c, (tr' @ a # l', s), n) \<in> multistep_k" and
        indom: "indom (PC sa) (\<union>\<^sub>s\<^sub>c c)" and
        stuff: "(0 < n \<longrightarrow> indom (PC sa') (\<union>\<^sub>s\<^sub>c c)) \<and> (indom (PC s) (\<union>\<^sub>s\<^sub>c c) \<longrightarrow> indom (PC sa') (\<union>\<^sub>s\<^sub>c c))"
        using multistep_k_prefix_closed'[OF csublis k[unfolded tr], simplified]
        by blast
      have "PC sa \<in> dom c"
        using msm indom dom_eq_indom by blast
      then obtain i where i: "(PC sa, i) \<in> lis" and succ: "(tr' @ [a], normal sa', \<Sigma>) \<in> \<lbrakk>(PC sa :: i)\<rbrakk>{(tr', normal sa, \<Sigma>)}"
        using smallstepImpliesSfsem[OF smstp[unfolded Stos], of "normal sa'" \<Sigma>] by blast
      hence i': "(PC sa, i) \<in> \<union>\<^sub>s\<^sub>c c"
        by (simp add: csublis indom indomSC_element)
      obtain ef sf where "i = comm ef sf"
        using sfsem_one_trace_append_implies_comm[OF succ]
        by fast
      with i' show ?thesis by fast
    qed
    obtain lil where
      sInLil: "(tr, s) \<in> densem_exec lil {(tr', s')}" and
      lilInC: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c"
      using densem_exec1[OF bo1]
      by blast
    obtain lilC where
      lilCinLil: "set lilC \<subseteq> set lil" and
      lenLilCLeqLil: "length lilC \<le> length lil" and
      lilC: "densem_exec_path_is_critical (tr, s) lilC {(tr', s')}"
      using densem_exec_critical_path[OF sInLil]
      by blast
    have sInLilC: "(tr, normal s, \<Sigma>) \<in> sfsem_exec lilC (trace_state_set_to_failure_set {(tr', s')})"
      using densem_critical_path_is_sfsem_path[OF lilC, of \<Sigma>]
      by simp
    have sInLilC: "(tr, normal s, \<Sigma>) \<in> sfsem_exec lilC S"
    proof -
      obtain sb Xb where
        bInb': "(tr', s') \<in> failureSetToTraceStateSet {(tr', sb, Xb)}" and
        b'inS: "(tr', sb, Xb) \<in> S" and
        "N sb"
        using bo2 allInSAreNormal
        unfolding trace_state_set_to_failure_set_def failureSetToTraceStateSet_def
        by force
      hence sb: "sb = normal s'"
        unfolding failureSetToTraceStateSet_def
        using Nnormal by blast
      have lNotEmpty: "l \<noteq> []"
        using l neq by auto
      have "(tr' @ l, normal s, \<Sigma>) \<in> sfsem_exec lilC {(tr', normal s', Xb)}"
        using sfsem_exec_transforming_start_is_possible[OF sInLilC[unfolded l]] lNotEmpty
        by simp
      thus ?thesis
        unfolding l sb[symmetric]
        using b'inS
      proof -
        assume a1: "(tr' @ l, normal s, \<Sigma>) \<in> sfsem_exec lilC {(tr', sb, Xb)}"
        have "insert (tr', sb, Xb) (S - {(tr', sb, Xb)}) = S"
          by (meson \<open>(tr', sb, Xb) \<in> S\<close> insert_Diff)
        then show "(tr' @ l, normal s, \<Sigma>) \<in> sfsem_exec lilC S"
          using a1 by (metis (no_types) UnCI insert_is_Un sfsem_exec_additive)
      qed
    qed
    obtain s' X' where
      succ: "(tr, s', X') \<in> \<lbrakk>c\<rbrakk>{(tr, normal s, \<Sigma>)}"
      using sfsem_not_empty_has_same_trace_successor[of c tr "normal s" \<Sigma>, OF _ csublis wffc]
        a
      by blast
    obtain lil' where
      lil': "(tr, s', X') \<in> sfsem_exec lil' {(tr, normal s, \<Sigma>)}" and
      lil'c: "set lil' \<subseteq> \<union>\<^sub>s\<^sub>c c"
      using sfsem_exec1[OF succ]
      by blast
    with sInLilC have lillil': "(tr, s', X') \<in> sfsem_exec (lil' @ lilC) S"
    proof -
      have "\<And>P. (tr, s', X') \<in> sfsem_exec lil' (insert (tr, normal s, \<Sigma>) P)"
        by (metis (no_types) Un_iff insert_is_Un lil' sfsem_exec_additive)
      then show ?thesis
        by (metis (no_types) insert_absorb sInLilC sfsem_exec_unfold2)
    qed
    have notPCN: "\<not> (NCPC s' \<in> dom c \<and> N s')"
    proof
      assume ba1: "NCPC s' \<in> dom c \<and> N s'"
      from sfsem_not_outside_means_cbr[OF succ ba1] comminc show "False"
        by force
    qed
    hence "\<exists> s X . (tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
      using sfsem_exec2[OF _ lillil' notPCN wffc] lilInC lilCinLil lil'c
      by auto
  }
  ultimately
  show ?thesis by auto
qed

text {* This is the stronger version of the main theorem being stated afterwards, as the precondition is weaker. *}

lemma sfsem_each_state_has_successor_imp_sf_refinement_implies_traces_refinement:
  assumes
    successors: "\<forall> tr s X . sfsem c {(tr, s, X)} \<noteq> {}" and
    wffc: "wff\<^sub>s\<^sub>c c" and
    wffc': "wff\<^sub>s\<^sub>c c'" and
    sf_refinement: "(sfsem c S) \<subseteq>\<^sub>S\<^sub>F (sfsem c' S)" and
    allInSAreNormal: "\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> N s" and
    csublis: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis" and
    c'sublis: "\<union>\<^sub>s\<^sub>c c' \<subseteq> lis"
  shows "(densem c (failureSetToTraceStateSet S)) \<subseteq>\<^sub>T (densem c' (failureSetToTraceStateSet S))"
unfolding densem_refinement_def
proof (rule ccontr, clarsimp)
  fix tr s
  assume
    aa1: "\<forall>s'. (tr, s') \<notin> \<lbrakk>c'\<rbrakk>failureSetToTraceStateSet S" and
    aa2: "(tr, s) \<in> \<lbrakk>c\<rbrakk>failureSetToTraceStateSet S"
  have a1: "\<forall> s X . (tr, s, X) \<notin> \<lbrakk>c'\<rbrakk> S"
    using aa1 SF1 by blast
  have trNotInS: "\<forall> s . (tr, s) \<notin> failureSetToTraceStateSet S"
    using aa1 densem_super_id_element by blast
  from tr_of_T_will_be_in_SF [OF successors wffc aa2 allInSAreNormal csublis]
    sf_refinement[unfolded sfsem_refinement_def] a1 show False
    by blast
qed

text {* If each execution path eventually executes a @{term comm} or leads out of the code, then
a SF refinement implies a traces refinement.
*}

theorem sfsem_no_divergencies_imp_sf_refinement_implies_traces_refinement:
  assumes
    no_div: "all_commands_are_followed_by_comm_or_leave c" and
    sf_refinement: "(sfsem c S) \<subseteq>\<^sub>S\<^sub>F (sfsem c' S)" and
    cSubsLis: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis" and
    c'SubsLis: "\<union>\<^sub>s\<^sub>c c' \<subseteq> lis" and
    wffc: "wff\<^sub>s\<^sub>c c" and
    wffc': "wff\<^sub>s\<^sub>c c'" and
    allInSAreNormal: "\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> N s" and
    doF: "\<forall> f . (\<exists> lbl . (lbl, do f) \<in> \<union>\<^sub>s\<^sub>c c) \<longrightarrow> (\<forall> x . f x \<noteq> {})" and
    commEF: "\<forall> ef . (\<exists> lbl sf . (lbl, comm ef sf) \<in> \<union>\<^sub>s\<^sub>c c) \<longrightarrow> (\<forall> x . ef x \<noteq> {})"
  shows "(densem c (failureSetToTraceStateSet S)) \<subseteq>\<^sub>T (densem c' (failureSetToTraceStateSet S))"
using sfsem_each_state_has_successor_imp_sf_refinement_implies_traces_refinement[OF
  sfsem_each_instruction_has_a_successor[OF no_div wffc doF commEF]
  wffc
  wffc'
  sf_refinement
  allInSAreNormal
  cSubsLis
  c'SubsLis
] .

end

end
