section {* Static Code analysis *}

theory static_code_analysis
imports definitions
begin

subsection {* Divergency Free *}

text {* This section paves the work towards the lemma @{text sfsem_no_divergencies_imp_sf_refinement_implies_traces_refinement}. This lemma proves that
if the code fulfills some conditions, then the stable failure refinement also implies a traces refinement.
The condition used here is that every instruction will be followed at some point by a @{term comm} instruction, or leaves the code. Thus, the code cannot diverge,
as each loop has to communicate in every iteration. *}

text {* Therefor, the following function checks in a depth first search, whether for a given instruction all possible execution paths
eventually lead to a @{term comm} instruction or leave the code's domain. The recursive search ends when a @{term comm} is found (good),
the following program counter would point out of the code's domain (good), or we find an already visited instruction (bad). This is bad
as there is a path from some previous instruction A to the current instruction B and leading back to A without a @{term comm} in between, thus
a possible candidate for a diverging loop. Of course this is not necessarily a diverging loop, but in this static analysis case we want to
exclude this case.
To detect the bad case, the @{text visited} argument is used, storing all already visited instructions (in a set).
Note: The hlpr argument is just for convenience, as proving termination of the function then is trivial. This function will be wrapped in
a more userfriendly definition, where this hlpr is set to the code size plus one, meaning we reach every possible instruction and step outside
of the code's domain.
*}

fun instr_followed_by_comm_or_leaves :: "label \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> label set \<Rightarrow> nat \<Rightarrow> bool"
where
"instr_followed_by_comm_or_leaves curLabel c visited (Suc hlpr) = (if (curLabel \<in> visited) then False else
(if (curLabel \<notin> dom c) then ( True ) else
(\<forall> ef f . (curLabel, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow> True (* This case is just for completeness and not required. *)) \<and>
(\<forall> f . (curLabel, do f) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow> instr_followed_by_comm_or_leaves (Suc curLabel) c (visited \<union> {curLabel}) hlpr) \<and>
(\<forall> b m n . (curLabel, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow> (instr_followed_by_comm_or_leaves m c (visited \<union> {curLabel}) hlpr) \<and> (instr_followed_by_comm_or_leaves n c (visited \<union> {curLabel}) hlpr))
))" |
"instr_followed_by_comm_or_leaves curLabel c visited 0 = False"

text {* A small check, whether it works. *}

lemma "instr_followed_by_comm_or_leaves 1 ((one 1 (do (\<lambda> x . {x})) \<oplus> one 2 (do (\<lambda> x . {x})))) {} (Suc (Suc (Suc 0)))"
by simp

text {* This wrapper definition defines the property we are looking for. For each instruction in the code, we want all its execution paths
to eventually reach a comm or point out of the code's domain. *}

definition "all_commands_are_followed_by_comm_or_leave c = (\<forall> l \<in> dom c . instr_followed_by_comm_or_leaves l c {} (Suc (size c)))"

text {* A small check, whether it works. *}

lemma "all_commands_are_followed_by_comm_or_leave ((one 1 (do (\<lambda> x . {x})) \<oplus> one 2 (do (\<lambda> x . {x}))))"
  unfolding all_commands_are_followed_by_comm_or_leave_def
by force

text {* To prove the final lemma more easily, we reformulate the initial @{term instr_followed_by_comm_or_leaves} function
and instead of returning @{prop True} or @{prop False}, return the trace which lead us to the communication instruction, out of
the code, or into a loop. These traces can be analyzed later on, and in fact are dense traces (if they are semantically valid paths,
otherwise they do not matter). But therefor, to be dense, the case of a self referencing branch has to be checked directly in the
found branch: The branch then is not allowed to be added to the path, as executing this self branch results in the same state.
Also another trick is used: To easily identify a path as leading outside of the code's domain, we append an artificial @{term do}
instruction. 
*}
fun create_all_possible_execution_paths :: "label \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> (label \<times> ('s, 'e) instr) list \<Rightarrow> nat \<Rightarrow> (label \<times> ('s, 'e) instr) list set"
where
"create_all_possible_execution_paths start c visited (Suc hlpr) = (if (\<not> indom start (set visited)) then
(if (start \<notin> dom c) then {(start, do (\<lambda> x . {x})) # visited} else
{ x . \<exists> f . (start, do f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, do f) # visited) hlpr} \<union>
{ x . \<exists> ef f . (start, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, comm ef f) # visited) hlpr} \<union>
{ x . \<exists> b m n . (start, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<and>
  x \<in> (if start = m then {visited} else create_all_possible_execution_paths m c ((start, cbr b m n) # visited) hlpr) \<union> (if start = n then {visited} else create_all_possible_execution_paths n c ((start, cbr b m n) # visited) hlpr)}
) else
({visited}))" |
--"This will never happen and is equivalent to @{prop False}, as we have not yet seen comm or an end."
"create_all_possible_execution_paths start c visited 0 = {visited}"

text {* For easier reasoning, use this rewrite rule. The @{text "if"} version in the functions is more readable, hence
failsafe rewriting is done in the proofs with Isabelle's help. *}

lemma create_all_possible_execution_paths_if_rewrite:
"(x \<in> (if P then S else T) \<union> (if Q then A else B)) = (((P \<longrightarrow> (x \<in> S)) \<and> (\<not> P \<longrightarrow> (x \<in> T))) \<or> ((Q \<longrightarrow> (x \<in> A)) \<and> (\<not> Q \<longrightarrow> (x \<in> B))))"
  by simp
lemma create_all_possible_execution_paths_if_rewrite':
"(\<forall> x \<in> (if P then S else T) \<union> (if Q then A else B). M x) = (((P \<longrightarrow> (\<forall> x \<in> S. M x)) \<and> (\<not> P \<longrightarrow> (\<forall> x \<in> T. M x))) \<and> ((Q \<longrightarrow> (\<forall> x \<in> A . M x)) \<and> (\<not> Q \<longrightarrow> (\<forall> x \<in> B . M x))))"
  by auto

text {* Whatever we get out of the function, its suffix is always the @{term visited} trace. *}
  
lemma create_all_possible_execution_paths_extends_visited:
  assumes "X \<in> create_all_possible_execution_paths start c visited CNT"
  shows "\<exists> prefix . prefix @ visited = X"
using assms
proof (induct CNT arbitrary: visited X start, auto)
  fix CNT visited X start 
  assume
    aa1: "\<And>visited X start. X \<in> create_all_possible_execution_paths start c visited CNT \<Longrightarrow> \<exists>prefix. prefix @ visited = X" and
    aa2: "X \<in> (if \<not> indom start (set visited)
        then if start \<notin> definitions.dom c then {(start, do (\<lambda>x. {x})) # visited}
             else {x. \<exists>f. (start, do f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, do f) # visited) CNT} \<union>
                  {x. \<exists>ef f. (start, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, comm ef f) # visited) CNT} \<union>
                  {x. \<exists>f m n. (start, cbr f m n) \<in> \<union>\<^sub>s\<^sub>c c \<and>
                                    x \<in> (if start = m then {visited} else create_all_possible_execution_paths m c ((start, cbr f m n) # visited) CNT) \<union>
                                         (if start = n then {visited} else create_all_possible_execution_paths n c ((start, cbr f m n) # visited) CNT)}
        else {visited})"
  let ?G = "\<exists>prefix. prefix @ visited = X"
  {
    assume ba1: "\<not>\<not>indom start (set visited)"
    from aa2[unfolded if_not_P[OF ba1]] have ?G by simp
  } moreover {
    assume ba1: "\<not>indom start (set visited)"
    {
      assume ca1: "start \<notin> definitions.dom c"
      from aa2[unfolded if_P[OF ba1] if_P[OF ca1]] have ?G by simp
    } moreover {
      assume ca1: "\<not> start \<notin> definitions.dom c"
      have startNotInside: "\<not> start \<in> {lbl. \<exists>i. (lbl, i) \<in> set visited}" using ba1 unfolding indom_def by blast
      obtain li where li: "li = (SOME x. x \<in> \<union>\<^sub>s\<^sub>c c \<and> fst x = start)"
        by blast
      obtain i where i : "i = snd li" by fastforce
      have b2: "li = (start, i)" using li i
        using ca1[simplified] eq_fst_iff[of start] not_in_sc_inj_imp_not_in_dom[of start c] snd_conv[of start _] someI
        by (metis (mono_tags, lifting))
      have inside: "(start, i) \<in> \<union>\<^sub>s\<^sub>c c" using li ca1 i b2
        by (metis (mono_tags, lifting) fst_conv not_in_sc_inj_imp_not_in_dom someI_ex)
      have X: "X \<in> {x. \<exists>f. (start, do f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, do f) # visited) CNT} \<union>
          {x. \<exists>ef f. (start, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, comm ef f) # visited) CNT} \<union>
          {x. \<exists>f m n. (start, cbr f m n) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> (if start = m then {visited} else create_all_possible_execution_paths m c ((start, cbr f m n) # visited) CNT) \<union> (if start = n then {visited} else create_all_possible_execution_paths n c ((start, cbr f m n) # visited) CNT)}"
          (is "X \<in> ?DO \<union> ?COMM \<union> ?CBR")
        using aa2
        unfolding if_P[OF ba1] if_not_P[OF ca1]
          li[symmetric] b2
        by force
      moreover
      {
        assume xInDo: "X \<in> ?DO"
        then obtain f where c3: "X \<in> create_all_possible_execution_paths (Suc start) c ((start, do f) # visited) CNT"
          by force
        hence ?G
          using aa1[OF c3]
          by force
      }
      moreover
      {
        assume xInDo: "X \<in> ?COMM"
        then obtain ef f where c3: "X \<in> create_all_possible_execution_paths (Suc start) c ((start, comm ef f) # visited) CNT"
          by force
        have ?G
          using aa1[OF c3]
          by auto
      }
      moreover
      {
        assume xInDo: "X \<in> ?CBR"
        then obtain b m n where c3: "X \<in> (if start = m then {visited} else create_all_possible_execution_paths m c ((start, cbr b m n) # visited) CNT) \<union>
                        (if start = n then {visited} else create_all_possible_execution_paths n c ((start, cbr b m n) # visited) CNT)"
          by force
        have ?G
          using c3[unfolded create_all_possible_execution_paths_if_rewrite]
            aa1[of X m "((start, cbr b m n) # visited)"]
            aa1[of X n "((start, cbr b m n) # visited)"]
            by auto
      }
      ultimately
      have ?G
        by fast
    }
    ultimately have ?G by blast
  }
  ultimately show "\<exists>prefix. prefix @ visited = X" by blast
qed

text {* For any generated non empty path, we know that the paths tail consists of instructions of the code. *}

lemma create_all_possible_execution_paths_tail_indom_c:
  assumes a: "lil \<in> create_all_possible_execution_paths start c visited CNT" and b:"lil \<noteq> []" and c:"set visited \<subseteq> \<union>\<^sub>s\<^sub>c c"
  shows "set (tl lil) \<subseteq> \<union>\<^sub>s\<^sub>c c"
using a c
proof (induct CNT arbitrary: visited start)
  show "\<And>visited start.
       lil \<in> create_all_possible_execution_paths start c visited 0 \<Longrightarrow> set visited \<subseteq> \<union>\<^sub>s\<^sub>c c \<Longrightarrow> set (tl lil) \<subseteq> \<union>\<^sub>s\<^sub>c c"
    unfolding create_all_possible_execution_paths.simps
    using list.set_sel(2) b
    by fastforce
next
  fix CNT visited start
  assume
    aa1: "\<And>visited start.
           lil \<in> create_all_possible_execution_paths start c visited CNT \<Longrightarrow>
           set visited \<subseteq> \<union>\<^sub>s\<^sub>c c \<Longrightarrow> set (tl lil) \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    aa2: "lil \<in> create_all_possible_execution_paths start c visited (Suc CNT)" and
    aa4: "set visited \<subseteq> \<union>\<^sub>s\<^sub>c c"
  let ?G = "set (tl lil) \<subseteq> \<union>\<^sub>s\<^sub>c c"
  have "True"
    using aa2 unfolding "create_all_possible_execution_paths.simps"
    by simp
  {
    assume ba1: "\<not> \<not> indom start (set visited)"
    have ?G
      using aa2 aa4
      unfolding create_all_possible_execution_paths.simps if_not_P[OF ba1]
      using b contra_subsetD list.set_sel(2) singletonD subsetI
      by fastforce
  }
  moreover
  {
    assume ba1: "start \<notin> definitions.dom c" and ba2: "\<not> indom start (set visited)"
    have ?G
      using aa2 aa4
      unfolding create_all_possible_execution_paths.simps if_P[OF ba2] if_P[OF ba1]
      by simp
  }
  moreover
  {
    assume b1: "\<not> indom start (set visited)" and ba2: "\<not> start \<notin> definitions.dom c"
    have ba1: "\<not> start \<in> {lbl. \<exists>i. (lbl, i) \<in> set visited}" using b1 unfolding indom_def by blast
    have X: "lil \<in> {x. \<exists>f. (start, do f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, do f) # visited) CNT} \<union>
        {x. \<exists>ef f. (start, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, comm ef f) # visited) CNT} \<union>
        {x. \<exists>b m n. (start, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> (if start = m then {visited} else create_all_possible_execution_paths m c ((start, cbr b m n) # visited) CNT) \<union> (if start = n then {visited} else create_all_possible_execution_paths n c ((start, cbr b m n) # visited) CNT)}"
        (is "lil \<in> ?DO \<union> ?COMM \<union> ?CBR")
      using aa2
      unfolding create_all_possible_execution_paths.simps if_P[OF ba1] if_P[OF b1] if_not_P[OF ba2]
      by force
    moreover
    {
      assume xInDo: "lil \<in> ?DO"
      then obtain f where c3: "lil \<in> create_all_possible_execution_paths (Suc start) c ((start, do f) # visited) CNT" and "(start, do f) \<in> \<union>\<^sub>s\<^sub>c c"
        by blast
      hence ?G
        using aa1[OF c3] aa4
        by force
    }
    moreover
    {
    assume xInDo: "lil \<in> ?COMM"
    then obtain ef f where c3: "lil \<in> create_all_possible_execution_paths (Suc start) c ((start, comm ef f) # visited) CNT" and "(start, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c"
      by force
    hence ?G
      using aa1[OF c3] aa4
      by force
    }
    moreover
    {
    assume xInDo: "lil \<in> ?CBR"
    then obtain b m n where c3: "lil \<in> (if start = m then {visited} else create_all_possible_execution_paths m c ((start, cbr b m n) # visited) CNT) \<union>
                          (if start = n then {visited} else create_all_possible_execution_paths n c ((start, cbr b m n) # visited) CNT)" and c4:"(start, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c"
      by blast
    hence inC: "set ((start, cbr b m n) # visited) \<subseteq> \<union>\<^sub>s\<^sub>c c"
      using c4 aa4 by auto
    have "lil \<in> {visited} \<longrightarrow> ?G"
      using list.set_sel(2) aa4 b singletonD subset_code(1) by fastforce
    hence ?G
      using aa1[of m "((start, cbr b m n) # visited)", OF _ inC] aa1[of n "((start, cbr b m n) # visited)", OF _ inC]
        aa4 b c3[unfolded create_all_possible_execution_paths_if_rewrite] c4
      by linarith
    }
    ultimately
    have ?G
    by fast
  }
  ultimately
  show ?G
    by blast
qed

text {* For any generated non empty path, we know that if the head's label is in code's domain,
the path consists solely of instructions of the code. *}

lemma create_all_possible_execution_paths_all_indom_c:
  assumes "lil \<in> create_all_possible_execution_paths start c visited n" and "lil \<noteq> []" and "set visited \<subseteq> \<union>\<^sub>s\<^sub>c c" and "fst (hd lil) \<in> dom c"
  shows "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c"
using assms
proof (induct n arbitrary: visited start)
  show "\<And>visited start.
       lil \<in> create_all_possible_execution_paths start c visited 0 \<Longrightarrow>
       lil \<noteq> [] \<Longrightarrow> set visited \<subseteq> \<union>\<^sub>s\<^sub>c c \<Longrightarrow> fst (hd lil) \<in> definitions.dom c \<Longrightarrow> set lil \<subseteq> \<union>\<^sub>s\<^sub>c c"
    unfolding create_all_possible_execution_paths.simps
    using list.set_sel(2)
    by fastforce
next
  fix CNT visited start
  assume
    aa1: "\<And>visited start.
           lil \<in> create_all_possible_execution_paths start c visited CNT \<Longrightarrow>
           lil \<noteq> [] \<Longrightarrow> set visited \<subseteq> \<union>\<^sub>s\<^sub>c c \<Longrightarrow>
           fst (hd lil) \<in> definitions.dom c \<Longrightarrow>
           set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    aa2: "lil \<in> create_all_possible_execution_paths start c visited (Suc CNT)" and
    aa3: "lil \<noteq> []" and
    aa4: "set visited \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    aa5: "fst (hd lil) \<in> definitions.dom c"
  let ?G = "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c"
  have "True"
    using aa2 unfolding "create_all_possible_execution_paths.simps"
    by simp
  {
    assume ba1: "\<not> \<not> indom start (set visited)"
    have ?G
      using aa2 aa4
      unfolding create_all_possible_execution_paths.simps if_not_P[OF ba1]
      using aa3 contra_subsetD list.set_sel(2) singletonD subsetI
      by fastforce
  }
  moreover
  {
    assume ba1: "start \<notin> definitions.dom c" and ba2: "\<not> indom start (set visited)"
    have ?G
      using aa2 aa4
      unfolding create_all_possible_execution_paths.simps if_P[OF ba2] if_P[OF ba1]
      using aa5 ba1
      by simp
  }
  moreover
  {
    assume b1: "\<not> indom start (set visited)" and ba2: "\<not> start \<notin> definitions.dom c"
    have ba1: "\<not> start \<in> {lbl. \<exists>i. (lbl, i) \<in> set visited}" using b1 unfolding indom_def by blast
    have X: "lil \<in> {x. \<exists>f. (start, do f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, do f) # visited) CNT} \<union>
        {x. \<exists>ef f. (start, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc start) c ((start, comm ef f) # visited) CNT} \<union>
        {x. \<exists>b m n. (start, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> (if start = m then {visited} else create_all_possible_execution_paths m c ((start, cbr b m n) # visited) CNT) \<union>
                          (if start = n then {visited} else create_all_possible_execution_paths n c ((start, cbr b m n) # visited) CNT)}"
        (is "lil \<in> ?DO \<union> ?COMM \<union> ?CBR")
      using aa2
      unfolding create_all_possible_execution_paths.simps if_P[OF ba1] if_P[OF b1] if_not_P[OF ba2]
      by force
    moreover
    {
      assume xInDo: "lil \<in> ?DO"
      then obtain f where c3: "lil \<in> create_all_possible_execution_paths (Suc start) c ((start, do f) # visited) CNT" and "(start, do f) \<in> \<union>\<^sub>s\<^sub>c c"
        by blast
      hence ?G
        using aa1[OF c3] aa4
        by (simp add: aa3 aa5)
    }
    moreover
    {
    assume xInDo: "lil \<in> ?COMM"
    then obtain ef f where c3: "lil \<in> create_all_possible_execution_paths (Suc start) c ((start, comm ef f) # visited) CNT" and "(start, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c"
      by force
    hence ?G
      using aa1[OF c3] aa4
      by (simp add: aa3 aa5)
    }
    moreover
    {
    assume xInDo: "lil \<in> ?CBR"
    then obtain b m n where c3: "lil \<in> (if start = m then {visited} else create_all_possible_execution_paths m c ((start, cbr b m n) # visited) CNT) \<union> (if start = n then {visited} else create_all_possible_execution_paths n c ((start, cbr b m n) # visited) CNT)" and "(start, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c"
      by blast
    hence ?G
      unfolding create_all_possible_execution_paths_if_rewrite
      using aa1[of m "((start, cbr b m n) # visited)"] aa1[of n "((start, cbr b m n) # visited)"]
        aa4 aa3 aa5
      by auto
    }
    ultimately
    have ?G
      by blast
  }
  ultimately
  show ?G
    by blast
qed

text {* The intuitive definition for \emph{all instructions being followed by a comm or leave} is equal to the
path reconstruction, where each path contains a @{term comm} or its end is outside of the code's domain. *}

lemma instr_followed_by_comm_or_leaves_eq_create_all_possible_execution_paths:
  fixes c :: "('s, 'e) structured_code"
  assumes
    a: "\<not> ((\<exists>sf f l. (l, comm sf f) \<in> set visited) \<or> (visited \<noteq> [] \<and> fst (hd visited) \<notin> definitions.dom c))"
  shows "instr_followed_by_comm_or_leaves l c {lbl . \<exists> i . (lbl, i) \<in> set visited} CNT = 
  (\<forall> X \<in> create_all_possible_execution_paths l c visited CNT . (\<exists> sf f l . (l, comm sf f) \<in> set X)
                                   \<or> (X \<noteq> [] \<and> fst (hd X) \<notin> dom c))"
using assms
proof (induct CNT arbitrary: l visited, simp)
  fix CNT l and visited :: "(nat \<times> ('s, 'e) instr) list"
  assume
    aa1: "\<And>l visited.
           \<not> ((\<exists>sf f l. (l, comm sf f) \<in> set visited) \<or> visited \<noteq> [] \<and> fst (hd visited) \<notin> definitions.dom c) \<Longrightarrow>
           instr_followed_by_comm_or_leaves l c {lbl. \<exists>i. (lbl, i) \<in> set visited} CNT = (\<forall>X\<in>create_all_possible_execution_paths l c visited CNT. (\<exists>sf f l. (l, comm sf f) \<in> set X) \<or> X \<noteq> [] \<and> fst (hd X) \<notin> definitions.dom c)" and
    aa2: "\<not> ((\<exists>sf f l. (l, comm sf f) \<in> set visited) \<or> visited \<noteq> [] \<and> fst (hd visited) \<notin> definitions.dom c)"
  let ?G = "instr_followed_by_comm_or_leaves l c {lbl. \<exists>i. (lbl, i) \<in> set visited} (Suc CNT) = (\<forall>X\<in>create_all_possible_execution_paths l c visited (Suc CNT). (\<exists>sf f l. (l, comm sf f) \<in> set X) \<or> X \<noteq> [] \<and> fst (hd X) \<notin> definitions.dom c)"
  (* show ?G *)
  {
    assume ba1: "l \<in> {lbl. \<exists>i. (lbl, i) \<in> set visited}"
    have b1: "\<not> \<not> indom l (set visited)" unfolding indom_def using ba1 by simp
    have ?G
      unfolding instr_followed_by_comm_or_leaves.simps
        if_P[OF ba1]
      using aa2
      by (simp add: b1)
  }
  moreover
  {
    assume ba1: "\<not> l \<in> {lbl. \<exists>i. (lbl, i) \<in> set visited}" and ba2: "l \<notin> definitions.dom c"
    have b1: "\<not> indom l (set visited)" unfolding indom_def using ba1 by blast
    have ?G
      unfolding create_all_possible_execution_paths.simps instr_followed_by_comm_or_leaves.simps
        unfolding if_not_P[OF ba1] if_P[OF ba2] if_P[OF b1]
      using ba2
      using ba1 empty_iff empty_set fst_conv list.sel(1) list.set_intros(1) singletonD
      by force
  }
  moreover
  {
    assume ba1: "\<not> l \<in> {lbl. \<exists>i. (lbl, i) \<in> set visited}" and ba2: "\<not> l \<notin> definitions.dom c"
    have b1: "\<not> indom l (set visited)" unfolding indom_def using ba1 by blast
    have ?G
        unfolding instr_followed_by_comm_or_leaves.simps create_all_possible_execution_paths.simps
        unfolding if_not_P[OF ba1] if_not_P[OF ba2] if_P[OF b1]
    proof (rule, rule)
      fix X
      assume
        ca1: "(\<forall>ef f. (l, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow> True) \<and>
         (\<forall>f. (l, do f) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow> instr_followed_by_comm_or_leaves (Suc l) c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT) \<and>
         (\<forall>b m n. (l, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow>
                  instr_followed_by_comm_or_leaves m c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT \<and> instr_followed_by_comm_or_leaves n c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT)" and
        ca2: "X \<in> {x. \<exists>f. (l, do f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc l) c ((l, do f) # visited) CNT} \<union>
              {x. \<exists>ef f. (l, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc l) c ((l, comm ef f) # visited) CNT} \<union>
              {x. \<exists>b m n. (l, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> (if l = m then {visited} else create_all_possible_execution_paths m c ((l, cbr b m n) # visited) CNT) \<union>
                               (if l = n then {visited} else create_all_possible_execution_paths n c ((l, cbr b m n) # visited) CNT)}"
          (is "X \<in> ?DO \<union> ?COMM \<union> ?CBR")
      let ?G = "(\<exists>sf f l. (l, comm sf f) \<in> set X) \<or> X \<noteq> [] \<and> fst (hd X) \<notin> definitions.dom c" (* show ?G *)
      {
        assume xInDo: "X \<in> ?DO"
        then obtain f where c3: "X \<in> create_all_possible_execution_paths (Suc l) c ((l, do f) # visited) CNT" and iInC: "(l, do f) \<in> \<union>\<^sub>s\<^sub>c c"
          by force
        have c1: "\<not> ((\<exists>sf fa l'. (l', comm sf fa) \<in> set ((l, do f) # visited)) \<or> (l, do f) # visited \<noteq> [] \<and> fst (hd ((l, do f) # visited)) \<notin> definitions.dom c)"
          using aa2 ba2
          by simp
        have c2: "({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) = ({lbl. \<exists>i. (lbl, i) \<in> set ((l, do f) # visited)})"
          by auto
        have ?G
          using aa1[of "(l, do f) # visited" "Suc l", OF c1] ca1[THEN cj3_2] c3 iInC
          unfolding indom_def c2
          by fast
      }
      moreover
      {
        assume xInDo: "X \<in> ?COMM"
        then obtain ef f where c3: "X \<in> create_all_possible_execution_paths (Suc l) c ((l, comm ef f) # visited) CNT" and iInC: "(l, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c"
          by force
        have ?G
          using create_all_possible_execution_paths_extends_visited[OF c3]
          by force
      }
      moreover
      {
        assume xInDo: "X \<in> ?CBR"
        then obtain b m n where c3: "X \<in> (if l = m then {visited} else create_all_possible_execution_paths m c ((l, cbr b m n) # visited) CNT) \<union>
                        (if l = n then {visited} else create_all_possible_execution_paths n c ((l, cbr b m n) # visited) CNT)" and
          iInC: "(l, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c"
          by force
        have c1: "\<not> ((\<exists>sf f l'. (l', comm sf f) \<in> set ((l, cbr b m n) # visited)) \<or> (l, cbr b m n) # visited \<noteq> [] \<and> fst (hd ((l, cbr b m n) # visited)) \<notin> definitions.dom c)"
          using aa2 ba2
          by auto
        have c2: "({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) = ({lbl. \<exists>i. (lbl, i) \<in> set ((l, cbr b m n) # visited)})"
          by auto
        have "X \<in> {visited} \<longrightarrow> False"
        proof (clarsimp)
          assume X: "X = visited"
          with c3 have lmn: "l = m \<or> l = n"
            using create_all_possible_execution_paths_extends_visited[of X _ c "((l, cbr b m n) # visited)" CNT]
            unfolding create_all_possible_execution_paths_if_rewrite
            by auto
          with ca1[THEN cj3_3] iInC show False
          proof (induct CNT, simp)
            fix CNT
            assume da1: "\<forall>b m n. (l, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow>
                   instr_followed_by_comm_or_leaves m c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) (Suc CNT) \<and>
                   instr_followed_by_comm_or_leaves n c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) (Suc CNT)"
            hence da1: "instr_followed_by_comm_or_leaves m c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) (Suc CNT) \<and>
                   instr_followed_by_comm_or_leaves n c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) (Suc CNT)"
              using iInC by blast
            have "m \<in> {lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l} \<or> n \<in> {lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}"
              using lmn by fast
            thus False
              using da1 lmn
              unfolding "instr_followed_by_comm_or_leaves.simps"
              by presburger
          qed
        qed
        hence ?G
          using
            aa1[of "(l, cbr b m n) # visited" "m", OF c1]
            aa1[of "(l, cbr b m n) # visited" "n", OF c1]
            ca1[THEN cj3_3] c3 iInC
          unfolding indom_def c2 create_all_possible_execution_paths_if_rewrite
          by blast
      }
      ultimately
      show ?G using ca2
        by blast
    next
      assume X: "\<forall>X\<in>{x. \<exists>f. (l, do f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc l) c ((l, do f) # visited) CNT} \<union>
        {x. \<exists>ef f. (l, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc l) c ((l, comm ef f) # visited) CNT} \<union>
        {x. \<exists>b m n. (l, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> (if l = m then {visited} else create_all_possible_execution_paths m c ((l, cbr b m n) # visited) CNT) \<union>
                         (if l = n then {visited} else create_all_possible_execution_paths n c ((l, cbr b m n) # visited) CNT)}.
       (\<exists>sf f l. (l, comm sf f) \<in> set X) \<or> X \<noteq> [] \<and> fst (hd X) \<notin> definitions.dom c"
      show "(\<forall>ef f. (l, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow> True) \<and>
    (\<forall>f. (l, do f) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow> instr_followed_by_comm_or_leaves (Suc l) c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT) \<and>
    (\<forall>b m n. (l, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow>
             instr_followed_by_comm_or_leaves m c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT \<and> instr_followed_by_comm_or_leaves n c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT)"
      proof(rule, simp, rule, rule, rule)
        fix f
        assume iInC: "(l, do f) \<in> \<union>\<^sub>s\<^sub>c c"
        have c1: "\<not> ((\<exists>sf fa l'. (l', comm sf fa) \<in> set ((l, do f) # visited)) \<or> (l, do f) # visited \<noteq> [] \<and> fst (hd ((l, do f) # visited)) \<notin> definitions.dom c)"
          using aa2 ba2
          by simp
        have c2: "({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) = ({lbl. \<exists>i. (lbl, i) \<in> set ((l, do f) # visited)})"
          by auto
        have " \<forall>X\<in>{x. \<exists>f. (l, do f) \<in> \<union>\<^sub>s\<^sub>c c \<and> x \<in> create_all_possible_execution_paths (Suc l) c ((l, do f) # visited) CNT}.
          (\<exists>sf f l. (l, comm sf f) \<in> set X) \<or> X \<noteq> [] \<and> fst (hd X) \<notin> definitions.dom c"
          using X by blast
        hence "\<forall>X\<in>create_all_possible_execution_paths (Suc l) c ((l, do f) # visited) CNT. (\<exists>sf f l. (l, comm sf f) \<in> set X) \<or> X \<noteq> [] \<and> fst (hd X) \<notin> definitions.dom c"
          using X iInC
          by blast
        thus "instr_followed_by_comm_or_leaves (Suc l) c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT"
          using aa1[of "(l, do f) # visited" "Suc l", OF c1] iInC
          unfolding indom_def c2
          by fast
      next
        show "\<forall>b m n. (l, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<longrightarrow> instr_followed_by_comm_or_leaves m c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT \<and> instr_followed_by_comm_or_leaves n c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT"
        proof (rule, rule, rule, rule)
          fix b m n
          assume iInC: "(l, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c"
          have c1: "\<not> ((\<exists>sf f l'. (l', comm sf f) \<in> set ((l, cbr b m n) # visited)) \<or> (l, cbr b m n) # visited \<noteq> [] \<and> fst (hd ((l, cbr b m n) # visited)) \<notin> definitions.dom c)"
            using aa2 ba2
            by auto
          have c2: "({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) = ({lbl. \<exists>i. (lbl, i) \<in> set ((l, cbr b m n) # visited)})"
            by auto
          have hlpr: "\<forall>X\<in>{x. \<exists>b m n. (l, cbr b m n) \<in> \<union>\<^sub>s\<^sub>c c \<and>
                    x \<in> (if l = m then {visited} else create_all_possible_execution_paths m c ((l, cbr b m n) # visited) CNT) \<union>
                         (if l = n then {visited} else create_all_possible_execution_paths n c ((l, cbr b m n) # visited) CNT)}.
                         (\<exists>sf f l. (l, comm sf f) \<in> set X) \<or> X \<noteq> [] \<and> fst (hd X) \<notin> definitions.dom c" using X by blast
          have c3: "(\<forall>X\<in>(if l = m then {visited} else create_all_possible_execution_paths m c ((l, cbr b m n) # visited) CNT) \<union>
                         (if l = n then {visited} else create_all_possible_execution_paths n c ((l, cbr b m n) # visited) CNT). (\<exists>sf f l. (l, comm sf f) \<in> set X) \<or> X \<noteq> [] \<and> fst (hd X) \<notin> definitions.dom c)"
            using hlpr iInC
            using mem_Collect_eq by blast
          have "m \<noteq> l \<and> n \<noteq> l"
          proof (rule ccontr, simp)
            assume eq: "m = l \<or> n = l"
            have "\<not>(\<forall>X\<in>{visited}. (\<exists>sf f l. (l, comm sf f) \<in> set X) \<or> X \<noteq> [] \<and> fst (hd X) \<notin> definitions.dom c)"
              using aa2 by blast
            with c3 eq show False
              by fastforce
          qed
          thus "instr_followed_by_comm_or_leaves m c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT \<and> instr_followed_by_comm_or_leaves n c ({lbl. \<exists>i. (lbl, i) \<in> set visited} \<union> {l}) CNT"
            using aa1[of "(l, cbr b m n) # visited" "m", OF c1] aa1[of "(l, cbr b m n) # visited" "n", OF c1] iInC c1 c3
            unfolding indom_def c2 create_all_possible_execution_paths_if_rewrite'
            by meson
        qed
      qed
    qed
  }
  ultimately
  show ?G
    unfolding instr_followed_by_comm_or_leaves.simps
    by fast
qed

subsection {* Infinite systems *}

text {* In an infinite system each statically possible execution path can be extended infinitely and still stays inside the code domain.
We will never at some point exit the code's domain.
*}

fun noPathOutsideFun :: "('s, 'e) structured_code \<Rightarrow> label set \<Rightarrow> bool" where
"noPathOutsideFun (one lbl (do f)) lbls = (Suc lbl \<in> lbls)" |
"noPathOutsideFun (one lbl (cbr b m n)) lbls = (m \<in> lbls \<and> n \<in> lbls)" |
"noPathOutsideFun (one lbl (comm ef sf)) lbls = (Suc lbl \<in> lbls)" |
"noPathOutsideFun (seq c1 c2) lbls = ((noPathOutsideFun c1 lbls) \<and> (noPathOutsideFun c2 lbls))"

definition "noPathOutside c = noPathOutsideFun c (dom c)"
lemma noPathOutside_For_Every_Subinstruction':
  assumes "noPathOutsideFun c lbls" and "(l, i) \<in> \<union>\<^sub>s\<^sub>c c"
  shows "noPathOutsideFun (one l i) lbls"
using assms unfolding noPathOutside_def
proof (induct c, auto) qed

lemma noPathOutside_For_Every_Subinstruction:
  assumes "noPathOutside c" and "(l, i) \<in> \<union>\<^sub>s\<^sub>c c"
  shows "noPathOutsideFun (one l i) (dom c)"
using noPathOutside_For_Every_Subinstruction'[OF assms[unfolded noPathOutside_def]]
.

subsection {* Allowed communication *}

text {* The property is @{prop True}, if every statically possible communicable event is contained in the supplied set. *}

definition wff_cuc_allowed_communication :: "('s, 'e) structured_code \<Rightarrow> 'e set \<Rightarrow>  bool" where
"wff_cuc_allowed_communication c E = (
  (\<Union>{ C . \<exists> ef f lbl . (lbl, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<and> C = \<Union>{v . \<exists> s . v = ef s}}) \<subseteq> E
)"

end
