section {* Parallel SF semantics *}

text{* This section introduces parallel concepts of the stable failure semantics and ultimately exercises it on a parallel buffer example. *}

theory "trace_restriction"
imports stable_failures parallel_base
begin

paragraph {* restrictTo - Trace restriction *}

text {* Restriction of traces (\emph{restrictTo}) works the same as it is defined for CSP. *}

primrec restrictTo :: "'e definitions.trace \<Rightarrow> 'e definitions.event set \<Rightarrow> 'e definitions.trace" ("_ \<restriction> _" [81, 80] 95) where
"(x # xs) \<restriction> \<alpha> = (if x \<in> \<alpha> then x # (restrictTo xs \<alpha>) else restrictTo xs \<alpha>)" |
"[] \<restriction> \<alpha> = []"

text {* Helper lemmas *}

lemma restrictTo_set1: "set (tr \<restriction> \<alpha>) \<subseteq> \<alpha>"
proof (induct tr, auto) qed
lemma restrictTo_set2: "(set tr) \<inter> \<alpha> = set (tr \<restriction> \<alpha>)"
proof (induct tr, auto) qed
lemma restrictTo_set3: "set (tr \<restriction> \<alpha>) \<subseteq> set tr"
proof (induct tr, auto) qed
lemma restrictTo_set4:
  assumes  "\<beta> \<subseteq> \<alpha>"
  shows "set (tr \<restriction> \<beta>) \<subseteq> set (tr \<restriction> \<alpha>)"
using assms proof (induct tr, auto) qed
lemma restrictTo_set5: "set tr \<subseteq> \<alpha> \<Longrightarrow> tr \<restriction> \<alpha> = tr"
proof (induct tr, auto) qed

text {* If we restrict a trace first with some set and then with a subset, we could have used the subset
in the first place. *}

lemma restrictTo_list1:
  assumes "\<beta> \<subseteq> \<alpha>"
  shows "(tr \<restriction> \<alpha>) \<restriction> \<beta> = tr \<restriction> \<beta>"
using assms
proof (induct tr, auto) qed

lemma restrictTo_list3:
  assumes "\<alpha> \<subseteq> \<beta>"
  shows "(tr \<restriction> \<alpha>) \<restriction> \<beta> = tr \<restriction> \<alpha>"
using assms
proof (induct tr, auto) qed

text {* If two restricted traces are equal, they stay equal when restricted with a subset. *}

lemma restrictTo_list2:
  assumes a: "tr \<restriction> \<alpha> = tr' \<restriction> \<alpha>" and subs: "\<beta> \<subseteq> \<alpha>"
  shows "tr \<restriction> \<beta> = tr' \<restriction> \<beta>"
using a proof (induct tr arbitrary:tr')
  fix tr'
  assume
    aa1: "[] \<restriction> \<alpha> = tr' \<restriction> \<alpha>"
  show "[] \<restriction> \<beta> = tr' \<restriction> \<beta>"
    using restrictTo_set4[OF subs, of tr'] aa1
    unfolding restrictTo.simps
    by force
next
  fix a tr tr'
  assume
    aa1: "\<And>tr'. tr \<restriction> \<alpha> = tr' \<restriction> \<alpha> \<Longrightarrow> tr \<restriction> \<beta> = tr' \<restriction> \<beta>" and
    aa2: "(a # tr) \<restriction> \<alpha> = tr' \<restriction> \<alpha>"
  consider (inAlpha) "a \<in> \<alpha>" | (notInAlpha) "a \<notin> \<alpha>" by blast
  thus "(a # tr) \<restriction> \<beta> = tr' \<restriction> \<beta>"
  proof (cases)
    case notInAlpha
    hence "a \<notin> \<beta>" using subs by blast
    with notInAlpha have
      eqA: "(a # tr) \<restriction> \<alpha> = tr \<restriction> \<alpha>" and
      eqB: "(a # tr) \<restriction> \<beta> = tr \<restriction> \<beta>"
      unfolding restrictTo.simps
      by auto
    from aa1[OF aa2[unfolded eqA]] show ?thesis
      unfolding eqB .
  next
    case inAlpha
    hence eq: "(a # tr) \<restriction> \<alpha> = a # (tr \<restriction> \<alpha>)"
      by simp
    obtain tr'' where
      o1: "tr \<restriction> \<alpha> = tr'' \<restriction> \<alpha>" and
      o2: "tr' \<restriction> \<alpha> = a # (tr'' \<restriction> \<alpha>)"
      using aa2 eq by presburger
    have eq2: "a # (tr \<restriction> \<beta>) = a # (tr'' \<restriction> \<beta>)"
      using aa1[OF o1] by simp
    consider (inBeta) "a \<in> \<beta>" | (notInBeta) "a \<notin> \<beta>" by blast
    thus ?thesis
    proof (cases)
      case inBeta
      have eq3: "a # (tr \<restriction> \<beta>) = (a # tr) \<restriction> \<beta>"
        using inBeta by fastforce
      have "a # (tr'' \<restriction> \<beta>) = a # ((tr'' \<restriction> \<alpha>) \<restriction> \<beta>)"
        using restrictTo_list1[OF subs] by simp
      also have "... = ((a # (tr'' \<restriction> \<alpha>)) \<restriction> \<beta>)"
        using inAlpha inBeta by simp
      also have "... = (tr' \<restriction> \<alpha>) \<restriction> \<beta>"
        using o2 by simp
      also have "... = tr' \<restriction> \<beta>"
        using restrictTo_list1[OF subs] by simp
      finally show ?thesis
        unfolding eq3[symmetric] eq2
        .
    next
      case notInBeta
      thus ?thesis
        by (metis aa2 restrictTo_list1 subs)
    qed
  qed
qed

lemma restrictTo_intersection1: "tr \<restriction> \<alpha> \<restriction> \<beta> = tr \<restriction> (\<alpha> \<inter> \<beta>)"
proof (induct tr, auto) qed

text {* Appending of an allowed event can be done inside or outside of the restriction. *}

lemma restrictTo_append_in_alpha: assumes "a \<in> \<alpha>" shows "(tr \<restriction> \<alpha>) @ [a] = (tr @ [a]) \<restriction> \<alpha>"
  proof (induct tr, auto simp add: assms) qed

lemma restrictTo_append_list_in_alpha: assumes "set as \<subseteq> \<alpha>" shows "(tr \<restriction> \<alpha>) @ as = (tr @ as) \<restriction> \<alpha>"
  using assms
apply (induct as rule: rev_induct)
apply (auto)
apply (metis append.assoc restrictTo_append_in_alpha)
done

text {* This does not apply for a forbidden event. *}
lemma restrictTo_append_not_in_alpha: shows "(a \<notin> \<alpha>) = ((tr @ [a]) \<restriction> \<alpha> = (tr \<restriction> \<alpha>))"
  proof (induct tr, auto) qed

lemma restrictTo_append_list_not_in_alpha:
  assumes "tr \<restriction> \<alpha> = (tr @ as) \<restriction> \<alpha>"
  shows "set as \<inter> \<alpha> = {}"
proof (rule ccontr)
  assume contr: "set as \<inter> \<alpha> \<noteq> {}"
  then obtain a where
    a_in_as: "a \<in> set as" and
    a_in_\<alpha>: "a \<in> \<alpha>"
    by auto
  have hlpr: "(tr @ as) \<restriction> \<alpha> = (tr \<restriction> \<alpha>) @ (as \<restriction> \<alpha>)"
  proof (induct as rule: rev_induct)
    case Nil
    then show ?case by simp
  next
    case (snoc x xs)
    then show ?case
      by (metis append.assoc restrictTo_append_in_alpha restrictTo_append_not_in_alpha)
  qed
  have "(as \<restriction> \<alpha>) \<noteq> []"
    using a_in_as a_in_\<alpha>
    by (metis contr list.set(1) restrictTo_set2)
  thus False
    using assms[unfolded hlpr] by simp
qed

end
