section {* Stable Failure Semantics *}

theory stable_failures
imports small_step static_code_analysis
begin

subsection {* Preparation *}

text {*
  In the stable failures semantics, we need an additional distinction whether to treat a state as a
  communication-failure, or a normal-failure.
  \begin{description}
    \item[Communication failures] occur after executing a comm instruction (being essentially
    now a copy of the normal state, which executed the comm instruction).
    \item[Normal failures] are everything else - If they are used at some point to execute an
    instruction, they are removed. Thus, after executing a program, all remaining
    normal states have a program counter outside of the program's domain. 
  \end{description}
    *}
datatype 's NCstate = normal "'s state"
                    | communication "'s state"

text {*
  A failure contains the trace of communicated events (thus, always becoming one
  piece larger upon communication, appended to the end), a state (comm or normal), and
  a set of events which could not be communicated. The general variables used for the three elements
  are @{text tr}, @{text s} or @{text t}, and @{text X}, respectively.
  As a normal failure does not relate to communication at all, @{text X} can be any subset of @{term \<Sigma>}.
  *}
type_synonym ('s, 'e) failure = "'e trace \<times> 's NCstate \<times> 'e event set"

paragraph {* Accessor and transformer functions *}

text{* Is a state a normal state? *}
primrec N :: "'a NCstate \<Rightarrow> bool" where
  "N (normal _) = True" |
  "N (communication _) = False"

text {* Extract the program counter out of a state. *}
primrec NCPC :: "'a NCstate \<Rightarrow> label" where
  "NCPC (normal s) = PC s" |
  "NCPC (communication s) = PC s"

text {* Extract the register set out of a state *}
primrec NCRS :: "'a NCstate \<Rightarrow> 'a" where
  "NCRS (normal s) = R s" |
  "NCRS (communication s) = R s"

text {* Convert any state into a communication state - required in comm instruction *}
primrec NCC :: "'a NCstate \<Rightarrow> 'a NCstate" where
  "NCC (normal s) = communication s" |
  "NCC (communication s) = communication s"

text {* Extracts the \emph{plain} state out of a (SF)state. The \emph{plain} state is
   the same state as used in the smallstep and trace semantics. *}
primrec NCS :: "'a NCstate \<Rightarrow> 'a state" where
  "NCS (normal s) = s" |
  "NCS (communication s) = s"


text{*
  Remove all normal states from S, whose program counter is contained in pcs.
  So afterwards, the only states whose pc still is contained in pcs,
  are communication states.
 *}
definition SClean :: "('s, 'e) failure set \<Rightarrow> label set \<Rightarrow> ('s, 'e) failure set" (infix "-\<^sub>s\<^sub>f" 65) where
"SClean S pcs = {(tr,  s,  X) . (tr, s, X) \<in> S \<and> (((NCPC s) \<notin> pcs) \<or> \<not> (N s))}"

declare SClean_def [simp]

paragraph {* Helper Lemmas *}

lemma NCCNotNormal: "\<not> N (NCC s)"
  by (metis N.simps(2) NCC.simps(1) NCC.simps(2) NCstate.exhaust)

lemma NCCNotNormal2: "N s \<Longrightarrow> s = NCC s \<Longrightarrow> False"
  using NCCNotNormal by metis

lemma NCRSh1: "NCRS (fst (snd (a, b, c))) = NCRS b"
  by simp

lemma NCRSh2: "NCRS (NCC b) = NCRS b"
  by (metis NCC.simps(1) NCC.simps(2) NCRS.simps(1) NCRS.simps(2) NCstate.exhaust)

lemma Nh1: "N (fst (snd (a, b, c))) = N b"
  by simp

lemma NCPCh2: "NCPC (NCC b) = NCPC b"
  by (metis NCC.simps(1) NCC.simps(2) NCPC.simps(1) NCPC.simps(2) NCstate.exhaust)

lemma PCNCS: "PC (NCS s) = NCPC s"
  by (metis NCPC.simps(1) NCPC.simps(2) NCS.simps(1) NCS.simps(2) NCstate.exhaust)

lemma RNCS: "R (NCS s) = NCRS s"
  by (metis NCRS.simps(1) NCRS.simps(2) NCS.simps(1) NCS.simps(2) NCstate.exhaust)

lemma NCCSameR: "t = NCC u \<Longrightarrow> NCRS t = NCRS u"
  by (metis NCRSh2)
  
lemma Nnormal: "\<lbrakk>N s; s' = NCS s\<rbrakk> \<Longrightarrow> s = normal s'"
  by (metis N.simps(2) NCS.simps(1) NCstate.exhaust)

lemma NCS_NCC: "NCS (NCC s) = NCS s"
  by (metis NCC.simps(1) NCC.simps(2) NCS.simps(1) NCS.simps(2) NCstate.exhaust)

text {* If the structure and content of two states are equal, then the two states are equal. *}
lemma NC_state_eq:
  assumes
    a: "NCPC u = NCPC s" and
    b: "N u" and
    c: "N s" and
    d: "NCRS u = NCRS s"
  shows "u = s"
proof -
  obtain t :: "'a state" where
    ao1: "R t = NCRS u \<and> PC t = NCPC u \<and> t = NCS u"
    using PCNCS RNCS 
    by blast
  obtain t' :: "'a state" where
    ao2: "R t' = NCRS s \<and> PC t' = NCPC s \<and> t' = NCS s"
    using PCNCS RNCS 
    by blast
  from ao1 have a1: "R t = NCRS u \<and> PC t = NCPC u"
    by auto
  from ao2 have a2: "R t' = NCRS s \<and> PC t' = NCPC s"
    by auto
  from  ao1 ao2 have "NCS s = NCS u"
    using a cases d equality select_convs(3) by auto
  thus ?thesis
    using b c
    by (metis Nnormal)
qed

lemma SClean_distr:
  "(S1 -\<^sub>s\<^sub>f pcs) \<union> (S2 -\<^sub>s\<^sub>f pcs) = (S1 \<union> S2) -\<^sub>s\<^sub>f pcs"
by auto

lemma SClean_union: "(\<Union>f\<in>A. f (S)) -\<^sub>s\<^sub>f definitions.dom (c) = (\<Union>f\<in>A. f (S) -\<^sub>s\<^sub>f definitions.dom (c))"
  by auto

subsection {* The Semantics *}

locale SFsemSigma=liSet lis
for lis :: "('s, 'e)labeled_instruction_set" +
fixes \<Sigma> :: "'e set"
begin

fun sfsem :: "('s, 'e) structured_code \<Rightarrow> ('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set" ("\<lbrakk>_\<rbrakk>(_)"  [700, 700] 999)
and extends :: "('s, 'e) structured_code \<Rightarrow> (('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set) \<Rightarrow> ('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set"
where
sfdof: "sfsem (one l (do f)) S = (S -\<^sub>s\<^sub>f {l}) \<union> 
    {(tr, t, X). \<exists>u X'. (tr, u, X') \<in> S \<and> NCPC u = l \<and> N u \<and> N t \<and> NCPC t = Suc l \<and> NCRS t \<in> f (NCRS u) \<and> X \<subseteq> \<Sigma>}"
|
sfcbr: "sfsem (one l (cbr b m n)) S = (S -\<^sub>s\<^sub>f {l}) \<union> 
    {(tr, t, X). \<exists>u X'. (tr, u, X') \<in> S \<and> NCPC u = l \<and> N u \<and> N t \<and> NCRS u = NCRS t \<and> X \<subseteq> \<Sigma> \<and> 
      ((b (NCRS u) \<and> NCPC t = m) \<or> (\<not>b (NCRS u) \<and> NCPC t = n))}"
|
sfcomm: "sfsem (one l (comm ef f)) S = (S -\<^sub>s\<^sub>f {l}) \<union> 
    {(tr, t, X). \<exists>u X'. (tr, u, X') \<in> S \<and> (NCPC u = l) \<and> N u \<and> X \<subseteq> \<Sigma> - ef (NCRS u) \<and> t = NCC u} \<union>
    {(tr', t, X). \<exists>u X' tr ev. (tr, u, X') \<in> S \<and> NCPC u = l \<and> N u \<and> N t \<and> (ev \<in> (ef (NCRS u))) \<and> NCPC t = Suc l \<and> NCRS t = (f (NCRS u) ev) \<and> tr' = tr@[ev] \<and> X \<subseteq> \<Sigma>}"
|
sfseq: "sfsem (sc_1 \<oplus> sc_2) S = (ccpo_class.fixp (extends (sc_1 \<oplus> sc_2)) S) -\<^sub>s\<^sub>f (dom (sc_1 \<oplus> sc_2))"
|
extends: "extends (sc_1 \<oplus> sc_2) f = (\<lambda>S. f(\<lbrakk>sc_1\<rbrakk>(S)) \<union> f(\<lbrakk>sc_2\<rbrakk>(S)) \<union> S)"

definition sfsem_refinement :: "('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set \<Rightarrow> bool"
("_ \<subseteq>\<^sub>S\<^sub>F _") where
"(S \<subseteq>\<^sub>S\<^sub>F S') = (\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> (\<exists> s' X' . (tr, s', X') \<in> S' \<and> X \<subseteq> X'))"

paragraph {* Some fixpoint related lemmas. *}

text{* The functor @{text extends} is monotone. *}
lemma func_monotone[simp]: "monotone (op \<le>) (op \<le>) (extends (sc_1 \<oplus> sc_2))"
apply(rule monotoneI) apply simp
unfolding le_fun_def by blast

text {* Thus, the induction rule can be simplified a little. *}
lemma fixp_induct_sf:
  "ccpo.admissible Sup op \<le> P \<Longrightarrow> P (Sup {}) \<Longrightarrow> (\<And>x. P x \<Longrightarrow> P (extends (c1 \<oplus> c2) x)) \<Longrightarrow> P (ccpo_class.fixp (extends (c1 \<oplus> c2)))"
  using "ccpo_class.fixp_induct"[of _ "extends (_ \<oplus> _)"] func_monotone
by blast

text {* Some convenient unfoldings. *}

lemma sfseq_unfold:
  "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S = extends (c1 \<oplus> c2) (ccpo_class.fixp (extends (c1 \<oplus> c2))) S -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)"
unfolding sfseq
using ccpo_class.fixp_unfold[OF func_monotone, of c1 c2]
by simp
  
lemma sfseq_unfold2: "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk> (\<lbrakk>c1\<rbrakk>S) \<union> \<lbrakk>(c1 \<oplus> c2)\<rbrakk> (\<lbrakk>c2\<rbrakk>S) \<union> (S -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2))"
  unfolding sfseq
  using fixp_unfold[OF func_monotone, unfolded extends, of c1 c2]
  by (metis SClean_distr extends.simps local.sfseq sfseq_unfold)

text {* With fixed S, we can simply stop the unfolding-recursion. *}
lemma sfseq_unfold3: "ccpo_class.fixp (extends (c1 \<oplus> c2)) S = extends (c1 \<oplus> c2) (ccpo_class.fixp (extends (c1 \<oplus> c2))) S"
  using fixp_unfold[OF func_monotone]
by simp

paragraph {* Semantic Lemmas *}

lemma oplus_commutative: "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>(S) = \<lbrakk>(c2 \<oplus> c1)\<rbrakk>(S)"
unfolding sfseq
proof -
  have hlpr: "\<And> f g S . \<forall> S . f S = g S \<Longrightarrow> ccpo_class.fixp f S = ccpo_class.fixp g S"
    by presburger
  have d: "dom (c1 \<oplus> c2) = dom (c2 \<oplus> c1)" by auto
  have eq: "\<forall>S. extends (c1 \<oplus> c2) S = extends (c2 \<oplus> c1) S" by (simp add: Un_assoc sup.left_commute)
  from hlpr[OF eq, of S] d
  show "ccpo_class.fixp (extends (c1 \<oplus> c2)) S -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2) = ccpo_class.fixp (extends (c2 \<oplus> c1)) S -\<^sub>s\<^sub>f definitions.dom (c2 \<oplus> c1)"
    by presburger
qed

text {* Imagine in the following @{text "g=id"}, then taking the union over the start states is distributive over densem. *}

lemma sfsem_big_union_additive: "\<lbrakk>c\<rbrakk>(\<Union>f\<in>PS . g f) = (\<Union>f\<in>PS. \<lbrakk>c\<rbrakk>(g f))"
proof (induct c arbitrary: PS g)
  fix l :: label and i :: "('s, 'e) instr" and PS :: "'b set" and g :: "'b \<Rightarrow> ('s, 'e) failure set"
  show "\<lbrakk>(one l i)\<rbrakk>(\<Union>f\<in>PS . g f) = (\<Union>f\<in>PS. \<lbrakk>(one l i)\<rbrakk>(g f))"
  proof (cases i, simp_all, fast+) qed
next
  fix c1 c2 :: "('s, 'e) structured_code" and PS :: "'b set" and g :: "'b \<Rightarrow> ('s, 'e) failure set"
  assume
    aa1: "\<And>(PS:: 'b set) (g::'b \<Rightarrow> ('s, 'e) failure set). \<lbrakk>c1\<rbrakk>(\<Union>f\<in>PS. g f) = (\<Union>f\<in>PS. \<lbrakk>c1\<rbrakk>g f)" and
    aa2: "\<And>(PS:: 'b set) (g::'b \<Rightarrow> ('s, 'e) failure set). \<lbrakk>c2\<rbrakk>(\<Union>f\<in>PS. g f) = (\<Union>f\<in>PS. \<lbrakk>c2\<rbrakk>g f)"
  have "ccpo_class.fixp (extends (c1 \<oplus> c2)) (\<Union>f\<in>PS. g f) = (\<Union>f\<in>PS. ccpo_class.fixp (extends (c1 \<oplus> c2)) (g f))"
  proof (induct arbitrary: PS g rule: fixp_induct_sf)
    show "ccpo.admissible Sup op \<le> (\<lambda>(a::('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set). \<forall>x (xa::'b \<Rightarrow> ('s, 'e) failure set). a (UNION x xa) = (\<Union>f\<in>x. a (xa f)))"
      unfolding "ccpo.admissible_def"
    proof (rule, rule, rule, rule, rule, rule)
      fix A :: "(('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set) set" and PS :: "'b set" and g :: "'b \<Rightarrow> ('s, 'e) failure set"
      assume
        ba1: "A \<noteq> {}" and
        ba2: "\<forall>a\<in>A. \<forall>PS (g :: 'b \<Rightarrow> ('s, 'e) failure set). a (\<Union>s\<in>PS. g s) = (\<Union>s\<in>PS. a (g s))"
      have "(\<Union>a\<in>A. a (\<Union>s\<in>PS. g s)) = (\<Union>a\<in>A. (\<Union>s\<in>PS. a (g s)))"
        using ba2
        by auto
      moreover
      have "(\<Union>a\<in>A. a (\<Union>s\<in>PS. g s)) = {t. \<exists>a\<in>A. t \<in> a {t. \<exists>s\<in>PS. t \<in> g s}}"
        unfolding UNION_eq by simp
      moreover
      have "(\<Union>a\<in>A. (\<Union>s\<in>PS. a (g s))) = {t. \<exists>s\<in>PS. \<exists>a\<in>A. t \<in> a (g s)}"
        unfolding UNION_eq by auto
      ultimately
      show "Sup A (UNION PS g) = (\<Union>f\<in>PS. Sup A (g f))"
        unfolding  UNION_eq
        by fastforce
    qed
  next
    fix x :: "('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set" and PS :: "'b set" and g :: "'b \<Rightarrow> ('s, 'e) failure set"
    assume ba1: "\<And>(PS::'b set) g. x (\<Union>s\<in>PS. g s) = (\<Union>s\<in>PS. x (g s))"
    have "x (\<lbrakk>c1\<rbrakk>(\<Union>s\<in>PS. g s)) = (\<Union>s\<in>PS. x (\<lbrakk>c1\<rbrakk>g s))"
    proof -
      have "x (\<lbrakk>c1\<rbrakk>(\<Union>s\<in>PS. g s)) = x (\<Union>s\<in>PS. \<lbrakk>c1\<rbrakk>g s)"
        using aa1[of g PS] by simp
      moreover
      have "x (\<Union>s\<in>PS. \<lbrakk>c1\<rbrakk>g s) = (\<Union>s\<in>PS. x (\<lbrakk>c1\<rbrakk>g s))"
        using ba1[of "\<lambda>s. \<lbrakk>c1\<rbrakk>g s"] .
      ultimately show ?thesis by simp
    qed
    moreover
    have "x (\<lbrakk>c2\<rbrakk>(\<Union>s\<in>PS. g s)) = (\<Union>s\<in>PS. x (\<lbrakk>c2\<rbrakk>g s))"
    proof -
      have "x (\<lbrakk>c2\<rbrakk>(\<Union>s\<in>PS. g s)) = x (\<Union>s\<in>PS. \<lbrakk>c2\<rbrakk>g s)"
        using aa2[of g] by simp
      moreover
      have "x (\<Union>s\<in>PS. \<lbrakk>c2\<rbrakk>g s) = (\<Union>s\<in>PS. x (\<lbrakk>c2\<rbrakk>g s))"
        using ba1[of "\<lambda>s. \<lbrakk>c2\<rbrakk>g s"] .
      ultimately show ?thesis by simp
    qed
    ultimately
    show "extends (c1 \<oplus> c2) x (UNION PS g) = (\<Union>f\<in>PS. extends (c1 \<oplus> c2) x (g f))"
      unfolding extends
      by auto
  qed simp
  thus "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>UNION PS g = (\<Union>f\<in>PS. \<lbrakk>(c1 \<oplus> c2)\<rbrakk>g f)"
    by auto
qed

text {* A convenient special case of the above for only two sets. *}

lemma sfsem_union_additive: "\<lbrakk>code\<rbrakk>(A \<union> B) =  \<lbrakk>code\<rbrakk>(A) \<union>  \<lbrakk>code\<rbrakk>(B)"
  using sfsem_big_union_additive[of code _ "{A, B}"]
  by fastforce

text {* If S only contains communication states, and normal states outside of @{term c}'s domain,
  then the application of @{text "\<lbrakk>c\<rbrakk>"} does not change anything, behaving like @{term id}. *}

lemma pc_does_not_point_to: 
  assumes
    a: "\<forall> (tr, t, X) \<in> S . \<not> ( NCPC t \<in> dom c \<and> N t)"
  shows
    "\<lbrakk>c\<rbrakk>S = S"
  using assms
proof (induct c arbitrary: S)
  fix
    c1 c2 :: "('s, 'e) structured_code" and
    S :: "('s, 'e) failure set"
  assume
    aa1: "(\<And>S. \<forall>(tr, t, X)\<in>S. \<not> (NCPC t \<in> definitions.dom c1 \<and> N t) \<Longrightarrow> \<lbrakk>c1\<rbrakk>S = S)" and
    aa2: "(\<And>S. \<forall>(tr, t, X)\<in>S. \<not> (NCPC t \<in> definitions.dom c2 \<and> N t) \<Longrightarrow> \<lbrakk>c2\<rbrakk>S = S)" and
    aa3: "\<forall>(tr, t, X)\<in>S. \<not> (NCPC t \<in> definitions.dom (c1 \<oplus> c2) \<and> N t)"
  from aa3 have a0: "\<forall>(tr, t, X)\<in>S.
    \<not> (NCPC t \<in> definitions.dom c1 \<and> N t) \<and> \<not> (NCPC t \<in> definitions.dom c2 \<and> N t)"
    by auto
  have "(extends (c1 \<oplus> c2) (ccpo_class.fixp (extends (c1 \<oplus> c2)))) S -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2) = S"
    using aa3
  proof (induct arbitrary: S rule: fixp_induct_sf)
    fix
      x :: "('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set" and
      S' :: "('s, 'e) failure set"
    assume
      ba1: "(\<And>S. \<forall>(tr, t, X)\<in>S.
                \<not> (NCPC t \<in> definitions.dom (c1 \<oplus> c2) \<and>
                    N t) \<Longrightarrow>
             extends (c1 \<oplus> c2) x S -\<^sub>s\<^sub>f
             definitions.dom (c1 \<oplus> c2) =
             S)" and
      ba2: "\<forall>(tr, t, X)\<in>S'. \<not> (NCPC t \<in> definitions.dom (c1 \<oplus> c2) \<and> N t)"
    from ba2 aa1[of S'] aa2[of S'] have b2: "\<lbrakk>c1\<rbrakk>S' = S' \<and> \<lbrakk>c2\<rbrakk>S' = S'"
      by fastforce
    from ba1 have b3: "(\<And>S. \<forall>(tr, t, X)\<in>S. \<not> (NCPC t \<in> definitions.dom (c1 \<oplus> c2) \<and> N t) \<Longrightarrow>
             (x (\<lbrakk>c1\<rbrakk>S) \<union> x (\<lbrakk>c2\<rbrakk>S) \<union> S) -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2) = S)" unfolding extends
      by simp
    show "extends (c1 \<oplus> c2) (extends (c1 \<oplus> c2) x) S' -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)
         = S'"
      using b3[of S', OF ba2]
      unfolding extends b2[THEN cj2_1] b2[THEN cj2_2]
      by simp
  next
    show "ccpo.admissible Sup op \<le>
     (\<lambda>a. \<forall>x. (\<forall>(tr, t, X)\<in>x. \<not> (NCPC t \<in> definitions.dom (c1 \<oplus> c2) \<and> N t)) \<longrightarrow> extends (c1 \<oplus> c2) a x -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2) = x)"
      unfolding "ccpo.admissible_def" chain_def SClean_def
      by (auto, blast+)
  qed auto
  thus "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S = S" using sfseq sfseq_unfold
    by blast
next
  fix
    l :: label and
    i :: "('s, 'e) instr" and
    S :: "('s, 'e) failure set"
  assume 
    a1: "\<forall>(tr, t, X)\<in>S. \<not> (NCPC t \<in> definitions.dom (one l i) \<and> N t)"
  show "\<forall>(tr, t, X)\<in>S. \<not> (NCPC t \<in> definitions.dom (one l i) \<and> N t) \<Longrightarrow> \<lbrakk>(one l i)\<rbrakk>S = S"
  proof (cases i, auto+) qed
qed

text {* For a concrete instance of a state in S: If it is a communication state or its program counter
outside of the code's domain, then it is not affected by the application of @{text "\<lbrakk>c\<rbrakk>"}. *}

lemma pc_does_not_point_to_single: 
  assumes
    not_pcN: "\<not> ( NCPC s \<in> dom c \<and> N s)" and
    inS: "((tr, s, X) \<in> S)"
  shows
    "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
proof -
  obtain SI SO :: "('s, 'e) failure set" where
     S: "S = SO \<union> SI" and
     SI: "SI = {(tr, s, X) . (tr, s, X) \<in> S \<and> ((NCPC s \<in> dom c) \<and> N s)}" and
     SO: "SO = {(tr, s, X) . (tr, s, X) \<in> S \<and> \<not> ((NCPC s \<in> dom c) \<and> N s)}"
    by auto
  have a1: "\<forall>(tr, t, X)\<in>SO. \<not> (NCPC t \<in> definitions.dom c \<and> N t)"
    using SO 
    by fast
  have a2: "\<forall>(tr, t, X)\<in>SI. (NCPC t \<in> definitions.dom c \<and> N t)"
    using SI 
    by fast
  have "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>SO \<union> \<lbrakk>c\<rbrakk>SI"
    using SO inS S not_pcN pc_does_not_point_to[OF a1]
    by auto
  thus ?thesis
    unfolding sfsem_union_additive[symmetric] S[symmetric]
    .
qed

lemma sfsem_remove_nondom_set:
  assumes
    b: "\<forall> (tr, t, X) \<in> S2 . \<not> ( NCPC t \<in> dom c \<and> N t)"
  shows
    "\<lbrakk>c\<rbrakk>(S1 \<union> S2) = ((\<lbrakk>c\<rbrakk>S1) \<union> S2)"
  using sfsem_union_additive[of c S1 S2] pc_does_not_point_to[of S2 c] b by simp

paragraph {* Monotonicity *}

lemma sfsem_monotone:
  assumes "S' \<subseteq> S"
  shows "\<lbrakk>c\<rbrakk> S' \<subseteq> \<lbrakk>c\<rbrakk> S"
by (metis assms sfsem_union_additive sup.orderE sup.orderI)

lemma sfsem_monotone_single:
  assumes "x' \<in> S" and "x \<in> \<lbrakk>c\<rbrakk>{x'}"
  shows "x \<in> \<lbrakk>c\<rbrakk>S"
using sfsem_monotone[of "{x'}" S c] assms(1) assms(2) by blast

paragraph {* Predecessor knowledge *}

text {*
  There has to exist a predecessor in S, from which we where created.
  Also, this can be (tr, s, X) itself, or something different, but then
  the predecessor has to point into c, thus has to be executed.
  NOTE: both postconditions can be true, if c is a single selfbranch.
*}

lemma sfsem_predecessor:
  assumes
    a: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
  shows "\<exists> tr' s' X' . (tr', s', X') \<in> S \<and> (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')} \<and> 
    ((tr = tr' \<and> s = s' \<and> X = X') \<or> (NCPC s' \<in> dom c \<and> N s'))"
proof -
  have goal: "\<exists> tr' s' X' . (tr', s', X') \<in> S \<and> (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}"
  proof (rule ccontr)
    (* we can create a contradiction with the sfsem_big_union_additive *)
    assume aa1: "\<not> (\<exists>tr' s' X'. (tr', s', X') \<in> S \<and> (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')})"
    from aa1 have a1: "\<forall> (tr', s', X') \<in> S. (tr, s, X) \<notin> \<lbrakk>c\<rbrakk>{(tr', s', X')}"
      by auto
    from a1 have "(tr, s, X) \<notin> (\<Union>x'\<in>S . \<lbrakk>c\<rbrakk>{x'})"
      by auto
    moreover
    have "(\<Union>x'\<in>S . \<lbrakk>c\<rbrakk>{x'}) = \<lbrakk>c\<rbrakk>(\<Union>x'\<in>S . {x'})"
      using sfsem_big_union_additive[of c _ S, symmetric]
      by presburger
    moreover
    have "(\<Union>x'\<in>S . {x'}) = S"
      by simp
    ultimately
    have "(tr, s, X) \<notin> \<lbrakk>c\<rbrakk>S"
      by simp
    thus "False"
      using a
      by auto
  qed
  then obtain tr' s' X' where o: "(tr', s', X') \<in> S \<and> (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}" by blast
  show ?thesis
    using pc_does_not_point_to[of "{(tr', s', X')}" c] goal a o by fastforce
qed

text {* A concrete instance of a predecessor. Saying: If they are not equal, then @{text "(tr', s', X')"}
has to be a normal state and in @{text c}'s domain, as it otherwise would stay. *}

lemma sfsem_predecessor2:
  assumes
    a: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}"
  shows
    "(tr = tr' \<and> s = s' \<and> X = X') \<or> (NCPC s' \<in> dom c \<and> N s')"
  using pc_does_not_point_to[of "{(tr', s', X')}" c] a
  by force

text {* If a state is in @{term S} and still exists after the application of a single instruction,
then we can apply the instruction directly onto the state: If the label equals the program counter,
then it has to be a selfbranch, as otherwise the state would have been cleared out in the first place.
If lbl does not equal the program counter, then it is not affected at all by the instruction.
*}
lemma sfsem_skip:
  assumes
    a: "(tr, s, X) \<in> S" and
    b: "(tr, s, X) \<in> \<lbrakk>(one lbl i)\<rbrakk>S"
  shows "(tr, s, X) \<in> \<lbrakk>(one lbl i)\<rbrakk>{(tr, s, X)}"
using assms NCCNotNormal
proof (induct i, auto) qed

text {* We cannot generate states from nothing. *}

lemma sfsem_empty: "sfsem code {} = {}"
  using sfsem_predecessor by fast

definition sfsem_fresh :: "('s, 'e) failure \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> ('s, 'e) failure set \<Rightarrow> bool" where
sfsem_fresh: "sfsem_fresh x c S = (\<exists> tr' s' X' . (tr', s', X') \<in> S \<and> NCPC s' \<in> dom c \<and> N s' \<and> x \<in> sfsem c {(tr', s', X')})"

lemma sfsem_freshI1:
  assumes
    sf: "(tr, s, X) \<in> sfsem c S" and not_in_S: "(tr, s, X) \<notin> S"
  shows "sfsem_fresh (tr, s, X) c S"
proof -
  obtain tr' s' X' where
    sf': "(tr, s, X) \<in> sfsem c {(tr', s', X')}" and
    inS: "(tr', s', X') \<in> S" and
    may_eq: "tr = tr' \<and> s = s' \<and> X = X' \<or> NCPC s' \<in> definitions.dom c \<and> N s'"
    using sfsem_predecessor[OF sf] by blast
  have "NCPC s' \<in> definitions.dom c \<and> N s'"
    using may_eq not_in_S inS by blast
  with sf' inS show ?thesis unfolding sfsem_fresh by blast
qed

text {* Selfbranches make some lemmas more complex (case distinctions, etc.) - with this lemma, we
can detect selfbranches easily. *}

lemma sfsem_not_outside_means_cbr:
  assumes
    ex: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S" and
    Nindom: "NCPC s \<in> dom c \<and> N s"
  shows
    "(\<exists> X' m n f l . (tr, s, X') \<in> S \<and> c = one l (cbr f m n) \<and> (f (NCRS s) \<longrightarrow> m = l) \<and> (\<not> f (NCRS s) \<longrightarrow> n = l) \<and> (\<forall> X \<subseteq> \<Sigma> . (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr, s, X')}) \<and> X \<subseteq> \<Sigma> \<and> l = NCPC s) \<and>
     sfsem_fresh (tr, s, X) c S
  "
  using assms
proof -
  have G: "\<exists> X' m n f l . (tr, s, X') \<in> S \<and> c = one l (cbr f m n) \<and> (f (NCRS s) \<longrightarrow> m = l) \<and> (\<not> f (NCRS s) \<longrightarrow> n = l) \<and> (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr, s, X')} \<and> X \<subseteq> \<Sigma> \<and> l = NCPC s"
    using assms
  proof (induct c arbitrary: S)
    case (one lbl i)
    show ?case
      using one NC_state_eq
    proof (induct i, auto simp add: NCCNotNormal) qed
  qed auto
  then obtain X' m n f l where
    c: "c = one l (cbr f m n)" and
    ex_c': "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr, s, X')}" and
    l: "l = NCPC s" and
    inS: "(tr, s, X') \<in> S"
    by blast
  hence "\<forall>X\<subseteq>\<Sigma>. (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr, s, X')}"
    using Nindom unfolding c l
    by simp
  thus ?thesis using G inS
    using Nindom sfsem_fresh by auto
qed

text {* A @{term communication} state is created by a @{term comm} instruction. *}

lemma sfsem_C_from_comm:
  assumes
    a: "(tr, s, X) \<in> sfsem (one l i) S" and
    b: "(tr, s, X) \<notin> S" and
    c: "\<not> N s"
  shows "\<exists> evfun sfun . i = comm evfun sfun \<and> (X \<subseteq> (\<Sigma> - evfun (NCRS s))) \<and> NCPC s = l"
using assms
proof (induct i, fastforce, simp, clarsimp simp add: NCPCh2 NCRSh2) qed

text {* A @{term communication} state has the same state content and we can exchange the refusal set
in compliance to the communication instruction. *}

lemma sfsem_notN_same_pred:
  assumes
    a: "(tr, s, X) \<in> \<lbrakk>one l (comm evf sf)\<rbrakk>{(trp, sp, Xp)}" and
    b: "\<not> N s" and
    c: "(tr, s, X) \<noteq> (trp, sp, Xp)" and
    d: "X' \<subseteq> \<Sigma> - evf (NCRS s)"
  shows "(tr, s, X') \<in> \<lbrakk>one l (comm evf sf)\<rbrakk>{(trp, sp, Xp)} \<and> tr = trp \<and> s = NCC sp \<and> X \<subseteq> \<Sigma> - evf (NCRS s)"
using assms proof (auto simp add: NCCSameR) qed

paragraph {* Code absorption - multiple code applications *}

text {* SF is idempotent. *}

lemma sfsem_idem: "\<lbrakk>c\<rbrakk>\<lbrakk>c\<rbrakk>S = \<lbrakk>c\<rbrakk>S"
proof (rule sym, induct c)
  fix
    l :: nat and i :: "('s, 'e) instr"
  show "\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>(one l i)\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S"
  proof (auto, induct i, auto simp add: NCCNotNormal)
    fix tr s X
    assume ex2: "(tr, s, X) \<in> \<lbrakk>(one l i)\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S"
    show "(tr, s, X) \<in> \<lbrakk>(one l i)\<rbrakk>S"
    proof (cases "NCPC s \<in> definitions.dom (one l i) \<and> N s")
      assume in_c: "NCPC s \<in> definitions.dom (one l i) \<and> N s"
      obtain X' X'' m n f where
        inliS: "(tr, s, X'') \<in> S" and
        pred2: "(tr, s, X') \<in> \<lbrakk>(one l i)\<rbrakk>{(tr, s, X'')}" and
        i: "i = cbr f m n" and
        X: "X \<subseteq> \<Sigma>"
        using sfsem_not_outside_means_cbr[OF ex2 in_c]
          sfsem_not_outside_means_cbr[OF _ in_c, of tr _ S]
        by fastforce
      show "(tr, s, X) \<in> \<lbrakk>(one l i)\<rbrakk>S" using pred2[unfolded i] inliS in_c X unfolding i by auto
    next
      assume not_in_c: "\<not> (NCPC s \<in> definitions.dom (one l i) \<and> N s)"
      obtain tr' s' X' where
        in'S: "(tr', s', X') \<in> \<lbrakk>(one l i)\<rbrakk>S" and
        li1: "(tr, s, X) \<in> \<lbrakk>one l i\<rbrakk>{(tr', s', X')}" and
        may_eq: "tr = tr' \<and> s = s' \<and> X = X' \<or> NCPC s' \<in> definitions.dom (one l i) \<and> N s'"
        using sfsem_predecessor[OF ex2] by blast
      have "tr = tr' \<and> s = s' \<and> X = X'"
        using not_in_c in'S li1
      proof (induct i, force, force, fastforce simp add: NCCNotNormal) qed
      with in'S show ?thesis by simp
    qed
  qed
next
  case (seq c1 c2)
  have "\<forall> (tr, t, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S . \<not> (NCPC t \<in> definitions.dom (c1 \<oplus> c2) \<and> N t)" by auto
  thus "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S" 
    using pc_does_not_point_to[of "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S" "(c1 \<oplus> c2)"]
    by blast
qed

text {*
  From a sequential composition, we can apply the left and right elements individually beforehand 
  without changing the result. In general, this can be generalized to any subset of code of the
  sequential composition, but this is far easier to show via @{term sfsem_exec}, thus here only
  the general case. Also, it is not required for the other proofs.
 *}
lemma sfsem_code_absorption_right_1:
  assumes wff: "wff\<^sub>s\<^sub>c (code1 \<oplus> code2)"
  shows "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code1\<rbrakk>S  = \<lbrakk>(code1 \<oplus> code2)\<rbrakk>S"
proof
  show "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code1\<rbrakk>S \<subseteq> \<lbrakk>(code1 \<oplus> code2)\<rbrakk>S"
    using UnCI sfseq_unfold2
    by fast
next
  obtain SI SO :: "('s, 'e) failure set" where
     S: "S = SI \<union> SO" and
     SI: "SI = {(tr, s, X) . (tr, s, X) \<in> S \<and> ((NCPC s \<in> dom code1) \<and> N s)}" and
     SO: "SO = {(tr, s, X) . (tr, s, X) \<in> S \<and> \<not> ((NCPC s \<in> dom code1) \<and> N s)}"
    by auto
  have a1: "\<forall>(tr, t, X)\<in>SO. \<not> (NCPC t \<in> definitions.dom code1 \<and> N t)"
    using SO 
    by fast
  have a2: "\<forall>(tr, t, X)\<in>SI. (NCPC t \<in> definitions.dom code1 \<and> N t)"
    using SI 
    by fast
  hence a3: "\<forall>(tr, t, X)\<in>SI. \<not> (NCPC t \<in> definitions.dom code2 \<and> N t)"
    using wff by auto
  have al1: "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>S = \<lbrakk>(code1 \<oplus> code2)\<rbrakk>SI \<union> \<lbrakk>(code1 \<oplus> code2)\<rbrakk>SO"
    using sfsem_union_additive[of "code1 \<oplus> code2" SI SO, unfolded S[symmetric]] .
  have al2: "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code1\<rbrakk>S = \<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code1\<rbrakk>SI \<union> \<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code1\<rbrakk>SO"
    using sfsem_union_additive[of "code1 \<oplus> code2" "\<lbrakk>code1\<rbrakk>SI" "\<lbrakk>code1\<rbrakk>SO", unfolded sfsem_union_additive[of code1 SI SO, unfolded S[symmetric], symmetric]]
    .
  have al3: "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>SO = \<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code1\<rbrakk>SO"
    using pc_does_not_point_to[of SO code1] SO
    by force
  have "extends (code1 \<oplus> code2) (ccpo_class.fixp (extends (code1 \<oplus> code2))) SI -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2) \<subseteq> \<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code1\<rbrakk>SI"
    unfolding sfseq
  proof (induct rule: fixp_induct_sf)
    show "ccpo.admissible Sup op \<le>
     (\<lambda>a. extends (code1 \<oplus> code2) a SI -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2)
          \<subseteq> a (\<lbrakk>code1\<rbrakk>SI) -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2))"
      unfolding "ccpo.admissible_def"
    proof (rule, rule, rule, rule)
      fix A ::"(('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set) set"
      assume
        aNotEmpty: "A \<noteq> {}" and
        ba2: "\<forall>a\<in>A. extends (code1 \<oplus> code2) a SI -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2) \<subseteq> a (\<lbrakk>code1\<rbrakk>SI) -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2)"
      have  b4:"((\<Union>f\<in>A. f (\<lbrakk>code1\<rbrakk>SI)) \<union> (\<Union>f\<in>A. f (\<lbrakk>code2\<rbrakk>SI)) \<union> SI) = (\<Union>f\<in>A.( f (\<lbrakk>code1\<rbrakk>SI) \<union> f (\<lbrakk>code2\<rbrakk>SI) \<union> SI))"
        using aNotEmpty
        by auto
      have "extends (code1 \<oplus> code2) (\<lambda>x. \<Union>f\<in>A. f x) SI -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2) \<subseteq> (\<Union>f\<in>A. f (\<lbrakk>code1\<rbrakk>SI) -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2))"
        unfolding extends b4
        using ba2
        by auto
      thus "extends (code1 \<oplus> code2) (Sup A) SI -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2)
         \<subseteq> Sup A (\<lbrakk>code1\<rbrakk>SI) -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2)"
        unfolding SClean_union[symmetric]
        by auto
    qed
  next
    show "extends (code1 \<oplus> code2) (Sup {}) SI -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2) \<subseteq> Sup {} (\<lbrakk>code1\<rbrakk>SI) -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2)"
      using a2
      by auto
  next
    fix x :: "('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set"
    assume ba1: "extends (code1 \<oplus> code2) x SI -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2) \<subseteq> x (\<lbrakk>code1\<rbrakk>SI) -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2)"
    from ba1 have b2: "(x (\<lbrakk>code1\<rbrakk>SI) \<union> x SI \<union> SI) -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2) \<subseteq> x (\<lbrakk>code1\<rbrakk>SI) -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2)"
      unfolding extends pc_does_not_point_to[OF a3]
      .
    show "extends (code1 \<oplus> code2) (extends (code1 \<oplus> code2) x) SI -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2)
            \<subseteq> extends (code1 \<oplus> code2) x (\<lbrakk>code1\<rbrakk>SI) -\<^sub>s\<^sub>f definitions.dom (code1 \<oplus> code2)"
      using b2
      unfolding extends sfsem_idem pc_does_not_point_to[OF a3]
      by auto
  qed
  thus "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>S \<subseteq> \<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code1\<rbrakk>S"
    unfolding sfsem_union_additive[of "code1 \<oplus> code2" SI SO,unfolded S[symmetric]]
      sfsem_union_additive[of "code1 \<oplus> code2" "\<lbrakk>code1\<rbrakk>SI" "\<lbrakk>code1\<rbrakk>SO"]
      sfsem_union_additive[of "code1" "SI" "SO",unfolded S[symmetric]]
      sfseq_unfold[symmetric]
    using al3
    by blast  
qed

lemma sfsem_code_absorption_right_2: 
  assumes "wff\<^sub>s\<^sub>c (code1 \<oplus> code2)"
  shows "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code2\<rbrakk>S  = \<lbrakk>(code1 \<oplus> code2)\<rbrakk>S"
proof -
  from assms have wff: "wff\<^sub>s\<^sub>c (code2 \<oplus> code1)"
    by auto
  from sfsem_code_absorption_right_1[OF wff, of S, unfolded oplus_commutative]
    show ?thesis .
qed

lemma sfsem_code_absorption_left_1: "\<lbrakk>code1\<rbrakk>\<lbrakk>(code1 \<oplus> code2)\<rbrakk>S = \<lbrakk>(code1 \<oplus> code2)\<rbrakk>S"
proof -
  have "\<forall> (tr, s, X) \<in> \<lbrakk>(code1 \<oplus> code2)\<rbrakk>S . \<not> (NCPC s \<in> definitions.dom (code1 \<oplus> code2) \<and> N s)"
    by auto
  hence "\<forall> (tr, s, X) \<in> \<lbrakk>(code1 \<oplus> code2)\<rbrakk>S . \<not> (NCPC s \<in> definitions.dom (code1) \<and> N s)"
    by auto
  thus ?thesis
    using pc_does_not_point_to 
    by blast
qed

lemma sfsem_code_absorption_left_2: "\<lbrakk>code2\<rbrakk>\<lbrakk>(code1 \<oplus> code2)\<rbrakk>S = \<lbrakk>(code1 \<oplus> code2)\<rbrakk>S"
proof -
  have "\<forall> (tr, s, X) \<in> \<lbrakk>(code1 \<oplus> code2)\<rbrakk>S . \<not> (NCPC s \<in> definitions.dom (code1 \<oplus> code2) \<and> N s)"
    by auto
  hence "\<forall> (tr, s, X) \<in> \<lbrakk>(code1 \<oplus> code2)\<rbrakk>S . \<not> (NCPC s \<in> definitions.dom (code2) \<and> N s)"
    by auto
  thus ?thesis
    using pc_does_not_point_to 
    by blast
qed

text {* This one will be necessary for sfsem_exec. *}
lemma sfsem_code_absorption_right_one:
  assumes a: "{(l, i)} \<subseteq> (\<union>\<^sub>s\<^sub>c c)" and b: "wff\<^sub>s\<^sub>c c"
  shows "\<lbrakk>c\<rbrakk>\<lbrakk>one l i\<rbrakk>S = \<lbrakk>c\<rbrakk>S"
  using assms
proof (induct c arbitrary: S, simp add: sfsem_idem)
  fix
    c1 c2 :: "('s, 'e) structured_code" and S :: "('s, 'e) failure set"
  assume
    aa1: "\<And>S. {(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c c1 \<Longrightarrow> wff\<^sub>s\<^sub>c c1 \<Longrightarrow> \<lbrakk>c1\<rbrakk>\<lbrakk>one l i\<rbrakk>S = \<lbrakk>c1\<rbrakk>S" and
    aa2: "\<And>S. {(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c c2 \<Longrightarrow> wff\<^sub>s\<^sub>c c2 \<Longrightarrow> \<lbrakk>c2\<rbrakk>\<lbrakk>one l i\<rbrakk>S = \<lbrakk>c2\<rbrakk>S" and
    aa3: "{(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)" and
    aa4: "wff\<^sub>s\<^sub>c (c1 \<oplus> c2)"
  from aa4 have a2: "wff\<^sub>s\<^sub>c c1 \<and> wff\<^sub>s\<^sub>c c2"
    by auto
  consider (inc1) "{(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c c1" | (inc2) "{(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c c2" using aa3 by fastforce
  thus "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>one l i\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
  proof cases
    case inc1
    have b1:  "\<And>S . \<lbrakk>c1\<rbrakk>\<lbrakk>one l i\<rbrakk>S = \<lbrakk>c1\<rbrakk>S"
      using aa1[OF inc1 a2[THEN conjunct1]] .
    show "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>one l i\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
      using sfsem_code_absorption_right_1[OF aa4] b1
      by metis
  next
    case inc2
    have b1:  "\<And>S . \<lbrakk>c2\<rbrakk>\<lbrakk>one l i\<rbrakk>S = \<lbrakk>c2\<rbrakk>S"
      using aa2[OF inc2 a2[THEN conjunct2]] .
    show "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>one l i\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
      using sfsem_code_absorption_right_2[OF aa4] b1
      by metis
  qed
qed

lemma sfsem_code_absorption_left_one: 
  assumes "\<union>\<^sub>s\<^sub>c (one l i) \<subseteq> \<union>\<^sub>s\<^sub>c c"
  shows "\<lbrakk>(one l i)\<rbrakk>\<lbrakk>c\<rbrakk>S = \<lbrakk>c\<rbrakk>S"
  using assms
proof (induct c)
  fix
    l' :: label and i' :: "('s, 'e) instr"
  assume
    aa1: "\<union>\<^sub>s\<^sub>c (one l i) \<subseteq> \<union>\<^sub>s\<^sub>c (one l' i')"
  from aa1 have a1: "(one l i) = (one l' i')"
    by auto
  show "\<lbrakk>(one l i)\<rbrakk>\<lbrakk>(one l' i')\<rbrakk>S = \<lbrakk>(one l' i')\<rbrakk>S"
    unfolding a1
    using sfsem_idem[of "one l' i'" S]
    .
next
  fix c1 c2 :: "('s, 'e) structured_code"
  show "\<union>\<^sub>s\<^sub>c (one l i) \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<Longrightarrow>
       \<lbrakk>one l i\<rbrakk>\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
    using pc_does_not_point_to[of "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S" "one l i"] indom_sc_subset_imp_indom_dom_subset[of "one l i" "seq c1 c2"]
    by fastforce
qed

paragraph {* Predecessor refusal set manipulation *}

text {* If we have a predecessor being normal and in the code domain, then its refusal set is irrelevant.
This means, we can exchange it with an arbitrary refusal set as we like.
*}

lemma sfsem_sigma_irrelevant_single:
  assumes
    a: "N s' \<and> NCPC s' \<in> dom c" and
    wff: "wff\<^sub>s\<^sub>c c" (* in general not necessary, but makes reasoning easier *) and
    in_c: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}"
  shows
    "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X'')}"
  using assms
proof (induct c arbitrary: tr s X)
  fix x1 x2 tr s X tr' s' X' X''
  assume
    inside: "N s' \<and> NCPC s' \<in> definitions.dom (one x1 x2)" and
    exec: "(tr, s, X) \<in> \<lbrakk>(one x1 x2)\<rbrakk>{(tr', s', X')}"
  thus "(tr, s, X) \<in> \<lbrakk>(one x1 x2)\<rbrakk>{(tr', s', X'')}"
  proof (induct x2, auto) qed
next
  fix c1 c2 tr s X
  assume
    IH1: "\<And>tr s X. N s' \<and> NCPC s' \<in> definitions.dom c1 \<Longrightarrow> wff\<^sub>s\<^sub>c c1 \<Longrightarrow> (tr, s, X) \<in> \<lbrakk>c1\<rbrakk>{(tr', s', X')} \<Longrightarrow> (tr, s, X) \<in> \<lbrakk>c1\<rbrakk>{(tr', s', X'')}" and
    IH2: "\<And>tr s X. N s' \<and> NCPC s' \<in> definitions.dom c2 \<Longrightarrow> wff\<^sub>s\<^sub>c c2 \<Longrightarrow> (tr, s, X) \<in> \<lbrakk>c2\<rbrakk>{(tr', s', X')} \<Longrightarrow> (tr, s, X) \<in> \<lbrakk>c2\<rbrakk>{(tr', s', X'')}" and
    in_c12: "N s' \<and> NCPC s' \<in> definitions.dom (c1 \<oplus> c2)" and
    wff12: "wff\<^sub>s\<^sub>c (c1 \<oplus> c2)" and
    exec_12: "(tr, s, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>{(tr', s', X')}"
  have wff1: "wff\<^sub>s\<^sub>c c1" using wff12 by simp
  have wff2: "wff\<^sub>s\<^sub>c c2" using wff12 by simp
  consider (in_c1) "N s' \<and> NCPC s' \<in> definitions.dom c1" | (in_c2) "N s' \<and> NCPC s' \<in> definitions.dom c2" using in_c12 wff12 by fastforce
  thus "(tr, s, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>{(tr', s', X'')}"
  proof cases
    case in_c1
    have ex: "(tr, s, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>c1\<rbrakk>{(tr', s', X')}"
      using exec_12
      unfolding sfsem_code_absorption_right_1[OF wff12] .
    then obtain tru su Xu where
      u_ex_c1: "(tru, su, Xu) \<in> \<lbrakk>c1\<rbrakk>{(tr', s', X')}" and
      ex_u: "(tr, s, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>{(tru, su, Xu)}"
      using sfsem_predecessor[OF ex] by blast
    have u_ex_c1': "(tru, su, Xu) \<in> \<lbrakk>c1\<rbrakk>{(tr', s', X'')}"
      using IH1[OF in_c1 wff1 u_ex_c1] by fastforce
    have "(tr, s, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>c1\<rbrakk>{(tr', s', X'')}"
      using sfsem_monotone_single[OF u_ex_c1' ex_u] .
    thus ?thesis 
      unfolding sfsem_code_absorption_right_1[OF wff12] .
  next
    case in_c2
    have ex: "(tr, s, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>c2\<rbrakk>{(tr', s', X')}"
      using exec_12
      unfolding sfsem_code_absorption_right_2[OF wff12] .
    then obtain tru su Xu where
      u_ex_c2: "(tru, su, Xu) \<in> \<lbrakk>c2\<rbrakk>{(tr', s', X')}" and
      ex_u: "(tr, s, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>{(tru, su, Xu)}"
      using sfsem_predecessor[OF ex] by blast
    have u_ex_c2': "(tru, su, Xu) \<in> \<lbrakk>c2\<rbrakk>{(tr', s', X'')}"
      using IH2[OF in_c2 wff2 u_ex_c2] by fastforce
    have "(tr, s, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>c2\<rbrakk>{(tr', s', X'')}"
      using sfsem_monotone_single[OF u_ex_c2' ex_u] .
    thus ?thesis 
      unfolding sfsem_code_absorption_right_2[OF wff12] .
  qed
qed

text {*
  For any given startset consisting only of normal states showing into the code (meaning, 
  that everything will be executed), we can ignore the refusal sets, ie we can just
  assign arbitrary refusal sets instead. This in general will be useful at a later point
  for proofing conformance with traces (SF1 - SF4).
 *}
lemma sfsem_sigma_irrelevant:
  assumes
    a: "\<forall> (tr, s, X) \<in> S . N s \<and> NCPC s \<in> dom c" and
    wff: "wff\<^sub>s\<^sub>c c" (* probably not necessary, but makes reasoning easier *) and
    S': "S' \<subseteq> {(tr, s, X') . \<exists> X . (tr, s, X) \<in> S}" and (* shrinking S can be done beforehand, extending afterwards. *)
    S'_all_contained: "\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> (\<exists> X' . (tr, s, X') \<in> S')"
  shows
    "\<lbrakk>c\<rbrakk>S = \<lbrakk>c\<rbrakk>S'"
proof auto
  fix tr s X
  assume exec_S: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
  obtain tr' s' X' where
    exec_': "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}" and
    in'S: "(tr', s', X') \<in> S"
    using sfsem_predecessor[OF exec_S] by fast
  obtain X'' where
    in''S': "(tr', s', X'') \<in> S'" using S' S'_all_contained in'S by blast
  have "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X'')}"
    using sfsem_sigma_irrelevant_single[of s' c, OF _ wff exec_', of X''] in'S a
    by fast
  thus "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S'" using in''S' sfsem_monotone_single by blast
next
  fix tr s X
  assume exec_S: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S'"
  obtain tr' s' X' where
    exec_': "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}" and
    in'S': "(tr', s', X') \<in> S'"
    using sfsem_predecessor[OF exec_S] by fast
  obtain X'' where
    in''S: "(tr', s', X'') \<in> S" using S' S'_all_contained in'S' by blast
  have "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X'')}"
    using sfsem_sigma_irrelevant_single[of s' c, OF _ wff exec_', of X''] in'S' a S'
    by fast
  thus "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S" using in''S sfsem_monotone_single by blast
qed

subsection {* SF instruction path execution *}
text {*
  The idea behind @{term sfsem_exec}: Just unroll the execution, which leads to a concrete successor.
  This means, create a list of "one" instructions, which will be executed one by one.
  We often just are interested in a single successor, so we do not need the full "parallel" execution 
  of everything via the iterates-fixpoint.
  This leads to a list structure, where the elements are labeled instructions, i.e. a tuple of type
  @{text "label x instr"}. They are then executed.
  To be fully useful, there will be the concept of density and criticality:
  * density: a list is dense wrt. starting set S and final state s, if each instruction in general
    is useful. So for each instruction in the list, there will be a successor state, and combining
    all those intermediate successor states shows how we come to the final state.
  * criticality: a list is critical wrt. starting set S and final state s, if it is dense and
    there are no shortcuts. Imagine an ifblock, which does not really do anything. This means,
    after the if block, the state could have been created due to taking the if, or without taking
    the if. Not taking the if is the shortcut, where we already created the state, which later on
    would be created again in executing the if branch. Thus, if the if branch would be part of the
    list, it is not critical. It will be clearer upon looking at the definitions.
  As density is the weaker concept, lemmas are tried to be shown for density, which implies they
  hold for criticality, too.
*}

text {*
   In the beginning is the introduction of the reversed execution, due to the orientation of lists.
   But as it is more intuitive to execute the instruction at the rightmost end, we then make an alias
   to correct the situation. Thus, first the introduction of @{term sfsem_exec_b}, and later on the corrected version.
*}

lemma exec_induct: "P [] \<Longrightarrow> (\<And>l i lil . P lil \<Longrightarrow> P ((l, i) # lil)) \<Longrightarrow> P lil'"
  by (metis list_nonempty_induct prod.collapse)

fun sfsem_exec_b :: "(label \<times> ('s, 'e) instr) list \<Rightarrow> ('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set" where
  "sfsem_exec_b [] S = S" |
  "sfsem_exec_b ((l, i) # lil) S = sfsem_exec_b lil (sfsem (one l i) S)"

text {*
  We can remove the last instruction and transform it into a separate sfsem call.
  It's just the opposite unroll as @{text "sfsem_exec_b.simps(1)"}.
 *}
lemma sfsem_exec_b_unfold:
  "sfsem_exec_b (lil @ [(l, i)]) S = (sfsem (one l i) (sfsem_exec_b (lil) S))"
proof (induct lil arbitrary: S l i)
  fix li :: "nat \<times> ('s, 'e) instr" and lil :: "(nat \<times> ('s, 'e) instr) list" and S :: "('s, 'e) failure set"
    and l :: label and i :: "('s, 'e) instr"
  assume aa1: "(\<And>S l i. sfsem_exec_b (lil @ [(l, i)]) S = \<lbrakk>(one l i)\<rbrakk>sfsem_exec_b lil S)"
  have a1: "sfsem_exec_b ((li # lil) @ [(l, i)]) S = sfsem_exec_b ((lil) @ [(l, i)]) (sfsem (one (fst li) (snd li)) S)"
    by (metis append_Cons sfsem_exec_b.simps(2) prod.collapse)
  have a2: "sfsem_exec_b ((lil) @ [(l, i)]) (sfsem (one (fst li) (snd li)) S) = \<lbrakk>(one l i)\<rbrakk>sfsem_exec_b lil (sfsem (one (fst li) (snd li)) S)"
    using aa1
    by auto
  have a3: "\<lbrakk>(one l i)\<rbrakk>sfsem_exec_b lil (sfsem (one (fst li) (snd li)) S) = \<lbrakk>(one l i)\<rbrakk>sfsem_exec_b (li # lil) S"
    by (metis sfsem_exec_b.simps(2) prod.exhaust_sel)
  show "sfsem_exec_b ((li # lil) @ [(l, i)]) S = \<lbrakk>(one l i)\<rbrakk>sfsem_exec_b (li # lil) S"
    using a1 a2 a3
    by simp
qed auto

text {* Unfolding of the execution for arbitrary lists. *}
lemma sfsem_exec_b_unfold2:
  "sfsem_exec_b (l @ r) S = (sfsem_exec_b (r) (sfsem_exec_b (l) S))"
proof (induct l arbitrary: S)
  show "\<And>S. sfsem_exec_b ([] @ r) S =
         sfsem_exec_b r (sfsem_exec_b [] S)"
    by auto
next
  fix a :: "nat \<times> ('s, 'e) instr" and l :: "(nat \<times> ('s, 'e) instr) list" and S :: "('s, 'e) failure set"
  assume aa1: "\<And>S. sfsem_exec_b (l @ r) S = sfsem_exec_b r (sfsem_exec_b l S)"
  obtain lbl :: label and i :: "('s, 'e) instr" where
    bo1: "a = (lbl, i)"
    by fastforce
  show "sfsem_exec_b ((a # l) @ r) S = sfsem_exec_b r (sfsem_exec_b (a # l) S)"
    using sfsem_exec_b.simps(2)[of lbl i l S] aa1[of "\<lbrakk>one lbl i\<rbrakk>S"] bo1
    by simp
qed

text {* Here we create the more intuitive way for @{term sfsem_exec}. *}
definition "sfsem_exec lil S \<equiv> sfsem_exec_b (rev lil) S"

lemma sfsem_exec_single: "(sfsem_exec [(l, i)] S) = (\<lbrakk>one l i\<rbrakk> S)"
  unfolding sfsem_exec_def "rev.simps" append_Nil sfsem_exec_b.simps ..

lemma sfsem_exec_empty [simp]: "sfsem_exec [] S = S"
  using sfsem_exec_def[of "[]" S]
  by auto

lemma sfsem_exec_unfold:
  "sfsem_exec ((l, i) # lil) S = (sfsem (one l i) (sfsem_exec (lil) S))"
using sfsem_exec_b_unfold sfsem_exec_def
  by (metis rev.simps(2))

lemma sfsem_exec_unfold2:
  "sfsem_exec (l @ r) S = (sfsem_exec (l) (sfsem_exec (r) S))"
by (simp add: sfsem_exec_b_unfold2 sfsem_exec_def)

lemma sfsem_exec_additive:
  "sfsem_exec lil (P \<union> Q) = sfsem_exec lil P \<union> sfsem_exec lil Q"
proof (induct lil)
  show "sfsem_exec [] (P \<union> Q) = sfsem_exec [] P \<union> sfsem_exec [] Q"
    unfolding sfsem_exec_empty ..
next
  fix li :: "nat \<times> ('s, 'e) instr" and lil :: "(nat \<times> ('s, 'e) instr) list"
  assume
    aa1: "sfsem_exec lil (P \<union> Q) = sfsem_exec lil P \<union> sfsem_exec lil Q"
  obtain l :: label and i :: "('s, 'e) instr" where
    ao1: "li = (l, i)"
    by fastforce
  show "sfsem_exec (li # lil) (P \<union> Q) = sfsem_exec (li # lil) P \<union> sfsem_exec (li # lil) Q"
    unfolding ao1 sfsem_exec_unfold sfsem_union_additive[of "one l i" "sfsem_exec lil P" "sfsem_exec lil Q", symmetric] aa1
    ..
qed

lemma sfsem_exec_pred1: 
  assumes "(tr, s, X) \<in> sfsem_exec ((l, i) # lil) S"
  shows "\<exists> tr' s' X' . (tr', s', X') \<in> sfsem_exec lil S \<and> (tr, s, X) \<in> \<lbrakk>one l i\<rbrakk>{(tr', s', X')} \<and> (tr = tr' \<and> s = s' \<and> X = X' \<or> 
  (NCPC s' \<in> definitions.dom (one l i) \<and> N s' \<and> 
    (N s \<longrightarrow> X \<subseteq> \<Sigma>) \<and>
    (\<not> N s \<longrightarrow> (\<exists> ef sf . i = comm ef sf \<and> X \<subseteq> \<Sigma> - ef (NCRS s) \<and> s = NCC s' \<and> tr = tr'))
  ))"
proof -
  have "\<exists> tr' s' X' . (tr', s', X') \<in> sfsem_exec lil S \<and> (tr, s, X) \<in> \<lbrakk>one l i\<rbrakk>{(tr', s', X')} \<and> (tr = tr' \<and> s = s' \<and> X = X' \<or> NCPC s' \<in> definitions.dom (one l i) \<and> N s')"
    using sfsem_exec_unfold[of l i lil S] assms sfsem_predecessor[of tr s X "one l i" "sfsem_exec lil S"]
    by blast
  then obtain tr' s' X' where
    exec'lil: "(tr', s', X') \<in> sfsem_exec lil S" and
    execli': "(tr, s, X) \<in> \<lbrakk>(one l i)\<rbrakk>{(tr', s', X')}" and
    maybe_eq: "(tr = tr' \<and> s = s' \<and> X = X' \<or> NCPC s' \<in> definitions.dom (one l i) \<and> N s')"
    by blast
  moreover have "(tr, s, X) \<noteq> (tr', s', X') \<longrightarrow> ((N s \<longrightarrow> X \<subseteq> \<Sigma>) \<and>
    (\<not> N s \<longrightarrow> (\<exists> ef sf . i = comm ef sf \<and> X \<subseteq> \<Sigma> - ef (NCRS s) \<and> s = NCC s' \<and> tr = tr')))"
    using execli'
  proof (induct i, auto simp add: NCCSameR) qed
  ultimately show ?thesis by blast
qed

lemma sfsem_exec_pred2:
  assumes "(tr, s, X) \<in> sfsem_exec (lil) S"
  shows "\<exists> tr' s' X' . (tr, s, X) \<in> sfsem_exec (lil) {(tr', s', X')} \<and> (tr', s', X') \<in> S"
using assms
proof(induct lil arbitrary: tr s X)
  show "\<And>tr s X. (tr, s, X) \<in> sfsem_exec [] S \<Longrightarrow> \<exists>tr' s' X'. (tr, s, X) \<in> sfsem_exec [] {(tr', s', X')} \<and> (tr', s', X') \<in> S"
    unfolding sfsem_exec_empty
    by blast
next
  fix li :: "nat \<times> ('s, 'e) instr" and lil :: "(nat \<times> ('s, 'e) instr) list" and
    tr :: "'e event list" and s :: "'s NCstate" and X :: "'e event set"
  assume
    aa1: "\<And>tr s X. (tr, s, X) \<in> sfsem_exec lil S \<Longrightarrow>
                  \<exists>tr' s' X'. (tr, s, X) \<in> sfsem_exec lil {(tr', s', X')} \<and> (tr', s', X') \<in> S"
  obtain l i where
    ao1: "li = (l, i)"
    by fastforce
  show "(tr, s, X) \<in> sfsem_exec (li # lil) S \<Longrightarrow> \<exists>tr' s' X'. (tr, s, X) \<in> sfsem_exec (li # lil) {(tr', s', X')} \<and> (tr', s', X') \<in> S"
    unfolding ao1 sfsem_exec_unfold
    using sfsem_predecessor[of tr s X "one l i" "sfsem_exec lil S"] aa1
      using sfsem_monotone_single[of _ _ "(tr, s, X)" "one l i"]
    by meson
qed

paragraph {* Transformation lemmas *}
text {*
  The transformation lemmas show, that we can transform @{term sfsem} and @{term sfsem_exec} into each other
  (if some minor conditions hold).
*}

text {*
  The first important transformation lemma. If a state is in the normal execution of code, then
  there exists a path of execution (being part of code), which also creates that state.
*}
theorem sfsem_exec1: 
  assumes
    a: "(tr, s, X) \<in> sfsem c S"
  shows "\<exists> lil . set lil \<subseteq> (\<union>\<^sub>s\<^sub>c c) \<and> (tr, s, X) \<in> sfsem_exec lil S"
using assms
proof(induct c arbitrary: S tr s X)
  show "\<And>x1 x2 S tr s X.
       (tr, s, X) \<in> \<lbrakk>(one x1 x2)\<rbrakk>S \<Longrightarrow>
       \<exists>lil.
          set lil \<subseteq> \<union>\<^sub>s\<^sub>c (one x1 x2) \<and>
          (tr, s, X) \<in> sfsem_exec lil S"
    by (metis sfsem_exec_empty append_Nil empty_set inf_sup_ord(4) list.simps(15) sc_inj.one set_append sfsem_exec_unfold)
next
  fix c1 c2 :: "('s, 'e) structured_code" and S ::"('s, 'e) failure set" and
      tr :: "'e event list" and s :: "'s NCstate" and X :: "'e event set"
  assume
    aa1: "\<And> S tr s X . (tr, s, X) \<in> \<lbrakk>c1\<rbrakk>S \<Longrightarrow>
        \<exists>lil.
           set lil \<subseteq> \<union>\<^sub>s\<^sub>c c1 \<and>
           (tr, s, X) \<in> sfsem_exec lil S" and
    aa2: "\<And> S tr s X . (tr, s, X) \<in> \<lbrakk>c2\<rbrakk>S \<Longrightarrow>
        \<exists>lil.
           set lil \<subseteq> \<union>\<^sub>s\<^sub>c c2 \<and>
           (tr, s, X) \<in> sfsem_exec lil S" and
    aa3: "(tr, s, X) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
  show "\<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec lil S"
    using aa3
    unfolding sfseq
  proof (induct arbitrary: S rule: fixp_induct_sf)
    show "ccpo.admissible Sup op \<le>
     (\<lambda>a. \<forall>x. (tr, s, X) \<in> a x -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2) \<longrightarrow>
              (\<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec lil x))"
      unfolding "ccpo.admissible_def" chain_def
      by auto
  next
    fix x :: "('s, 'e) failure set \<Rightarrow> ('s, 'e) failure set" and S :: "('s, 'e) failure set"
    assume
      ba1: "\<And>S. (tr, s, X) \<in> x S -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2) \<Longrightarrow>
             \<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec lil S" and
      ba2: "(tr, s, X) \<in> extends (c1 \<oplus> c2) x S -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)"
    from ba2 have "(tr, s, X) \<in> (x (\<lbrakk>c1\<rbrakk>S) \<union> x (\<lbrakk>c2\<rbrakk>S) \<union> S) -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)"
      unfolding extends .
    moreover
    {
      assume ca1: "(tr, s, X) \<in> x (\<lbrakk>c1\<rbrakk>S) -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)"
      obtain lil :: "(nat \<times> ('s, 'e) instr) list" where
        co1: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec lil (\<lbrakk>c1\<rbrakk>S)"
        using ba1[of "\<lbrakk>c1\<rbrakk>S"] ca1
        by blast
      obtain tr' s' X' where
        co2: "(tr, s, X) \<in> sfsem_exec lil ({(tr', s', X')}) \<and> (tr', s', X') \<in> \<lbrakk>c1\<rbrakk>S"
        using sfsem_exec_pred2[of tr s X lil "\<lbrakk>c1\<rbrakk>S"] co1
        by fast
      obtain lil' where
        co3: "set lil' \<subseteq> \<union>\<^sub>s\<^sub>c c1 \<and> (tr', s', X') \<in> sfsem_exec lil' S"
        using co2 aa1[of tr' s' X' S]
        by blast
      have c1: "{(tr', s', X')} \<union> sfsem_exec lil' S = sfsem_exec lil' S"
        using co3 
        by blast
      have "(tr, s, X) \<in> sfsem_exec lil (sfsem_exec lil' S \<union> {(tr', s', X')})"
        using co2 co3 sfsem_exec_additive[of lil "sfsem_exec lil' S" "{(tr', s', X')}"]
        by blast
      hence "(tr, s, X) \<in> sfsem_exec lil (sfsem_exec lil' S)"
        using c1
        by simp
      hence c2: "(tr, s, X) \<in> sfsem_exec (lil @ lil') S"
        using sfsem_exec_unfold2[of lil lil' S]
        by blast
      have "set (lil @ lil') \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)"
        using co1 co3
        by auto
      with c2 have "\<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec lil S"
        by blast
    }
    moreover
    {
      assume ca1: "(tr, s, X) \<in> x (\<lbrakk>c2\<rbrakk>S) -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)"
      obtain lil where
        co1: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec lil (\<lbrakk>c2\<rbrakk>S)"
        using ba1[of "\<lbrakk>c2\<rbrakk>S"] ca1
        by blast
      obtain tr' s' X' where
        co2: "(tr, s, X) \<in> sfsem_exec lil ({(tr', s', X')}) \<and> (tr', s', X') \<in> \<lbrakk>c2\<rbrakk>S"
        using sfsem_exec_pred2[of tr s X lil "\<lbrakk>c2\<rbrakk>S"] co1
        by fast
      obtain lil' where
        co3: "set lil' \<subseteq> \<union>\<^sub>s\<^sub>c c2 \<and> (tr', s', X') \<in> sfsem_exec lil' S"
        using co2 aa2[of tr' s' X' S]
        by blast
      have c1: "{(tr', s', X')} \<union> sfsem_exec lil' S = sfsem_exec lil' S"
        using co3 
        by blast
      have "(tr, s, X) \<in> sfsem_exec lil (sfsem_exec lil' S \<union> {(tr', s', X')})"
        using co2 co3 sfsem_exec_additive[of lil "sfsem_exec lil' S" "{(tr', s', X')}"]
        by blast
      hence "(tr, s, X) \<in> sfsem_exec lil (sfsem_exec lil' S)"
        using c1
        by simp
      hence c2: "(tr, s, X) \<in> sfsem_exec (lil @ lil') S"
        using sfsem_exec_unfold2[of lil lil' S]
        by blast
      have "set (lil @ lil') \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)"
        using co1 co3
        by auto
      with c2 have "\<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec lil S"
        by blast
    }
    moreover
    {
      assume "(tr, s, X) \<in> S -\<^sub>s\<^sub>f definitions.dom (c1 \<oplus> c2)"
      hence "set [] \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec [] S"
        unfolding sfsem_exec_def
        by auto
      hence "\<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec lil S"
        by blast
    }
    ultimately
    show "\<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s, X) \<in> sfsem_exec lil S"
      by auto
  qed simp
qed

text {* A version where the triple is not expanded. *}
lemma sfsem_exec1': 
  assumes
    a: "x \<in> sfsem c S"
  shows "\<exists> lil . set lil \<subseteq> (\<union>\<^sub>s\<^sub>c c) \<and> x \<in> sfsem_exec lil S"
proof -
  obtain tr s X where
    ao1: "x = (tr, s, X)"
    using prod_cases3 by blast
  show ?thesis
    unfolding ao1
    using sfsem_exec1[OF a[unfolded ao1]]
    .
qed

text {* The reverse part: If we have a list of instructions and the final state is not in the code's
domain, then we can also get it via a normal sfsem application. *}

theorem sfsem_exec2: 
  assumes 
    lilSubC: "set (lil) \<subseteq> (\<union>\<^sub>s\<^sub>c c)" and
    exec_lil: "(tr, s, X) \<in> sfsem_exec lil S" and
    not_pcN: "\<not> (NCPC s \<in> dom c \<and> N s)" and
    wff: "wff\<^sub>s\<^sub>c c"
  shows "(tr, s, X) \<in> sfsem c S"
proof (cases "(tr, s, X) \<in> S")
  case True
  thus ?thesis
    using not_pcN
    by (simp add: pc_does_not_point_to_single)
next
  case False
  hence "lil \<noteq> []" using exec_lil by fastforce
  then obtain l i right where lil: "lil = (l, i) # right" by (meson sfsem_exec_b.elims)
  then obtain right l i tr' s' X' where
    lil: "lil = [(l, i)] @ right" and
    step: "(tr, s, X) \<in> \<lbrakk>(one l i)\<rbrakk>{(tr', s', X')}" and
    execRight: "(tr', s', X') \<in> sfsem_exec right S"
    by (metis append_Cons append_Nil exec_lil sfsem_exec_unfold sfsem_predecessor)
  have a2: "{(l, i)} \<subseteq> set lil"
    using lil
    by auto
  have "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<lbrakk>(one l i)\<rbrakk>{(tr', s', X')}"
    using pc_does_not_point_to_single [of s "c" tr X "\<lbrakk>(one l i)\<rbrakk>{(tr', s', X')}"] step not_pcN
    by auto
  hence a4: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}"
    using sfsem_code_absorption_right_one[of l i c "{(tr', s', X')}"] a2 wff lil lilSubC
    by blast
  have a5: "set right \<subseteq> (\<union>\<^sub>s\<^sub>c c)"
    using lil lilSubC
    by auto
  show "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
    using a5 a4 execRight
  proof (induct right arbitrary: tr' s' X')
    show "\<And>tr' s' X'.
       set [] \<subseteq> \<union>\<^sub>s\<^sub>c c \<Longrightarrow>
       (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')} \<Longrightarrow>
       (tr', s', X') \<in> sfsem_exec [] S \<Longrightarrow>
       (tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
      unfolding sfsem_exec_empty
      using sfsem_monotone_single by blast
  next
    fix li' right tr' s' X'
    assume
      aa1: "\<And>tr' s' X'.
           set right \<subseteq> \<union>\<^sub>s\<^sub>c c \<Longrightarrow>
           (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')} \<Longrightarrow>
           (tr', s', X') \<in> sfsem_exec right S \<Longrightarrow>
           (tr, s, X) \<in> \<lbrakk>c\<rbrakk>S" and
      aa2: "set (li' # right) \<subseteq> \<union>\<^sub>s\<^sub>c c" and
      aa3: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}" and
      aa4: "(tr', s', X') \<in> sfsem_exec (li' # right) S"
    obtain l' i' where
      ao1: "li' = (l', i')" by fastforce
    from aa4 have a1: "(tr', s', X') \<in> \<lbrakk>(one l' i')\<rbrakk>sfsem_exec right S"
      using sfsem_exec_unfold[of l' i' right S] ao1
      by blast
    from aa2 have a2: "set (right) \<subseteq> \<union>\<^sub>s\<^sub>c c \<and> {li'} \<subseteq> \<union>\<^sub>s\<^sub>c c"
      by auto
    obtain tr'' s'' X'' where
      ao2: "(tr'', s'', X'') \<in> sfsem_exec right S \<and> (tr', s', X') \<in> \<lbrakk>(one l' i')\<rbrakk>{(tr'', s'', X'')}"
      using a1 sfsem_predecessor[of tr' s' X' "one l' i'" "sfsem_exec right S"] by blast
    from ao2 have "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>\<lbrakk>(one l' i')\<rbrakk>{(tr'', s'', X'')}"
      by (metis (no_types, lifting) Set.set_insert UnI1 aa3 insert_is_Un sfsem_union_additive)
    hence "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr'', s'', X'')}"
      using aa2 ao1
      by (simp add: a2 wff sfsem_code_absorption_right_one)
    thus "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
      using a2 aa1 ao2 by auto
  qed
qed

text {* A version where the triple is not expanded. *}
lemma sfsem_exec2': 
  assumes 
    a: "set (lil) \<subseteq> (\<union>\<^sub>s\<^sub>c c)" and
    b: "x \<in> sfsem_exec lil S" and
    c: "\<not> (NCPC (fst (snd x)) \<in> dom c \<and> N (fst (snd x)))" and
    d: "wff\<^sub>s\<^sub>c c"
  shows "x \<in> sfsem c S"
proof -
  obtain tr s X where
    ao1: "x = (tr, s, X)"
    using prod_cases3 by blast
  have a1: "fst (snd (tr, s, X)) = s"
    by auto
  show ?thesis
    using assms
    unfolding ao1 a1
    using sfsem_exec2[of lil c tr s X S]
    by blast
qed

paragraph {* Critical and Dense paths *}

text{* A critical path is a path, where each single instruction on the path is crucial to reach to the final state. This means,
   we could not leave out a slice of the path. *}
fun sfsem_exec_path_is_critical :: "('s, 'e) failure \<Rightarrow> (label \<times> ('s, 'e) instr) list \<Rightarrow> ('s, 'e) failure set \<Rightarrow> bool" where
  "sfsem_exec_path_is_critical x [] S = (x \<in> S)" |
  "sfsem_exec_path_is_critical x ((l, i) # lil) S = (x \<notin> sfsem_exec lil S \<and> (\<exists> x' . x \<noteq> x' \<and> x \<in> \<lbrakk>one l i\<rbrakk>{x'} \<and> sfsem_exec_path_is_critical x' lil S))"

text {* A dense path is a path, where we don't care for shortcuts in the execution.
   So, we just ask, if we now could be created. It does not matter, if we already exist (except for the case of a self jump).
   Imagine a loop, which can be exited. Due to nondeterminism, the same state (which is outside of the loop) could be created in multiple iterations.
   The critical path property would not allow a later creation, as the escaped state already exists due to a previous loop iteration.
   Here, we just ask, if it could be created via this path, no matter if a suffix of the execution path already created that state. *}
fun sfsem_exec_path_is_dense :: "('s, 'e) failure \<Rightarrow> (label \<times> ('s, 'e) instr) list \<Rightarrow> ('s, 'e) failure set \<Rightarrow> bool" where
  "sfsem_exec_path_is_dense x [] S = (x \<in> S)" |
  "sfsem_exec_path_is_dense x ((l, i) # lil) S = (x \<notin> {x . x \<in> sfsem_exec lil S \<and> NCPC (fst (snd x)) = l} \<and> (\<exists> x' . x \<noteq> x' \<and> x \<in> \<lbrakk>one l i\<rbrakk>{x'} \<and> sfsem_exec_path_is_dense x' lil S))"

lemma sfsem_exec_critical_implies_dense:
  assumes critical: "sfsem_exec_path_is_critical x lil S"
  shows "sfsem_exec_path_is_dense x lil S"
using assms
proof (induct lil arbitrary: x, simp)
  fix a lil x
  assume
    aa1: "\<And>x. sfsem_exec_path_is_critical x lil S \<Longrightarrow> sfsem_exec_path_is_dense x lil S" and
    aa2: "sfsem_exec_path_is_critical x (a # lil) S"
  obtain l i where a: "a = (l, i)" by fastforce
  show "sfsem_exec_path_is_dense x (a # lil) S"
    using aa1 aa2
    unfolding a sfsem_exec_path_is_critical.simps sfsem_exec_path_is_dense.simps
    by blast
qed

lemma list_append_singleton:
  "[x] @ xs = x # xs"
  unfolding append_Cons
  unfolding append_Nil
by simp

text {* Each path can be transformed into a (shorter) critical path, and thus also into a dense path. Effectively, we just
  remove the unnecessary instructions.
*}

lemma sfsem_exec_criticalPath:
  assumes
    a: "x \<in> sfsem_exec lil S"
  shows "\<exists> lil' . set lil' \<subseteq> set lil \<and> length lil' \<le> length lil \<and> sfsem_exec_path_is_critical x lil' S"
  using a
proof (induct lil arbitrary: x rule: length_induct)
  fix lil :: "(nat \<times> ('s, 'e) instr) list" and
    x :: "('s, 'e) failure"
  assume
    IH: "\<forall>lil'. length lil' < length lil \<longrightarrow>
            (\<forall>x. x \<in> sfsem_exec lil' S \<longrightarrow>
                 (\<exists>lil''.
                     set lil'' \<subseteq> set lil' \<and>
                     length lil'' \<le> length lil' \<and>
                     sfsem_exec_path_is_critical x
                      lil'' S))" and
    step: "x \<in> sfsem_exec lil S"
  show "\<exists>lil'.
          set lil' \<subseteq> set lil \<and>
          length lil' \<le> length lil \<and>
          sfsem_exec_path_is_critical x lil' S" (is ?A)
  proof (cases "lil = []")
    assume ba1: "lil = []"
    thus ?thesis
      using step[unfolded ba1, unfolded sfsem_exec_empty]
      by simp
  next
    assume ba1: "lil \<noteq> []"
    obtain l i right where
      lil: "lil = [(l, i)] @ right"
      using ba1
      by (metis append_Cons append_Nil neq_Nil_conv prod.collapse)
    obtain tr s X where x: "x = (tr, s, X)" using prod_cases3 by blast
    obtain x' tr' s' X' where
      x': "x' = (tr', s', X')" and
      exec_step: "x \<in> \<lbrakk>one l i\<rbrakk>{x'}" and
      exec_right: "x' \<in> sfsem_exec right S"
      using sfsem_predecessor[of tr s X "one l i" "sfsem_exec right S", OF
          step[unfolded lil, unfolded list_append_singleton sfsem_exec_unfold x]
        ] x
      by blast
    have b1: "length right < length lil \<and> set right \<subseteq> set lil"
      using lil
      by auto
    obtain rightC where
      subs: "set rightC \<subseteq> set right" and
      len: "length rightC \<le> length right" and
      exec_rightC: "sfsem_exec_path_is_critical x' rightC S"
      using IH b1 exec_right
      by blast
    consider (ex_right) "x \<in> sfsem_exec right S" | (ex_rightC) "x \<in> sfsem_exec rightC S" | (no_ex) "x \<notin> sfsem_exec right S \<and> x \<notin> sfsem_exec rightC S"
      by blast
    thus ?thesis
    proof cases
      case ex_right
      thus ?thesis
        by (meson IH b1 dual_order.strict_implies_order order_trans)
    next
      case ex_rightC
      thus ?thesis
        by (meson IH b1 subs len dual_order.strict_trans2 order.strict_implies_order order_trans)
    next
      case no_ex
      hence d1: "x \<noteq> x'"
        using exec_right
        by blast
      have "(x \<noteq> x' \<and> x \<in> \<lbrakk>one l i\<rbrakk>{x'} \<and> sfsem_exec_path_is_critical x' rightC S)"
        using exec_rightC d1 exec_step
        by simp
      hence "sfsem_exec_path_is_critical x ((l, i) # rightC) S"
        using no_ex sfsem_exec_path_is_critical.simps(2)[symmetric, of x rightC S l i]
        by blast
      moreover
      have "set ((l, i) # rightC) \<subseteq> set lil \<and> length ((l, i) # rightC) \<le> length lil"
        using lil subs len
        by auto
      ultimately
      show ?thesis
        by blast
    qed
  qed
qed

lemma sfsem_exec_critical_path_unfold_single:
  assumes
    "sfsem_exec_path_is_critical x (li # lil) S"
  shows
    "\<exists> x' . sfsem_exec_path_is_critical x [li] {x'} \<and> sfsem_exec_path_is_critical x' lil S"
proof -
  obtain l i where
    ao1: "li = (l, i)"
    by fastforce
  from assms show ?thesis
    unfolding sfsem_exec_path_is_critical.simps ao1
    by simp
qed

lemma sfsem_exec_is_dense_is_exec:
  assumes "sfsem_exec_path_is_dense x lil S"
  shows "x \<in> sfsem_exec lil S"
using assms
proof (induct lil arbitrary: x)
  show "\<And>x. sfsem_exec_path_is_dense x [] S \<Longrightarrow>
         x \<in> sfsem_exec [] S"
    unfolding sfsem_exec_empty sfsem_exec_path_is_dense.simps .
next
  fix a :: "label \<times> ('s, 'e) instr" and lil :: "(label \<times> ('s, 'e) instr) list" and x :: "('s, 'e) failure"
  assume
    aa1: "\<And>x. sfsem_exec_path_is_dense x lil S \<Longrightarrow>
             x \<in> sfsem_exec lil S" and
    aa2: "sfsem_exec_path_is_dense x (a # lil) S"
  obtain l i where
    ao1: "a = (l, i)"
    by fastforce
  obtain x' where
    ao2: "sfsem_exec_path_is_dense x [a] {x'} \<and>
       sfsem_exec_path_is_dense x' lil S"
    using aa2
    unfolding ao1 sfsem_exec_path_is_dense.simps sfsem_exec_empty
    by blast
  have a1: "x' \<in> sfsem_exec lil S"
    using aa1[OF ao2[THEN conjunct2]]
    .
  have a2: "x \<in> \<lbrakk>(one l i)\<rbrakk>{x'}"
    using ao2[THEN conjunct1, unfolded ao1 sfsem_exec_path_is_dense.simps]
    by blast
  have a3: "x \<in> \<lbrakk>(one l i)\<rbrakk>sfsem_exec lil S"
    using sfsem_monotone_single[OF a1 a2] .
  show "x \<in> sfsem_exec (a # lil) S"
    using a3[unfolded sfsem_exec_unfold[symmetric] ao1[symmetric]]
    .
qed

lemma sfsem_exec_is_critical_is_exec:
  assumes "sfsem_exec_path_is_critical x lil S"
  shows "x \<in> sfsem_exec lil S"
using sfsem_exec_is_dense_is_exec[OF sfsem_exec_critical_implies_dense[OF assms]]
.

lemma sfsem_exec_dense_path_unfold:
  assumes
    "sfsem_exec_path_is_dense x (l @ r) S"
  shows
    "\<exists> x' . sfsem_exec_path_is_dense x l {x'} \<and> sfsem_exec_path_is_dense x' r S"
  using assms
proof (induct l arbitrary: x)
  show "\<And>x. sfsem_exec_path_is_dense x ([] @ r) S \<Longrightarrow> 
  \<exists>x'. sfsem_exec_path_is_dense x [] {x'} \<and> sfsem_exec_path_is_dense x' r S"
    unfolding sfsem_exec_path_is_dense.simps
    by simp
next
  fix a :: "label \<times> ('s, 'e) instr" and l :: "(label \<times> ('s, 'e) instr) list" and x :: "('s, 'e) failure"
  assume
    IH: "\<And>x. sfsem_exec_path_is_dense x (l @ r) S \<Longrightarrow>
             \<exists>x'. sfsem_exec_path_is_dense x l {x'} 
                 \<and> sfsem_exec_path_is_dense x' r S" and
    step: "sfsem_exec_path_is_dense x ((a # l) @ r) S"
  from step have step': "sfsem_exec_path_is_dense x (a # (l @ r)) S"
    by simp
  obtain lbl i where
    a: "a = (lbl, i)"
    by fastforce
  obtain x' where
    dense_a: "sfsem_exec_path_is_dense x [a] {x'}" and
    dense_lr: "sfsem_exec_path_is_dense x' (l @ r) S"
    using step'
    unfolding a sfsem_exec_path_is_dense.simps sfsem_exec_empty
    by blast
  obtain x'' where
    dense_l: "sfsem_exec_path_is_dense x' l {x''}" and
    dense_r: "sfsem_exec_path_is_dense x'' r S"
    using IH[OF dense_lr]
    by blast
  from step'[unfolded a sfsem_exec_path_is_dense.simps, THEN conjunct1] have
    a1: "x \<notin> {x \<in> sfsem_exec (l @ r) S. NCPC (fst (snd x)) = lbl}" .
  hence "x \<notin> {x \<in> sfsem_exec l (sfsem_exec r S). NCPC (fst (snd x)) = lbl}"
    unfolding sfsem_exec_unfold2[of l r S] .
  hence a2: "\<forall> x' \<in> sfsem_exec r S . x \<notin> {x \<in> sfsem_exec l {x'} . NCPC (fst (snd x)) = lbl}"
    using sfsem_exec_additive[of l] UnCI[of x "sfsem_exec l _"] insert_is_Un mk_disjoint_insert insert_is_Un mem_Collect_eq
    by (metis (no_types, lifting))
  hence a3: "x \<notin> {x \<in> sfsem_exec l {x''}. NCPC (fst (snd x)) = lbl}"
    using sfsem_exec_is_dense_is_exec[OF dense_r]
    by blast
  have a4: "x \<noteq> x' \<and> x \<in> \<lbrakk>(one lbl i)\<rbrakk>{x'} \<and> sfsem_exec_path_is_dense x' l {x''}"
    using dense_a[unfolded a sfsem_exec_path_is_dense.simps sfsem_exec_empty]
          dense_l
    by blast
  have a5: "sfsem_exec_path_is_dense x (a # l) {x''}"
    unfolding a sfsem_exec_path_is_dense.simps
    using a3 a4
    by blast
  show "\<exists>x'. sfsem_exec_path_is_dense x (a # l) {x'} \<and>
             sfsem_exec_path_is_dense x' r S"
    using a5 dense_r
    by blast
qed

lemma sfsem_exec_critical_path_unfold:
  assumes
    "sfsem_exec_path_is_critical x (l @ r) S"
  shows
    "\<exists> x' . sfsem_exec_path_is_critical x l {x'} \<and> sfsem_exec_path_is_critical x' r S"
  using assms
proof (induct l arbitrary: x)
  show "\<And>x. sfsem_exec_path_is_critical x ([] @ r) S \<Longrightarrow> 
  \<exists>x'. sfsem_exec_path_is_critical x [] {x'} \<and> sfsem_exec_path_is_critical x' r S"
    unfolding sfsem_exec_path_is_critical.simps
    by simp
next
  fix a :: "label \<times> ('s, 'e) instr" and l :: "(label \<times> ('s, 'e) instr) list" and x :: "('s, 'e) failure"
  assume
    IH: "\<And>x. sfsem_exec_path_is_critical x (l @ r) S \<Longrightarrow>
             \<exists>x'. sfsem_exec_path_is_critical x l {x'} 
                 \<and> sfsem_exec_path_is_critical x' r S" and
    step: "sfsem_exec_path_is_critical x ((a # l) @ r) S"
  from step have aa2': "sfsem_exec_path_is_critical x (a # (l @ r)) S"
    by simp
  obtain x' where
    crit_a: "sfsem_exec_path_is_critical x [a] {x'}" and
    crit_lr: "sfsem_exec_path_is_critical x' (l @ r) S"
    using sfsem_exec_critical_path_unfold_single[OF aa2']
    by blast
  obtain x'' where
    crit_l: "sfsem_exec_path_is_critical x' l {x''}" and
    crit_r: "sfsem_exec_path_is_critical x'' r S"
    using IH[OF crit_lr]
    by blast
  obtain lbl i where
    a: "a = (lbl, i)"
    by fastforce
  from aa2'[unfolded a sfsem_exec_path_is_critical.simps, THEN conjunct1] have
    a1: "x \<notin> sfsem_exec (l @ r) S" .
  hence "x \<notin> sfsem_exec l (sfsem_exec r S)"
    unfolding sfsem_exec_unfold2[of l r S] .
  hence a2: "\<forall> x' \<in> sfsem_exec r S . x \<notin> sfsem_exec l {x'}"
    using sfsem_exec_additive[of l] UnCI[of x "sfsem_exec l _"] insert_is_Un mk_disjoint_insert
    by (metis (no_types, lifting))
  hence a3: "x \<notin> sfsem_exec l {x''}"
    using sfsem_exec_is_critical_is_exec[OF crit_r]
    by blast
  have a4: "x \<noteq> x' \<and> x \<in> \<lbrakk>(one lbl i)\<rbrakk>{x'} \<and> sfsem_exec_path_is_critical x' l {x''}"
    using crit_a[unfolded a sfsem_exec_path_is_critical.simps sfsem_exec_empty]
          crit_l
    by blast
  have a5: "sfsem_exec_path_is_critical x (a # l) {x''}"
    unfolding a sfsem_exec_path_is_critical.simps
    using a3 a4
    by blast
  show "\<exists>x'. sfsem_exec_path_is_critical x (a # l) {x'} \<and>
             sfsem_exec_path_is_critical x' r S"
    using a5 crit_r
    by blast
qed

text {* Gathering of facts. *}

text {* Append to @{term sfsem_exec1'}, that one of the existing paths definitely is critical. *}
lemma sfsem_exec1'_crit:
  assumes
    a: "x \<in> sfsem c S"
  shows "\<exists> lil . set lil \<subseteq> (\<union>\<^sub>s\<^sub>c c) \<and> x \<in> sfsem_exec lil S \<and> sfsem_exec_path_is_critical x lil S \<and> (x \<notin> S \<longrightarrow> (\<exists> l i lil'. lil = (l, i) # lil'))"
proof -
  obtain lil where
    ao1: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<and> x \<in> sfsem_exec lil S"
    using sfsem_exec1'[OF a]
    by blast
  obtain lil' where
    ao2: "set lil' \<subseteq> set lil \<and> length lil' \<le> length lil \<and> sfsem_exec_path_is_critical x lil' S"
    using sfsem_exec_criticalPath[OF ao1[THEN conjunct2]]
    by blast
  have a1: "set lil' \<subseteq> \<union>\<^sub>s\<^sub>c c"
    using ao1[THEN conjunct1] ao2[THEN conjunct1]
    by simp
  moreover have "(x \<notin> S \<longrightarrow> (\<exists> l i lil. lil' = (l, i) # lil))"
    by (metis ao2 neq_Nil_conv sfsem_exec_empty sfsem_exec_is_critical_is_exec surj_pair)
  ultimately show ?thesis
    using ao2[THEN conjunct2, THEN conjunct2] sfsem_exec_is_critical_is_exec[OF ao2[THEN conjunct2, THEN conjunct2]]
    by blast
qed

lemma sfsem_exec1_crit:
  assumes
    a: "(tr, s, X) \<in> sfsem c S"
  shows "\<exists> lil . set lil \<subseteq> (\<union>\<^sub>s\<^sub>c c) \<and> (tr, s, X) \<in> sfsem_exec lil S \<and> sfsem_exec_path_is_critical (tr, s, X) lil S \<and> ((tr, s, X) \<notin> S \<longrightarrow> (\<exists> l i lil'. lil = (l, i) # lil'))"
using sfsem_exec1'_crit[OF a] .

paragraph {* Fresh states *}

text {* A fresh variable is created on the way and is not part of S. *}
fun sfsem_exec_fresh :: "('s, 'e) failure \<Rightarrow> (label \<times> ('s, 'e) instr) list \<Rightarrow> ('s, 'e) failure set \<Rightarrow> bool" where
  "sfsem_exec_fresh x [] S = False" |
  "sfsem_exec_fresh x (a # lil) S = (x \<notin> sfsem_exec lil S \<or> sfsem_exec_fresh x lil S)"

lemma sfsem_exec_freshI1:
  assumes "x \<notin> S" and "x \<in> sfsem_exec lil S"
  shows "sfsem_exec_fresh x lil S"
using assms sfsem_exec_empty
proof (induct lil, auto) qed

lemma sfsem_exec_freshI4:
  assumes "x \<notin> S"
  shows "sfsem_exec_fresh x (a # lil) S"
using assms sfsem_exec_empty
proof (induct lil, auto) qed

lemma sfsem_exec_freshI2:
  assumes "x \<notin> sfsem_exec lil S"
  shows "sfsem_exec_fresh x (a # lil) S"
using assms by auto

lemma sfsem_exec_freshI3:
  assumes
    a: "lil = l @ cl" and
    b: "sfsem_exec_fresh x cl S"
  shows "sfsem_exec_fresh x lil S"
  using a b
proof (induct l arbitrary: lil, auto) qed

lemma sfsem_exec_freshI5:
  assumes "sfsem_exec_path_is_critical x ((l, i) # lil) S"
  shows "sfsem_exec_fresh x ((l, i) # lil) S"
using assms unfolding sfsem_exec_path_is_critical.simps sfsem_exec_fresh.simps by blast

subsection {* Manipulation and knowledge of the refusal sets *}

text {* If the @{term normal} successor state is not contained in @{term S}, we can exchange its refusal set
with any refusal set being a subset of @{term \<Sigma>}, and its original refusal set also is a subset
of @{term \<Sigma>}.*}

lemma sfsem_N_has_all_X:
  assumes
    a: "(tr, s, X) \<in> sfsem (one l i) S" and
    b: "N s" and
    c: "(tr, s, X) \<notin> S" and
    d: "X' \<subseteq> \<Sigma>"
  shows "(tr, s, X') \<in> sfsem (one l i) S \<and> X \<subseteq> \<Sigma>"
using assms NCCNotNormal
proof(cases i, auto, blast) qed

text {* A rather general version, which merely helps in the normal state case.
Any newly created successor state has a refusal set being a subset of @{term \<Sigma>}.*}

lemma sfsem_refusal_subset_of_sigma:
  assumes
    a: "(tr, s, X) \<in> sfsem c S" and
    b: "(tr, s, X) \<notin> S"
  shows
    "X \<subseteq> \<Sigma>"
using assms
proof -
  obtain lil where
    exec_lil: "(tr, s, X) \<in> sfsem_exec lil S" and
    crit_lil: "sfsem_exec_path_is_critical (tr, s, X) lil S"
    using sfsem_exec1_crit[OF a]
    by blast
  have lilNotEmpty: "lil \<noteq> []"
  proof (rule ccontr, simp)
    assume ba1: "lil = []"
    show False
      using exec_lil[unfolded ba1 sfsem_exec_empty] b
      by simp
  qed
  obtain i l lil' where
    ao2: "lil = (l, i) # lil'"
    using lilNotEmpty
    by (metis neq_Nil_conv surj_pair)
  show "X \<subseteq> \<Sigma>"
    using exec_lil[unfolded ao2 sfsem_exec_unfold] 
    using crit_lil[unfolded ao2 sfsem_exec_path_is_critical.simps, THEN conjunct1]
  proof (induct i, auto) qed
qed

text {* If the  @{term normal} successor state is not contained in @{term S}, we can exchange its refusal set
with any refusal set being a subset of @{term \<Sigma>}, and its original refusal set also is a subset
of @{term \<Sigma>}. *}

lemma sfsem_exec_fresh_N_all_X:
  assumes
    a: "sfsem_exec_fresh (tr, s, X) lil S" and
    b: "(tr, s, X) \<in> sfsem_exec lil S" and
    c: "N s" and
    d: "X' \<subseteq> \<Sigma>"
  shows "(tr, s, X') \<in> sfsem_exec lil S \<and> X \<subseteq> \<Sigma>"
using a b c
proof (induct lil arbitrary: tr s X rule: exec_induct, simp)
  fix l i lil tr s X
  assume
    IH: "\<And>tr s X. sfsem_exec_fresh (tr, s, X) lil S \<Longrightarrow> (tr, s, X) \<in> sfsem_exec lil S \<Longrightarrow> N s \<Longrightarrow> (tr, s, X') \<in> sfsem_exec lil S \<and> X \<subseteq> \<Sigma>" and
    fresh_lilil: "sfsem_exec_fresh (tr, s, X) ((l, i) # lil) S" and
    exec_lilil: "(tr, s, X) \<in> sfsem_exec ((l, i) # lil) S" and
    Ns: "N s"
  consider (exec_lil) "(tr, s, X) \<in> sfsem_exec lil S" | (not_exec_lil) "(tr, s, X) \<notin> sfsem_exec lil S" by blast
  thus "(tr, s, X') \<in> sfsem_exec ((l, i) # lil) S \<and> X \<subseteq> \<Sigma>"
  proof cases
    case exec_lil 
    hence b1: "sfsem_exec_fresh (tr, s, X) lil S"
      using fresh_lilil
      by auto
    from IH[OF b1 exec_lil Ns] have "(tr, s, X') \<in> sfsem_exec lil S \<and> X \<subseteq> \<Sigma>" .
    thus ?thesis
      using d exec_lilil
      unfolding sfsem_exec_unfold
    proof (induct i, auto simp add: NCCNotNormal) qed
  next
    case not_exec_lil
    thus ?thesis
      using d exec_lilil Ns
      unfolding sfsem_exec_unfold
    proof (induct i, auto simp add: NCCNotNormal, blast) qed
  qed
qed

text {* If the  @{term communication} successor state is not contained in @{term S}, then there has to
be the according communication instruction somewhere in the instruction path. We then can exchange its refusal set
with any refusal set being compliant to that instruction, and its original refusal set also was compliant. *}

lemma sfsem_exec_fresh_notN_all_X:
  assumes
    a: "sfsem_exec_fresh (tr, s, X) lil S" and
    b: "(tr, s, X) \<in> sfsem_exec lil S" and
    c: "\<not> N s"
  shows "\<exists> ef sf . (NCPC s, comm ef sf) \<in> (set lil) \<and> (\<forall> X' \<subseteq> \<Sigma> - ef (NCRS s) . (tr, s, X') \<in> sfsem_exec lil S) \<and> X \<subseteq> \<Sigma> - ef (NCRS s)"
using a b c
proof (induct lil arbitrary: tr s X rule: exec_induct, simp)
  fix l i lil tr s X
  assume
    IH: "\<And>tr s X. sfsem_exec_fresh (tr, s, X) lil S \<Longrightarrow> (tr, s, X) \<in> sfsem_exec lil S \<Longrightarrow> \<not> N s \<Longrightarrow> \<exists>ef sf. (NCPC s, comm ef sf) \<in> set lil \<and> (\<forall>X'\<subseteq>\<Sigma> - ef (NCRS s). (tr, s, X') \<in> sfsem_exec lil S) \<and> X \<subseteq> \<Sigma> - ef (NCRS s)" and
    fresh_lilil: "sfsem_exec_fresh (tr, s, X) ((l, i) # lil) S" and
    exec_lilil: "(tr, s, X) \<in> sfsem_exec ((l, i) # lil) S" and
    notNs: "\<not> N s"
  consider (exec_lil) "(tr, s, X) \<in> sfsem_exec lil S" | (not_exec_lil) "(tr, s, X) \<notin> sfsem_exec lil S" by blast
  thus "\<exists>ef sf. (NCPC s, comm ef sf) \<in> set ((l, i) # lil) \<and> (\<forall>X'\<subseteq>\<Sigma> - ef (NCRS s). (tr, s, X') \<in> sfsem_exec ((l, i) # lil) S) \<and> X \<subseteq> \<Sigma> - ef (NCRS s)"
  proof cases
    case exec_lil 
    hence b1: "sfsem_exec_fresh (tr, s, X) lil S"
      using fresh_lilil
      by auto
    from IH[OF b1 exec_lil notNs] have "\<exists>ef sf. (NCPC s, comm ef sf) \<in> set lil \<and> (\<forall>X'\<subseteq>\<Sigma> - ef (NCRS s). (tr, s, X') \<in> sfsem_exec lil S) \<and> X \<subseteq> \<Sigma> - ef (NCRS s)" .
    thus ?thesis
      using exec_lilil
      unfolding sfsem_exec_unfold
      using pc_does_not_point_to_single[of s "one l i" tr _ "sfsem_exec lil S"] notNs
      by auto
  next
    case not_exec_lil
    thus ?thesis
      using exec_lilil notNs
      unfolding sfsem_exec_unfold
    proof (induct i, auto simp add: NCCNotNormal, metis NCCSameR NCPCh2) qed
  qed
qed

text {* A selfbranch can only be part of a critical path, if it is the only
instruction on the path, and if for some refusal set the trace and state are
already part of the starting set.
The lemma is formulated in the contraposition to the above statement.
*}

lemma sfsem_N_no_selfcbr:
  assumes 
    a: "sfsem_exec_path_is_critical (tr, s, X) ((l, i) # lil) S"
    and b: "(\<forall> X . (tr, s, X) \<notin> S) \<or> lil \<noteq> []"
  shows "(i = cbr b m n) \<longrightarrow> ((b (NCRS s) \<longrightarrow> l \<noteq> m) \<and> (\<not> b (NCRS s) \<longrightarrow> l \<noteq> n))"
proof(rule, rule ccontr, simp)
  assume
    i: "i = cbr b m n" and
    label_imp: "b (NCRS s) \<and> l = m \<or> \<not> b (NCRS s) \<and> l = n"
  (* The predecessor has to have the same label as s, that's why it's working. *)
  have "(\<forall> X . (tr, s, X) \<notin> S) \<longrightarrow> lil \<noteq> []"
  proof (rule, rule ccontr, simp)
    assume
      notInS: "\<forall>X. (tr, s, X) \<notin> S" and
      lil: "lil = []"
    from sfsem_exec_is_critical_is_exec[OF a, unfolded lil i sfsem_exec_unfold sfsem_exec_empty] notInS label_imp NC_state_eq
      show False
      proof (fastforce+) qed
  qed
  moreover have "lil \<noteq> [] \<longrightarrow> False"
  proof
    assume lil_not_empty: "lil \<noteq> []"
    then obtain l' i' lil' where lil: "lil = (l', i') # lil'" by (metis list.exhaust prod.collapse)
    obtain tr'1 s'1 X'1 where
      li_1: "(tr, s, X) \<in> \<lbrakk>(one l i)\<rbrakk>{(tr'1, s'1, X'1)}" and
      n: "(tr'1, s'1, X'1) \<in> sfsem_exec ((l', i') # lil') S" and
      not_prev: "(tr'1, s'1, X'1) \<notin> sfsem_exec lil' S" and
      neq: "(tr, s, X) \<noteq> (tr'1, s'1, X'1)"
      using lil a prod_cases3 sfsem_exec_is_critical_is_exec sfsem_exec_path_is_critical.simps(2)
      by metis
    have "\<forall> X \<subseteq> \<Sigma> . (tr'1, s'1, X) \<in> sfsem_exec ((l', i') # lil') S"
      by (metis sfsem_exec_path_is_critical.simps(2) sfsem_exec_unfold a li_1 lil n not_prev sfsem_N_has_all_X sfsem_predecessor2)
    moreover have "tr'1 = tr \<and> s'1 = s \<and> X \<subseteq> \<Sigma>"
      using li_1 neq label_imp  NC_state_eq
      unfolding i sfcbr SClean_def
      proof (fastforce) qed
    ultimately have "(tr, s, X) \<in> sfsem_exec lil S" using lil by blast
    with a show False
      using sfsem_exec_path_is_critical.simps(2) by blast
  qed
  ultimately show False using b by blast
qed

text {* Gathering as many information as possible of the pure fact of having a critical path
of length at least one. *}

lemma sfsem_exec_lil_split2:
  assumes
    b: "sfsem_exec_path_is_critical (tr, s, X) ((l, i) # lil') S"
  shows
    "\<exists> tr' s' X' . (tr, s, X) \<in> \<lbrakk>one l i\<rbrakk>sfsem_exec lil' S \<and>
    (tr, s, X) \<notin> sfsem_exec lil' S \<and>
    (tr, s, X) \<in> \<lbrakk>one l i\<rbrakk>{(tr', s', X')} \<and>
    (tr, s, X) \<noteq> (tr', s', X') \<and>
    (tr', s', X') \<in> sfsem_exec lil' S \<and>
    sfsem_exec_path_is_critical (tr, s, X) [(l, i)] {(tr', s', X')} \<and>
    sfsem_exec_path_is_critical (tr', s', X') lil' S \<and>
    N s' \<and>
    NCPC s' = l \<and>
    ((\<forall> X . (tr, s, X) \<notin> S) \<or> lil' \<noteq> [] \<longrightarrow> s \<noteq> s')"
proof -
  have exec_lilil': "(tr, s, X) \<in> \<lbrakk>(one l i)\<rbrakk>sfsem_exec lil' S"
    using sfsem_exec_is_critical_is_exec[OF b, unfolded sfsem_exec_unfold]
    .
  have not_exec_lil': "(tr, s, X) \<notin> sfsem_exec lil' S"
    using b[unfolded sfsem_exec_path_is_critical.simps, THEN conjunct1]
    .
  obtain x' where
    crit_li_x': "sfsem_exec_path_is_critical (tr, s, X) [(l, i)] {x'}" and
    x'_crit_lil': "sfsem_exec_path_is_critical x' lil' S"
    using sfsem_exec_critical_path_unfold_single[OF b] by blast
  obtain tr' s' X' where x': "x' = (tr', s', X')"
    using prod_cases3 by blast
  have G2: "(tr, s, X) \<in> \<lbrakk>one l i\<rbrakk>{(tr', s', X')} \<and> (tr, s, X) \<noteq> (tr', s', X') \<and> (tr', s', X') \<in> sfsem_exec lil' S"
    using
      crit_li_x'[unfolded sfsem_exec_path_is_critical.simps sfsem_exec_empty x']
      sfsem_exec_is_critical_is_exec[OF x'_crit_lil', unfolded x']
      by auto
  moreover have Ns': "N s' \<and> NCPC s' = l"
    using G2[THEN cj2_1] G2[THEN cj3_2]
  proof (induct i, auto) qed
  {
    assume a: "\<forall> X . (tr, s, X) \<notin> S"
    have "s \<noteq> s'"
    proof (cases "N s", rule ccontr, simp)
      assume
        eq: "s = s'"
      hence "\<exists> b m n . i = cbr b m n \<and> (b (NCRS s') \<longrightarrow> m = (NCPC s)) \<and> (\<not> b (NCRS s') \<longrightarrow> n = (NCPC s)) \<and> tr = tr'"
        using G2 Ns'  NCCNotNormal2
      proof (cases i, auto simp add: NC_state_eq) qed
      thus False 
        using Ns' sfsem_N_no_selfcbr[OF b] a eq
        by blast
    next
      show "\<not> N s \<Longrightarrow> s \<noteq> s'"
        using Ns'
        by auto
    qed
  }
  moreover
  {
    assume a: "lil' \<noteq> []"
    obtain li' lil'' where bo1: "lil' = li' # lil''"
      using list.exhaust a
      by blast
    obtain l' i' where bo2: "li' = (l', i')"
      by fastforce
    have "s \<noteq> s'"
    proof (cases "N s", rule ccontr, simp)
      assume
        eq: "s = s'"
      have d0: "\<exists> b m n . i = cbr b m n \<and> (b (NCRS s') \<longrightarrow> m = (NCPC s)) \<and> (\<not> b (NCRS s') \<longrightarrow> n = (NCPC s)) \<and> tr = tr'"
        using G2 Ns' eq NCCNotNormal2
      proof (cases i, auto) qed
      have d1: "(tr', s', X') \<in> sfsem_exec lil' S"
        using sfsem_exec_is_critical_is_exec[OF x'_crit_lil', unfolded x']
        .
      have "(tr', s', X) \<in> \<lbrakk>(one l' i')\<rbrakk>sfsem_exec lil'' S"
        using sfsem_N_has_all_X[OF d1[unfolded bo1 bo2 sfsem_exec_unfold] Ns'[THEN cj2_1] x'_crit_lil'[unfolded x' bo1 bo2 sfsem_exec_path_is_critical.simps, THEN conjunct1] sfsem_refusal_subset_of_sigma[OF G2[THEN conjunct1] G2[THEN conjunct2, THEN conjunct1, unfolded singleton_iff[symmetric]]]]
        by simp
      hence G3: "(tr', s', X) \<in> sfsem_exec lil' S"
        unfolding sfsem_exec_unfold[symmetric] bo2[symmetric] bo1[symmetric]
        .
      thus False
        using G3 eq d0 not_exec_lil'
        by fast
    next
      show "\<not> N s \<Longrightarrow> s \<noteq> s'"
        using Ns'
        by auto
    qed
  }
  ultimately
  show ?thesis
    using G2 x'_crit_lil' exec_lilil' Ns' not_exec_lil' crit_li_x'
    unfolding x'
    by metis
qed

text {* If we know that for some element in a list a property @{prop P} does not hold, we know that
there has to exist some (possibly empty) prefix in the list, for which every element satisfies @{prop P},
but whose successor does not. Note: This successor does not have to be the element for which we already
know, that it does not satisfy @{prop P}*}

lemma list_split_by_P:
  assumes
    x_in_xs: "x \<in> set xs" and
    notPx: "\<not> P x"
  shows "\<exists> hd x tl . xs = hd @ x # tl \<and> (\<forall> x . x \<in> set hd \<longrightarrow> P x) \<and> \<not> P x"
using x_in_xs
proof (induct xs, simp)
  fix a xs
  assume
    IH: "x \<in> set xs \<Longrightarrow> \<exists>hd x tl. xs = hd @ x # tl \<and> (\<forall>x. x \<in> set hd \<longrightarrow> P x) \<and> \<not> P x" and
    step: "x \<in> set (a # xs)"
  show "\<exists>hd x tl. a # xs = hd @ x # tl \<and> (\<forall>x. x \<in> set hd \<longrightarrow> P x) \<and> \<not> P x"
  proof (cases "x = a \<or> \<not> P a")
    assume x: "x = a \<or> \<not> P a"
    have "a # xs = [] @ a # xs \<and> (\<forall> x . x \<in> set [] \<longrightarrow> P x) \<and> \<not> P x"
      using notPx by simp
    thus ?thesis using x by blast
  next
    assume a: "\<not> (x = a \<or> \<not> P a)"
    from a have xneqa: "x \<noteq> a" by simp
    from a have Pa: "P a" by simp
    obtain hd x tl where
      xs: "xs = hd @ x # tl \<and> (\<forall>x. x \<in> set hd \<longrightarrow> P x)" and
      notPx: "\<not> P x"
      using step IH xneqa by auto
    thus ?thesis using Pa
      by (metis Cons_eq_appendI set_ConsD)
  qed
qed

text {* We can exchange the starting state's refusal set and still know, that some refusal set exists,
such that the original successor's trace and state with that new refusal set are still a successor.
Actually, we could gain more knowledge out of it: We can say that the new refusal sets match, or the
original successor's refusal set is still contained. But here, only the basic variant is formulated.
The extensions will be added later in a more finegrained way, as those basic formulations already
are useful.
*}

lemma sfsem_exec_general_sigma_exchange:
  assumes "(tr, s, X) \<in> sfsem_exec lil {(trs, ss, Xs)}"
  shows "\<exists> X' . (tr, s, X') \<in> sfsem_exec lil {(trs, ss, Xs')}"
using assms
proof (induct lil arbitrary: tr s X, auto simp add: sfsem_exec_unfold)
  fix a b lil tr s X
  assume
    IH : "\<And>tr s X. (tr, s, X) \<in> sfsem_exec lil {(trs, ss, Xs)} \<Longrightarrow> \<exists>X'. (tr, s, X') \<in> sfsem_exec lil {(trs, ss, Xs')}" and
    exec_ablil_s: "(tr, s, X) \<in> \<lbrakk>(one a b)\<rbrakk>sfsem_exec lil {(trs, ss, Xs)}"
  show "\<exists>X'. (tr, s, X') \<in> \<lbrakk>(one a b)\<rbrakk>sfsem_exec lil {(trs, ss, Xs')}"
  proof (cases "(tr, s, X) \<in> sfsem_exec lil {(trs, ss, Xs)}")
    assume not_exec_lil_s: "(tr, s, X) \<notin> sfsem_exec lil {(trs, ss, Xs)}"
    then obtain trp sp Xp where
      step_p: "(tr, s, X) \<in> \<lbrakk>one a b\<rbrakk>{(trp, sp, Xp)}" and
      p_exec_s: "(trp, sp, Xp) \<in> sfsem_exec lil {(trs, ss, Xs)}" and
      neqp: "(tr, s, X) \<noteq> (trp, sp, Xp)" and
      Nsp: "N sp"
      using sfsem_predecessor[OF exec_ablil_s] by blast
      obtain Xp' where p'_exec_lil_s': "(trp, sp, Xp') \<in> sfsem_exec lil {(trs, ss, Xs')}"
        using IH[OF p_exec_s] by blast
      thus ?thesis using step_p
      proof (induct b, auto, blast) qed
  next
    assume exec_lil_s: "(tr, s, X) \<in> sfsem_exec lil {(trs, ss, Xs)}"
    obtain X' where exec'_lil_s': "(tr, s, X') \<in> sfsem_exec lil {(trs, ss, Xs')}"
      using IH[OF exec_lil_s] by blast
    thus ?thesis using exec_ablil_s
    proof (induct b, auto,blast, blast, metis IH) qed
  qed
qed

text {* The above formulation also hold for the standard SF semantics. *}

lemma sfsem_general_sigma_exchange:
  assumes
    inside: "(tr, s, X) \<in> sfsem c {(trs, ss, Xs)}" and
    wff: "wff\<^sub>s\<^sub>c c"
  shows "\<exists> X' . (tr, s, X') \<in> sfsem c {(trs, ss, Xs')}"
proof -
  obtain lil where
    subs: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    ex: "(tr, s, X) \<in> sfsem_exec lil {(trs, ss, Xs)}"
    using sfsem_exec1[OF inside] by blast
  obtain X' where
    ex': "(tr, s, X') \<in> sfsem_exec lil {(trs, ss, Xs')}"
    using sfsem_exec_general_sigma_exchange[OF ex] by blast
  thus ?thesis
    using sfsem_exec2[OF subs ex' _ wff] inside
      sfsem_not_outside_means_cbr[OF inside]
    by fastforce
qed

text {* We can substract any part of the predecessor and successor refusal set. *}

lemma sfsem_refusal_substraction:
  fixes R
  assumes ex: "(tr, s, X) \<in> sfsem c {(tr', s', X')}" and wff: "wff\<^sub>s\<^sub>c c"
  shows "(tr, s, X - R) \<in> sfsem c {(tr', s', X' - R)}"
proof (cases "NCPC s \<in> dom c \<and> N s")
  case True
  show ?thesis
    using sfsem_not_outside_means_cbr[OF ex True] by force
next
  case False
  obtain lil where
    subs: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    ex_lil: "(tr, s, X) \<in> sfsem_exec lil {(tr', s', X')}" and
    crit: "sfsem_exec_path_is_critical (tr, s, X) lil {(tr', s', X')}"
    using sfsem_exec1_crit[OF ex] by blast
  consider (empty) "lil = []" | (notempty) "\<exists> l i lil' . lil = (l, i) # lil'"
    by (metis list.exhaust prod.collapse)
  hence ex: "(tr, s, X - R) \<in> sfsem_exec lil {(tr', s', X' - R)}"
  proof (cases)
    case empty
    thus ?thesis
      using ex_lil unfolding empty by simp
  next
    case notempty
    then obtain l i lil' where lil: "lil = (l, i) # lil'" by blast
    obtain tru su Xu where
      ex_one: "(tr, s, X) \<in> \<lbrakk>one l i\<rbrakk>{(tru, su, Xu)}" and
      ex': "(tru, su, Xu) \<in> sfsem_exec lil' {(tr', s', X')}" and
      neq: "(tr, s, X) \<noteq> (tru, su, Xu)"
      using crit[unfolded lil sfsem_exec_path_is_critical.simps] prod_cases3
        sfsem_exec_is_critical_is_exec
      by metis
    obtain Xt where Xt: "(tru, su, Xt) \<in> sfsem_exec lil' {(tr', s', X' - R)}"
      using sfsem_exec_general_sigma_exchange[OF ex']
      by blast
    have ex_one': "(tr, s, X - R) \<in> \<lbrakk>one l i\<rbrakk>{(tru, su, Xt)}"
      using ex_one neq proof (induct i, auto) qed
    show ?thesis
      unfolding lil sfsem_exec_unfold 
      using sfsem_monotone_single[OF Xt ex_one'] .
  qed
  show ?thesis
    using sfsem_exec2[OF subs ex False wff] .
qed

text {* A sanity lemma. The following formulation is kind of symmetric, although on the first view
it does not look like that. *}

lemma S'S_subset_equality:
  assumes
    S': "S' \<subseteq> {(tr, s, X') . \<exists> X . (tr, s, X) \<in> S}" and (* shrinking S can be done beforehand, extending afterwards. *)
    S'_all_contained: "\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> (\<exists> X' . (tr, s, X') \<in> S')"
  shows "S \<subseteq> {(tr, s, X') . \<exists> X . (tr, s, X) \<in> S'} \<and> (\<forall> tr s X . (tr, s, X) \<in> S' \<longrightarrow> (\<exists> X' . (tr, s, X') \<in> S))"
using assms
by blast

text{* If a fresh state (for the starting states with changed refusal sets) actually is not created by
the execution of an instruction path, it won't be with the exchanged starting set, either.
A contrapositional formulation with a single starting state can be found afterwards. This is just
a rather general formulation.
*}

lemma sfsem_exec_fresh_wont_appear:
  assumes
    fresh: "sfsem_exec_fresh (tr, s, X) lil S'" and
    not_exec: "(tr, s, X) \<notin> sfsem_exec lil S" and
    S': "S' \<subseteq> {(tr, s, X') . \<exists> X . (tr, s, X) \<in> S}" and (* shrinking S can be done beforehand, extending afterwards. *)
    S'_all_contained: "\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> (\<exists> X' . (tr, s, X') \<in> S')"
  shows "(tr, s, X) \<notin> sfsem_exec lil S'"
proof (rule ccontr, simp)
  show "(tr, s, X) \<in> sfsem_exec lil S' \<Longrightarrow> False"
    using fresh not_exec
  proof (induct lil arbitrary: tr s X, simp)
    fix a lila tr s X
    assume
      IH: "\<And> tr s X . (tr, s, X) \<in> sfsem_exec lila S' \<Longrightarrow> sfsem_exec_fresh (tr, s, X) lila S' \<Longrightarrow> (tr, s, X) \<notin> sfsem_exec lila S \<Longrightarrow> False" and
      exec_alS': "(tr, s, X) \<in> sfsem_exec (a # lila) S'" and
      fresh_alS': "sfsem_exec_fresh (tr, s, X) (a # lila) S'" and
      not_exec_alS: "(tr, s, X) \<notin> sfsem_exec (a # lila) S"
    obtain l i where a: "a = (l, i)" by fastforce
    show "False"
    proof (cases "lila = []")
      assume lila: "lila = []"
      show False
        using S' exec_alS' fresh_alS' not_exec_alS unfolding a lila sfsem_exec_single sfsem_exec_fresh.simps sfsem_exec_empty
      proof (induct i, auto, blast) qed
    next
      assume "lila \<noteq> []"
      then obtain a' lila' where lila: "lila = a' # lila'"
        by (meson neq_Nil_conv)
      obtain l' i' where a': "a' = (l', i')" by fastforce
      show False
      proof (cases "(tr, s, X) \<in> sfsem_exec lila S'")
        assume exec_lila_S': "(tr, s, X) \<in> sfsem_exec lila S'"
        moreover have fresh: "sfsem_exec_fresh (tr, s, X) lila S'" using exec_lila_S' fresh_alS' by auto
        moreover have "(tr, s, X) \<notin> sfsem_exec lila S"
        proof (rule ccontr, simp)
          assume exec_lila: "(tr, s, X) \<in> sfsem_exec lila S"
          hence "NCPC s = l \<and> N s"
            using not_exec_alS[unfolded a sfsem_exec_unfold]
          proof (induct i, auto) qed
          hence "(tr, s, X) \<notin> sfsem_exec (a # lila) S'"
            using exec_lila_S' unfolding a sfsem_exec_unfold
          proof (induct i, auto simp add: NCCNotNormal)
            have G: "(tr, s, X) \<in> sfsem_exec lila S' \<Longrightarrow> False"
              using exec_lila exec_alS' not_exec_alS 
              using sfsem_monotone_single[of "(tr, s, X)" "sfsem_exec lila S" "(tr, s, X)" "one l i"]
              using sfsem_skip[of tr s X "sfsem_exec lila S'" l i]
              unfolding a sfsem_exec_unfold
              by blast
            show "(tr, s, X) \<in> sfsem_exec lila S' \<Longrightarrow> False" using G .
            show "(tr, s, X) \<in> sfsem_exec lila S' \<Longrightarrow> False" using G .
          qed
          with exec_alS' show False by contradiction
        qed
        ultimately show False using IH by fast
      next
        assume not_exec_lila_S': "(tr, s, X) \<notin> sfsem_exec lila S'"
        obtain tru su Xu where
          exec_li_u: "(tr, s, X) \<in> \<lbrakk>one l i\<rbrakk> {(tru, su, Xu)}" and
          u_exec_lila_S': "(tru, su, Xu) \<in> sfsem_exec lila S'" and
          nequ: "(tr, s, X) \<noteq> (tru, su, Xu)"
          using exec_alS' not_exec_lila_S' sfsem_predecessor
          unfolding a sfsem_exec_unfold
          by blast
        obtain trs ss Xs where
          u_exec_lila_s: "(tru, su, Xu) \<in> sfsem_exec lila {(trs, ss, Xs)}" and
          s_in_S': "(trs, ss, Xs) \<in> S'"
          using u_exec_lila_S' sfsem_exec_pred2
          by blast
        obtain Xs' where s'_in_S: "(trs, ss, Xs') \<in> S" using S' s_in_S' by blast
        have S_union_s': "S \<union> {(trs, ss, Xs')} = S" using s'_in_S by blast
        obtain Xu' where u'_exec_lila_s': "(tru, su, Xu') \<in> sfsem_exec lila {(trs, ss, Xs')}"
          using sfsem_exec_general_sigma_exchange[OF u_exec_lila_s]
          by blast
        hence "(tr, s, X) \<in> \<lbrakk>one l i\<rbrakk>{(tru, su, Xu')}"
          using exec_li_u nequ proof (induct i, auto) qed
        hence "(tr, s, X) \<in> sfsem_exec ((l, i) # lila) {(trs, ss, Xs')}"
          using u'_exec_lila_s'
          unfolding sfsem_exec_unfold
          using sfsem_monotone_single by blast
        hence "(tr, s, X) \<in> sfsem_exec ((l, i) # lila) S"
          using sfsem_exec_additive[of "(l, i) # lila" S "{(trs, ss, Xs')}"] 
          unfolding S_union_s'
          by blast
        thus False using not_exec_alS[unfolded a] by contradiction
      qed
    qed
  qed
qed

text {* For a fresh state which is created by execution of an instruction path, we can simply exchange
its starting state's refusal set by any other refusal set. *}

lemma sfsem_exec_fresh_X_irrelevant:
  assumes
    exec_lil: "(tr, s, X) \<in> sfsem_exec lil {(tr', s', X')}" and
    fresh_lil: "sfsem_exec_fresh (tr, s, X) lil  {(tr', s', X')}"
  shows "(tr, s, X) \<in> sfsem_exec lil {(tr', s', X'')}"
  using assms sfsem_exec_fresh_wont_appear[OF fresh_lil, of "{(tr', s', X'')}", simplified]
  by blast

text {* A communication state can be created by some critical path with a @{term comm} instruction as
its head, and every critical path for this state has this very same head. *}

theorem sfsem_notN_refusal_crit_has_same_head:
  assumes
    sfc: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S" and
    notinS: "(tr, s, X) \<notin> S" and
    notNs: "\<not> N s" and
    wff: "wff\<^sub>s\<^sub>c c"
  shows "(\<forall> lil . (sfsem_exec_path_is_critical (tr, s, X) lil S \<and> (tr, s, X) \<in> sfsem_exec lil S \<and> (set lil) \<subseteq> \<union>\<^sub>s\<^sub>c c) \<longrightarrow>
(\<exists> evf sf . hd lil = (NCPC s, comm evf sf) \<and> X \<subseteq> \<Sigma> - evf (NCRS s))) \<and>
(\<forall> lil1 lil2 . (sfsem_exec_path_is_critical (tr, s, X) lil1 S \<and> (tr, s, X) \<in> sfsem_exec lil1 S \<and> (set lil1) \<subseteq> \<union>\<^sub>s\<^sub>c c \<and>
                sfsem_exec_path_is_critical (tr, s, X) lil2 S \<and> (tr, s, X) \<in> sfsem_exec lil2 S \<and> (set lil2) \<subseteq> \<union>\<^sub>s\<^sub>c c) \<longrightarrow>
               (hd lil1 = hd lil2))
"
proof (auto)
  fix lil
  assume
    crit: "sfsem_exec_path_is_critical (tr, s, X) lil S" and
    exec_lil: "(tr, s, X) \<in> sfsem_exec lil S" and
    lil_subs: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c"
  let ?G = "\<exists>evf. (\<exists>sf. hd lil = (NCPC s, comm evf sf)) \<and> X \<subseteq> \<Sigma> - evf (NCRS s)"
  have "lil \<noteq> []"
    using exec_lil notinS by auto
  then obtain l i lil' where lil: "lil = (l, i) # lil'" by (metis list.exhaust prod.collapse)
  show ?G
    using exec_lil crit[unfolded lil sfsem_exec_path_is_critical.simps, THEN cj2_1] notNs
    unfolding lil sfsem_exec_unfold hd_def
  proof (induct i, simp, simp, auto simp add: NCPCh2 NCRSh2) qed
next
  fix lil1 lil2
  assume
    crit1: "sfsem_exec_path_is_critical (tr, s, X) lil1 S" and
    exec_lil1: "(tr, s, X) \<in> sfsem_exec lil1 S" and
    lil1_subs: "set lil1 \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    crit2: "sfsem_exec_path_is_critical (tr, s, X) lil2 S" and
    exec_lil2: "(tr, s, X) \<in> sfsem_exec lil2 S" and
    lil2_subs: "set lil2 \<subseteq> \<union>\<^sub>s\<^sub>c c"
  have "lil1 \<noteq> [] \<and> lil2 \<noteq> []"
    using exec_lil1 exec_lil2 notinS by fastforce
  then obtain l1 i1 lil'1 l2 i2 lil'2 where
    lil1: "lil1 = (l1, i1) # lil'1" and
    lil2: "lil2 = (l2, i2) # lil'2"
    by (metis list.exhaust prod.collapse)
  have "NCPC s = l1 \<and> NCPC s = l2"
    by (metis SFsemSigma.sfsem_C_from_comm SFsemSigma.sfsem_exec_path_is_critical.simps(2) SFsemSigma.sfsem_exec_unfold SFsemSigma_axioms crit1 crit2 exec_lil1 exec_lil2 lil1 lil2 notNs)
  thus "hd lil1 = hd lil2"
    using wff wff_code_implies_unique_labels[OF wff, of l1 i1 i2] lil1_subs lil2_subs
    unfolding lil1 lil2 hd_def
    by force
qed

text{* In the SF context it holds, too, that for any newly created @{term communication} state, we
can find the according @{term comm} instruction in the code. It is unique, as the code is wellformed
and for every label there exists only one instruction. *}

theorem sfsem_notN_refusal_restricted:
  assumes
    sfc: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S" and
    notinS: "(tr, s, X) \<notin> S" and
    notNs: "\<not> N s" and
    wff: "wff\<^sub>s\<^sub>c c"
  shows "\<exists> ef sf . (NCPC s, comm ef sf) \<in> \<union>\<^sub>s\<^sub>c c \<and> X \<subseteq> \<Sigma> - ef (NCRS s)"
proof (rule ccontr, auto)
  assume contr: "\<forall>ef. (\<forall>sf. (NCPC s, comm ef sf) \<notin> \<union>\<^sub>s\<^sub>c c) \<or> \<not> X \<subseteq> \<Sigma> - ef (NCRS s)"
  obtain l i lil where
    lil_subs_c: "set ((l, i) # lil) \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    exec_lil: "(tr, s, X) \<in> sfsem_exec ((l, i) # lil) S" and
    crit_lil: "sfsem_exec_path_is_critical (tr, s, X) ((l, i) # lil) S"
    using sfsem_exec1_crit[OF sfc] notinS
    by blast
  then obtain ef sf where
    lieq: "(l, i) = (NCPC s, comm ef sf)" and
    X: "X \<subseteq> \<Sigma> - ef (NCRS s)"
    using sfsem_notN_refusal_crit_has_same_head[OF sfc notinS notNs wff]
      list.sel(1)
    by metis
  thus False
    using contr contra_subsetD lil_subs_c list.set_intros(1) by auto
qed

text {* For any newly created state, whose trace and state are not in the set of starting states,
we can find an according greatest refusal set. for that trace / state combination.
If the state is a @{term normal} state, this greatest refusal set is @{term \<Sigma>}.
Otherwise, if the state is a @{term communication} state, this greatest refusal set depends
on the according @{term comm} instruction.
*}

theorem sfsem_greatest_refusal:
  fixes X
  assumes
    a: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S" and
    c: "\<forall> X . (tr, s, X) \<notin> S" (* For N s it can be restricted to \<forall> X . X - \<Sigma> \<noteq> {} \<longrightarrow> *) and
    d: "wff\<^sub>s\<^sub>c c"
  shows
    "\<exists> X' . (tr, s, X') \<in> \<lbrakk>c\<rbrakk>S \<and> (\<forall> X'' . ((tr, s, X'') \<in> \<lbrakk>c\<rbrakk>S) = (X'' \<subseteq> X')) \<and>
      (N s \<longrightarrow> X' = \<Sigma>) \<and>
      (\<not> N s \<longrightarrow> (\<exists> evf sf . (NCPC s, comm evf sf) \<in> \<union>\<^sub>s\<^sub>c c \<and> X' = \<Sigma> - evf (NCRS s)))"
proof (cases "(NCPC s \<in> dom c \<and> N s)")
  assume aa1: "NCPC s \<in> definitions.dom c \<and> N s"
  --"This would mean, c is a selfbranch, thus conflicting with the assumptions."
  have "False"
    using sfsem_not_outside_means_cbr[OF a aa1] c
    by blast
  thus ?thesis
    by blast
next
  assume ba1: "\<not> (NCPC s \<in> definitions.dom c \<and> N s)"
  obtain l i lil where
    lil_subs_c: "set ((l, i) # lil) \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    exec_lil: "(tr, s, X) \<in> sfsem_exec ((l, i) # lil) S" and
    crit_lil: "sfsem_exec_path_is_critical (tr, s, X) ((l, i) # lil) S"
    using sfsem_exec1_crit[OF a] c
    by blast
  have fresh: "sfsem_exec_fresh (tr, s, X) ((l, i) # lil) S"
    by (simp add: c exec_lil sfsem_exec_freshI1)
    consider (Ns) "N s" | (notNs) "\<not> N s" by blast
  thus ?thesis
  proof cases
    case Ns
    have "\<forall> X \<subseteq> \<Sigma> . (tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
      using 
        sfsem_exec2[OF lil_subs_c
          sfsem_exec_fresh_N_all_X[OF fresh exec_lil Ns, THEN cj2_1]
          ba1 d
        ]
      by simp
    thus ?thesis
      using Ns
      using c sfsem_refusal_subset_of_sigma[of tr s _ c S] by fastforce
  next
    case notNs
    obtain ef sf where
      inC: "(NCPC s, comm ef sf) \<in> \<union>\<^sub>s\<^sub>c c" and
      X: "X \<subseteq> \<Sigma> - ef (NCRS s)"
      using sfsem_notN_refusal_restricted[OF a _ notNs d] c by blast
    moreover have "\<forall> X \<subseteq> \<Sigma> - ef (NCRS s) . (tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
      using 
        sfsem_exec_fresh_notN_all_X[OF fresh exec_lil notNs]
        sfsem_exec2[OF lil_subs_c _ ba1 d, of tr _ S]
      by (metis (no_types, lifting) d inC wff_code_implies_unique_labels instr.inject(3) lil_subs_c subsetCE)
    moreover have "\<forall> X . \<not> X \<subseteq> \<Sigma> - ef (NCRS s) \<longrightarrow> (tr, s, X) \<notin> \<lbrakk>c\<rbrakk>S"
    proof (rule, rule, rule ccontr, simp)
      fix X
      assume
        notSubs: "\<not> X \<subseteq> \<Sigma> - ef (NCRS s)"and
        sfc: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
      obtain ef' sf' where
        in'C: "(NCPC s, comm ef' sf') \<in> \<union>\<^sub>s\<^sub>c c" and
        Xsubs': "X \<subseteq> \<Sigma> - ef' (NCRS s)"
        using sfsem_notN_refusal_restricted[OF sfc _ notNs d] c by blast
      have "ef = ef'"
        using d in'C inC wff_code_implies_unique_labels by blast
      thus False using notSubs Xsubs' by simp
    qed
    ultimately show ?thesis
      using notNs c
      by blast
  qed
qed

text {* When there exist multiple sets of refused elements, then they are also refused all together.
Note that this lemma is not used afterwards. However, it is a nice to have sanity check
 *}
lemma sfsem_refusal_set_Union:
  fixes X
  assumes sf: "\<forall> a \<in> XA . (tr, s, a) \<in> sfsem c {(tr', s', X')}" and
      not_empty: "XA \<noteq> {}" and
      wff: "wff\<^sub>s\<^sub>c c"
  shows "(tr, s, \<Union>XA) \<in> sfsem c {(tr', s', X')}"
proof (cases "NCPC s \<in> dom c \<and> N s")
  case True
  obtain X'a m n f l where
    inS: "(tr, s, X'a) \<in> {(tr', s', X')}" and
    c: "c = (l :: cbr f m n)" and
    f: "f (NCRS s) \<longrightarrow> m = l" and
    not_f: "\<not> f (NCRS s) \<longrightarrow> n = l" and
    sigma: "\<forall>X\<subseteq>\<Sigma>. (tr, s, X) \<in> \<lbrakk>c\<rbrakk>{(tr, s, X'a)}" and
    allSubs: "\<forall> a \<in> XA . a \<subseteq> \<Sigma>" and
    l: "l = NCPC s"
    using sfsem_not_outside_means_cbr[OF _ True, of tr _ "{(tr', s', X')}"]
      sf not_empty
    by (metis equals0I)
  show ?thesis using allSubs sigma inS
    by (meson Sup_least Un_subset_iff equals0I not_empty sfsem_monotone_single)
next
  case False
  define XA' where XA': "XA' = {xa' . \<exists> xa . xa \<in> XA \<and> xa \<noteq> {} \<and> xa' = xa}"
  have sf': "\<forall> a \<in> XA' . (tr, s, a) \<in> sfsem c {(tr', s', X')}" using sf unfolding XA' by auto
  consider (empty) "XA' = {}" | (singleton) "\<exists> xa . XA' = {xa}" | (big) "\<exists> xa1 xa2 . xa1 \<in> XA' \<and> xa2 \<in> XA' \<and> xa1 \<noteq> xa2" by blast
  thus ?thesis
  proof cases
    case empty
    hence "\<Union>XA = {}" using XA' by auto
    thus ?thesis using sf not_empty empty
      by (metis (no_types, hide_lams) Sup_insert Sup_le_iff Un_insert_right Union_Un_distrib all_not_in_conv subset_antisym sup_bot.right_neutral sup_ge1)
  next
    case singleton
    hence g: "(tr, s, \<Union>XA') \<in> sfsem c {(tr', s', X')}" using sf' by blast
    have eq: "\<Union>XA' = \<Union>XA" using singleton unfolding XA' by blast
    show ?thesis using g unfolding eq .
  next
    case big
    then obtain xa1 xa2 where
      xa1: "xa1 \<in> XA'" and
      xa2: "xa2 \<in> XA'" and
      neq: "xa1 \<noteq> xa2" by blast
    have sf1: "(tr, s, xa1) \<in> sfsem c {(tr', s', X')}" using xa1 sf' by blast
    have sf2: "(tr, s, xa2) \<in> sfsem c {(tr', s', X')}" using xa2 sf' by blast
    (* To show this, we need two distinct elements - hence the three cases. *)
    have neq: "(tr, s) \<noteq> (tr', s')"
    proof
      assume eq: "(tr, s) = (tr', s')"
      hence "xa1 = xa2" using sf1 sf2 False
        by (metis prod.inject sfsem_predecessor2)
      thus False using neq
        using XA' xa1 xa2 by auto
    qed
    hence tmp: "\<forall>X. (tr, s, X) \<notin> {(tr', s', X')}" by blast
    obtain Xg where
      sfg: "(tr, s, Xg) \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}" and
      greatest: "(\<forall>X''. ((tr, s, X'') \<in> \<lbrakk>c\<rbrakk>{(tr', s', X')}) = (X'' \<subseteq> Xg))"
      using sfsem_greatest_refusal[OF sf1 tmp wff]
      by blast
    have "\<forall> a \<in> XA . a \<subseteq> Xg" using sf greatest by simp
    hence "\<Union>XA \<subseteq> Xg" using not_empty by blast
    thus ?thesis using greatest by blast
  qed
qed

subsection {* Connecting SF with multistep and other theorems. *}

text {* Executing any single instruction in SF can be related to executing that instruction in the
smallstep world. *}

lemma sfsemImpliesSmallstep:
  assumes
    a: "(tr, s, X) \<in> sfsem (one (NCPC s') i) {(tr', s', X')}" and
    b: "(NCPC s', i) \<in> lis" and
    c: "N s" (* otherwise we do not progress - in contrast to smallstep *)
  shows "((tr', NCS s'), tr, NCS s) \<in> smallstep(lis)"
using a b 
proof (induct i, simp)
  fix x
  assume
    aa1: "tr = tr' \<and> s = s' \<and> X = X' \<and> (NCPC s = NCPC s' \<longrightarrow> \<not> N s) \<or>
         tr = tr' \<and> N s' \<and> N s \<and> NCPC s = Suc (NCPC s') \<and> NCRS s \<in> x (NCRS s') \<and> X \<subseteq> \<Sigma>" and
    aa2: "(NCPC s', do x) \<in> lis" and
    aa3: "(tr, s, X) \<in> \<lbrakk>(one (NCPC s') i)\<rbrakk>{(tr', s', X')}" and
    aa4: "(NCPC s', i) \<in> lis"
  have aa1: "tr = tr' \<and> N s \<and> NCPC s = Suc (NCPC s') \<and> NCRS s \<in> x (NCRS s') \<and> X \<subseteq> \<Sigma>"
    using c aa1 
    by fast
  have a1: "R (NCS s) \<in> x (R (NCS s'))"
    using aa1[THEN cj5_4, unfolded RNCS[symmetric]] .
  have a2: "NCS s = NCS s' \<lparr>R := R (NCS s), PC := PC (NCS s') + 1\<rparr>"
    using aa1[THEN cj4_3, unfolded PCNCS[symmetric]]
    unfolding Nat.Suc_eq_plus1[symmetric]
    by simp
  show "((tr', NCS s'), tr, NCS s) \<in> smallstep lis"
    using sdof[of "NCPC s'" x lis "NCRS s" "NCS s'" "NCS s" tr, unfolded RNCS[symmetric]]
      PCNCS[of s'] a1 a2 aa1[THEN cj2_1] aa2
    by blast
next
  fix b m n
  assume
    aa1: "(tr, s, X) \<in> \<lbrakk>(one (NCPC s') (cbr b m n))\<rbrakk>{(tr', s', X')}" and
    aa2: "(NCPC s', cbr b m n) \<in> lis"
  have a1: "tr = tr' \<and> N s' \<and> N s \<and> NCRS s' = NCRS s \<and> X \<subseteq> \<Sigma> \<and> (b (NCRS s') \<and> NCPC s = m \<or> \<not> b (NCRS s') \<and> NCPC s = n)"
    using aa1[unfolded sfcbr, simplified] c
    by blast
  show "((tr', NCS s'), tr, NCS s) \<in> smallstep lis"
  proof (cases "b (NCRS s')")
    assume T: "b (NCRS s')"
    have "PC (NCS s) = m"
      using T a1 unfolding PCNCS
      by fast
    thus ?thesis
      using scbrt[of "NCPC s'" b m n lis "NCS s'" "NCS s" tr']
        aa2 T a1
      unfolding PCNCS[symmetric] RNCS[symmetric]
      by fastforce
  next
    assume F: "\<not> b (NCRS s')"
    have "PC (NCS s) = n"
      using F a1 unfolding PCNCS
      by fast
    thus ?thesis
      using scbrf[of "NCPC s'" b m n lis "NCS s'" "NCS s" tr']
        aa2 F a1
      unfolding PCNCS[symmetric] RNCS[symmetric]
      by fastforce
  qed
next
  fix evf :: "('s, 'e) event_fun" and sf :: "('s, 'e) state_fun"
  assume
    aa1: "(tr, s, X) \<in> \<lbrakk>(one (NCPC s') (comm evf sf))\<rbrakk>{(tr', s', X')}" and
    aa2: "(NCPC s', comm evf sf) \<in> lis"
  have aa1: "tr = tr' \<and> N s' \<and> X \<subseteq> \<Sigma> - evf (NCRS s') \<and> s = NCC s' \<or>
  N s' \<and>
  N s \<and>
  (\<exists>a. a \<in> evf (NCRS s') \<and>
         NCPC s = Suc (NCPC s') \<and>
         NCRS s = sf (NCRS s') a \<and> tr = tr' @ [a] \<and> X \<subseteq> \<Sigma>)"
    using aa1[simplified] c
    by meson
  show "((tr', NCS s'), tr, NCS s) \<in> smallstep lis"
    using aa1
  proof
    assume aa1: "N s' \<and> N s \<and>
      (\<exists>a. a \<in> evf (NCRS s') \<and> NCPC s = Suc (NCPC s') \<and> NCRS s = sf (NCRS s') a \<and> tr = tr' @ [a] \<and> X \<subseteq> \<Sigma>)"    
    obtain a where
      ao1: "a \<in> evf (NCRS s') \<and> NCPC s = Suc (NCPC s') \<and>
           NCRS s = sf (NCRS s') a \<and> tr = tr' @ [a] \<and> X \<subseteq> \<Sigma>"
      using aa1
      by blast
    show ?thesis
      using scomm[of "NCPC s'" evf sf lis "a" "NCS s'" tr tr' "NCS s"]
        aa2 ao1
      unfolding RNCS[symmetric] PCNCS[symmetric]
      by simp
  next
    assume aa1: "tr = tr' \<and> N s' \<and> X \<subseteq> \<Sigma> - evf (NCRS s') \<and> s = NCC s'"
    with c NCCNotNormal[of s'] show ?thesis
      by blast
  qed
qed

text {* Executing a smallstep can be related to executing that single instruction in the SF world. *}

lemma smallstepImpliesSfsem:
  assumes
    a: "((tr', s'), tr, t) \<in> smallstep(lis)" and
    (* b: "(PC s', i) \<in> lis" and *)
    c: "s = normal t" and (* otherwise we do not progress - in contrast to smallstep *)
    d: "X \<subseteq> \<Sigma>"
  shows "\<exists> i . (PC s', i) \<in> lis \<and> (tr, s, X) \<in> sfsem (one (PC s') i) {(tr', normal s', X')}"
  using a c d
proof (induct rule: smallstep.induct, auto) qed

text {* Executing an instruction path can be related to executing a multistep. *}

theorem sfsem_exec_implies_multistep: 
  assumes 
    a: "(tr, s, X) \<in> sfsem_exec lil {(tro, so, Xo)}" and
    b: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    c: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis"
  shows
    "((tro, NCS so), \<union>\<^sub>s\<^sub>c c, (tr, NCS s)) \<in> multistep"
using a b
proof (induct lil arbitrary: tr s X)
  fix tr s X
  assume
    aa1: "(tr, s, X) \<in> sfsem_exec [] {(tro, so, Xo)}" and
    aa2: "set [] \<subseteq> \<union>\<^sub>s\<^sub>c c"
  have allEq: "tr = tro \<and> s = so \<and> X = Xo"
    using aa1[unfolded sfsem_exec_empty]
    by fast
  show "((tro, NCS so), \<union>\<^sub>s\<^sub>c c, tr, NCS s) \<in> multistep"
    unfolding allEq[THEN cj3_1] allEq[THEN cj3_2]
    using multistep.Mterm[OF c]
    .
next
  fix a lil tr s X
  assume
    IH: "\<And>tr s X.
           (tr, s, X) \<in> sfsem_exec lil {(tro, so, Xo)} \<Longrightarrow>
           set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<Longrightarrow>
           ((tro, NCS so), \<union>\<^sub>s\<^sub>c c, tr, NCS s) \<in> multistep" and
    aa1: "(tr, s, X) \<in> sfsem_exec (a # lil) {(tro, so, Xo)}" and
    aa2: "set (a # lil) \<subseteq> \<union>\<^sub>s\<^sub>c c"
  from aa2 have a1: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c"
    by simp
  obtain l i where ao1: "a = (l, i)" by fastforce
  obtain tr' s' X' where
    ao2: "(tr, s, X) \<in> \<lbrakk>one l i\<rbrakk> {(tr', s', X')} \<and>
      (tr', s', X') \<in> sfsem_exec lil {(tro, so, Xo)} \<and>
     (tr = tr' \<and> s = s' \<and> X = X' \<or>
      NCPC s' \<in> definitions.dom (one l i) \<and> N s')"
    using sfsem_predecessor[OF aa1 [unfolded ao1 sfsem_exec_unfold]]
    by blast
  consider (eq) "(tr', s', X') = (tr, s, X)" | (neq) "(tr, s, X) \<noteq> (tr', s', X')" by blast
  thus "((tro, NCS so), \<union>\<^sub>s\<^sub>c c, tr, NCS s) \<in> multistep"
  proof (cases)
    case eq
    show b1: "?thesis"
      using IH[OF ao2[unfolded eq, THEN cj3_2] a1]
      .
  next
    case neq
    show ?thesis
    proof (cases "N s")
      assume Ns: "N s"
      have b1: "(l, i) \<in> \<union>\<^sub>s\<^sub>c c"
        using aa2[unfolded ao1]
        by force
      moreover
      have "PC (NCS s') = l"
        using ao2[THEN cj3_3] neq
        unfolding PCNCS dom.one singleton_iff
        by fastforce
      moreover
      have "((tr', NCS s'), tr, NCS s) \<in> Smallstep"
        unfolding Stos
        using calculation[unfolded PCNCS] c 
         Ns sfsemImpliesSmallstep[OF ao2[THEN cj2_1, unfolded calculation(2)[symmetric] PCNCS]]
        by blast
      ultimately
      show ?thesis
        using IH[OF ao2[THEN cj3_2] a1]
          multistep_seq_other_end[OF c, of l i "NCS s'" tr' tr "NCS s" tro "NCS so"]
        by blast
    next
      assume notNs: "\<not> N s"
      obtain evf sf where
        co1: "i = comm evf sf \<and>
        X \<subseteq> \<Sigma> - evf (NCRS s) \<and> NCPC s = l"
        using sfsem_C_from_comm[OF ao2[THEN cj2_1] neq[unfolded singleton_iff[symmetric]] notNs]
        by blast
      have s'stuff: "NCPC s' = l \<and> N s'"
        using neq ao2
        unfolding dom.one
        by blast
      have c1: "s = NCC s'"
        using ao2[THEN cj2_1, unfolded co1[THEN cj2_1] sfcomm]
          neq
        unfolding SClean_def
        using UnE case_prod_conv mem_Collect_eq notNs prod.inject singletonD
        by auto
      have G1: "NCS s = NCS s'"
        unfolding c1
        by (simp add: NCS_NCC)
      have G2: "tr = tr'"
        using ao2[THEN cj2_1, unfolded co1[THEN cj2_1] sfcomm SClean_def]
        s'stuff notNs
        by fast
      show ?thesis
        using IH[OF ao2[THEN cj3_2] a1, unfolded G2[symmetric] G1[symmetric]]
        .
    qed
  qed
qed

text {* Executing a multistep can be related to executing an instruction path. We also find some
refusal set.*}
    
theorem multistep_implies_sfsem_exec: 
  assumes 
    a: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr, s)) \<in> multistep" and
    c: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis"
  shows "\<exists> lil X . (tr, normal s, X) \<in> sfsem_exec lil {(tr', normal s', X')} \<and> set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<and>
         (lil = [] \<longrightarrow> X = X') \<and> ((tr, s) \<noteq> (tr', s') \<longrightarrow> X \<subseteq> \<Sigma>)"
proof -
  obtain k where
    k: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr, s), k) \<in> multistep_k"
    using multistep_implies_multistep_k[OF a]
    by blast
  have a1: "\<exists> lil X . (tr, normal s, X) \<in> sfsem_exec lil {(tr', normal s', X')} \<and> set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<and>
    (lil = [] \<longrightarrow> X = X') \<and> ((tr, s) \<noteq> (tr', s') \<longrightarrow> X \<subseteq> \<Sigma>) \<and> length lil = k"
    using k 
  proof (induct k arbitrary: tr s rule:nat_induct)
    fix tr s
    assume
      ba2: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr, s), 0) \<in> multistep_k"
    have b1: "tr' = tr \<and> s' = s" using multistep_k_0_same[OF ba2]
      by presburger
    hence b2: "(tr, s) = (tr', s')" by blast
    have b3: "(tr, normal s, X') \<in> sfsem_exec [] {(tr', normal s', X')}"
      unfolding sfsem_exec_empty
      using b1
      by fastforce
    show "\<exists>lil X. (tr, normal s, X) \<in> sfsem_exec lil {(tr', normal s', X')} \<and>
            set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<and> (lil = [] \<longrightarrow> X = X') \<and> ((tr, s) \<noteq> (tr', s') \<longrightarrow> X \<subseteq> \<Sigma>) \<and> length lil = 0"
      using b3 b2 empty_set empty_subsetI length_greater_0_conv neq0_conv
      by force
  next
    fix n tr s
    assume
      ba1: "\<And> tr s . ((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr, s), n) \<in> multistep_k \<Longrightarrow>
          \<exists>lil X. (tr, normal s, X) \<in> sfsem_exec lil {(tr', normal s', X')} \<and>
                  set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<and> (lil = [] \<longrightarrow> X = X') \<and> ((tr, s) \<noteq> (tr', s') \<longrightarrow> X \<subseteq> \<Sigma>) \<and> length lil = n" and
      ba2: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr, s), Suc n) \<in> multistep_k"
    obtain tra sa trb sb where
      bo1: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tra, sa), n) \<in> multistep_k" and
      bo2: "((tra, sa), trb, sb) \<in> Smallstep" and
      bo3: "((trb, sb), \<union>\<^sub>s\<^sub>c c, (tr, s), 0) \<in> multistep_k" and
      bo4: "indom (PC sa) (\<union>\<^sub>s\<^sub>c c)" and
      bo5: "indom (PC s) (\<union>\<^sub>s\<^sub>c c) \<longrightarrow> indom (PC sb) (\<union>\<^sub>s\<^sub>c c)"
      using multistep_k_split2[OF ba2 c, of n 0, unfolded add_0_right]
      by blast
    have b1: "trb = tr \<and> sb = s" using multistep_k_0_same[OF bo3]
      by blast
    obtain lil X where 
      bo6: "(tra, normal sa, X) \<in> sfsem_exec lil {(tr', normal s', X')}" and
      bo7: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
      bo8: "lil = [] \<longrightarrow> X = X'" and
      bo9: "(tra, sa) \<noteq> (tr', s') \<longrightarrow> X \<subseteq> \<Sigma>" and
      bo10: "length lil = n"
      using ba1[OF bo1]
      by blast
    obtain ins X'' where b2: "(PC sa, ins) \<in> lis" and
      bo13: "(tr, normal s, X'') \<in> \<lbrakk>(one (PC sa) ins)\<rbrakk>{(tra, normal sa, X)}" and
      bo14: "X'' \<subseteq> \<Sigma>"
      using smallstepImpliesSfsem[OF bo2[unfolded Stos], of "normal sb" _ X] 
      using bo4 unfolding indom_def
      unfolding b1[THEN cj2_1] b1[THEN cj2_2]
      by blast
    obtain lil' where bo12: "lil' = (PC sa, ins) # lil"
      by blast
    have "(tr, normal s, X'') \<in> sfsem_exec lil' {(tr', normal s', X')}"
      using sfsem_monotone_single[OF bo6 bo13]
      unfolding sfsem_exec_unfold[symmetric] bo12[symmetric]
      .
    moreover
    have "set lil' \<subseteq> \<union>\<^sub>s\<^sub>c c"
      unfolding bo12
      using bo2 bo7
    proof -
      obtain ii :: "(nat \<times> ('s, 'e) instr) set \<Rightarrow> nat \<Rightarrow> ('s, 'e) instr" where
        f1: "(PC sa, ii (\<union>\<^sub>s\<^sub>c c) (PC sa)) \<in> \<union>\<^sub>s\<^sub>c c"
        by (meson bo4 indom_def)
      then have "ii (\<union>\<^sub>s\<^sub>c c) (PC sa) = ins"
        by (meson b2 c sameinstruction subset_iff)
      then show "set ((PC sa, ins) # lil) \<subseteq> \<union>\<^sub>s\<^sub>c c"
        using f1 bo7 by auto
    qed
    moreover
    have "lil' = [] \<longrightarrow> X'' = X'"
      unfolding bo12 by simp
    moreover
    have "(tr, s) \<noteq> (tr', s') \<longrightarrow> X'' \<subseteq> \<Sigma>"
      using bo14 by blast
    moreover
    have "length lil' = Suc n"
      unfolding bo12
      using bo10
      by auto
    ultimately
    show "\<exists>lil X. (tr, normal s, X) \<in> sfsem_exec lil {(tr', normal s', X')} \<and>
                 set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<and> (lil = [] \<longrightarrow> X = X') \<and> ((tr, s) \<noteq> (tr', s') \<longrightarrow> X \<subseteq> \<Sigma>) \<and> length lil = Suc n"
      by blast
  qed
  thus ?thesis
    by auto
qed

text {* Executing something in the SF world can be related to a multistep. *}

theorem sfsem_implies_multistep: 
  assumes 
    a: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>({(tro, so, Xo)})" and
    b: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis" and
    c: "wff\<^sub>s\<^sub>c c"
  shows
    "((tro, NCS so), \<union>\<^sub>s\<^sub>c c, (tr, NCS s)) \<in> multistep"
using sfsem_exec1[OF a] sfsem_exec_implies_multistep b c
  by force

text {* Executing something in the SF world can be related to an exec step. *}

theorem sfsem_implies_exec: 
  assumes 
    a: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>({(tro, so, Xo)})" and
    b: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis" and
    c: "wff\<^sub>s\<^sub>c c" and
    d: "\<not> indom (NCPC s) (\<union>\<^sub>s\<^sub>c c)"
  shows
    "((tro, NCS so), \<union>\<^sub>s\<^sub>c c, (tr, NCS s)) \<in> Exec"
by (metis PCNCS a b c d notIndomIsStuck sfsem_implies_multistep)

text {* Executing an exec step can be related to an execution in the SF world. *}

theorem Exec_implies_sfsem:
  assumes
    a: "(\<union>\<^sub>s\<^sub>c c \<subseteq> lis)" and
    b: "(wff\<^sub>s\<^sub>c c)" and
    c: "(tr', normal s', X')\<in>S" and
    d: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr, s)) \<in> Exec" and (* so, (tr, s) lies outside of code, thus is a normal state / terminal failure *)
    e: "(tr', s') \<noteq> (tr, s)" and (* otherwise, we have not moved, as then s' also is out of dom *)
    X'S: "X \<subseteq> \<Sigma>"
  shows "(tr, normal s, X) \<in> \<lbrakk>c\<rbrakk>(S)"
proof -
  obtain k where
    ao1: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr, s), k) \<in> Exec_k"
    using Exec_implies_Exec_k d by blast
  have kNot0: "k \<noteq> 0"
    using Exec_k_0_same[of tr' s' "\<union>\<^sub>s\<^sub>c c" tr s] e ao1
    by fast
  show "(tr, normal s, X) \<in> \<lbrakk>c\<rbrakk>(S)"
  proof (cases "k = 1")
    assume k1: "k = 1"
    hence kSuc0: "k = Suc 0" by auto
    obtain tra sa where
      bo1: "((tr', s'), tra, sa) \<in> Smallstep \<and>
        ((tra, sa), \<union>\<^sub>s\<^sub>c c, (tr, s), 0) \<in> Exec_k \<and> \<not> indom (PC sa) (\<union>\<^sub>s\<^sub>c c)"
      using Exec_k_split[OF ao1[unfolded kSuc0]]
      by blast
    have b1: "(tra, sa) = (tr, s)"
      using Exec_k_0_same[OF bo1[THEN cj3_2]]
      by blast
    obtain ins where bo2: "(PC s', ins) \<in> lis" and b2: "(tra, normal sa, X) \<in> \<lbrakk>(one (PC s') ins)\<rbrakk>{(tr', normal s', X')}"
      using smallstepImpliesSfsem[OF bo1[THEN cj2_1, unfolded Stos] _ X'S, of "normal sa" X']
      unfolding indom_def
      by auto
    hence bo2': "(PC s', ins) \<in> \<union>\<^sub>s\<^sub>c c"
      using a
      using One_nat_def SFsemSigma_axioms SFsemSigma_def ao1 indomSC_element k1 lessI liSet.Exec_k_indom
      by fastforce
    have b3: "\<not> (NCPC (normal sa) \<in> dom c \<and> N (normal sa))"
      using bo1
      by (simp add: Exec_lastPC dom_eq_indom)
    have "(tra, normal sa, X) \<in> \<lbrakk>c\<rbrakk>\<lbrakk>(one (PC s') ins)\<rbrakk>{(tr', normal s', X')}"
      using pc_does_not_point_to_single[OF b3 b2] .
    hence "(tra, normal sa, X) \<in> \<lbrakk>c\<rbrakk>{(tr', normal s', X')}"
      using sfsem_code_absorption_right_one[of "PC s'" ins c "{(tr', normal s', X')}"] bo2' b
      by blast
    thus "(tr, normal s, X) \<in> \<lbrakk>c\<rbrakk>(S)"
      using c Pair_inject[of tra sa tr s] b1 sfsem_monotone_single[of "(tr', normal s', X')" S " (tra, normal sa, X)" c]
      by fast
  next
    assume ba1: "k \<noteq> 1"
    obtain k' where kSuc: "k = Suc k'" and k'Gt0: "k' > 0"
      using ba1 One_nat_def kNot0 lessE neq0_conv
      by (metis)
    obtain tra sa where
      bo1: "((tr', s'), tra, sa) \<in> Smallstep \<and>
        ((tra, sa), \<union>\<^sub>s\<^sub>c c, (tr, s), k') \<in> Exec_k \<and> indom (PC sa) (\<union>\<^sub>s\<^sub>c c)"
      using Exec_k_split[OF ao1[unfolded kSuc]] k'Gt0
      by blast
    obtain ins where bo2: "(PC s', ins) \<in> lis" and b2: "(tra, normal sa, X) \<in> \<lbrakk>(one (PC s') ins)\<rbrakk>{(tr', normal s', X')}"
      using smallstepImpliesSfsem[OF bo1[THEN cj2_1, unfolded Stos] _ X'S, of "normal sa" X']
      using Exec_k_indom[OF ao1] kSuc
      unfolding indom_def
      by auto
    hence bo2': "(PC s', ins) \<in> \<union>\<^sub>s\<^sub>c c"
      using a
      using One_nat_def SFsemSigma_axioms SFsemSigma_def ao1 indomSC_element kSuc lessI liSet.Exec_k_indom
      by fastforce
    have b3: "\<not> (NCPC (normal s) \<in> dom c \<and> N (normal s))"
      using Exec_lastPC[OF d] dom_eq_indom
      by fastforce
    have a1: "\<exists> lil . set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<and>
      (tr, normal s, X) \<in> sfsem_exec lil {(tra, normal sa, X)}"
      using bo1[THEN cj3_2] a b
    proof (induct k' arbitrary:)
      fix instr_set l ins s trs trs' s' trt t k X'
      assume
        aa1: "instr_set \<subseteq> lis" and
        aa2: "(l, ins) \<in> instr_set \<and> PC s = l \<and>
         ((trs, s), trs', s') \<in> Smallstep \<and>
         ((trs', s'), instr_set, (trt, t), k) \<in> Exec_k \<and>
         (instr_set \<subseteq> lis \<longrightarrow> wff\<^sub>s\<^sub>c c \<longrightarrow>
          (\<exists>lil. set lil \<subseteq> instr_set \<and>
              (trt, normal t, X) \<in> sfsem_exec lil {(trs', normal s', X)}))" and
        aa3: "instr_set \<subseteq> lis" and
        aa4: "wff\<^sub>s\<^sub>c c"
      obtain X'' where  bo1: "X'' \<subseteq> \<Sigma>" by blast
      obtain lil where
        bo2: "set lil \<subseteq> instr_set \<and>
              (trt, normal t, X) \<in> sfsem_exec lil {(trs', normal s', X)}"
        using aa1 aa2 aa3 aa4 bo1 X'S
        by blast
      obtain ins' where pcsinlis: "(PC s, ins') \<in> lis" and b1: "(trs', normal s', X) \<in> \<lbrakk>(one (PC s) ins')\<rbrakk>{(trs, normal s, X)}"
        using smallstepImpliesSfsem[OF aa2[THEN cj4_3, unfolded Stos], of "normal s'" X, simplified, OF X'S]
        by blast
      have ininstr_set: "(l, ins') \<in> instr_set" using aa2 [THEN cj3_2]
        using aa2 aa3 pcsinlis sameinstruction by blast
      have hlpr: "{(trs', normal s', X)} \<union> \<lbrakk>(one (PC s) ins')\<rbrakk>{(trs, normal s, X)} = \<lbrakk>(one (PC s) ins')\<rbrakk>{(trs, normal s, X)}"
        using b1
        by blast
      hence hlpr: "(trt, normal t, X) \<in> sfsem_exec lil (\<lbrakk>(one (PC s) ins')\<rbrakk>{(trs, normal s, X)}) = ((trt, normal t, X) \<in> sfsem_exec lil ({(trs', normal s', X)} \<union> \<lbrakk>(one (PC s) ins')\<rbrakk>{(trs, normal s, X)}))" by presburger
      have b2: "(trt, normal t, X) \<in> sfsem_exec (lil @ [(PC s, ins')]) {(trs, normal s, X)}"
        unfolding sfsem_exec_unfold2[of lil "[(PC s, ins')]" "{(trs, normal s, X)}"]
        unfolding sfsem_exec_unfold sfsem_exec_empty
        unfolding hlpr
        unfolding sfsem_exec_additive[of lil "{(trs', normal s', X)}" "\<lbrakk>(one (PC s) ins')\<rbrakk>{(trs, normal s, X)}"]
        using bo2 by blast
      moreover
      have "set (lil @ [(PC s, ins')]) \<subseteq> instr_set" using bo2 ininstr_set aa2
        by auto
      ultimately
      show "\<exists>lil. set lil \<subseteq> instr_set \<and>
            (trt, normal t, X) \<in> sfsem_exec lil {(trs, normal s, X)}"
        using bo2[THEN cj2_2]
        by blast
    next
      show "\<And>instr_set s tr.
       instr_set \<subseteq> lis \<Longrightarrow>
       \<not> indom (PC s) instr_set \<Longrightarrow>
       instr_set \<subseteq> lis \<Longrightarrow>
       wff\<^sub>s\<^sub>c c \<Longrightarrow>
       \<exists>lil. set lil \<subseteq> instr_set \<and> (tr, normal s, X) \<in> sfsem_exec lil {(tr, normal s, X)}"
        using sfsem_exec_empty
        by (metis empty_set empty_subsetI singletonI)
    qed
    obtain lil where
      ao2: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<and> (tr, normal s, X) \<in> sfsem_exec lil {(tra, normal sa, X)}"
      using a1
      by blast
    moreover
    have "\<not> (NCPC (normal s) \<in> dom c \<and> N (normal s))"
      using d
      by (simp add: Exec_lastPC dom_eq_indom)
    ultimately
    show "(tr, normal s, X) \<in> \<lbrakk>c\<rbrakk>S"
      using sfsem_exec2 b
    proof -
      have f1: "(tra, normal sa, X) \<in> \<lbrakk>(one (PC s') ins)\<rbrakk>S"
        using b2 c sfsem_monotone_single by blast
      have f2: "\<And>P. \<lbrakk>c\<rbrakk>\<lbrakk>(one (PC s') ins)\<rbrakk>P = \<lbrakk>c\<rbrakk>P"
        by (simp add: b bo2' sfsem_code_absorption_right_one)
      have f3: "N (normal s) \<or> (tr, normal s, X) \<in> \<lbrakk>c\<rbrakk>S"
        by simp
      have "NCPC (normal s) \<in> definitions.dom c \<or> (tr, normal s, X) \<in> \<lbrakk>c\<rbrakk>{(tra, normal sa, X)}"
        using ao2 b sfsem_exec2 by blast
      then show ?thesis
        using f3 f2 f1 b3 sfsem_monotone_single by blast
    qed
  qed
qed

text {* We can find any trace expansion in the SF semantics and the according @{term comm} instruction. *}

theorem multistep_trace_implies_comm_failure:
  assumes
    a: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr' @ r @ [a] @ l, s)) \<in> multistep" and
    b: "(tr', normal s', X') \<in> S" and
    c: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis" and
    wff: "wff\<^sub>s\<^sub>c c"
  shows
    "\<exists> sx X evf sf . (tr' @ r, communication sx, X) \<in> \<lbrakk>c\<rbrakk>S \<and>
    (PC sx, comm evf sf) \<in> \<union>\<^sub>s\<^sub>c c \<and>
    X \<subseteq> \<Sigma> - evf (R sx) \<and>
    ((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr' @ r, sx)) \<in> multistep \<and>
    ((tr' @ r, sx), \<union>\<^sub>s\<^sub>c c, (tr' @ r @ [a] @ l, s)) \<in> multistep \<and>
    indom (PC sx) (\<union>\<^sub>s\<^sub>c c)"
proof -
  obtain k where k: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr' @ r @ [a] @ l, s), k) \<in> multistep_k"
    using multistep_implies_multistep_k[OF a]
    by blast
  obtain m n sa sa' where
    ao1: "Suc (m + n) = k" and
    ao2: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr' @ r, sa), m) \<in> multistep_k" and
    ao3: "((tr' @ r, sa), tr' @ r @ [a], sa') \<in> Smallstep" and
    ao4: "((tr' @ r @ [a], sa'), \<union>\<^sub>s\<^sub>c c, (tr' @ r @ [a] @ l, s), n) \<in> multistep_k" and
    ao5: "indom (PC sa) (\<union>\<^sub>s\<^sub>c c)" and
    ao6: "0 < n \<or> indom (PC s) (\<union>\<^sub>s\<^sub>c c) \<longrightarrow> indom (PC sa') (\<union>\<^sub>s\<^sub>c c)"
    using multistep_k_prefix_closed'[OF c k]
    by blast
  obtain ins where ao7: "(PC sa, ins) \<in> \<union>\<^sub>s\<^sub>c c" using ao5
    unfolding indom_def by blast
  with c have a1: "(PC sa, ins) \<in> lis" by blast
  obtain evf sf where
    ao10: "ins = comm evf sf"
    using Smallstep_rev a1 ao3 lis_wellformed by blast
  obtain lil Xm where
    ao11: "(tr' @ r, normal sa, Xm) \<in> sfsem_exec lil {(tr', normal s', X')}" and
    ao12: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    ao13: "lil = [] \<longrightarrow> Xm = X'" and
    ao14: "(tr' @ r, sa) \<noteq> (tr', s') \<longrightarrow> Xm \<subseteq> \<Sigma>"
    using multistep_implies_sfsem_exec[OF multistep_k_implies_multistep[OF ao2] c, of X']
    by auto
  obtain X where
    ao8: "X \<subseteq> \<Sigma> - evf (R sa)" and
    ao9: "(tr' @ r, communication sa, X) \<in>
            \<lbrakk>(one (PC sa) ins)\<rbrakk>{(tr' @ r, normal sa, Xm)}"
    unfolding ao10 sfcomm
    by auto
  have a2: "((tr' @ r, sa), \<union>\<^sub>s\<^sub>c c, (tr' @ r @ [a] @ l, s)) \<in> multistep"
    by (meson ao3 ao4 ao7 c multistep.Mseq multistep_k_implies_multistep)
  have a3: "(tr' @ r, communication sa, X) \<in> sfsem_exec ((PC sa, ins) # lil) {(tr', normal s', X')}"
    using ao11 ao9
    unfolding sfsem_exec_unfold
    using sfsem_monotone_single
    by blast
  have a4: "(tr' @ r, communication sa, X) \<in> \<lbrakk>c\<rbrakk>{(tr', normal s', X')}"
    using sfsem_exec2[OF _ a3 _ wff]
      ao7 ao12
    by simp
  show ?thesis
    using a2 ao10 ao5 ao7 ao8
      multistep_k_implies_multistep[OF ao2] sfsem_monotone_single[OF b a4]
    by blast
qed

paragraph {* Commutativity *}

text {*
  This is the ultimate formulation of commutativity. They only have to contain the same
  @{text "label \<times> instr"} pairs, and be wellformed. This is sufficient for commutativity - any other
  structural information implied by the tree structure is irrelevant.
*}
theorem sfsemCommutative:
  assumes
    wffa: "wff\<^sub>s\<^sub>c (a)" and
    wffb: "wff\<^sub>s\<^sub>c (b)" and
    same_instrs: "\<union>\<^sub>s\<^sub>c (a) = \<union>\<^sub>s\<^sub>c (b)"
  shows "\<lbrakk>a\<rbrakk>S = \<lbrakk>b\<rbrakk>S"
  using wffa same_instrs
proof (induct a)
  fix l i
  assume
    aa1: "wff\<^sub>s\<^sub>c (one l i)" and
    aa2: "\<union>\<^sub>s\<^sub>c (one l i) = \<union>\<^sub>s\<^sub>c b"
  have bIsOne: "b = one l i"
    using aa2[unfolded sc_inj.one] sc_inj_singleton_equal[of b l i] wffb
    by simp
  show "\<lbrakk>one l i\<rbrakk>S = \<lbrakk>b\<rbrakk>S"
    unfolding bIsOne ..
next
  fix l r :: "('s, 'e) structured_code"
  assume
    wfflr: "wff\<^sub>s\<^sub>c (l \<oplus> r)" and
    same_instrs: "\<union>\<^sub>s\<^sub>c (l \<oplus> r) = \<union>\<^sub>s\<^sub>c b"
  show "\<lbrakk>(l \<oplus> r)\<rbrakk>S = \<lbrakk>b\<rbrakk>S"
    using same_instrs wffb
  proof (induct b)
    fix lbl i
    assume ba1: "\<union>\<^sub>s\<^sub>c (l \<oplus> r) = \<union>\<^sub>s\<^sub>c (one lbl i)"
    from ba1 have False
      using wfflr sc_inj_num[of l r "(lbl, i)"]
      unfolding sc_inj.one
      by blast
    thus "\<And>x1 x2.
       \<union>\<^sub>s\<^sub>c (l \<oplus> r) = \<union>\<^sub>s\<^sub>c (one x1 x2) \<Longrightarrow> wff\<^sub>s\<^sub>c (one x1 x2) \<Longrightarrow> \<lbrakk>(l \<oplus> r)\<rbrakk>S = \<lbrakk>(one x1 x2)\<rbrakk>S"
       by auto
  next
    fix p q :: "('s, 'e) structured_code"
    assume
      b: "wff\<^sub>s\<^sub>c (p \<oplus> q)" and
      c: "\<union>\<^sub>s\<^sub>c (l \<oplus> r) = \<union>\<^sub>s\<^sub>c (p \<oplus> q)"
    {
      fix x :: "('s, 'e) failure"
      assume inLr: "x \<in> \<lbrakk>(l \<oplus> r)\<rbrakk>S"
      obtain lil where
        subs: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c (l \<oplus> r)" and
        exec_lil: "x \<in> sfsem_exec lil S"
        using sfsem_exec1'[OF inLr]
        by blast
      from subs have a1: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c (p \<oplus> q)"
        using c
        by blast
      from inLr have "\<not>(NCPC (fst (snd x)) \<in> dom (l \<oplus> r) \<and> N (fst (snd x)))"
        by auto
      hence a2: "\<not>(NCPC (fst (snd x)) \<in> dom (p \<oplus> q) \<and> N (fst (snd x)))"
        using c indom_sc_eq_imp_indom_dom_eq[of "l \<oplus> r" "p \<oplus> q"]
        by auto
      have "x \<in> \<lbrakk>(p \<oplus> q)\<rbrakk>S"
        using sfsem_exec2'[OF a1, OF exec_lil, OF a2, OF b] .
    }
    moreover
    {
      fix x :: "('s, 'e) failure"
      assume inPq: "x \<in> \<lbrakk>(p \<oplus> q)\<rbrakk>S"
      obtain lil where
        subs: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c (p \<oplus> q)" and
        exec_lil: "x \<in> sfsem_exec lil S"
        using sfsem_exec1'[OF inPq]
        by blast
      from subs have a1: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c (l \<oplus> r)"
        using c
        by blast
      from inPq have "\<not>(NCPC (fst (snd x)) \<in> dom (p \<oplus> q) \<and> N (fst (snd x)))"
        by auto
      hence a2: "\<not>(NCPC (fst (snd x)) \<in> dom (l \<oplus> r) \<and> N (fst (snd x)))"
        using c indom_sc_eq_imp_indom_dom_eq[of "l \<oplus> r" "p \<oplus> q"]
        by auto
      have "x \<in> \<lbrakk>(l \<oplus> r)\<rbrakk>S"
        using sfsem_exec2'[OF a1, OF exec_lil, OF a2, OF wfflr] .
    }
    ultimately
    show "\<lbrakk>(l \<oplus> r)\<rbrakk>S = \<lbrakk>(p \<oplus> q)\<rbrakk>S"
      by blast
  qed
qed

paragraph {*General Code Absorption*}

text {* Code absorption is shown via @{text sfsemCommutative}. The lemma will be: @{text "\<lbrakk>big\<rbrakk>\<lbrakk>small\<rbrakk>S = \<lbrakk>big\<rbrakk>S"}. The overall idea is: If the "smaller" part is actually as big as
the bigger part, then we can directly apply @{text sfsemCommutativeX}. Otherwise, we remove from the bigger part the smaller code and get some remaining code @{text other}, such that we can
replace @{text "\<lbrakk>big\<rbrakk>"} with @{text "\<lbrakk>other \<oplus> small\<rbrakk>"} and hence use the restricted version of code absorption.*}

context
begin

text {* We therefor define first the removal of code. @{text removeCode} takes some structured code @{text c} and a labeled instruction set @{text s}, and removes all instructions in @{text s} out of @{text c}.
  @{text removeCode} assumes, that there is at least a single instruction in @{text c} which is not part of @{text s}.
*}

text {* Note: Code removal is not required anywhere outside, thus keep in private to this scope. *}

private fun removeCode :: "('s, 'e) structured_code \<Rightarrow> ('s, 'e) labeled_instruction_set \<Rightarrow> ('s, 'e) structured_code" where
  "removeCode (one l i) s = one l i" |
  --"Three possibilities: @{text l} and @{text r} contain remaining instructions, or either @{text l} or @{text r}."
  "removeCode (l \<oplus> r) s = (if \<not>(\<union>\<^sub>s\<^sub>c l \<subseteq> s) \<and> \<not>(\<union>\<^sub>s\<^sub>c r \<subseteq> s) then ((removeCode l s) \<oplus> (removeCode r s)) else
            if \<not>(\<union>\<^sub>s\<^sub>c l \<subseteq> s) \<and> (\<union>\<^sub>s\<^sub>c r \<subseteq> s) then (removeCode l s) else (removeCode r s) )"

text {* This is a correctness check of the definition: All instructions from @{text s} are removed, and only instructions from @{text s} are removed. *}

private lemma removeCode_correct:
  assumes "\<not> (\<union>\<^sub>s\<^sub>c c \<subseteq> s)" --"Some instruction is not in s."
  shows "\<union>\<^sub>s\<^sub>c (removeCode c s) = \<union>\<^sub>s\<^sub>c c - s"
  using assms
proof (induct c, auto) qed

private lemma removeCode_smaller:
  shows "\<union>\<^sub>s\<^sub>c (removeCode c s) \<subseteq> \<union>\<^sub>s\<^sub>c c"
proof (induct c, auto) qed

private lemma removeCode_wff_hlpr:
  assumes a: "definitions.dom c1 \<inter> definitions.dom c2 = {}"
  shows "definitions.dom (removeCode c1 s) \<inter> definitions.dom (removeCode c2 s) = {}"
using assms indom_sc_subset_imp_indom_dom_subset[OF removeCode_smaller]
proof (induct c1, auto) qed

private lemma removeCode_wff:
  assumes a: "wff\<^sub>s\<^sub>c c"
  shows "wff\<^sub>s\<^sub>c (removeCode c s)"
using a removeCode_wff_hlpr
proof (induct c, auto) qed

theorem sfsem_code_absorbtion_right:
  assumes
    wffsmall: "wff\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l" and
    wffbig: "wff\<^sub>s\<^sub>c c\<^sub>b\<^sub>i\<^sub>g" and
    subs: "\<union>\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l \<subseteq> \<union>\<^sub>s\<^sub>c c\<^sub>b\<^sub>i\<^sub>g"
  shows "\<lbrakk>c\<^sub>b\<^sub>i\<^sub>g\<rbrakk>S = \<lbrakk>c\<^sub>b\<^sub>i\<^sub>g\<rbrakk>\<lbrakk>c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l\<rbrakk>S"
using wffbig subs
proof (induct c\<^sub>b\<^sub>i\<^sub>g)
  fix l :: label and i :: "('s, 'e) instr"
  assume
    wffli: "wff\<^sub>s\<^sub>c (one l i)" and
    subs: "\<union>\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l \<subseteq> \<union>\<^sub>s\<^sub>c (one l i)"
  have c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l: "c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l = one l i"
    using sc_inj_singleton_equal[OF subs[unfolded sc_inj.one] wffsmall] .
  show "\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>(one l i)\<rbrakk>\<lbrakk>c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l\<rbrakk>S"
    unfolding c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l
    using sfsem_idem[of "one l i" S, symmetric] .
next
  fix c\<^sub>b\<^sub>i\<^sub>g1 c\<^sub>b\<^sub>i\<^sub>g2 :: "('s, 'e) structured_code"
  assume
    wffbig: "wff\<^sub>s\<^sub>c (c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)" and
    subs: "\<union>\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l \<subseteq> \<union>\<^sub>s\<^sub>c (c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)"
  {
    assume ba1: "\<union>\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l \<subset> \<union>\<^sub>s\<^sub>c (c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)"
    obtain c' :: "('s, 'e) structured_code" where
      bo1: "c' = removeCode (c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2) (\<union>\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l)"
      by blast
    have b1: "wff\<^sub>s\<^sub>c c'"
      using bo1 removeCode_wff wffbig
      by blast
    have b2: "\<union>\<^sub>s\<^sub>c c' = \<union>\<^sub>s\<^sub>c (c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2) - \<union>\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l"
      using removeCode_correct[of "c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2" "\<union>\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l"] 
      unfolding bo1 using ba1
      by fast
    have b2': "\<union>\<^sub>s\<^sub>c c' \<subset> \<union>\<^sub>s\<^sub>c (c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)"
      using b2 sc_inj_not_empty[of c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l]
      using Diff_eq_empty_iff Diff_subset subs double_diff le_imp_less_or_eq by auto
    have b3: "\<union>\<^sub>s\<^sub>c (c' \<oplus> c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l) = \<union>\<^sub>s\<^sub>c (c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)"
      using ba1 b2
      by auto
    have b4: "\<union>\<^sub>s\<^sub>c c' \<inter> \<union>\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l = {}"
      using ba1 b2
      by auto
    have b5: "wff\<^sub>s\<^sub>c (c' \<oplus> c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l)"
      using wff_reconstruction[OF wffbig b2' ba1 b4 b1 wffsmall] .
    have b6: "(\<lbrakk>(c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)\<rbrakk>S = \<lbrakk>(c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)\<rbrakk>\<lbrakk>c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l\<rbrakk>S) = (\<lbrakk>(c' \<oplus> c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l)\<rbrakk>S = \<lbrakk>(c' \<oplus> c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l)\<rbrakk>\<lbrakk>c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l\<rbrakk>S)"
      unfolding sfsemCommutative[OF b5 wffbig b3] ..
    have "\<lbrakk>(c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)\<rbrakk>S = \<lbrakk>(c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)\<rbrakk>\<lbrakk>c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l\<rbrakk>S"
      unfolding b6
      using sfsem_code_absorption_right_2[OF b5, of S]
      by simp
  }
  moreover
  {
    assume ba1: "\<union>\<^sub>s\<^sub>c c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l = \<union>\<^sub>s\<^sub>c (c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)"
    have "\<lbrakk>(c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)\<rbrakk>S = \<lbrakk>(c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)\<rbrakk>\<lbrakk>c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l\<rbrakk>S"
      unfolding sfsemCommutative[OF wffsmall wffbig ba1, symmetric] sfsem_idem
      ..
  }
  ultimately
  show "\<lbrakk>(c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)\<rbrakk>S = \<lbrakk>(c\<^sub>b\<^sub>i\<^sub>g1 \<oplus> c\<^sub>b\<^sub>i\<^sub>g2)\<rbrakk>\<lbrakk>c\<^sub>s\<^sub>m\<^sub>a\<^sub>l\<^sub>l\<rbrakk>S"
    using subs
    by blast
qed

end

subsection {* Trace lemmas *}

text {* Only a comm instruction can enlarge the trace. *}

lemma sfsem_one_trace_append_implies_comm:
  fixes X X'
  assumes "(tr @ [a], s, X) \<in> \<lbrakk>(one lbl i)\<rbrakk>{(tr, s', X')}"
  shows "\<exists> ef sf X . i = comm ef sf \<and> X \<subseteq> \<Sigma> - ef (NCRS s') \<and> (tr, NCC s', X) \<in> \<lbrakk>(one lbl i)\<rbrakk>{(tr, s', X')}"
proof -
  have "\<exists> ef sf . i = comm ef sf"
    using assms
  proof (induct i, auto) qed
  then obtain ef sf where i: "i = comm ef sf"
    by blast
  obtain X where X: "X \<subseteq> \<Sigma> - ef (NCRS s')"
    by blast
  have "NCPC s' = lbl \<and> N s'"
  proof (rule ccontr)
    assume ba1: "\<not> (NCPC s' = lbl \<and> N s')"
    hence "NCPC s' \<noteq> lbl \<or> \<not> N s'"
      by linarith
    then consider (Uneq) "NCPC s' \<noteq> lbl" | (NotN) "\<not> N s'"
      by blast
    thus False
    proof (cases)
      case Uneq
      with assms show False
        using append_self_conv dom.one not_Cons_self2 sfsem_predecessor2 singletonD
        by fastforce
    next
      case NotN
      with assms show False
      by (meson append_self_conv not_Cons_self2 sfsem_predecessor2)
    qed
  qed
  hence "(tr, NCC s', X) \<in> \<lbrakk>(one lbl i)\<rbrakk>{(tr, s', X')}"
    using X unfolding i 
    unfolding sfcomm
    by fastforce
  thus ?thesis
    using i X
    by blast
qed

text {* Each start trace is a prefix of its successor traces.  *}

lemma sfsem_trace_growing:
  assumes "(tr, s, X) \<in> \<lbrakk>c\<rbrakk> {(tr', s', X')}" and "\<union>\<^sub>s\<^sub>c c \<subseteq> lis" and "wff\<^sub>s\<^sub>c c"
  shows "\<exists> l . tr = tr' @ l"
using multistep_implies_multistep_k[OF sfsem_implies_multistep[OF assms]]
  multistep_k_growing_trace[of tr' "NCS s'" "\<union>\<^sub>s\<^sub>c c" tr "NCS s"]
by blast

text {* Each start trace also appears as a successor, if there is at least one successor. *}

lemma sfsem_not_empty_has_same_trace_successor:
  assumes "\<lbrakk>c\<rbrakk> {(tr, s, X)} \<noteq> {}" and clis: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis" and wffc: "wff\<^sub>s\<^sub>c c"
  shows "\<exists> s' X' . (tr, s', X') \<in> \<lbrakk>c\<rbrakk> {(tr, s, X)}"
proof(cases "NCPC s \<in> dom c \<and> N s")
  assume aa1: "NCPC s \<in> dom c \<and> N s"
  hence normalNCSs: "normal (NCS s) = s"
    using Nnormal[of s "NCS s"] by presburger
  obtain trF sF XF where
    ao1: "(trF, sF, XF) \<in> \<lbrakk>c\<rbrakk>{(tr, s, X)}"
    using assms by fast
  obtain l where
    trF: "trF = tr @ l"
    using sfsem_trace_growing[OF ao1 clis wffc]
    by blast
  have a1: "((tr, NCS s), \<union>\<^sub>s\<^sub>c c, trF, NCS sF) \<in> multistep"
    using sfsem_implies_multistep[OF ao1 clis wffc] .
  let ?G = "\<exists>s' X'. (tr, s', X') \<in> \<lbrakk>c\<rbrakk>{(tr, s, X)}"
  {
    assume "trF = tr"
    with ao1 have ?G by blast
  }
  moreover
  {
    --"Proof steps:"
    --"1) Find the small step, where the trace is expanded with the first event"
    --"2) Reconstruct the instruction list leading to that smallstep"
    --"3) Extracting the comm state, which contains the original non-expanded trace."
    --"4) Show, that this comm state also is in the original sfsem."
    assume neq: "trF \<noteq> tr"
    hence lnotEmpty: "l \<noteq> []"
      using trF
      by simp
    then obtain a l' where l: "l = a # l'"
      by (meson neq_Nil_conv)
    obtain k where
      k: "((tr, NCS s), \<union>\<^sub>s\<^sub>c c, (tr @ [] @ [a] @ l', NCS sF), k) \<in> multistep_k"
      using multistep_implies_multistep_k[OF a1] l trF
      by auto
    obtain sa sa' m n where
      sucmn: "Suc (m + n) = k" and
      msm: "((tr, NCS s), \<union>\<^sub>s\<^sub>c c, (tr, sa), m) \<in> multistep_k" and
      smstp: "((tr, sa), tr @ [a], sa') \<in> Smallstep" and
      msn: "((tr @ [a], sa'), \<union>\<^sub>s\<^sub>c c, (tr @ a # l', NCS sF), n) \<in> multistep_k" and
      indom: "indom (PC sa) (\<union>\<^sub>s\<^sub>c c)" and
      stuff: "(0 < n \<longrightarrow> indom (PC sa') (\<union>\<^sub>s\<^sub>c c)) \<and> (indom (PC (NCS sF)) (\<union>\<^sub>s\<^sub>c c) \<longrightarrow> indom (PC sa') (\<union>\<^sub>s\<^sub>c c))"
      using multistep_k_prefix_closed'[OF clis k, simplified]
      by blast
    obtain lil Xa where
      lil: "(tr, normal sa, Xa) \<in> sfsem_exec lil {(tr, s, X)}" and
      lilc: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
      bo1: "lil = [] \<longrightarrow> Xa = X" and
      bo2: "(tr, sa) \<noteq> (tr, NCS s) \<longrightarrow> Xa \<subseteq> \<Sigma>"
      using multistep_implies_sfsem_exec[OF multistep_k_implies_multistep[OF msm] clis, of X, unfolded normalNCSs]
      by auto
    have "PC sa \<in> dom c"
      using msm
      by (simp add: indom dom_eq_indom)
    then obtain i where i: "(PC sa, i) \<in> lis" and succ: "(tr @ [a], normal sa', \<Sigma>) \<in> \<lbrakk>(one (PC sa) i)\<rbrakk>{(tr, normal sa, Xa)}"
      using smallstepImpliesSfsem[OF smstp[unfolded Stos], of "normal sa'" \<Sigma>, simplified]
      by blast
    hence i': "(PC sa, i) \<in> \<union>\<^sub>s\<^sub>c c"
      using clis indom indomSC_element by blast
    obtain Xu where
      Xu: "(tr, NCC (normal sa), Xu) \<in> \<lbrakk>(one (PC sa) i)\<rbrakk>{(tr, normal sa, Xa)}"
      using sfsem_one_trace_append_implies_comm[OF succ]
      by blast
    have notN: "\<not> N (NCC (normal sa))" by auto
    have "(tr, NCC (normal sa), Xu) \<in> \<lbrakk>(one (PC sa) i)\<rbrakk>sfsem_exec lil {(tr, s, X)}"
      using Xu lil sfsem_monotone_single
      by blast
    hence lilex: "(tr, NCC (normal sa), Xu) \<in> sfsem_exec ((PC sa, i) # lil) {(tr, s, X)}"
      unfolding sfsem_exec_unfold[symmetric]
      .
    have ?G  
      using sfsem_exec2[OF _ lilex _ wffc] i' lilc notN
      by auto
  }
  ultimately
  show ?G
    by fast
next
  case False
  thus ?thesis
    using pc_does_not_point_to_single by blast
qed

text {* We can exchange the trace of a starting state with any other trace and find this modified 
trace then again as successor state. *}

lemma sfsem_exec_prefix_exchange:
  assumes
    in_sf: "(tr @ new, s, X) \<in> sfsem_exec lil {(tr, s', X')}" and
    wff: "wff\<^sub>s\<^sub>c c" and
    subs_lis: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis" and
    subs_lil: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c"
  shows "(tr' @ new, s, X) \<in> sfsem_exec lil {(tr', s', X')}"
    using assms
proof (induct lil arbitrary: new s X rule: exec_induct, simp)
  fix l i lil new s X
  assume
    IH: "\<And>  new s X . (tr @ new, s, X) \<in> sfsem_exec lil {(tr, s', X')} \<Longrightarrow>
        wff\<^sub>s\<^sub>c c \<Longrightarrow> \<union>\<^sub>s\<^sub>c c \<subseteq> lis \<Longrightarrow> set lil \<subseteq> \<union>\<^sub>s\<^sub>c c \<Longrightarrow> (tr' @ new, s, X) \<in> sfsem_exec lil {(tr', s', X')}" and
    ex: "(tr @ new, s, X) \<in> sfsem_exec ((l, i) # lil) {(tr, s', X')}" and
    wff: "wff\<^sub>s\<^sub>c c" and
    subs_lis: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis" and
    subs_lil: "set ((l, i) # lil) \<subseteq> \<union>\<^sub>s\<^sub>c c"
  obtain trz sz Xz where
    ex_lil: "(trz, sz, Xz) \<in> sfsem_exec lil {(tr, s', X')}" and
    ex_one: "(tr @ new, s, X) \<in> \<lbrakk>(l :: i)\<rbrakk>{(trz, sz, Xz)}" and
    may_eq: "(tr @ new = trz \<and> s = sz \<and> X = Xz \<or>
      NCPC sz \<in> definitions.dom (l :: i) \<and>
      N sz \<and> (N s \<longrightarrow> X \<subseteq> \<Sigma>) \<and> (\<not> N s \<longrightarrow> (\<exists>ef sf. i = comm ef sf \<and> X \<subseteq> \<Sigma> - ef (NCRS s) \<and> s = NCC sz)))"
    using sfsem_exec_pred1[OF ex] by blast
  show "(tr' @ new, s, X) \<in> sfsem_exec ((l, i) # lil) {(tr', s', X')}"
    using may_eq
  proof rule
    assume eq: "tr @ new = trz \<and> s = sz \<and> X = Xz"
    have "(tr' @ new, s, X) \<in> sfsem_exec lil {(tr', s', X')}"
      using IH ex_lil eq wff subs_lis subs_lil by auto
    hence set_eq: "{(tr' @ new, s, X)} \<union> sfsem_exec lil {(tr', s', X')} = sfsem_exec lil {(tr', s', X')}"
      by blast
    have "(tr' @ new, s, X) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr' @ new, s, X)}"
      using ex_one[unfolded eq[THEN cj3_1, symmetric] eq[THEN cj3_2, symmetric] eq[THEN cj3_3, symmetric]]
      by (induct i, auto)
    thus ?thesis
      using sfsem_union_additive[of "one l i" "{(tr' @ new, s, X)}" "sfsem_exec lil {(tr', s', X')}", unfolded set_eq]
      unfolding sfsem_exec_unfold by blast
  next
    assume
      as: "NCPC sz \<in> definitions.dom (l :: i) \<and> N sz \<and> (N s \<longrightarrow> X \<subseteq> \<Sigma>) \<and> (\<not> N s \<longrightarrow> (\<exists>ef sf. i = comm ef sf \<and> X \<subseteq> \<Sigma> - ef (NCRS s) \<and> s = NCC sz))"
    obtain new' where trz: "trz = tr @ new'"
      using sfsem_exec_implies_multistep[OF ex_lil _ subs_lis] multistep_growing_trace subs_lil
    by force
    have "(tr' @ new', sz, Xz) \<in> sfsem_exec lil {(tr', s', X')}"
      using IH ex_lil wff trz subs_lis subs_lil by auto
    hence set_eq: "{(tr' @ new', sz, Xz)} \<union> sfsem_exec lil {(tr', s', X')} = sfsem_exec lil {(tr', s', X')}"
      by blast
    have "(tr' @ new, s, X) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr' @ new', sz, Xz)}"
      using ex_one trz
      by (induct i, auto)
    thus ?thesis
      using sfsem_union_additive[of "one l i" "{(tr' @ new', sz, Xz)}" "sfsem_exec lil {(tr', s', X')}", unfolded set_eq]
      unfolding sfsem_exec_unfold by blast
  qed
qed

text {* We can exchange the trace of a starting state with any other trace and find this modified 
trace then again as successor state. *}

lemma sfsem_prefix_exchange:
  assumes
    in_sf: "(tr @ new, s, X) \<in> \<lbrakk>c\<rbrakk> {(tr, s', X')}" and
    wff: "wff\<^sub>s\<^sub>c c" and
    subs: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis"
  shows "(tr' @ new, s, X) \<in> \<lbrakk>c\<rbrakk> {(tr', s', X')}"
    using assms
proof (cases "NCPC s \<in> definitions.dom c \<and> N s")
  assume as: "\<not> (NCPC s \<in> definitions.dom c \<and> N s)"
  obtain lil where
    subs_lil: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    ex_lil: "(tr @ new, s, X) \<in> sfsem_exec lil {(tr, s', X')}"
    using sfsem_exec1[OF in_sf] by blast
  have ex_lil: "(tr' @ new, s, X) \<in> sfsem_exec lil {(tr', s', X')}"
    using sfsem_exec_prefix_exchange using subs_lil wff subs ex_lil by blast
  from sfsem_exec2[OF subs_lil ex_lil as wff] show ?thesis .
next
  assume as: "NCPC s \<in> definitions.dom c \<and> N s"
  from sfsem_not_outside_means_cbr[OF in_sf as] show ?thesis
    by force
qed

text {* If there exists some successor, then still some successor will exist if we change the starting
set's trace and refusal set. *}

theorem trace_and_refusal_do_not_influence_existence_of_successors:
  assumes
    not_empty_for_tr: "\<lbrakk>c\<rbrakk> {(tr, s, X)} \<noteq> {}" and
    wff: "wff\<^sub>s\<^sub>c c" and
    subs: "\<union>\<^sub>s\<^sub>c c \<subseteq> lis"
  shows "\<lbrakk>c\<rbrakk>{(tr', s, X')} \<noteq> {}"
proof (auto)
  assume not_in_trX: "\<lbrakk>c\<rbrakk>{(tr', s, X')} = {}"
  obtain trp s' Xp where in_sf: "(trp, s', Xp) \<in> \<lbrakk>c\<rbrakk>{(tr, s, X)}" using not_empty_for_tr by auto
  obtain new where tr': "trp = tr @ new"
    using multistep_growing_trace sfsem_implies_multistep[OF in_sf subs wff] by blast
  obtain Xp' where in_sf: "(tr' @ new, s', Xp') \<in> \<lbrakk>c\<rbrakk>{(tr', s, X')}"
    using sfsem_general_sigma_exchange wff sfsem_prefix_exchange[OF in_sf[unfolded tr'] wff subs, of tr']
    by blast
  thus False
    using  not_in_trX
    by blast
qed

subsection {* Static Code analysis*}

paragraph {* Divergency Free *}

context
begin

text {*
Each reconstructed path is a dense path. We have to take care for the nondeterministic functions in @{term do}
and @{term comm}, as they have to return at least one successor state for each input. Otherwise, the execution
would just stop. Having functions returning no state probably is a niche requirement.
*}

private lemma create_all_possible_execution_paths_for_each_start_there_is_always_an_executable_dense_path':
  assumes
    b: "N sa" and
    --"Required for recursion"
    e: "sfsem_exec_path_is_dense (tra, sa, Xa) visited {(tr', s', X')}" and
    --"Required for recursion"
    f: "visited \<noteq> [] \<longrightarrow> fst (hd visited) \<in> dom c" and
    --"Required for recursion"
    g: "NCPC sa = start" (* otherwise we have no connection to the ongoing list *) and
    j: "\<forall> f . (\<exists> lbl . (lbl, do f) \<in> \<union>\<^sub>s\<^sub>c c) \<longrightarrow> (\<forall> x . f x \<noteq> {})" and
    k: "\<forall> ef . (\<exists> lbl sf . (lbl, comm ef sf) \<in> \<union>\<^sub>s\<^sub>c c) \<longrightarrow> (\<forall> x . ef x \<noteq> {})"
  shows "\<exists> lil . lil \<in> create_all_possible_execution_paths start c visited n \<and> (lil \<noteq> [] \<longrightarrow> ((fst (hd lil) \<in> dom c \<longrightarrow> (\<exists> tr s X . sfsem_exec_path_is_dense (tr, s, X) lil {(tr', s', X')})) \<and>
         (fst (hd lil) \<notin> dom c \<longrightarrow> (\<exists> tr s X . NCPC s \<notin> dom c \<and> sfsem_exec_path_is_dense (tr, s, X) (tl lil) {(tr', s', X')}))))"

using b e f g
proof (induct n arbitrary: visited tra sa Xa start)
  show "\<And>visited tra sa Xa start.
       N sa \<Longrightarrow>
       sfsem_exec_path_is_dense (tra, sa, Xa) visited {(tr', s', X')} \<Longrightarrow>
       visited \<noteq> [] \<longrightarrow> fst (hd visited) \<in> definitions.dom c \<Longrightarrow>
       NCPC sa = start \<Longrightarrow>
       \<exists>lil. lil \<in> create_all_possible_execution_paths start c visited 0 \<and>
             (lil \<noteq> [] \<longrightarrow>
              (fst (hd lil) \<in> definitions.dom c \<longrightarrow> (\<exists>tr s X. sfsem_exec_path_is_dense (tr, s, X) lil {(tr', s', X')})) \<and>
              (fst (hd lil) \<notin> definitions.dom c \<longrightarrow> (\<exists>tr s X. NCPC s \<notin> dom c \<and> sfsem_exec_path_is_dense (tr, s, X) (tl lil) {(tr', s', X')})))"
    by auto
next
  fix CNT lil visited tra sa Xa start
  assume
    aa1: "\<And>visited tra sa Xa start.
           N sa \<Longrightarrow>
           sfsem_exec_path_is_dense (tra, sa, Xa) visited {(tr', s', X')} \<Longrightarrow>
           visited \<noteq> [] \<longrightarrow> fst (hd visited) \<in> definitions.dom c \<Longrightarrow>
           NCPC sa = start \<Longrightarrow>
           \<exists>lil. lil \<in> create_all_possible_execution_paths start c visited CNT \<and>
                 (lil \<noteq> [] \<longrightarrow>
                  (fst (hd lil) \<in> definitions.dom c \<longrightarrow> (\<exists>tr s X. sfsem_exec_path_is_dense (tr, s, X) lil {(tr', s', X')})) \<and>
                  (fst (hd lil) \<notin> definitions.dom c \<longrightarrow> (\<exists>tr s X. NCPC s \<notin> dom c \<and> sfsem_exec_path_is_dense (tr, s, X) (tl lil) {(tr', s', X')})))" and
    aa3: "N sa" and
    aa5: "sfsem_exec_path_is_dense (tra, sa, Xa) visited {(tr', s', X')}" and
    aa6: "visited \<noteq> [] \<longrightarrow> fst (hd visited) \<in> definitions.dom c" and
    aa7: "NCPC sa = start"
  let ?G = "\<exists>lil. lil \<in> create_all_possible_execution_paths start c visited (Suc CNT) \<and>
             (lil \<noteq> [] \<longrightarrow>
              (fst (hd lil) \<in> definitions.dom c \<longrightarrow> (\<exists>tr s X. sfsem_exec_path_is_dense (tr, s, X) lil {(tr', s', X')})) \<and>
              (fst (hd lil) \<notin> definitions.dom c \<longrightarrow> (\<exists>tr s X. NCPC s \<notin> dom c \<and> sfsem_exec_path_is_dense (tr, s, X) (tl lil) {(tr', s', X')})))"
  {
    assume ba1: "indom start (set visited)"
    hence b1: "\<not> \<not> indom start (set visited)" by simp
    have ?G
      using aa5 aa6
      unfolding create_all_possible_execution_paths.simps if_not_P[OF b1]
      by auto
  }
  moreover
  {
    assume ba1: "\<not> indom start (set visited)" and ba2: "start \<notin> definitions.dom c"
    hence ?G
      using ba2 aa5 aa7
      by fastforce
  }
  moreover
  {
    assume ba1: "\<not> indom start (set visited)" and ba2: "start \<in> definitions.dom c"
    hence b1: "\<not> start \<notin> dom c" by simp
    have hd_unfold: "\<And> x xs . hd (x # xs) = x"
      by simp
    obtain i where startiInC: "(start, i) \<in> \<union>\<^sub>s\<^sub>c c" using ba2
      by (metis (mono_tags, lifting) not_in_sc_inj_imp_not_in_dom)
    {
      assume ca1: "\<exists> f . i = do f"
      obtain f where i: "i = do f" using ca1 by blast
      have hlpr: "\<forall> x . f x \<noteq> {}"
        using j ba2 startiInC[unfolded i]
        by (metis (mono_tags, lifting))
      obtain su Xu suR where
        suR: "suR \<in> f (NCRS sa)" and
        su: "su = \<lparr> R = suR, PC = Suc start \<rparr>" and
        Xu: "Xu \<subseteq> \<Sigma>"
        using aa7 j hlpr 
        by blast
      have Nnormalsu: "N (normal su)" by auto
      moreover
      have NCPCnormalsu: "NCPC (normal su) = Suc start" unfolding NCPC.simps using su by simp
      moreover
      have "NCRS (normal su) \<in> f (NCRS sa)" unfolding NCRS.simps using su suR by simp
      ultimately
      have "(tra, normal su, Xu) \<in> {(tr, t, X).
                  \<exists>u X'. (tr, u, X') \<in> {(tra, sa, Xa)} \<and>
                         NCPC u = start \<and> N u \<and> N t \<and> NCPC t = Suc start \<and> NCRS t \<in> f (NCRS u) \<and> X \<subseteq> \<Sigma>}"
        using suR su Xu aa3 aa7
        by blast
      hence "(tra, normal su, Xu) \<in> \<lbrakk>one start (do f)\<rbrakk>{(tra, sa, Xa)}"
        unfolding sfdof
        by blast
      moreover have "(tra, normal su, Xu) \<noteq> (tra, sa, Xa)"
        using NCPCnormalsu aa7
        by fastforce
      ultimately have "(\<exists>x'. (tra, normal su, Xu) \<noteq> x' \<and> (tra, normal su, Xu) \<in> \<lbrakk>(one start (do f))\<rbrakk>{x'} \<and> sfsem_exec_path_is_dense x' visited {(tr', s', X')})"
        using aa5
        by meson
      hence dense: "sfsem_exec_path_is_dense (tra, normal su, Xu) ((start, do f) # visited) {(tr', s', X')}"
        unfolding sfsem_exec_path_is_dense.simps
        using NCPCnormalsu
        by auto
      obtain lil where
        lil: "lil \<in> create_all_possible_execution_paths (Suc start) c ((start, do f) # visited) CNT \<and>
          (lil \<noteq> [] \<longrightarrow>
           (fst (hd lil) \<in> definitions.dom c \<longrightarrow> (\<exists>tr s X. sfsem_exec_path_is_dense (tr, s, X) lil {(tr', s', X')})) \<and>
           (fst (hd lil) \<notin> definitions.dom c \<longrightarrow> (\<exists>tr s X. NCPC s \<notin> dom c \<and> sfsem_exec_path_is_dense (tr, s, X) (tl lil) {(tr', s', X')})))"
        using aa1[OF Nnormalsu dense _ NCPCnormalsu] ba2
        unfolding hd_unfold fst_conv
        by blast
      hence "lil \<in> create_all_possible_execution_paths start c (visited) (Suc CNT)"
        unfolding create_all_possible_execution_paths.simps if_P[OF ba1] if_not_P[OF b1] 
        using startiInC[unfolded i]
        by blast
      hence ?G
        using lil
        by blast
    }
    moreover
    {
      assume ca1: "\<exists> ef sf . i = comm ef sf"
      obtain ef sf where i: "i = comm ef sf" using ca1 by blast
      have hlpr: "\<forall> x . ef x \<noteq> {}"
        using k ba2 startiInC[unfolded i]
        by blast
      obtain su Xu ev where
        ev: "ev \<in> ef (NCRS sa)" and
        su: "su = \<lparr> R = sf (NCRS sa) ev, PC = Suc start \<rparr>" and
        Xu: "Xu \<subseteq> \<Sigma>"
        using aa7 hlpr  
        by blast
      have Nnormalsu: "N (normal su)" by auto
      moreover
      have NCPCnormalsu: "NCPC (normal su) = Suc start" unfolding NCPC.simps using su by simp
      moreover
      have "NCRS (normal su) = sf (NCRS sa) ev" unfolding NCRS.simps using su by simp
      ultimately
      have "(tra @ [ev], normal su, Xu) \<in> {(tr', t, X).
        \<exists>u X' tr ev.
           (tr, u, X') \<in> {(tra, sa, Xa)} \<and>
           NCPC u = start \<and> N u \<and> N t \<and> ev \<in> ef (NCRS u) \<and> NCPC t = Suc start \<and> NCRS t = sf (NCRS u) ev \<and> tr' = tr @ [ev] \<and> X \<subseteq> \<Sigma>}"
        using ev su Xu aa3 aa7
        by blast
      hence "(tra @ [ev], normal su, Xu) \<in> \<lbrakk>one start (comm ef sf)\<rbrakk>{(tra, sa, Xa)}"
        unfolding sfcomm
        by blast
      moreover have "(tra @ [ev], normal su, Xu) \<noteq> (tra, sa, Xa)"
        by fastforce
      ultimately have "(\<exists>x'. (tra @ [ev], normal su, Xu) \<noteq> x' \<and> (tra @ [ev], normal su, Xu) \<in> \<lbrakk>(one start (comm ef sf))\<rbrakk>{x'} \<and> sfsem_exec_path_is_dense x' visited {(tr', s', X')})"
        using aa5
        by meson
      hence dense: "sfsem_exec_path_is_dense (tra @ [ev], normal su, Xu) ((start, comm ef sf) # visited) {(tr', s', X')}"
        unfolding sfsem_exec_path_is_dense.simps
        using NCPCnormalsu
        by auto
      obtain lil where
        lil: "lil \<in> create_all_possible_execution_paths (Suc start) c ((start, comm ef sf) # visited) CNT \<and>
          (lil \<noteq> [] \<longrightarrow>
           (fst (hd lil) \<in> definitions.dom c \<longrightarrow> (\<exists>tr s X. sfsem_exec_path_is_dense (tr, s, X) lil {(tr', s', X')})) \<and>
           (fst (hd lil) \<notin> definitions.dom c \<longrightarrow> (\<exists>tr s X. NCPC s \<notin> dom c \<and> sfsem_exec_path_is_dense (tr, s, X) (tl lil) {(tr', s', X')})))"
        using aa1[OF Nnormalsu dense _ NCPCnormalsu, unfolded hd_unfold fst_conv] ba2
        by auto
      hence "lil \<in> create_all_possible_execution_paths start c (visited) (Suc CNT)"
        using startiInC
        unfolding create_all_possible_execution_paths.simps if_P[OF ba1] if_not_P[OF b1] i
        by blast
      hence ?G
        using lil
        by blast
    }
    moreover
    {
      assume ca1: "\<exists> bf M N . i = cbr bf M N"
      obtain bf tM tN where i: "i = cbr bf tM tN" using ca1 by blast
      obtain su Xu where
        su: "su = \<lparr> R = NCRS sa, PC = (if bf (NCRS sa) then tM else tN) \<rparr>" and
        Xu: "Xu \<subseteq> \<Sigma>"
        by blast
      have Nnormalsu: "N (normal su)" by auto
      moreover
      have NCPCnormalsu: "NCPC (normal su) = (if bf (NCRS sa) then tM else tN)" unfolding NCPC.simps using su by simp
      moreover
      have "NCRS (normal su) = NCRS sa" unfolding NCRS.simps using su by simp
      ultimately
      have "(tra, normal su, Xu) \<in> {(tr, t, X).
        \<exists>u X'. (tr, u, X') \<in> {(tra, sa, Xa)} \<and>
               NCPC u = start \<and> N u \<and> N t \<and> NCRS u = NCRS t \<and> X \<subseteq> \<Sigma> \<and> (bf (NCRS u) \<and> NCPC t = tM \<or> \<not> bf (NCRS u) \<and> NCPC t = tN)}"
        using su Xu aa3 aa7
        by fastforce
      hence step: "(tra, normal su, Xu) \<in> \<lbrakk>one start (cbr bf tM tN)\<rbrakk>{(tra, sa, Xa)}"
        unfolding sfcbr
        by blast 
      {
        assume da1: "PC su \<noteq> NCPC sa"
        hence "(tra, normal su, Xu) \<noteq> (tra, sa, Xa)"
          by fastforce
        hence "(\<exists>x'. (tra, normal su, Xu) \<noteq> x' \<and> (tra, normal su, Xu) \<in> \<lbrakk>(one start (cbr bf tM tN))\<rbrakk>{x'} \<and> sfsem_exec_path_is_dense x' visited {(tr', s', X')})"
          using aa5 step
          by meson
        hence dense: "sfsem_exec_path_is_dense (tra, normal su, Xu) ((start, cbr bf tM tN) # visited) {(tr', s', X')}"
          unfolding sfsem_exec_path_is_dense.simps
          using NCPCnormalsu da1 aa7
          by auto
        obtain lil where
          lil: "lil \<in> create_all_possible_execution_paths (if bf (NCRS sa) then tM else tN) c ((start, cbr bf tM tN) # visited) CNT \<and>
            (lil \<noteq> [] \<longrightarrow>
             (fst (hd lil) \<in> definitions.dom c \<longrightarrow> (\<exists>tr s X. sfsem_exec_path_is_dense (tr, s, X) lil {(tr', s', X')})) \<and>
             (fst (hd lil) \<notin> definitions.dom c \<longrightarrow> (\<exists>tr s X. NCPC s \<notin> dom c \<and> sfsem_exec_path_is_dense (tr, s, X) (tl lil) {(tr', s', X')})))"
          using aa1[OF Nnormalsu dense _ NCPCnormalsu, unfolded hd_unfold fst_conv] ba2 startiInC
          by auto
        {
          assume bf_true: "bf (NCRS sa)"
          have "lil \<in> create_all_possible_execution_paths tM c ((start, cbr bf tM tN) # visited) CNT"
            using lil[THEN cj2_1, unfolded if_P[OF bf_true]]
            .
          moreover
          have "tM \<noteq> start"
            using aa7 bf_true da1 su by auto
          ultimately
          have "lil \<in> create_all_possible_execution_paths start c (visited) (Suc CNT)"
            using startiInC
            unfolding create_all_possible_execution_paths.simps if_P[OF ba1] if_not_P[OF b1] i
            using da1 aa7
            by fastforce
          hence ?G
            using lil
            by blast
        }
        moreover
        {
          assume bf_false: "\<not> bf (NCRS sa)"
          have "lil \<in> create_all_possible_execution_paths tN c ((start, cbr bf tM tN) # visited) CNT"
            using lil[THEN cj2_1, unfolded if_not_P[OF bf_false]]
            .
          moreover
          have "tN \<noteq> start"
            using aa7 bf_false da1 su by auto
          ultimately
          have "lil \<in> create_all_possible_execution_paths start c (visited) (Suc CNT)"
            using startiInC
            unfolding create_all_possible_execution_paths.simps if_P[OF ba1] if_not_P[OF b1] i
            using da1 aa7
            by fastforce
          hence ?G
            using lil
            by blast
        }
        ultimately
        have ?G
          by blast
      }
      moreover
      {
        assume da1: "PC su = NCPC sa"
        hence PCsustart: "PC su = start"
          using aa7
          by blast
        hence start: "start = (if bf (NCRS sa) then tM else tN)"
          using su by auto
        have "(start, cbr bf tM tN) # visited \<in> create_all_possible_execution_paths tM c ((start, cbr bf tM tN) # visited) CNT \<union> create_all_possible_execution_paths tN c ((start, cbr bf tM tN) # visited) CNT"
        proof (induct CNT, simp, cases "bf (NCRS sa)")
          fix CNT
          assume bf: "bf (NCRS sa)"
          have notIndom: "\<not>\<not>indom tM (set ((start, cbr bf tM tN) # visited))"
            using start[unfolded if_P[OF bf]]
            unfolding indom_def
            by auto 
          show "(start, cbr bf tM tN) # visited
           \<in> create_all_possible_execution_paths tM c ((start, cbr bf tM tN) # visited) (Suc CNT) \<union>
              create_all_possible_execution_paths tN c ((start, cbr bf tM tN) # visited) (Suc CNT)"
            unfolding "create_all_possible_execution_paths.simps" if_not_P[OF notIndom]
            by blast
        next
          fix CNT
          assume bf: "\<not> bf (NCRS sa)"
          have notIndom: "\<not>\<not>indom tN (set ((start, cbr bf tM tN) # visited))"
            using start[unfolded if_not_P[OF bf]]
            unfolding indom_def
            by auto 
          show "(start, cbr bf tM tN) # visited
           \<in> create_all_possible_execution_paths tM c ((start, cbr bf tM tN) # visited) (Suc CNT) \<union>
              create_all_possible_execution_paths tN c ((start, cbr bf tM tN) # visited) (Suc CNT)"
            unfolding "create_all_possible_execution_paths.simps" if_not_P[OF notIndom]
            by blast
        qed
        hence "visited \<in> create_all_possible_execution_paths start c ((* (start, cbr bf tM tN) #  *)visited) (Suc CNT)"
          using startiInC 
          unfolding create_all_possible_execution_paths.simps if_P[OF ba1] if_not_P[OF b1] i
          using NCPCnormalsu PCsustart
          by fastforce
        hence ?G
          using aa5 aa6 by blast
      }
      ultimately
      have ?G
        by blast
    }
    ultimately
    have ?G
      using instr.exhaust[of i ?G]
      by blast
  }
  ultimately
  show ?G
    by fast
qed

text {* Specialize the previous general lemma to the version of starting with no visited instructions. *}

private lemma create_all_possible_execution_paths_for_each_start_there_is_always_an_executable_dense_path:
  assumes
    b: "N s'" and
    j: "\<forall> f . (\<exists> lbl . (lbl, do f) \<in> \<union>\<^sub>s\<^sub>c c) \<longrightarrow> (\<forall> x . f x \<noteq> {})" and
    k: "\<forall> ef . (\<exists> lbl sf . (lbl, comm ef sf) \<in> \<union>\<^sub>s\<^sub>c c) \<longrightarrow> (\<forall> x . ef x \<noteq> {})"
  shows "\<exists> lil . lil \<in> create_all_possible_execution_paths (NCPC s') c [] n \<and> (lil \<noteq> [] \<longrightarrow> (
         (fst (hd lil) \<in> dom c \<longrightarrow> (\<exists> tr s X . sfsem_exec_path_is_dense (tr, s, X) lil {(tr', s', X')})) \<and>
         (fst (hd lil) \<notin> dom c \<longrightarrow> (\<exists> tr s X . NCPC s \<notin> dom c \<and> sfsem_exec_path_is_dense (tr, s, X) (tl lil) {(tr', s', X')}))))"
using create_all_possible_execution_paths_for_each_start_there_is_always_an_executable_dense_path'
  [OF b, of tr' X' "[]" tr' s' X' c "NCPC s'" n, OF _ _ _ j k]
unfolding sfsem_exec_path_is_dense.simps
by blast

text {* So, if each instruction is followed by a @{term comm} or leaves, then for each possible starting state, there always
is a successor. This result is weaker than the assumptions, as we now still could have loops that can be run infinite often,
but also can be left nondeterministically at any iteration. The assumptions would not allow that, but
this shows a relation between syntax and semantics, which is sufficient to proof the big lemma (SF refinement implies T refinement).
*}

lemma sfsem_each_instruction_has_a_successor:
  assumes
    a: "all_commands_are_followed_by_comm_or_leave c" and
    wff: "wff\<^sub>s\<^sub>c c" and
    doF: "\<forall> f . (\<exists> lbl . (lbl, do f) \<in> \<union>\<^sub>s\<^sub>c c) \<longrightarrow> (\<forall> x . f x \<noteq> {})" and
    commEF: "\<forall> ef . (\<exists> lbl sf . (lbl, comm ef sf) \<in> \<union>\<^sub>s\<^sub>c c) \<longrightarrow> (\<forall> x . ef x \<noteq> {})"
  shows "\<forall> tr s X . \<lbrakk>c\<rbrakk>{(tr, s, X)} \<noteq> {}"
proof(rule, rule, rule)
  fix tr s X
  let ?G = "\<lbrakk>c\<rbrakk>{(tr, s, X)} \<noteq> {}"
  {
    assume "\<not> (NCPC s \<in> dom c \<and> N s)"
    hence ?G
      using pc_does_not_point_to
      by simp
  }
  moreover
  {
    assume inside: "NCPC s \<in> dom c" and normal: "N s"
    obtain lil where
      lil1: "lil \<in> create_all_possible_execution_paths (NCPC s) c [] (Suc (size c))" and
      lil2: "lil \<noteq> [] \<longrightarrow>
         (fst (hd lil) \<in> definitions.dom c \<longrightarrow> (\<exists>tra sa Xa. sfsem_exec_path_is_dense (tra, sa, Xa) lil {(tr, s, X)})) \<and>
         (fst (hd lil) \<notin> definitions.dom c \<longrightarrow> (\<exists>tra sa Xa. NCPC sa \<notin> dom c \<and> sfsem_exec_path_is_dense (tra, sa, Xa) (tl lil) {(tr, s, X)}))"
      using create_all_possible_execution_paths_for_each_start_there_is_always_an_executable_dense_path[OF normal doF commEF, of "Suc (size c)" tr X]
      by blast
    have b1: "(\<exists>sf f l. (l, comm sf f) \<in> set lil) \<or> lil \<noteq> [] \<and> fst (hd lil) \<notin> definitions.dom c"
      using instr_followed_by_comm_or_leaves_eq_create_all_possible_execution_paths[of "[]", simplified, of "NCPC s" c "Suc (size c)"]
        a[unfolded all_commands_are_followed_by_comm_or_leave_def]
        inside lil1
      by blast
    moreover
    have "(\<exists>ef sf l. (l, comm ef sf) \<in> set lil) \<longrightarrow> lil \<noteq> []"
      by fastforce
    moreover
    {
      assume ca1: "\<exists>ef sf l. (l, comm ef sf) \<in> set lil" and ca2: "fst (hd lil) \<in> dom c"
      obtain ef sf lbl where li : "(lbl, comm ef sf) \<in> set lil"
        using ca1 by blast
      hence "\<exists> l r . lil = l @ (lbl, comm ef sf) # r"
        by (simp add: in_set_conv_decomp)
      then obtain l r where lil: "lil = l @ (lbl, comm ef sf) # r" by blast
      then obtain tra sa Xa where
        co1: "sfsem_exec_path_is_dense (tra, sa, Xa) lil {(tr, s, X)}"
        using lil2 ca2 by blast
      obtain x' where
        co2: "sfsem_exec_path_is_dense (tra, sa, Xa) l {x'}" and
        co3: "sfsem_exec_path_is_dense x' ((lbl, comm ef sf) # r) {(tr, s, X)}"
        using sfsem_exec_dense_path_unfold[OF co1[unfolded lil]]
        by blast
      obtain x'' where
        co4: "x' \<notin> {x \<in> sfsem_exec r {(tr, s, X)}. NCPC (fst (snd x)) = lbl}" and
        co5: "x' \<noteq> x''" and
        co6: "x' \<in> \<lbrakk>(one lbl (comm ef sf))\<rbrakk>{x''}" and
        co7: "sfsem_exec_path_is_dense x'' r {(tr, s, X)}"
        using co3[unfolded sfsem_exec_path_is_dense.simps]
        by blast
      obtain trx'' sx'' Xx'' where x'': "x'' = (trx'', sx'', Xx'')" using prod_cases3 by blast
      have sx'': "N sx'' \<and> NCPC sx'' = lbl"
        using co5 co6
        unfolding x''
        by force
      then obtain XCx'' where "(trx'', NCC sx'', XCx'') \<in> \<lbrakk>(one lbl (comm ef sf))\<rbrakk>{x''}"
        unfolding x''
        unfolding sfcomm
        using sx''
        by auto
      with co7 have c1: "(trx'', NCC sx'', XCx'') \<in> sfsem_exec ((lbl, (comm ef sf)) # r) {(tr, s, X)}"
        unfolding sfsem_exec_unfold
        using sfsem_exec_is_dense_is_exec[OF co7]
        using sfsem_monotone_single
        by blast
      have lilInC: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c"
        using create_all_possible_execution_paths_all_indom_c[OF lil1 _ _ ca2] ca1
        by fastforce
      hence c2: "set ((lbl, (comm ef sf)) # r) \<subseteq> \<union>\<^sub>s\<^sub>c c"
        using lil by simp
      hence "(trx'', NCC sx'', XCx'') \<in> \<lbrakk>c\<rbrakk>{(tr, s, X)}"
        using c1 NCCNotNormal[of sx''] sfsem_exec2[OF c2 c1 _ wff]
        by fast
      hence ?G
        by blast
    }
    moreover
    {
      assume ca1: "lil \<noteq> []" and ca2: "fst (hd lil) \<notin> dom c"
      have tlInC: "set (tl lil) \<subseteq> \<union>\<^sub>s\<^sub>c c"
        by (metis create_all_possible_execution_paths_tail_indom_c empty_subsetI lil1 list.sel(2) list.set(1))
      obtain tra sa Xa where
        co1: "NCPC sa \<notin> dom c" and
        co2: "sfsem_exec_path_is_dense (tra, sa, Xa) (tl lil) {(tr, s, X)}"
        using lil2 ca1 ca2
        by blast
      have c1: "(tra, sa, Xa) \<in> sfsem_exec (tl lil) {(tr, s, X)}"
        using sfsem_exec_is_dense_is_exec[OF co2]
        .
      have "(tra, sa, Xa) \<in> \<lbrakk>c\<rbrakk>{(tr, s, X)}"
        using sfsem_exec2[OF tlInC c1 _ wff] co1
        by fast
      hence ?G
        by blast
    }
    ultimately
    have ?G
      by fast
  }
  ultimately
  show ?G
    by blast
qed

end

paragraph {* Infinite systems *}

text {* If no execution path leads out of the code's domain, then we cannot generate a normal state. *}

theorem CannotHaveNormalOutput:
  assumes
    notSingleSelfCbr: "\<not> (\<exists> l b m n . c = (one l (cbr b m n)))" and
    allSInDom: "\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> NCPC s \<in> dom c" and
    noPathOutside: "noPathOutside c" and
    result: "(tr, s, X) \<in> \<lbrakk>c\<rbrakk>S"
  shows "\<not> N s"
proof
  assume Ns: "N s"
  obtain lil where
    lilSubset: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    execLil: "(tr, s, X) \<in> sfsem_exec lil S" and
    critical: "sfsem_exec_path_is_critical (tr, s, X) lil S"
    using sfsem_exec1'_crit[OF result]
    by blast
  have notPCInside: "\<not> NCPC s \<in> dom c"
    using Ns notSingleSelfCbr sfsem_not_outside_means_cbr[OF result]
    by blast
  hence notInS: "(tr, s, X) \<notin> S"
    using allSInDom
    by blast
  hence "lil \<noteq> []"
    using execLil sfsem_exec_empty
    by force
  then obtain li lil' where
    lil: "lil = li # lil'"
    by (meson neq_Nil_conv)
  obtain l i where li: "li = (l, i)" by fastforce
  obtain tr' s' X' where
    unfolded_exec: "(tr, s, X) \<in> \<lbrakk>(one l i)\<rbrakk>sfsem_exec lil' S" and
    sNotInLil': "(tr, s, X) \<notin> sfsem_exec lil' S" and
    sSuccOfs': "(tr, s, X) \<in> \<lbrakk>(one l i)\<rbrakk>{(tr', s', X')}" and
    sNeqs': "(tr, s, X) \<noteq> (tr', s', X')" and
    s'ExecLil': "(tr', s', X') \<in> sfsem_exec lil' S" and
    critical_pred: "sfsem_exec_path_is_critical (tr, s, X) [(l, i)] {(tr', s', X')}" and
    critical_s': "sfsem_exec_path_is_critical (tr', s', X') lil' S" and
    Ns': "N s'" and
    NCPCs'InDom: "NCPC s' = l"
    using sfsem_exec_lil_split2[OF critical[unfolded lil li]]
    by blast
  have lInDom: "l \<in> dom c"
    using in_sc_inj_imp_in_dom li lil lilSubset by auto
  hence "NCPC s' \<in> dom c"
    using NCPCs'InDom by blast
  have s'InDomL: "NCPC s' \<in> dom (one l i)"
    using NCPCs'InDom
    unfolding dom.simps by simp
  have liInSC: "(l, i) \<in> \<union>\<^sub>s\<^sub>c c"
    using li lil lilSubset
    by simp
  have noPathOutsideFunl: "noPathOutsideFun (one l i) (definitions.dom c)"
    using noPathOutside_For_Every_Subinstruction[OF noPathOutside liInSC]
    .
  show False 
  proof (cases i)
    fix f
    assume i: "i = do f"
    have "NCPC s = Suc (NCPC s')"
      using sSuccOfs' sNeqs'
      unfolding i
      by auto
    thus False
      using noPathOutsideFunl[unfolded i "noPathOutsideFun.simps"] notPCInside NCPCs'InDom
      by simp
  next
    fix ef sf
    assume i: "i = comm ef sf"
    have "NCPC s = Suc (NCPC s')"
      using sSuccOfs' sNeqs' Ns NCCNotNormal
      unfolding i sfcomm SClean_def
      by fast
    thus False
      using noPathOutsideFunl[unfolded i "noPathOutsideFun.simps"] notPCInside NCPCs'InDom
      by simp
  next
    fix b m n
    assume i: "i = cbr b m n"
    have "NCPC s = m \<or> NCPC s = n"
      using sSuccOfs' sNeqs' Ns NCCNotNormal
      unfolding i sfcbr SClean_def
      by fast
    thus False
      using noPathOutsideFunl[unfolded i "noPathOutsideFun.simps"] notPCInside NCPCs'InDom
      by blast
  qed
qed

paragraph {* Allowed communication *}

text {* If we restrict the allowed set of communicable events, then those won't be communicated. *}

lemma sfsem_allowed_communication:
  assumes
    allowed_comm: "wff_cuc_allowed_communication c E" and
    wff: "wff\<^sub>s\<^sub>c c" and
    base: "\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> set tr \<subseteq> E" and
    sf: "(tr, s, X) \<in> sfsem c S"
  shows "set tr \<subseteq> E"
proof -
  obtain lil where
    subs: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c c" and
    ex: "(tr, s, X) \<in> sfsem_exec lil S"
    using sfsem_exec1_crit[OF sf] by blast
  show ?thesis
    using ex subs
  proof (induct lil arbitrary: tr s X)
    case Nil
    then show ?case using base by auto
  next
    case (Cons a lil)
    obtain l i where a: "a = (l, i)" by fastforce
    obtain tr' s' X' where
      ex: "(tr', s', X') \<in> sfsem_exec lil S" and
      ex_one: "(tr, s, X) \<in> \<lbrakk>(l :: i)\<rbrakk>{(tr', s', X')}"
      using sfsem_exec_pred1[OF Cons(2)[unfolded a]] by force
    have base: "set tr' \<subseteq> E" using Cons(1)[OF ex] Cons by simp
    consider (treq) "tr = tr'" | (trneq) "tr \<noteq> tr'" by blast
    thus ?case
    proof cases
      case treq thus ?thesis using base by simp
    next
      case trneq
      then obtain ef f e where i: "i = comm ef f" and tr: "tr = tr' @ [e]" and e: "e \<in> ef (NCRS s')"
        using ex_one proof (induct i, auto) qed
      have instr_in_c: "(l, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c" using i Cons(3) a by simp
      define C where C: "C = \<Union>{v. \<exists>s. v = ef s}"
      have "e \<in> C" using C e by blast
      moreover have "C \<subseteq> \<Union>{C. \<exists>ef f lbl. (lbl, comm ef f) \<in> \<union>\<^sub>s\<^sub>c c \<and> C = \<Union>{v. \<exists>s. v = ef s}}"
        using C instr_in_c by blast
      ultimately have "e \<in> E"
        using allowed_comm[unfolded wff_cuc_allowed_communication_def] by fast
      thus ?thesis using base tr by simp
    qed
  qed
qed

end -- sfsemSigma

end
